
;;;*********************************************************************
;;;
;;;                  ANALOGY.LISP    In KCL
;;;                            
;;;               W. W. Bledsoe     13 July 1988
;;;
;;;               For use by  IMPLY.LISP
;;;
;;;***********************************************************************
;;;
;;;    This is used in conjuction with IMPLY.LISP and PROOFS.LISP
;;;
;;;    A PROOF, as found in PROOFS.LISP, is the result of proving a theorem
;;;      using a plan. (See PROOF.LISP for a description of the PLAN elements.)
;;;
;;;    The function ANALOGY-PROVE  found here attempts to use such a proof of 
;;;      an earlier theorem ("guiding theorem"), to guide the discovery of a 
;;;      proof of a new theorem.
;;;
;;;***********************************************************************

(defun ap (new-th gui-pr) (analogy-prove new-th gui-pr))

(defun analogy-prove (new-theorem guiding-proof)
  (let (%al% theta)                   ;%al% stands for Analogy-light
       (setq %proof% guiding-proof)
       (setq %al% t)
       (cond ((setq theta (prove new-theorem %proof% nil 'ap)) theta))))
          
(defun A-options (H C TL ANC LT YET %proof%) ;called by IMPLY,if %al% not nil
  ;;This is a lot like OPTIONS in the interactive prover, except
  ;;  here there is no **-IMPLY-STOP-**, the commands are always taken
  ;;  from the current value of %proof%.

 (declare  (special h c tl anc lt yet))
 (let (action-pair action state action-result read)
  (clear-in)
  (format t "~%~%~s~%" tl)
  (format t "Current Goal: ~s~%" C)

  (cond (*A-options-stop-light*               ;Stop if stop light is on.
	 ;;Wait for a user input from the keyboard.  Else go below to the
	 ;; second next cond statement for commands from %proof%.
         (format t "~%@@@A-OPTIONS-STOP-@@@@@@@@@: ")
         (setq read (testr))
         (cond ((eq 'carriage-return read) (format t "PROCEED~%"))   ;proceed
               ((eq 'US (setq read (read)))  ;Turn off Stop Light and proceed.
                (format t "Turn Off Stop Light and Proceed~%") 
                (setq *A-OPTIONS-STOP-LIGHT* nil))  
               ((and (format t "~s~%" read) nil))
               ((eq 'save-args read) 
                (setq *imply&prover-arguments*       ;Save-args for restarting.
		      (list H C TL ANC LT YET %proof% %al% %NON-STOP-LIGHT% nil
			    %REDUCE-LIGHT% 'NON-STOP-LIGHT %GOAL-MESSAGE% 
			    %TIME% %theorem-variables% %object-list%) ) 
                (return-from A-options (A-ops %proof%)))
               ((eq 'HH read) (number&print-h H)
                (return-from A-options (A-ops %proof%)))
               ((eq 'E read)      ;break to LISP level, :R to proceed.
                (format t "~%EVAL: ~%") (break)
                (return-from A-options (A-ops %proof%)))
               ((eq 'A read) 
                (setq %proof% (cons '(assume t) (cdr %proof%)))
                (setq *A-OPTIONS-STOP-LIGHT* nil))
               (t (setq %proof% (cons read %proof%))))))
  (cond ((not %proof%)
         (format t "~% In A-OPTIONS %proof% is NIL, return NIL.~%") 
         (return-from A-options nil)))

  (setq action-pair (car %proof%))
  (setq %proof% (cdr %proof%))
  (setq action (car action-pair))
  (setq state (cadr action-pair))
  (format t "~%%%--ACTION--%%%%%%%%%%%%%%%%%%%%%%%%%%: ")
  (cond ((< (symbol-length action) 12) (format t "~s~%" action))
        (t                             (format t "~%  ~s~%" action)))
  (format t "~%%%--INTENDED-RESULTANT-STATE--%%%%%%%%: ")
  (cond ((< (symbol-length state) 12) (format t "~s~%" state))
        (t                             (format t "~%  ~s~%" state)))

  (cond ((eq 'no-op action) (A-Options-stop)
         (return-from A-options (A-options H C TL ANC LT YET %proof%))))

  (setq state (remove-add&subtract state h c))

  (cond ((setq action-result (carry-out-the-action action))
         (cond ((eq t state) 
		(format t "~% Action Succeeded: Return sigma~%")
		action-result)
	       ((or (eq 'tt state) (state-check state (list H C)))
                (format t "~% Action Succeeded, Proceed with another Action~%")
                (A-options H C TL ANC LT YET %PROOF%))
	       (t (format t "~%Action State-check fails.~%") nil)) )
        (t (format t "~%  FAILED TO CARRY OUT THE ACTION: ~s ~%" action)
           (break) nil)) ))

(defun A-ops (pr)       ;called by A-Options
 (declare  (special h c tl anc lt yet))
 (A-options H C TL ANC LT YET pr))

(defun A-Options-stop (&optional (flag t))
  (setq *A-options-stop-light* flag))

(defun carry-out-the-action (action)       ;Called by A-options.
  (declare  (special h c tl anc lt yet))
  (let (C*)

   (cond ((eq 'DC action) (setq tl (append tl '(DC)))
          (setq C (DEFINE C t)) )

	 ((EQ 'PROMOTE action) (setq tl (append tl '(P->)))	    
	  (setq H (and-on-h H (cadr C))) (setq C (caddr C))
	  t)

	 ((eq 'PROVE action)
	  (Imply H C TL nil 'prove nil))

	 ((eq 'R* action) (setq tl (append tl '(R))) (setq C (reduc** C)))

         ((eq 'HH action) (number&print-H h) t)

         ((eq 'assume action)   ;Return TRUE for the current goal
          (format t "~%Current Goal assumed by the USER~%") t)

	 ((atom action) 
	  (format t "~%UNRECOGNIZED A-OPTION:  ~s ~%" action)        
	  nil)

	 ((member (car action)  '(USE USE-A))
	  (use-a (cadr action) (cond ((cddr action) (caddr action)))))

	 ((eq 'USE-BC  (car action)) 
	  (cond ((setq C* (use-bc (cadr action) 
                                  (cond ((cddr action) (caddr action)))))
                 (setq c (car C*)) (cadr C*))
                (t nil)))
 
	 ((eq 'CLAIM   (car action)) 
          (claim (cadr action) (caddr action))) 

	 ((eq 'CASES   (car action)) (cases2 (cdr action)))

	 ((eq 'INDUCTION (car action)) (Induction (cdr action)))

	 ((eq 'LET     (car action)) (lett (cdr action)))

	 ((eq 'SPLIT-C (car action)) (split-c (cdr action)))

	 ((eq 'SPLIT-H (car action)) (split-h (cdr action)))

	 ((eq 'SUB=    (car action))
	  (sub== (cadr action) (cond ((cddr action) (caddr action)))))

         ((eq 'SUB==   (car action)) (sub==-in (cadr action)))

         ((eq 'RR      (car action)) (setq C* (reduc** (cadr action)))
          (setq H (subst C* (cadr action) (copy-tree H)))
          (setq C (subst C* (cadr action) (copy-tree C))))

	 (t (format t "~%UNRECOGNIZED A-OPTION:  ~s ~%" action)  nil))))

(defun equivalent-& (A B)                      ;Called by state-check
  ;;Test whether A and B are set-equal.
  (or (equal A B)
      (and (not (atom A)) (eq '& (car A)) (not (atom A)) (eq '& (car A)) 
           (set-equal A B))))

(defun options-stop (&optional (flag t))
   (cond (flag (setq *A-options-stop-light* t))
         (t    (setq *A-options-stop-light* nil))))

(defun remove-add&subtract (A h c)             ;Called by state-check
  (let (h2 c2 u)
       (cond ((member  A '(t tt)) (return-from remove-add&subtract A)))
       (cond ((not (occur 'add A)) (setq H2 h)))
       (cond ((not (occur 'new-value A)) (setq C2 c)))
       (cond ((and h2 c2) (list h2 c2))
             (h2 (list h2 (caddar A)))               ;A = ((new-value c XXX))
             (c2 (list (and-on-h h (caddar A)) c2))  ;A = ((add h XXX))
             (t (setq u (assoc 'add A)) 
                (setq h2 (caddr u))
                (setq u (assoc 'new-value A))
                (setq c2 (caddr u))
                (list (and-on-h h h2) c2)))))

(defun state-check (State State*)       ;Called by A-options
  ;;State and State* each has the form ((U h H= Hv) C).
  (let (L1 L2)
       (setq L1 (append (car State)  (cdr State)))
       (setq L2 (append (car State*) (cdr State*)))
       (every #'(lambda (x y) (equivalent-& x y)) L1 L2)))

