;;;         grd-sb.lisp

;;;  From Larry Hines    28 July 1989
;;; -*- Fonts: HL12 -*-

(require "SLOOP")
;;;(in-package "SLOOP" :use '(LISP))
;;;
;;;(defmacro loop (&body body) (parse-loop body))
;;;    is assumed to have been defined.
;;;
;(require "sb1.lisp")
;(require "sb2.lisp")
;(require "cnf.lisp")
;;(require "double-unify")

(defun run-examples (&key (examples grd-sb-examples) (verbose nil) (time nil))
  "Run the ground subset prover on a list of the names of examples, returning
 the names along with the results.  If the keyword :TIME is supplied and is
 nonNIL, then the run-time taken by the example is also returned with each
 example.  The list of examples can be specified by the key :examples.  
 The default is  GRD-SB-EXAMPLES.  Also, the prover's verbose mode can be
 turned on by supplying :VERBOSE T."
  (loop for ex-name in examples for ex = (eval ex-name)
	for tm = (get-internal-run-time) do (format t "~a " ex-name)
	for results = (prog1 (grd-sb-top ex :verbose verbose)
			(setq tm (- (get-internal-run-time) tm)))
	if time collect (list ex-name results 
			      (read-from-string
				   (format nil "~5,2f" (/ (float tm) 60.))))
	else collect (list ex-name results)))

(defun grd-sb-top (cs &key (converted nil) (verbose t))
 " Top level function for calling the ground prover.  Returns either T
 (indicating that a contradiction was found) or a set of clauses which
 subsumes the conjunctive normal form of the formula.  
 If the formula has already been converted to cnf, then the user should
 supply :CONVERTED T.   The verbose mode of the prover can be turned
 off with :VERBOSE NIL."
  (let ((x (grd-sb-incremental cs :verbose verbose :converted converted)))
    (cond ((arrayp x) (loop with d = (1- (array-dimension x 0))
			    for i from 1 to d thereis (eq t (aref x i 0))
			    finally (return (aref x d 0))))
	  (t x))))

(defun grd-sb-incremental (cs &key (converted nil) (verbose t))
 " Calls the ground subset prover successively, starting with
 the first two clauses of the cnform of the formula and adding
 the next clause in each successive call.  Returns T if any
 of the calls finds a contradiction; otherwise, returns the
 clause set returned by the last call."
  (declare (special verbose *nonempty-universe*))
  ;;(setq *nonempty-universe* (occur-singleton cs))
  (unless converted (setq cs (negate-prep&cnf-formula cs)))
  (cond ((null cs) nil)
	((member nil cs) t)
	(t (let ((a (make-array (list (length cs) 2))))
	     (setf (aref a 0 0) cs) (setf (aref a 0 1) 2)
	     (loop for i from 2 to (length cs)
		   do (setf (aref a (1- i) 0) (grd-sb (aref a (- i 2) 0)
						      (aref a (- i 2) 1)))
		   thereis (eq t (aref a (1- i) 0))
		   do (setf (aref a (1- i) 1) (1+ (length (aref a (- i 1) 0))))
		   do (and verbose
			   (format t "~%~d clauses after ~d iterations."
				   (1- (aref a (1- i) 1)) (1- i)))
		   do (setf (aref a (1- i) 0)
			    (append (aref a (- i 1) 0)
				    (nthcdr (aref a (- i 2) 1)
					    (aref a (- i 2) 0)))))
	     a))))
(defun grd-sb (clause-set &optional n)
 " The ground subset prover.  Return T if box is found; otherwise,
 returns a clause set which subsumes the input clause set."
  (unless n (setq n (length clause-set)))
  (let ((cs (butlast clause-set (- (length clause-set) n))))
    (grd-sb-resolution (last cs) (butlast cs))))

(defun grd-sb-resolution (sos input-set)
  " Returns t if box is found.  Otherwise it returns the set of input 
 clauses including the set-of-support, but each clause may have been
 replaced by a subsuming clause."
  (declare (special verbose))
  (if (null sos) input-set
    (let* ((n 0) (cs (loop for c in input-set
			   collect (make-clause (incf n) 'input c)
			   do (and verbose (format t "~%~13d. ~a." n c)))))
      (loop for c in sos for top-cl = (make-clause (incf n) 'input c)
	    do (and verbose (format t "~%~13d. ~a." n c))
	    do (setq cs (append cs (list top-cl)))
	    thereis (grd-sb-refute n top-cl cs nil)
	    finally (return (loop for c in cs unless (subsumed? c)
				  collect (literals c)))))))

(defun grd-sb-refute (n central-cl input-set ancestors)
  " Returns t if top-cl can be refuted by cs using ground sis-resolution.
 (Ground sis-resolution is the same as input resolution with replacement
 of subsumed input clause).  Otherwise it returns nil.  As a side-effect, 
 each subsumed clause in CS is replaced by the subsuming clause."
  (loop with literals = (literals central-cl)
	for lit1 in literals thereis
	(loop for side-cl in input-set
	      unless (or (eq side-cl central-cl) (subsumed? side-cl))
	      thereis
	      (loop for lit2 in (literals side-cl) 
		     ;; In the ground case, there can be at most one
		     ;; resolvent given two literals in two clauses.
		    for r = (grd-sb-resolve-lits lit1 central-cl 
						 lit2 side-cl input-set
						 ancestors)
		    if r thereis 
		    (or (eq t r)				; Box.
			(multiple-value-bind 
			  (cl is anc) (grd-sb-resolvent-set
					r n central-cl input-set ancestors)
			 ;; CL is the resolvent or an input clause with
			 ;; its literals replaced by R.
			 ;; IS
			  (printout-resolvent (cl-number cl) r 
					      central-cl side-cl)
			  (grd-sb-refute (1+ n) cl is anc)))
		    if (subsumed? side-cl) return nil
		    ;; The side clause has been subsumed by R
		    ;; or some descendant of R. 
		    ;; So, we jump out of the loop.
		    if (or (subsumed? central-cl)
			   (not (eq literals (literals central-cl))))
		      ;; The central clause has been subsumed.
		      ;; So, we jump out of the outer loop.
		    do (return-from grd-sb-refute nil)))))
(defun grd-sb-resolvent-set (r n central-cl input-set ancestors) 
  (multiple-value-bind (is i-cl) (grd-sb-clauses-not-subsumed-by input-set r)
    (cond (i-cl (loop for a in ancestors do (setf (cl-type a) 'subsumed))
		(values i-cl is nil))
	  (t (values (make-clause (1+ n) nil r) is
		     (if (input? (cl-type central-cl))
			 ancestors (cons central-cl ancestors)))))))
(defun grd-sb-resolve-lits (lit1 central-cl lit2 side-cl input-set ancestors)
  (let ((term (intersection (right-side lit1) (left-side lit2) :test #'equal)))
    (cond (term (unless (or (cdr term) (intersection (left-side lit1)
						     (right-side lit2)
						     :test #'equal))
		  (grd-sb-resolve 
		    (grd-make-new-literal 'right (car term) lit1 lit2)
		    lit1 central-cl lit2 side-cl input-set ancestors)))
	  ((and (setq term (intersection (left-side lit1) (right-side lit2)
				      :test #'equal))
		(null (cdr term)))
	   (grd-sb-resolve (grd-make-new-literal 'left (car term) lit1 lit2)
			   lit1 central-cl lit2 side-cl input-set
			   ancestors)))))
(defun grd-sb-resolve (new-lit lit1 central-cl lit2 side-cl 
			       input-set ancestors)
  (let ((r (remove-if 
	     #'false-literal 
	     (union (list new-lit) (union (remove lit1 (literals central-cl))
					  (remove lit2 (literals side-cl))
					  :test #'equal)
		    :test #'(lambda (l1 l2)
			      (if (and (subset=? l1) (subset=? l2))
				  (and (null (set-exclusive-or
					       (left-side l1) (left-side l2)
					       :test #'equal))
				       (null
					 (set-exclusive-or
					   (right-side l1) (right-side l2)
					   :test #'equal)))
				(equal l1 l2)))))))
    (if (null r) t
      (unless (or (grd-sb-tautology r)
		  (grd-sb-subsumed-by-clause-in r input-set)
		  (grd-sb-subsumed-by-clause-in r ancestors))
	      r))))
(defun printout-resolvent (num lits central-cl side-cl)
  (declare (special verbose))
  (and verbose (format t "~%[~3d ~3d] ~3d. ~a"
		       (cl-number central-cl) (cl-number side-cl)
		       num lits)))
(defun grd-sb-subsumed-by-clause-in (literals clause-set-of-subsumers)
  (loop for c2 in clause-set-of-subsumers
	if (grd-sb-subsumed-by literals (literals c2)) return t))
(defun grd-sb-clauses-not-subsumed-by (cl-set subsumer-lits)
  " Returns clauses not subsumed by SUBSUMER." 
  (declare (special verbose))
  (let (i-cl two-subsumed)
    (values (loop for c1 in cl-set 
		  if (grd-sb-subsumed-by (literals c1) subsumer-lits) 
		  do (progn
		       (and verbose (format t "~%~d replaced by resolvent."
					    (cl-number c1)))
		       (cond (i-cl (setf (cl-type c1) 'subsumed) 
				   (setq two-subsumed t))
			     (t (setf (literals c1) subsumer-lits)
				(setq i-cl c1))))
		  else collect c1 into new-set
		  finally (return (if two-subsumed new-set cl-set)))
	    i-cl)))
(defun old-grd-sb-resolve-lits (lit1 central-cl lit2 side-cl 
				     input-set ancestors)
  (or (loop for term in (left-side lit1)
	    if (member term (right-side lit2) :test #'equal)
	    return (grd-sb-resolve (grd-make-new-literal 'left term lit1 lit2)
				   lit1 central-cl lit2 side-cl input-set
				   ancestors))
      (loop for term in (right-side lit1)
	    if (member term (left-side lit2) :test #'equal)
	    return (grd-sb-resolve (grd-make-new-literal 'right term lit1 lit2)
				   lit1 central-cl lit2 side-cl input-set
				   ancestors))))
(defun grd-sb-subsumed-by (literals subsumer-lits) 
  " Returns T if CLAUSE is subsumed by SUBSUMER."
  (every #'(lambda (lit) (grd-sb-lit-subsumed-by lit literals)) subsumer-lits))
(defun grd-sb-lit-subsumed-by (s-lit literals)
  (some #'(lambda (lit2) 
	    (and (subsetp (left-side s-lit) (left-side lit2) :test #'equal)
		 (subsetp (right-side s-lit) (right-side lit2) :test #'equal)))
	literals))

(defun grd-sb-tautology (lits)
  " Returns T if RES contains a literal and its negation."
  (not (null (some #'(lambda (lit) (intersection (left-side lit)
						 (right-side lit)))
		   lits))))


(defun grd-make-new-literal (side term lit1 lit2)
  (if (eq 'right side)
      (make-literal '[=
	(delete-duplicates
	  (append (left-side lit1)
		  (remove term (left-side lit2) :test #'equal :count 1))
	   :test #'equal)
	(delete-duplicates
	  (append (remove term (right-side lit1) :test #'equal :count 1)
		  (right-side lit2))
	   :test #'equal))
    (make-literal '[=
      (delete-duplicates
	(append (remove term (left-side lit1) :test #'equal :count 1)
		(left-side lit2))
	 :test #'equal)
      (delete-duplicates
	(append (right-side lit1)
		(remove term (right-side lit2) :test #'equal :count 1))
	 :test #'equal))))

(defun compile-sb-all () 
  (compile-file "cnf.lisp")
  (compile-file "sb1.lisp")
  (compile-file "sb2.lisp")
  (compile-file "grd-sb.lisp") )

(defun load-sb-o-all () 
  (load "cnf.o")
  (load "sb1.o")
  (load "sb2.o")
  (load "grd-sb.o") )

#|
(load "imply.lisp") (load-lisp-all) (load "grd-sb.lisp") (load-sb-o-all)
  (lcp1) (load "prove21.o") (load "prove22.o") (lcp3)
(lco)(lcu)(lcs) (load "ve1.o")(load "ve2.o") (load "replace1.o")
(load "replace2.o") (lcan)(lcd)
(lcu)(lcs)(lcr)
|#

(defun convert-for-grd-sb (A) ;convert IMPLY symbols for Larry's Set prover.
  (setq A (subst '[= 'subset A)) 
  (setq A (subst '[  'proper-subset  A))
  (sublis '((sng singleton) (non-empty nonempty) (non-unit nonempty)) A))

(defun LSP (cs &key (converted nil) (verbose nil))     ;Larry's set prover
   ;;eg, (LSP thc1)
   (setq cs (convert-for-grd-sb cs))
   (grd-sb-top cs :converted converted :verbose verbose))

(setq thc1 '(-> (= (C1) (set-diff (CC) (sng (L)))) (subset (C1) (CC))))
