
;;;**********************************************************************
;;;
;;;                    INFIX.LISP    In KCL
;;;
;;;               W. W. Bledsoe      27 Apr 1988
;;;
;;;                 Called by  IMPLY.LISP
;;;
;;;              For Infix Printing from OPTIONS
;;;
;;;***********************************************************************


(defun infix (A &optional (n '*))  ;Called by NUMBER&PRINT-INFIX, OPTIONS
  ;;It prints A in infix notation.  n is an entry number (default *).
  ;;eg,  (< A) goes to  A < 0,     (|| A)  goes to |A|
  ;;     (< (+ a b (- c) (- d))) goes to  a + b  <  c + d
  ;;     (+ a b (- c) (- d))  goes to         a + b  -c -d

  (let (L)
    (cond ((setq L (infix-parts A (spaces 3)))
           (format t "~s:~{~a~}" n (cdddr L)))

          (t (pprint A)) )))

(defun infix-parts (A &optional HEAD)  ;called by infix
  ;;This builds the list L for INFIX to print with the FORMAT instruction

  (let (L+ L- L L2)
    (cond ((atom A) (setq L (list '\  A)))

          ((eq '/ (car A))		;A = (/ a b) goes to  a/b
           (setq L (list '\  (cadr A) '/ (caddr A))))
     
          ((eq '- (car A))              ;A = (- a)   goes to  - a
           (setq L (append (list '\  '-) (infix-parts (cadr A)))))

	  ((eq '+ (car A))
	   (setq L (mapcan #'(lambda (x) 
               (setq L2 (infix-parts x))
               (cond ((eq '- (cadr L2)) L2) 
                     (t (append (list '\   '+) L2))))
                                 (cdr A)))
           (setq L (cond ((eq '- (cadr L)) L) (t (cddr L)))))

          ((eq '* (car A))
	   (setq L (cddr (mapcan #'(lambda (x) (append (list '\   '*)
                                                       (infix-parts x))) 
                                 (cdr A)))))

          ((eq '|| (car A))		;A = (|| z)
           (setq L (infix-parts (cadr A)))
           (setq L (append (list '\  '\|) (cdr L) (list '\|))) )

          ((member (car A) '(< <=))
           (cond ((atom (cadr A))
                  (setq L+ (list '\  (cadr A))) (setq L- (list '\  0)))

                 ((eq '- (caadr A))     ;A = (< (- a))  goes to  0 < a
                  (setq L+ (list '\  0))
                  (setq L- (infix-parts (cadadr A))))

                 ((eq '+ (caadr A))	           ;A = (< (+ a (- b) c))
                  (setq L  (copy-tree (cdadr A)))  ;L2=      (a (- b) c)
		  (setq L- (remove-if-not #'(lambda (x) 
				      (and (not (atom x)) (eq '- (car x)))) L))
		  (setq L+ (set-difference L L-))
		  (setq L- (mapcan #'cdr L-))
                  (setq L+ (cond (L+ (infix-parts (cons '+ L+))) 
                                 (t (list '\  0))))
                  (setq L- (cond (L- (infix-parts (cons '+ L-)))
                                 (t (list '\  0)))) )
                 (t (return-from infix-parts 
                        (infix-parts (list (car A) (list '+ (cadr A)))))))

	   (setq L (append L+ 
                           (cond ((< (length L+) 4) (list '\  ))
                                  (t (list '\  '\  )))
                           (list (car A)) 
                           (cond ((< (length L-) 4) nil)
                                  (t (list '\  )))
                           L-)) )

         ((eq '& (car A)) 
          (setq L (mapcan #'(lambda (x) 
                    (infix-parts x (append cr cr HEAD (list '& '\  ))))
                          (cdr A))) 
          (setq L (cddr L)) )

         ((eq '-> (car A))
          (setq L- (infix-parts (cadr A) HEAD))
          (setq L+ (infix-parts (caddr A)(append (spaces 5) (remove-&-cr HEAD))))
          (setq L2 (cond ((> (length HEAD) 15) (list '->))
                         ((> (length HEAD) 7) (list '--->))
                         (t                   (list '----->))))
          (setq L2 (append (spaces 2) L2))
          (setq L (append L- L2 cr cr L+)) )
 
         (t (setq L (list '\  A))) )

    (cond ((and (not (atom A)) (member (car A) '(& ->))) L)
          (t (append HEAD L))) ))

(defun spaces (n)  ;input for a format test-argument
   (cond ((zerop n) nil)
         (t (append (list '\  ) (spaces (1- n))))))

(setq cr (list '\
                         ))

(defun remove-&-cr (H)  ;called by infix-parts
  ;;This removes the &'s and CR's from H if there are any
  (remove '& (remove '\
 H)) )
