

;;;**********************************************************************
;;;
;;;               JOURNAL-READY.LISP    In kcl
;;;
;;;               W. W. Bledsoe      7 August 1989,  31 Oct 89
;;;                                                
;;;            Uses the IMPLY.LISP Interactive Prover  
;;;                   See ATP 88
;;;
;;;   A call is made to (prover) so the user can enter proof parts.
;;;
;;;     or to (prover script), where the users entries are in SCRIPT.
;;;
;;;***********************************************************************
;;;
;;;  Many of these functions are like similar ones in OPTIONS1,2,3.LISP
;;;***********************************************************************
;;;
;;;  PRINTING OPTIONS:  The *jr-light* is turned on in PROVER.
;;;  When the *jr-light* is "on" (non-nil), much of the ordinary printing 
;;;  used by IMPLY for interactive proofs, is suppressed. Furthermore, in
;;;  September 1989 version of this file, JOURNAL-READY.LISP, any user input
;;;  is preceeded by the symbols  |:, whereas the computer produced output
;;;  is printed with no such prefix.  (Note below)
;;;
;;;  31 Oct 89.  Now we want another option, where the user input is
;;;  is prefixed with nothing whatever, and the computer produced output
;;;  is given the prefix "- ".  Also the . printed by the user, tells the
;;;  program that not further input in forthcoming for the present goal,
;;;  and that it should go ahead and finish the proof.  When the computer
;;;  finished the proof (or any proof) it signals that by printing a simple
;;;  blank space (ie, advances the curser one space on the screen) to let
;;;  the user know that the previous required action has been completed by
;;;  the program and that further user input is now possible.
;;;
;;;  This new option is in effect if the parameter *-* is not nill.
;;;
;;;  A list of all user commands, and their affects, are found in       .
;;;************************************************************************

;(defvar *-* t)                     ;for 31Oct89 interactive proofs.
;(defvar *-c* t)                    ; "  for suppressing printing "claim etc"
;(defvar *iprn* t)          ;22Nov89 for using (PROVER script) to infix print

(defun prover (&optional script script-theta)   
  ;;Top level command. Used as part of the Journal-Ready Proofs.  
  (let (theorem)
    (cond (*-* (format t "~%  - Theorem to be Proved:"))
	  (t   (format t "~%Theorem to be Proved:")))
    (cond (*-* (format t "~%")) (t (format t "~%|: ")))
    (cond (script (setq theorem (car script)) 
		  (format t "~s~%" theorem)
		  (setq script (cdr script)))
	  (t (setq theorem (read))))
    (setq *jr-light* t)
    (setq *indent* 0)
    (setq *claim-number* 0)
    (setq *suppress-printing-c* nil)
    (setq *reduce-light* nil)
    (setq *in* 0) (setq *in-lt* nil)  ;For limiting calls to IMPLY-RULES
    (setq *Hn* nil) (setq *nh* 0) (setq *after-label* nil)
    (prove (eval theorem) script nil nil t)))

(defun assume-that (P &optional label)  ;15Nov89.  Called by Options
  ;;If P is assumed (and given a number nA) and later at number nC, C'
  ;; is deduced, and at nC+1  P->C' is claimed, then OK
  ;;See "Consequence" rule in IMPLY-RULES.
  (declare  (special h c tl anc lt yet))  
  (let (u)
    (cond ((ground P) nil)
	  (t (setq P (universalize-free-variables-in P))
	     (setq %reduce-light% t)
	     (setq P (skolemize P nil nil))))
    (cond (label nil)   ;if P does not have a label, give it one, *nh*
	  (t (setq label (setq *nh* (1+ *nh*)))))
    (setq u (list label P 'assume))
    (setq *Hn* (cons u *Hn*))      ;Put the labeled P in *Hn* 
    (setq H (and-on-h H P))        ;Put the UNlabeled P in H
    (setq *suppress-printing-c* t)))

(defun cases-jr (P)       ;called by options, when doing Journal Ready proofs
 (declare  (special h c tl anc lt yet))
 (let (Q h1 h2 theta lambda)
    (setq Q (list 'not P))

;;%%Put P and Q into h-gu, H= and Hv positions of H.
    (setq h1 (and-on-h H P))
    (setq h2 (and-on-h H Q))

;;%%Prove the current Goal for the two cases.
    (setq tl1 (append tl '(case-1)))
    (setq tl2 (append tl '(case-2)))
    (setq lt* (cond (%al% 'ao) (t 'prove)))
    (-%)(ind2)(--) (format t "Case-~s:  " 1)    ;later allow for case 1.2, etc
    (cond (*iprn* (print-infix P *indent*))
	  (t (format t "~s" P))) 

    (setq *indent* (+ *indent* 2)) (setq *suppress-printing-c* t)

    (cond ((setq theta (imply h1 c tl1 anc lt*))  ;dont use YET ??
	   (setq *indent* (- *indent* 2))

	   (-%)(ind2)(--) (format t "Case-~s:  " 1)
	   (cond (*iprn* (print-infix P *indent*))
		 (t (format t "~s" P)))
	   (-%)(ind2)(--) (format t "Finished for the Current Goal:~%")
	       (ind2)(--)(sp 2) 
	   (cond (*iprn* (print-infix C *indent*))
		 (t (format t "~s" C)))
	   (-%)(-%)

	   (ind2)(--) (format t "Case-~s:  "2)
	   (cond (*iprn* (print-infix Q *indent*))
		 (t (format t "~s" Q)))

	   (setq *indent* (+ *indent* 2)) (setq *suppress-printing-c* t)
	   (setq H2 (apply-substitution theta H2))
           (cond ((setq lambda (imply h2 c tl2 anc lt*))
		  (setq *indent* (- *indent* 2))

		  (-%)(ind2)(--) (format t "Case-~s:  " 2)
		  (cond (*iprn* (print-infix Q *indent*))
			(t (format t "~s" Q)))
		  (-%)(ind2)(--) (format t "Finished for Current Goal:~%")
		      (ind2)(--)(sp 2)
		  (cond (*iprn* (print-infix C *indent*))
			(t (format t "~s" C)))
		  (-%)(-%)
		  (prove-yet (compose theta lambda)))
                  
		 (t (format t "~% >>>>>>>>>>>>>>>> ~%  CASE-2 Failed~%")
		    nil)))
             
          (t (format t "~% >>>>>>>>>>>>>>>> ~%  CASE-1 Failed~%")
	     nil))))

(defun cases-h-jr ()     ;Called by options, when doing journal ready proofs
  ;;Split on a disjunction in the hypothesis.
  (declare  (special h c tl anc lt yet))
  (let (h-gu Hv W P Q H1 H2 tl1 tl2 theta lambda)
    (setq h-gu (cadr H)) (setq Hv (cadddr H))

    (cond ((eq-car h-gu 'or) (setq P (cadr h-gu)) (setq Q (last-part h-gu))
	   (setq h-gu nil))
	  ((eq-car Hv 'or) (setq P (cadr Hv)) (setq Q (caddr Hv))
	   (setq Hv nil))
	  ((and (eq '& (car h-gu)) (setq W (assoc 'or (cdr h-gu))))
	   (setq P (cadr W)) (setq Q (last-part W))
	   (setq h-gu (clear-& (remove W h-gu :test 'equal))))
	  ((and (eq '& (car Hv)) (setq W (assoc 'or (cdr Hv))))
	   (setq P (cadr W)) (setq Q (last-part W))
	   (setq Hv (clear-& (remove W Hv :test 'equal))))
	  (t (format t "~%H cannot be OR-Split~%") 
	     (return-from cases-h-jr nil)))
    (setq h (list (car H) h-gu (caddr H) Hv))

;;%%Put P and Q into h-gu and Hv positions of H.
    (setq H1 (and-on-h H P))  (setq H2 (and-on-h H Q))
    (setq tl1 (append tl '(H1))) (setq tl2 (append tl '(H2)))

;;%%Prove the current Goal for the two cases.
    (-%)(ind2)(--) (format t "Case-~s:  " 1)
    (cond (*iprn* (print-infix P *indent*))
	  (t (format t "~s" P)))  ;later allow for case 1.2, etc

    (setq *indent* (+ *indent* 2)) (setq *suppress-printing-c* t)

    (cond ((setq theta (imply H1 C tl1 NIL lt))
	   (setq *indent* (- *indent* 2))
	   (-%)(ind2)(--) (format t "Case-~s:  " 1)
	   (cond (*iprn* (print-infix P *indent*))
		 (t (format t "~s" P)))
	   (-%)(ind2)(--) (format t "Finished for the Current Goal:~%")
	       (ind2)(--)(sp 2) 
	   (cond (*iprn* (print-infix C (+ *indent* 6)))
		 (t (format t "~s" C)))
	   (-%)(-%)

	   (ind2)(--) (format t "Case-~s:  "2)
	   (cond (*iprn* (print-infix Q *indent*))
		 (t (format t "~s" Q)))
	   
	   (setq *indent* (+ *indent* 2)) (setq *suppress-printing-c* t)
	   (setq H2 (apply-substitution theta H2))
	   (cond ((setq lambda (imply H2 C tl2 NIL lt))
		  (setq *indent* (- *indent* 2))
		  (-%)(ind2)(--) (format t "Case-~s:  " 2)
		  (cond (*iprn* (print-infix Q *indent*))
			(t (format t "~s" Q)))
		  (-%)(ind2)(--) (format t "Finished for Current Goal:~%")
		      (ind2)(--)(sp 2)
		  (cond (*iprn* (print-infix C (+ *indent* 6)))
			(t (format t "~s" C)))
		  (-%)(-%)
		  (prove-yet (compose theta lambda)))

		 (t (format t "~%Second Part of Split-h Failed~%") nil)))
	  (t (format t "~%First Part of Split-h Failed~%") nil))))

(DEFUN CLAIM-jr (Lemma)    ;Called by OPTIONS, used in journal ready proofs
 (declare  (special h c tl anc lt yet))
 (let (LEMMA* lemma-q V X pl claim-plan theta plan* lt* u)
  ;; this is called from options 
  ;; the new goal - which is given in the claim statement - is proved and 
  ;; used to help prove the current goal being addressed by IMPLY.

  (cond (*-c* (ind2)(--)         ;Not usually invoked
       (format t "Claim-~s"(setq *claim-number* (1+ *claim-number*)))))

;;%% Prove the Claim-Lemma
  (setq lemma-q (cond ((setq V (free-variables-in Lemma))
                       (unique-bound-variables (universally-quantify lemma 
								     V)))
                      (t lemma)))
  (SETQ LEMMA* (SKOLEMIZE LEMMA-Q nil T))
  (cond ((eq-car Lemma* '&) (setq Lemma* (and-c-reorder-c Lemma*))))
; (SETQ %REDUCE-LIGHT% T)  
  (setq lt* 'prove)
  (setq *indent* (+ *indent* 2))
  (setq *suppress-printing-c* t)
  (COND ((SETQ X (IMPLY H LEMMA* (APPEND TL '(CLAIM)) ANC lt* nil))
	 (setq *indent* (- *indent* 2))
	 (cond (*-c* (ind2)(--)
		(format t "Claim-~s  Established~%"*claim-number*))
	       (t (-#)))
;;%% Now continue the proof using the Claim-Lemma.
         (SETQ LEMMA* (SKOLEMIZE LEMMA-Q nil NIL))
         (setq tl (append tl '(CLAIM)))
         (setq H (and-on-h H Lemma*))      ;Put UNlabeled Lemma in H
	 (cond (*after-label*              ;Put   Labeled Lemma in *Hn*
		(setq *Hn* (cons (list *after-label* Lemma*) *Hn*))
		(setq *after-label* nil))
	       (t (setq *Hn* (cons Lemma* *Hn*)))) ;Put UNlabeled Lem in *Hn*
	 (setq *suppress-printing-c* t)
	 X)
	(T 
         (format t "~% &&&-COULD-NOT-ESTABLISH-THE CLAIM-&&&&&: ")
         (cond ((< (symbol-length Lemma) 12) (format t "~s~%" Lemma))
               (t                            (format t "~%  ~s~%" Lemma)))
	 nil))))

(defun contradicts (refs)  ;16Nov89 called by OPTIONS 
  ;;Is the list of formulas represented by the list REFS contradictory?
  ;;These Labeled formulas are in *Hn* only. Also in H but not labeled there.
  ;;16Nov89.  For now we will assume that there are only two, negations.
  (let (A B)
    (cond ((setq A (assoc (car refs) *Hn* :test 'equal))
	   (setq A (cadr A)))
	  (t (format t "~%Incorrect References: Refs= ~s~%" refs) 
	     (return-from contradicts nil)))
    (cond ((setq B (assoc (cadr refs) *Hn* :test 'equal))
	   (setq B (cadr B)))
	  (t (format t "~%Incorrect References: Refs= ~s~%" refs)
	     (return-from contradicts nil)))
    (cond ((or (equal A (list 'not B)) (equal B (list 'not A))) (-#) t))))

(defun explode (a)
	(map 'list #'(lambda (char) (intern (string char)))
		 (symbol-name a)))
     
(defun fuze (L) (intern (apply #'concatenate 
			       (cons 'string (mapcar #'symbol-name L)))))

(defun ind () (sp *indent*) t)
(defun ind2 () (sp (+ *indent* 2)) t)

(DEFUN induction-jr (args) ;USED-IN OPTIONS for Journal Ready Proofs
  ;;Also called from PROVE in Script-prover.lisp

  ;;See INDUCTION, in file OPTIONS2.LISP,  for a general description of how 
  ;; Induction works.  This version of induction is called only when the 
  ;; *jr-light* is on. (Or from PROVER in script-prover.lisp)
  ;;There are two cases: 
  ;;  Ordinary Induction, where args = N  or  ('Ordinary N IH).
  ;;   If args = ('Ordinary N IH)it inducts on N using induction hypothesis IH.
  ;;   If args = N, it automatically finds the ind hyp, IH.
  ;;   If args = nil, it automatically finds both I and IH>
  ;;  Complete Induction, where args = ('Complete F).
  ;;   Then it does complete induction on the parameter F.
  ;;   (Eg, (Complete (Excess-Literal (S)))).

   (declare  (special h c tl anc lt yet))
   (let (N F CI W N* C* H* X Pr FY TX TQ TY IH* C0 C1 C2 Theta Theta Lambda)

     (cond ((not args) (setq N nil) (setq IH nil))           ;Ordinary Ind case
	   ((and (not (atom args)) (eq 'Ordinary (car args)));Ordinary Ind case
	    (setq N (cadr args)) (setq IH (caddr args)))
	   ((and (not (atom args)) (eq 'complete (car args)));Complete Ind case
	    (setq F (cadr args)) (setq CI t))
	   (t (setq N args) (setq IH nil)))                  ;Ordinary Ind case
	   

;;%% Prepare for Induction, Ordinary Induction Case
    (cond 
     ((not CI)		;The Ordinary induction case (not complete-induction)
      (cond (IH (setq N* N) (setq H* H))
            ((or 
              (and (setq W (prepare-for-induction (cadr H) C N))
		   ;;First try h-gu = (cadr H)
		   ;;(H -> C) is rewritten as ((element N (omega)) -> C*),
		   ;;                         W=(N C*)
		   (setq N* (car W)) (setq C* (cadr W)) 
                   (setq H* (subst nil (cadr H) H :test 'equal)))
              (and (setq W (prepare-for-induction (cadddr H) C N))
		   ;;and then try hv = (cadddr H)
		   ;;(H -> C) is rewritten as ((element N (omega)) -> C*),
		   ;;                         W=(N C*)
		   (setq N* (car W)) (setq C* (cadr W))
                   (setq H* (subst nil (cadddr H) H :test 'equal)))))
	    (t (format t  "~%For your N, ie, ~s , the Hypothesis: (element N (omega)) is not in H ~%" N)               
	       (format t "~%Do you want to try Again?
                      (Type  Y  or  N):  ")
	       (if (y-or-n-pw) (return-from induction (INDUCTION))
		 (progn (format t"~%Exit INDUCTION & Go Back to OPTIONS~%") 
			(return-from induction (OPTIONS))))))

      (setq N N*)

      (setq IH* (cond (IH IH) (t C*)))

      (setq IH* (cond ((setq V (free-variables-in ih*))
		       (existentially-quantify IH* V))
                      (t IH*)))

      (setq C0 nil)
      (SETQ C1 (SUBST= 0 N IH*))
      (SETQ C2 (LIST '-> (List '& (List 'element N '(omega)) IH*)
		     (SUBST= (LIST 'SCSR N) N IH*)))

      (setq W (list '& C0 C1 C2))
      (SETQ W (UNIQUE-BOUND-VARIABLES W))
      (setq C0 (cadr W)) (setq C1 (caddr W)) (setq C2 (cadddr W))

      (cond (V (setq C0 (skolem C0 Nil t))
	       (setq C1 (skolem C1 Nil t))
	       (setq C2 (skolem C2 Nil t)) )))

;;%% Prepare for Complete Induction Case.
     (CI		
      (setq H* H)
      (setq X (bound-variables-induction F))

      (setq pr (mapcar #'(lambda (z) (cons z (new-symbol (car z)))) X))
      (setq FY (sublis= pr F))
      (setq TX (list '-> (and-on (cadr H) (cadddr H)) C)) ;(h-gu & Hv -> C)
      (setq TQ (cond ((setq V (free-variables-in TX))
		      (existentially-quantify TX V))
		     (t TX)))
      (setq TY (sublis= pr TQ))
      (setq IH* (list '-> (list '< (list '+ FY (list '- F))) TY))
      (setq Y (mapcar #'cdr pr))
      (setq IH* (universally-quantify IH* Y))

      (setq C0 (list '<= (list '- F)))	;(element of (omega)) later?
      (setq C1 (list '-> (list '= F 0) C))
      (setq C2 (list '-> (list '& IH* (list '< (list '- F))) C))

      (setq W (list '& C0 C1 C2))
      (SETQ W (UNIQUE-BOUND-VARIABLES W))
      (setq C0 (cadr W)) (setq C1 (caddr W)) (setq C2 (cadddr W))

      (setq C2 (skolem C2 Nil t))))
                           
    (cond ((occur-in '(< > + *) C)		;arithemetic case
           (setq c2 (repl-scsr c2))
           (setq h  (repl-scsr h))))

;;Most of the Printing is suppressed in this version (see INDUCTION).

    (-%)(ind2)(--)(format t "Induction Hypothesis:~%")
    (ind2)(--)(cond (*iprn* (print-infix IH* (+ *indent* 4)) (-%)(-%))
		    (t (format t "~s~%~%" IH*)))

    ;;%% Prove the Cases: Induction-0, Induction-1, and Induction-2.

;   (SETQ %reduce-light% T)
    (COND 
     ((AND 
       (cond (CI (ind2)(--)(format t "Induction-0 Being Proved~%")
		 (ind2)(--)(sp 2)
		 (cond (*iprn* (print-infix C0 (+ *indent* 6)))
		       (t (format t "~s" C0)))
		 (setq *indent* (+ *indent* 2))
		 (setq *suppress-printing-c* t)
		 (cond ((SETQ THETA0 (IMPLY H* C0 (APPEND TL '(INDUCTION-0)) 
					    NIL LT* nil))
			(setq *indent* (- *indent* 2)) (-%)
			(ind2)(--)(format t "Induction-0 Has Been Proved~%")
			(ind2)(--)(sp 2)
			(cond (*iprn* (print-infix C0 (+ *indent* 6)) (-%)(-%))
			      (t (format t "~s~%~%" C0)))
			t)

		       (t (format t "~% $$$$$$$$$$-Induction-0 Failed-&&&~%") 
			  nil)))
	     (T (SETQ THETA0 T)))
       
       (prog2 
	(ind2)(--) (format t "Induction-1 Being Proved~%")
	(ind2)(--)(sp 2)
	(cond (*iprn* (print-infix C1 (+ *indent* 6)) t)
	      (t (format t "~s" C1) t))
	(setq *indent* (+ *indent* 2))
	(setq *suppress-printing-c* t))
       (cond ((SETQ THETA (IMPLY H* C1 (APPEND TL '(INDUCTION-1)) 
				 NIL LT* nil))
	      (setq *indent* (- *indent* 2)) (-%)
	      (ind2)(--)(format t "Induction-1 Has Been Proved~%")
	      (ind2)(--)(sp 2)
	      (cond (*iprn* (print-infix C1 (+ *indent* 6)))
		    (t (format t "~s" C1))) 
	      (-%)(-%)   t)
	     (t (format t "~% $$$$$$$$$$-Induction-1 Failed-&&&~%") nil))

       (prog2
	(ind2)(--) (format t "Induction-2 Being Proved~%")
	(ind2)(--)(sp 2)
	(cond (*iprn* (print-infix C2 (+ *indent* 6)))
	      (t (format t "   ~s" C2)))
	(setq *indent* (+ *indent* 2))
	(setq *suppress-printing-c* t))
       (cond ((SETQ LAMBDA (IMPLY H* C2 (APPEND TL '(INDUCTION-2)) 
				  NIL LT* nil))
	      (setq *indent* (- *indent* 2))
	      (ind2)(--)(format t "Induction-2 Has Been Proved~%")
	      (ind2)(--)(sp 2)
	      (cond (*iprn* (print-infix C2 (+ *indent* 6)))
		    (t (format t "   ~s~%~%" C2)))
	      t)

	     (t (format t "~% $$$$$$$$$$-Induction-2 Failed-&&&~%") nil)) )

      ;;Generalized Induction Case not in operation 7Aug89
      (COMPOSE THETA0 (COMPOSE THETA LAMBDA)))

     (t (format t "~% $$$$$$$$$$-Induction Failed-&&&~%") nil)) ))

(defun last-assummed-in (Hn) ;15Nov89 Called by IMPLY-RULES
 ;;In the list Hn, return the last entry of the form ( n A 'assume)
 (cond ((null Hn) nil)
       ((eq 'assume (caddar Hn)) (car Hn))
       (t (last-assummed-in (cdr Hn)))))

(defun match-part (lem c2) ;called by USE-jr
  ;;does the conclusion part of lem match a conclusion part of c2
 (let (lem* c2*)
  (cond ((unify lem c2) t)
	((eq-car lem '->) (match-part (caddr lem) c2))
	((eq-car c2  '->) (match-part lem (caddr c2)))
	((eq-car lem '&)  (some #'(lambda (u) (match-part u c2))  (cdr lem)))
	((eq-car c2  '&)  (some #'(lambda (u) (match-part lem u)) (cdr c2)))
	((eq-car lem 'or) (some #'(lambda (u) (match-part u c2))  (cdr lem)))
	((eq-car c2  'or) (some #'(lambda (u) (match-part lem u)) (cdr c2)))
	((eq-car lem 'not) 
	 (cond ((equal lem (setq lem* (push-nots-match lem))) nil)
	       (t (match-part lem* c2))))
        ((eq-car c2 'not) 
	 (cond ((equal c2 (setq c2* (push-nots-match c2))) nil)
	       (t (match-part lem c2*))))
	(t nil))))

(defun purge (HH A) ;called by IMPLY-RULES for the CONSEQUENCE rule
  ;;Remove the entry A from the four members of HH
  (mapcar #'(lambda (u) (remove A u)) HH))

(defun push-nots-match (A)  ;called by MATCH-PART
 (let (A*)
  (cond ((eq-car A 'not)
	 (cond ((eq-car (setq A* (cadr A)) '->)	 ;~(P->Q) goes to  P&~Q
		(push-nots-match (List '& (cadr A*) (list 'not (caddr A*)))))
	       ((eq-car A* '&)		;~(P&Q) goes to ~P or ~Q
		(push-nots-match (List 'or (list 'not (cadr A*))
				       (list 'not (caddr A*)))))
	       ((eq-car A* 'or)		;~(P or Q) goes to ~P & ~Q
		(push-nots-match (List '& (list 'not (cadr A*))
				       (list 'not (caddr A*)))))
	       ((eq-car A* 'not) (push-nots-match (cadr A*)))
	       (t A)))
	((eq-car A '->) (list '-> (push-nots-match (cadr A))
			          (push-nots-match (caddr A))))
	((eq-car A '&) 
	 (cons '& (mapcar #'push-nots-match (cdr A))))
	((eq-car A 'or) 
	 (cons 'or (mapcar #'push-nots-match (cdr A))))
	(t A))))
        
(defun sp (n)   ;used in Journal Ready Proofs.  sp means "spaces"
  (cond ((not (numberp n)) nil)
	((< n 1) nil)
	((< n 3) (format t "  "))
	((< n 5) (format t "    "))
	((< n 7) (format t "      "))
	((< n 9) (format t "        "))
	((< n 11)(format t "          "))
	((< n 13)(format t "            "))
	((< n 15)(format t "              "))
	))

(defun split-jr ()     ;Called by options, when doing journal-ready proofs
  (declare  (special h c tl anc lt yet))
  (let (c1 c2 tl1 tl2 theta lambda)
    (setq c1 (cadr c))           (setq c2 (last-part c))
    (setq tl1 (append tl '(C1))) (setq tl2 (append tl '(C2)))
    (cond ((setq theta (imply H c1 tl1 NIL lt))
	   (format t "~%First Part of Split-c Proved with MGU: ~s~%" theta)
	   (setq C2 (apply-substitution theta C2))
	   (cond ((setq lambda (imply H c2 tl2 NIL lt))
		  (format t "~%Second Part of Split-c Proved with MGU: ~s~%"
			  lambda)
		  (prove-yet (compose theta lambda)))
		 (t (format t "~%Second Part of Split-c Failed~%") nil)))

	  (t (format t "~%First Part of Split-c Failed~%") nil))))

(defun sublis-Q (theta A) ;Called by PROVER
 ;;This is like SUBLIS except that it does not replace a variable x
 ;; within the scope of a quantifier for which x is the quantified variable.
 (mapc #'(lambda (u) (setq A (subst-Q (cdr u) (car u) A))) theta) A)

(defun subst-Q (term x A) ;called by sublis-Q
  (cond ((eq x A) term)
	((atom A) A)
	((member (car A) '(all some let product sum)) ;quantifiers
	 (cond ((eq x (cadr A)) A) ;dont substitute 
	       (t (cons (car A) (cons (cadr A) (subst-Q term x (cddr A)))))))
	(t (cons (subst-Q term x (car A)) (subst-Q term x (cdr A))))))

(DEFUN USE-jr (Lemma &optional H-lt);used in OPTIONS, for Journal Ready proofs.
;;%%%%%%%%%%%% 7Sept89 This is a revision of the above. More like USE.LISP.
;;This version: Finds THETA and Applys it to the Lemma; Propositionally 
;; forward chains the ground units of H into it; and stores it in H;
;; and "proves" part or all of C by deletion.
  (declare  (special h c tl anc lt yet))
  (let (theta Lemma* GU USE h-gu h= hv c* c**)

;;%% *iprn* light is used for infix printing only, so no proving is necessary.
    (cond (*iprn* (setq *suppress-printing-c* t) (return-from use-jr t)))  

;;%%Recursively recall USE-jr if Lemma has the form (& L1 L2 etc)
    (cond ((eq-car Lemma '&) 
	   (mapc #'(lambda (lem) (use-jr lem h-lt)) (cdr Lemma))
 	  (return-from use-jr t)))

;;%% Read in and evaluate the Lemma
    (setq Lemma (eval Lemma))
    (setq tl (append tl '(U)))
    ;;(setq lemma (binary-to-unary< lemma))
    ;;must do this in the inequalify routines  2Jan90
    

;;%% Store it in h-gu if it is ground and return, else Slolemize it,
;;   find theta and apply it to the lemma.
    (cond ((ground Lemma) (setq H (and-on-h H lemma))
	   (and *-* (setq *suppress-printing-c* t))
	   (return-from Use-jr t)))

    (setq lemma (universalize-free-variables-in lemma))
    (setq %reduce-light% t)
    (setq lemma (skolemize lemma nil nil))



    (setq H-lt t)  ;;;%%%%%%%%%%%%%%% This is for 13Dec89%% and 2Jan90 only
                   ;;;%%%%%%%%%%%%%%% Remove it when finished.
    (cond (H-lt (setq H (and-on-h H lemma))   ;just store and return
		(setq *suppress-printing-c* t)(return-from Use-jr t)))

    (cond ((eq-car lemma '&) (setq H (and-on-h H lemma)) ;13Dec89
	   (setq *suppress-printing-c* t)(return-from Use-jr t)))

    (setq theta (find-theta Lemma))
    (and *v* (format t "~%It has found THETA= ~s ~%" theta)) ;if verbose,print
    (setq Lemma* (apply-substitution theta lemma))

;;%% Propositionally FC into the lemma, and store the results in h-gu, Hv,etc,
    (and (not (eq t theta)) 
         (setq GU (and-on (cadr H) (caddr H)))  ;(h-gu U H=)
         (setq GU (cond ((eq-car GU '&) (cdr GU))
                        (t (list GU))))
         (setq Lemma* (propositionally-forward-chain GU Lemma*)))

    (cond ((and (eq-car Lemma* '->) (in* 20 t)  ;control depth of IMPLY-RULES
		(setq theta (imply H (cadr Lemma*) tl nil 'qed nil)))
	   ;;If can prove the premise of the Lemma*, then store only the conc.
	   (in* *in-limit**)
	   (setq Lemma* (apply-substitution theta (caddr Lemma*)))))

    (cond ((eq-car Lemma* '&)
	   (setq Lemma* (cons '& (mapcar #'(lambda (u) 
                (cond ((and (eq-car u '->) (in* 20 t)
			    (setq theta (imply H (cadr u) tl nil 'qed nil)))
		       (in* *in-limit**)
		       (apply-substitution theta (caddr u)))
		      (t u))) (cdr Lemma*))))))

    (setq H (and-on-h H lemma* nil 'USE))
    (setq USE (car H)) (setq h-gu (cadr H)) (setq H= (caddr H))
                                            (setq Hv (cadddr H))

;;%% "Prove" part or all of C by Deletion.
    (cond ((setq GU (ground-units Lemma*))
	   (and *v* (format t "~%The following units have been added to H:~%")
		    (number&print GU)
		    (format t "~%(As a result of Forward-Chaining on the USE Lemma)~%")
		    (format t "~%The Hypotheses now contain:~%")
		    (number&print-list USE h-gu H= Hv))

           (cond ((null (setq c* (and-off c h-gu))) (setq c '(true)) 
                  (and *v* (format t "~%C is a subset of h-gu~%")))
                 ((null (setq c** (and-off c* Hv))) (setq c '(true))
                  (and *v* (format t "~%C is a subset of ~s~%" 
                          (cond ((equal c* c) 'Hv) (t '(h-gu & Hv))) )))
                 ((equal c** c) nil)
                 ((equal c* c) (setq c c**) 
                  (and *v* (format t "~% Part of C is a subset of Hv~%")))
                 ((equal c** c*) (setq c c**)
                  (and *v* (format t "~% Part of C is a subset of h-gu~%")))
                 (t (and *v* 
                     (format t "~% Part of C is a subset of (h-gu & Hv)~%")))))
	  (t (and *v* (format t "~%The USE Lemma has been put in the USE arg of IMPLY.~%"))))

    (setq *suppress-printing-c* t) t ))

(defun verify (C) ;called by options to CHECK-CASES in journal-ready proofs.
  ;;later we must fill in this code.  See my 8Dec89 notes.
  t)


(defun in* (n &optional lt) ;Called by USE-JR and FIND-THETA
   (setq *in-lt* lt) (setq *in* 0) (setq *in-limit* n))


;;;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
(setq th
   '(-> (& (NOT (ELEMENT [] (S))) (CLAUSES (S)) (GROUND (S))
	   (UNSATISFIABLE (S))
	   (ALL S3
		(-> (& (CLAUSES S3) (NOT (ELEMENT [] S3)))
		    (<= (- (EXCESS-LITERAL S3))))))
	(<= (- (EXCESS-LITERAL (S))))))

(setq th* 
  '(-> (& (not (p (a))) (q (a))
	  (all x (-> (& (not (p x)) (q x)) (r x))))
       (r (a))))

(defun -- () (or (cond (*-* (format t "- "))) t))
(defun -% () (or (cond (*-* (format t "~%"))) t))
(defun -. () (or (cond (*-* (format t " ."))) t))
(defun -# () (or (cond (*-* (format t " #"))) t))
       
;;;*******************************************************************
;;;
;;;      Some Scripts for Journal Ready Proofs
;;;
;;;*******************************************************************

;(setq cgr-jr-script cgs)
(setq s1 '('(-> A B) assume-the-hypothesis hh (use L14* t) hh a))
(setq s2 '(th assume-the-hypothesis split a a))
(setq s3 '(ex3 *))
(setq s3* '(ex3 a cr))
(setq s4 '('(a) a))
(setq s5 '(th assume-the-hypothesis (claim (clauses (s))) * 
                                    (claim (ground  (s))) * a))

(setq th6 '(-> (P (a)) (& (some x (P x)) (R (a)))))
(setq Lem6 '(-> (P x) (R x)))
(setq s6 '(th6 assume-the-hypothesis
               (claim (some x (P x))) *  
               (claim (R (a))) (use Lem6) * *))
(setq s6* '(th6 assume-the-hypothesis
               (some x (P x)) *  
               (R (a)) (use Lem6) * *))
(setq th7 '(-> (& (P(a)) (P(b))) (& (Q (b)) (P (a)))))
(setq Lm7 '(-> (P x) (Q x)))
(setq s7 '(th7 assume-the-hypothesis (Q (b)) (use Lm7) * *))

(setq th8 '(-> (& (R(a)) (P(b))) (& (Q (b)) (R (a)))))
(setq Lm8 '(-> (P x) (Q x)))
(setq Lm8* '(-> (R x)(Q x)))
(setq s8 '(th8 assume-the-hypothesis (Q (b)) (use Lm8*)(use Lm8) * *))
(setq s8* '(th8 assume-the-hypothesis (Q (b)) Lm8* Lm8 * *))
(setq s9  '(th8 assume-the-hypothesis (1)(Q (b)) Lm8* Lm8 * (2)(R (a)) *
            (& (R (a)) (Q (b))) [1 3 2] * *))
(setq s$ '(TH$ ASSUME-THE-HYPOTHESIS L13 L5 * *))

(setq cgs-6Nov89  ;Same as cgs-13, except that (claim x) is changed to x.
;;and (for now) "A" is used instead of (use L5) * for use L13,L5  and L14,L5
 `(Completeness-of-Ground-Resolution assume-the-hypotheses 
   (cases (element [] (S))) (use (& L17 L6)) * 
   (induction (complete (excess-literal (S)))) (use L7) *
   (use (& L17 L4)) * assume-the-hypotheses 
   (some CC (& (element CC (S)) (non-unit CC))) (use L8) *
   (some L (element L (CC))) (use L12) *  
   (let C1 (set-diff (CC) (sng (L)))) (let C2 (sng (L))) 
   (let S1 (union (set-diff (S) (sng (CC))) (sng (C1)))) 
   (let S2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
   (& (clauses (S1)) (ground (S1)) (unsatisfiable (S1))
	     (not (element [] (S1)))
	     (< (+ (excess-literal (S1))
		   (- (excess-literal (S))))))  
   (use L13 t) (use L5) *
   (& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
	     (not (element [] (S2)))
	     (< (+ (excess-literal (S2))
		   (- (excess-literal (S)))))) 
   (use L14 t) hh (e (setq *v* t)) (use L5) *    ;to "trace" Imply, etc
   (some N1 (& (element [] (res (S1) N1)) (p-integer N1))) *
   (some N2 (& (element [] (res (S2) N2)) (p-integer N2))) *

   (or (element [] (res (S) (N12)))
	      (element (sng (L)) (res (S) (N12))))   (use L1) *
   cases-h * 
   (subset (S)  (res (S) (N12)))  (use L6) *
   (subset (S2) (res (S) (N12)))  *
   (subset (res (S2) (N2)) (res (res (S) (N12)) (N2))) 
   (use L2 t) (use L6) *
   (element [] (res (S)(+ (N12) (N2)))) (use L3) *
   *))

(setq cgs-7Nov89      ;Save this one
  ;;Same as cgs-13, except that (claim x) is changed to x, and 
  ;; (use LM) AFTER A CLAIM, is changed to just LM , (& L1 L2) to (L1 L2), and
  ;;  in this case (USE LM), (USE (L1 L2)) is printed to the right by USE-jr
 `(Completeness-of-Ground-Resolution assume-the-hypotheses 
   (cases (element [] (S))) (use (& L17 L6)) * 
   (induction (complete (excess-literal (S)))) (use L7) *
   (use (& L17 L4)) * assume-the-hypotheses 
   (some CC (& (element CC (S)) (non-unit CC))) L8 *
   (some L (element L (CC))) L12 *  
   (let C1 (set-diff (CC) (sng (L)))) (let C2 (sng (L))) 
   (let S1 (union (set-diff (S) (sng (CC))) (sng (C1)))) 
   (let S2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
   (& (clauses (S1)) (ground (S1)) (unsatisfiable (S1))
	     (not (element [] (S1)))
	     (< (+ (excess-literal (S1))
		   (- (excess-literal (S))))))  
   L13 L5 *
   (& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
	     (not (element [] (S2)))
	     (< (+ (excess-literal (S2))
		   (- (excess-literal (S)))))) 
   L14 L5 * 
   (some N1 (& (element [] (res (S1) N1)) (p-integer N1))) *
   (some N2 (& (element [] (res (S2) N2)) (p-integer N2))) *

   (or (element [] (res (S) (N12)))
	      (element (sng (L)) (res (S) (N12))))  L1 *
   cases-h * 
   (subset (S)  (res (S) (N12)))  L6 *
   (subset (S2) (res (S) (N12)))  *
   (subset (res (S2) (N2)) (res (res (S) (N12)) (N2))) 
   L2 L6 *
   (element [] (res (S) (+ (N12) (N2)))) L3 *
   *))

(setq cgs-9Nov89     ;I presume that this is the final one 12Jan90
  ;;Same as cgs-7Nov89, except for adding Lemma L2c

 `(Completeness-of-Ground-Resolution assume-the-hypotheses 
   (cases (element [] (S))) (use (& L17 L6)) * 
   (induction (complete (excess-literal (S)))) (use L7) *
   (use (& L17 L4)) * assume-the-hypotheses 
   (some CC (& (element CC (S)) (non-unit CC))) L8 *
   (some L (element L (CC))) L12 *  
   (let C1 (set-diff (CC) (sng (L)))) (let C2 (sng (L))) 
   (let S1 (union (set-diff (S) (sng (CC))) (sng (C1)))) 
   (let S2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
   (& (clauses (S1)) (ground (S1)) (unsatisfiable (S1))
	     (not (element [] (S1)))
	     (< (+ (excess-literal (S1))
		   (- (excess-literal (S))))))  
   L13 L5 *
   (& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
	     (not (element [] (S2)))
	     (< (+ (excess-literal (S2))
		   (- (excess-literal (S)))))) 
   L14 L5 * 
   (some N1 (& (element [] (res (S1) N1)) (p-integer N1))) *
   (some N2 (& (element [] (res (S2) N2)) (p-integer N2))) *

   (or (element [] (res (S) (N12)))
	      (element (sng (L)) (res (S) (N12))))  L1 *
   cases-h * 
   (subset (S)  (res (S) (N12)))  L6 *
   (subset (S2) (res (S) (N12)))  *
   (subset (res (S2) (N2)) (res (res (S) (N12)) (N2))) L2 L2c L6 *
   (element [] (res (S) (+ (N12) (N2)))) L3 *
   *))

(setq cgs-22Nov89  
  ;;Same as cgs-9Nov89, except for putting aa for * so that
  ;; the printed output is the same but every IMPLY goal is
  ;; is assumed.  This for printing using PRINT-INFIX.

 `(Completeness-of-Ground-Resolution assume-the-hypotheses 
   (cases (element [] (S))) (use (& L17 L6)) aa 
   (induction (complete (excess-literal (S)))) (use L7) aa
   (use (& L17 L4)) aa assume-the-hypotheses 
   (some CC (& (element CC (S)) (non-unit CC))) L8 aa
   (some L (element L (CC))) L12 aa  
   (let C1 (set-diff (CC) (sng (L)))) (let C2 (sng (L))) 
   (let S1 (union (set-diff (S) (sng (CC))) (sng (C1)))) 
   (let S2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
   (& (clauses (S1)) (ground (S1)) (unsatisfiable (S1))
	     (not (element [] (S1)))
	     (< (+ (excess-literal (S1))
		   (- (excess-literal (S))))))  
   L13 L5 aa
   (& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
	     (not (element [] (S2)))
	     (< (+ (excess-literal (S2))
		   (- (excess-literal (S)))))) 
   L14 L5 aa 
   (some N1 (& (element [] (res (S1) N1)) (p-integer N1))) aa
   (some N2 (& (element [] (res (S2) N2)) (p-integer N2))) aa

   (or (element [] (res (S) (N12)))
	      (element (sng (L)) (res (S) (N12))))  L1 aa
   cases-h aa 
   (subset (S)  (res (S) (N12)))  L6 aa
   (subset (S2) (res (S) (N12)))  aa
   (subset (res (S2) (N2)) (res (res (S) (N12)) (N2))) L2 L2c L6 aa
   (element [] (res (S) (+ (N12) (N2)))) L3 aa
   aa))

(setq cgs-25Nov89  
  ;;Same as cgs-9Nov89, except for putting hh c a for aa so that
  ;; we get a listing of the subgoals send to the stand-alone prover
  ;;For the papers  ATP 88e  and  ATP 88b

 `(Completeness-of-Ground-Resolution assume-the-hypotheses 
   (cases (element [] (S))) (use (& L17 L6)) hh c a 
   (induction (complete (excess-literal (S)))) (use L7) hh c a
   (use (& L17 L4)) hh c a assume-the-hypotheses 
   (some CC (& (element CC (S)) (non-unit CC))) L8 hh c a
   (some L (element L (CC))) L12 hh c a  
   (let C1 (set-diff (CC) (sng (L)))) (let C2 (sng (L))) 
   (let S1 (union (set-diff (S) (sng (CC))) (sng (C1)))) 
   (let S2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
   (& (clauses (S1)) (ground (S1)) (unsatisfiable (S1))
	     (not (element [] (S1)))
	     (< (+ (excess-literal (S1))
		   (- (excess-literal (S))))))  
   L13 L5 hh c a
   (& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
	     (not (element [] (S2)))
	     (< (+ (excess-literal (S2))
		   (- (excess-literal (S)))))) 
   L14 L5 hh c a 
   (some N1 (& (element [] (res (S1) N1)) (p-integer N1))) hh c a
   (some N2 (& (element [] (res (S2) N2)) (p-integer N2))) hh c a

   (or (element [] (res (S) (N12)))
	      (element (sng (L)) (res (S) (N12))))  L1 hh c a
   cases-h hh c a 
   (subset (S)  (res (S) (N12)))  L6 hh c a
   (subset (S2) (res (S) (N12)))  hh c a
   (subset (res (S2) (N2)) (res (res (S) (N12)) (N2))) L2 L2c L6 hh c a
   (element [] (res (S) (+ (N12) (N2)))) L3 hh c a
   hh c a))

(setq cgs-9Nov89-Labels       
  ;;Same as cgs-7Nov89, except for adding Lemma L2c

 `(Completeness-of-Ground-Resolution assume-the-hypotheses 
   (cases (element [] (S))) (use (& L17 L6)) * 
   (induction (complete (excess-literal (S)))) (use L7) *
   (use (& L17 L4)) * assume-the-hypotheses 
   (some CC (& (element CC (S)) (non-unit CC))) L8 *
   (some L (element L (CC))) L12 *  
   (let C1 (set-diff (CC) (sng (L)))) (let C2 (sng (L))) 
   (let S1 (union (set-diff (S) (sng (CC))) (sng (C1)))) 
   (let S2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
   (& (clauses (S1)) (ground (S1)) (unsatisfiable (S1))
      (not (element [] (S1)))
      (< (+ (excess-literal (S1))(- (excess-literal (S)))))) L13 L5 *

   (& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
      (not (element [] (S2)))
      (< (+ (excess-literal (S2))(- (excess-literal (S)))))) L14 L5 * 

   (1) (some N1 (& (element [] (res (S1) N1)) (p-integer N1))) *
   (2) (some N2 (& (element [] (res (S2) N2)) (p-integer N2))) *

   (or (element [] (res (S) (N12)))
       (element (sng (L)) (res (S) (N12))))  L1 [1] *
   cases-h * 
   (3) (subset (S)  (res (S) (N12)))  L6 *
       (subset (S2) (res (S) (N12)))  *
   (4) (subset (res (S2) (N2)) (res (res (S) (N12)) (N2))) L2 L2c L6 [3] *
   (element [] (res (S) (+ (N12) (N2)))) L3 [2 4] *
   *))

(setq thap '(-> (& (PA) (-> (PA) (or(Q1) (Q2) (Q3)))) (R)))
(setq LA1 '(-> (Q1) (R)))
(setq LA2 '(-> (Q2) (R)))
(setq LA3 '(-> (Q3) (R)))
(setq sap 
 '(thap assume-the-hypothesis 
	(assume-that (1)(Q1)) (R) LA1 *   (-> (Q1) (R)) *
	(assume-that (2)(Q2)) (R) LA2 *   (-> (Q2) (R)) *
	(assume-that (3)(Q3)) (R) LA3 *   (-> (Q3) (R)) *
	*))

;;%% A simple example to test PROOF-BY-CONTRADICTION
(setq th-contra '(-> (& (P (a)) (all x (-> (P x) (Q x)))) (Q (a))))
(setq s-contra 
 '(th-contra assume-the-hypotheses proof-by-contradiction
	     (assume-that (1) (not (Q (a))))
	     (2) (Q (a)) *
	     (contradicts (1) (2)) *
	     *))

;;%% A Harder Example:  The Square Root of 2 is irrartional.
(setq th-sq2 '(not (rational (sqrt 2))))
(setq Lr0 '(integer 2))
(setq Lr1 '(all x (-> (rational x) 
		      (some m (some n (& (integer m) (integer n) 
					 (not (= n 0))
					 (= x (/ m n))))))))
(setq Lr2 '(-> (& (integer m) (integer n)) 
	       (some k (& (integer k) (divisors m n k)))))
(setq Lr3 '(-> (divisors N p d) (divisors (* N N) p (* 2 d))))
(setq Lr4 '(-> (divisors N p d) (divisors (* p N) p (+ d 1))))
(setq Lr5 '(-> (& (divisors N p d1)(divisors N p d2)) (= d1 d2)))
 ;;Lr5 is from the unique factorization theorem
(setq Lr6 '(-> (& (integer m) (integer n))  ;an even number is not odd
	       (not (= (* 2 m) (+ (* 2 n) 1)))))
(setq Lr7 '(-> (& (not (= r 0)) (= x (/ s r))) (= (* x r) s))) ;remove later
(setq Lr8 '(-> (= (* (sqrt x) y) z)  (= (* x y y) (* z z))))   ;  "     "
(setq s-sq2
 '(th-sq2 proof-by-contradiction 
	  (assume-that (1) (rational (sqrt 2)))
	  (2) (some m (some n (& (integer m) (integer n) (not (= n 0))
                                 (= (sqrt 2) (/ m n)))))  Lr1 Lr0 *
          (3) (= (* (sqrt 2) (n)) (m)) Lr7 *
	  (4) (= (* 2 (n) (n)) (* (m) (m))) Lr8 *
	  (5) (some k (divisors (n) 2 k))  Lr2 Lr0 *
	  (6) (some L (divisors (m) 2 L))  Lr2 Lr0 *
	  (7) (divisors (* (n) (n)) 2 (* 2 (k))) Lr3 *
	  (8) (divisors (* (m) (m)) 2 (* 2 (L))) Lr3 *
	  (9) (divisors (* 2 (n) (n)) 2 (+ (* 2 (k)) 1)) Lr4 *
	  (10)(divisors (* (m) (m)) 2 (+ (* 2 (k)) 1)) [4 9] *
	  (11) (= (* 2 (L)) (+ (* 2 (k)) 1)) [8 10] Lr5 *
	  (12) (not (= (* 2 (L)) (+ (* 2 (k)) 1)))  Lr6 *
	  (13) (contradicts (11) (12)) *
	  (14) (not (rational (sqrt 2))) [1 13] *
	  *))
	  