
;;;*********************************************************************
;;;
;;;                  PLANS.LISP    In KCL
;;;                            
;;;               W. W. Bledsoe      10 June 1988   24 June 88
;;;
;;;
;;;               For use by  IMPLY.LISP
;;;
;;;
;;;***********************************************************************
;;;
;;;    This contains some PLANS and procedures for using them,
;;;      in order to remotely supply the normally user-supplied
;;;      commands to IMPLY.LISP in proving a theorem interactively.
;;;
;;;    If p is such a plan, then (Prove TH p) causes the theorem TH
;;;      to be proved employing the commands from p (instead of the
;;;      corresponding user input).
;;;
;;;    cr stands for Carriage-return; crs stands for repeated Carriage returns
;;;    a  stands for "apply".  Thus (a cases-plan) means to apply cases-plan.
;;;
;;;*********************************************************************
;;;             PLANS

(setq Ex70-plan '(dc promote (a cases-plan) crs))

(setq cases-plan
     '((cases  ((element [] (S)) 
		((use L17) (USEq L6))               ;A Plan for proving Case-1
		((a Case-2-plan) crs) ))))          ;A Plan for proving Case-2

(setq Case-2-plan
      '((Induction ((complete (excess-literal (S)))
		    ((USEq L7) crs)	   ;A Plan for proving Induction-0
		    (promote (use L17) (USEq L4) crs) 
		                           ;A Plan for proving Induction-1
		    (promote (a claims-1) (a Let-variables) (a claims-2) 
                               (a finish) crs) ))))
                                           ;A Plan for proving Induction-2

(setq claims-1 '((claim ('(some cc (& (element cc (S)) (non-unit cc)))
                         ((USEq L8) hh)))  ;A Plan for proving the Claim
                 (claim ('(some L (element L (CC)))
	                 ((use L12) bc (use L11) bc (USEq L9) crs hh))) ))
                           ;;A Plan for proving the Claim
(setq let-variables
    '((let c1 (set-diff (cc) (sng (L))))
      (let c2 (sng (L)))
      (let s1 (union (set-diff (S) (sng (CC))) (sng (C1))))
      (let s2 (union (set-diff (S) (sng (CC))) (sng (C2)))) 
      hh))

(setq claims-2 '((claim ('(element (C1) (S1)) ((sub= (S1)) R* crs))) hh
                 (claim ('(element (C2) (S2)) ((sub= (S2)) R* crs))) hh
                 (claim ('(& (subset (C1) (CC)) 
                             (< (+ (length (C1))(- (length (CC))))))
			 ((USEq L13) crs))) hh
                 (claim ('(& (clauses (S1)) (ground (S1)) (unsatisfiable (S1)) 
		             (finite (S1)) (not (element [] (S1)))
		             (< (+ (excess-literal (S1))
                                   (- (excess-literal (S))))))
			 ((useq L5) crs))) hh 

                 (claim ('(& (subset (C2) (CC)) 
                             (< (+ (length (C2)) (- (length (CC))))))
			 ((USEq L14) crs))) hh
                 (claim ('(& (clauses (S2)) (ground (S2)) (unsatisfiable (S2))
			     (finite (S2)) (not (element [] (S2)))
			     (< (+ (excess-literal (S2))
				   (- (excess-literal (S))))))
			 ((useq L5) crs))) hh
                 (claim ('(some n1 (& (element [] (res (S1) n1)) 
                                      (p-integer n1)))
			 (qed crs)))
                 (claim ('(some n2 (& (element [] (res (S2) n2)) 
                                      (p-integer n2)))
			 (qed crs))) hh

                 (claim ('(or (element [] (res (S) (N1))) 
                              (element (sng (L)) (res (S) (N1))))
			 ((useq L1) (sub= (C1) 'H) crs)))
                 hh))

(setq finish '((split-h ((a split-h-plan-1)) ((a split-h-plan-2)))))

(setq split-h-plan-1 '(qed crs))

(setq split-h-plan-2 
     '((claim ('(subset (S2) (res (S) (N1)))
                ((sub= (S2)) dc r* promote 
                 (split-h ((useq L6) crs) 
                           (  crs)) crs)))
       (use-h L6)  hh  
       (claim ('(SUBSET (RES (S2) (N2)) (RES (RES (S) (N1)) (N2))) 
               ((useq L2) crs))) hh  
       (claim ('(element [] (res (S) (+ (N1) (N2))))
               ((useq L3) crs))) hh  
       (use-h L18) crs))

(setq %plan-list% '(ex70-plan cases-plan cases-2-plan claims-1 let-variables
                    claims-2 finish split-h-plan-1 split-h-plan-2) )
;;;********************************************************************
;;;         PROCEDURES
;;;********************************************************************

(defun read-h ()
  ;;This is ordinary Read in case %plan% is nil
  ;;Otherwise it "reads" the first element from %plan%
  (let (action)

   (cond ((null %plan%) (read))
         (t (setq action (car %plan%)) 
	    (format t "~s~%" 
		    (cond ((eq 'cr action) 'carriage-return)
			  ((eq 'bc action) 'Back-Chain)
			  (t action)))
            (cond ((member action '(promote back-chain bc))
                   (setq action 'carriage-return))
                  ((eq 'crs action) 
                   ;;This causes carriage-returns indefinitely
                   (return-from read-h 'cr)))
	    (setq %plan% (cdr %plan%)) 
	    action))))

(defun first-n (n L &optional entry) 
  ;;This rerurns the list of the first n (or less) members of L,
  ;; or the part before (including) the entry ENTRY, if it is not NIL
  (cond ((null L) L)
         ((< n 1) nil)
         ((and entry (equal (car L) entry)) (list (car L)))
         (t (cons (car L) (first-n (1- n) (cdr L) entry)))))


(defun next-n (m n L &optional entry1 entry2) 
  ;;This rerurns the list of members of L, starting with the mth
  ;; and ending with the nth members of L.  
  ;; or the part of L between entry1 and entry2 (inclusive).
  (cond ((null L) nil)
        ((and entry1 (equal (car L) entry1))
         (cons (car L) (next-n 0 (1- n) (cdr L) nil entry2)))
        ((< 1 m) (next-n (1- m) (1- n) (cdr L) entry1 entry2))       
	((< n 1) nil)
	((and entry2 (equal (car L) entry2)) (list (car L)))
	(t (cons (car L) (next-n 0 (1- n) (cdr L) entry1 entry2)))))


(defun y-or-n-pw (&optional s)
   (cond ((not %plan%) (y-or-n-p s))
         (s (format t "~% ~s " s) (eq 'y (read-h)))
         (t (eq 'y (read-h)))))

(defun clear-in ()            ; Used in options
  (do () ((null (read-char-no-hang)))))
