
;;;*********************************************************************
;;;
;;;                     REDUCE.LISP    In KCL
;;;                                 26 Aug 1987                        
;;;               W. W. Bledsoe     20 Feb 1987    3 Mar 87  21 Jan 88
;;;
;;;                 Called by  IMPLY.LISP
;;;
;;;
;;;***********************************************************************

(defun reduc (A &optional (parity t) arr)
  ;;For the sake of tracing I am putting in an intermidiate Reduce,
  ;;  this one, whose only job (almost) is to call Reduce2.

  ;;The *reduce-light* is ONLY set by the USER, when it is on absolutely NO
  ;; reducing is possible via REDUC.  A direct call to REDUC*  or REDUC**
  ;; circumvents the *reduce-light* and allows reductin.

;;  (COND ((not *reduce-light*) (return-from reduc A))
;;        ((and (occur-in '(< <=) A) (not arr))
;;         (setq A (apply-reduce-rule '(=--> (scsr n) (+ n 1)) A nil))))

  (reduce2 A parity ))

(defun reduc* (A &optional (parity t) arr )
  ;;Essentially Reduc, except that this version ALWAYS calls 
  ;; Reduce2 (whether or not the *reduce-light* is on)

  (cond ((and (occur-in '(< <=) A) (not arr))
         (setq A (apply-reduce-rule '(=--> (scsr n) (+ n 1)) A nil))))
  (reduce2 a parity ))

(defun reduc** (A &optional (parity t) )
  ;;Keeps calling reduc* until there is no change in A
  (cond ((equal A (setq A (reduc* A parity ))) A)
        (t (reduc** A parity ))))

(DEFUN REDUCE2 (A &optional (PARITY t) )
  ;Since REDUCE is a system function, our main reduce call is REDUC
 (and (not *jr-light*) (format t "."))

 (let (THETA A* A** NN x fg s V)
   (COND
    ((NULL A) NIL)			
    ((ATOM A) A)
    ((and (eq '= (car a)) (eval (list 'equal (list 'quote (cadr a))
				      (list 'quote (caddr a)))))
     t)
    ((NULL (CDR A)) (COND ((EQ '+ (CAR A)) 0) ;(+) goes to 0
			  ((EQ '* (CAR A)) 1)
			  ((EQ '& (CAR A)) '(TRUE))
			  ((EQ 'OR(CAR A)) '(FALSE))
			  (T A)))    
    ((AND (NULL (CDDR A)) (MEMBER (CAR A) '(& OR + *))) (CADR A))
					;eg  (+ a)  goes to a
    ((EQ 'NOT (CAR A))
     (COND ((ATOM (CADR A)) A)
	   ((EQ 'NOT (CAADR A)) (CADADR A)) ;(not (not B)) 
					;goes to B
	   ((eq '& (caadr a)) 
	    (reduce2 
	     (cons 'or (mapcar #'not-on (cdadr a))) parity ))
	   ;;(not (& B C D)) goes to (or (not B) (not C) (not D))
	   ((eq 'or (caadr a)) 
	    (reduce2 
	     (cons '& (mapcar #'not-on (cdadr a))) parity ))
	   ;;(not (or B C D)) goes to (& (not B) (not C) (not D))
           ((eq '< (caadr a)) (reduce2 (list '<= (list '- (cadadr a)))))
           ((eq '<= (caadr a)) (reduce2 (list '< (list '- (cadadr a)))))
	   ((eq-car (cadr A) 'all) 
	    (list 'some (cadadr A) (reduce2 (list `not (caddr (cadr A)))
					    (not parity))))
	   ((eq-car (cadr A) 'some) 
	    (list 'all (cadadr A) (reduce2 (list `not (caddr (cadr A)))
					   (not parity))))
	   ((eq-car (cadr A) `->)
	    (and-on (cadadr A) (not-on (caddr (cadr A)))))

	   (T (LIST 'NOT (REDUCE2 (CADR A) (NOT PARITY) )))))
 
   ((EQ '- (CAR A))
     (COND ((numberp (cadr a)) (- (cadr a)))
	   ((ATOM (CADR A)) A)
	   ((EQ '- (CAADR A)) (CADADR A)) ;(- (- B)) goes to B
	   ((eq '+ (caadr a)) 
	    (reduce2 
	     (cons '+ (mapcar #'neg-on (cdadr a))) parity ))
	    ;;(- (+ B C D)) goes to (+ (- B) (- C) (- D))

	   (T (LIST '- (REDUCE2 (CADR A)  PARITY )))))


    ((and (member (car A) '(< <=)) (cddr a)) (reduce2 (binary-to-unary< a)))
    ((and (eq '< (car A))  (numberp (cadr a)) (<  (cadr a) 0)) '(TRUE))
    ((and (eq '<= (car A)) (numberp (cadr a)) (<= (cadr a) 0)) '(TRUE))

    ((and (eq '< (car A))  (numberp (cadr a)) (<= 0 (cadr a))) '(FALSE))
    ((and (eq '<= (car A)) (numberp (cadr a)) (<  0 (cadr a))) '(FALSE))

    ((and (member (car A) '(< <=)) (listp (setq A* (cadr A)))
          (or (and (eq 'min (car A*)) (setq A** A*))
              (and (eq '+ (car A*)) 
		   (setq A** (find-if #'(lambda (x) 
				  (and (listp x) (eq 'min (car x)))) A*)))))
     ;; (a + min (b,c)) < 0  goes to  [a + b < 0  or  a + c < 0]
     (or-on (subst (cadr  A**) A** A :test 'equal)
	    (subst (caddr A**) A** A :test 'equal)))
    
    ((and (member (car A) '(< <=)) (listp (setq A* (cadr A)))
          (or (and (eq '- (car A*)) (listp (cadr A*)) (eq 'min (caadr A*))
                   (setq A** A*))
              (and (eq '+ (car A*)) 
		   (setq A** (find-if #'(lambda (x) 
			  (and (listp x) (eq '- (car x)) (listp (cadr x))
			       (eq 'min (caadr x)))) A*)))))
     ;; (a - min (b,c)) < 0  goes to  [a + -b < 0  &  a + -c < 0]
     (and-on (subst (list '- (cadr  (cadr A**))) A** A :test 'equal)
             (subst (list '- (caddr (cadr A**))) A** A :test 'equal)))

    ((member (car A) '(< <=)) 
     (cond ((equal A (setq A (list (car a) (reduce2 (cadr a) parity))))
            (reduce-algebraically A))
           (t A)))

    ((and (not (atom (car a))) (eq 'lambda (caar a))) ;Lambda reduction
     (subst= (cadr a) (cadar a) (caddar a)))
					;eg, ((lambda x (P x)) t) --> (P t)
    ((and (member (car a) '(lim cont cont-de cont-N diff diff-de diff-L))
	  (not (atom (cadr a)))
	  (member (caadr a) '(+ * /)))	;eg, (cadr a) = (+ f g)
     (setq x (new-symbol 'x ))
     (setq fg (cons (caadr a) (mapcar #'(lambda (f) (list f x)) (cdadr a))))
     (setq fg (list 'lambda x fg))
     ;;eg, replace (+ f g) with (lambda x (+ (f x) (g x)))
     (setq a* (subst= fg (cadr a) a)) )

     ;;eg, (Lim          (+ f g)           a (+ l k)) goes to 
     ;;    (Lim (lambda x (+ (f x) (g x))) a (+ l k))

;;%%

    ((SETQ A*
	   (SOME
	    #'(lambda(Q) (and
			  (setq THETA (instance A (car Q)))
			  (apply-substitution THETA (cadr Q)) ))
	    *reduce-table*))
     (reduce2 A* parity ))	  

    ((SETQ A*
	   (SOME
	    #'(lambda(Q) (and
			  (setq THETA (instance-ellipses A (car Q)))
					;if we are to use ellipses, make sure that 
					;the parts of A are reduced first.
			  (setq A** (cons (car a) (mapcar #'(lambda (u) 
							      (reduce2 u parity )) (cdr a))))
			  (setq THETA (instance-ellipses A** (car Q)))
			  (apply-ellipses THETA (cadr Q)) ))
	    *reduce-table-ellipses*))
     (reduce2 A* parity ))   

    ;; We probably must be concerned about NEW-BINDINGS later
    ;; See old program for some of what would be needed  3/2/87

    ((EQ 'ELEMENT (CAR A)) 
     (SETQ A* (CADDR A)) 
     (cond ((OR (eq-car A* 'set-of)          ; x E (set-of t (P t)) --> (P x)
		(AND (not (atom A*)) 
                     (member (car A*) '(union intersection 
                                       union-of intersection-of 
                                       union-of-sets intersection-of-sets
                                       subsets range))
                     (SETQ A* (DEFINE A* PARITY )) (eq-car A* 'set-of))
               )
	    ;; THIS IS REDUCE RULES 3 AND 4 
	    (SKOLEMize (SUBST= (CADR A) (CADR A*) (CADDR A*))
		       NIL PARITY ))
          ((eq-car A* 'intersection)         ; x E A/\B --> x E A  &  x E B
	   (and-on (list 'element (cadr A) (cadr A*))
		   (list 'element (cadr A) (cadr A*)) ))

          ((or (eq-car A* 'union)            ; x E A\/B --> x E A  or  x E B
#|
	       (and (not (atom A*)) (null (cdr A*)) ;x E A* & (= A* (union R))
	            (setq A** (some #'(lambda (u) 
			(and (not (atom u)) (eq '= (car u)) (equal A* (cadr u))
					     (eq 'union (caaddr u)) (caddr u)))
				    H))
	            (setq A* A**))      |# 
	       )
	   (or-on (list 'element (cadr A) (cadr A*))
		  (list 'element (cadr A) (caddr A*))) )
          ((eq-car A* 'sng)	             ; x E (sng A) --> x = A 
	   (cond ((equal (cadr A) (cadr A*)) '(true)) 
		 (t (list '= (cadr A) (cadr A*)) )) )
                
	   ((eq-car A* 'set-diff)	     ; x E A~B --> x E A  &  x ~E B
	    (and-on (list 'element (cadr A) (cadr A*))
		    (list 'not (list 'element (cadr A) (caddr A*))) ))
;           ((eq-car A* 'res)                ;disabled for now
;           (setq V (free-variables-in A))
;           (setq A** (define A* parity V))
;           (reduce2 (list 'element (cadr A) A**)) )

           (t (list 'element (cadr A) (reduce2 (caddr A)))) ))

     ((EQ '-> (CAR A))
      (COND ((EQUAL '(FALSE) (CADR A)) '(TRUE))
	    ((EQUAL '(TRUE) (CADDR A)) '(TRUE))
	    ((equal '(true) (cadr A)) (reduce2 (caddr A) parity ))
	    ((equal '(false) (caddr A))
	     (reduce2 (list 'not (cadr A)) (not parity) ))
	    (T (LIST '-> (REDUCE2 (CADR A) (NOT PARITY) ) 
		     (REDUCE2 (CADDR A) PARITY )))))
     ((eq 'union (car A))
      (cond ((eq '[] (cadr A)) (caddr A))
            ((eq '[] (caddr A)) (cadr A))
            (t A)))

     ((eq 'res (car A))
      (cond ((equal 0 (caddr A)) (cadr A))
            (t A)))

     ((and (eq '& (car a))
	   (progn (setq a* (crunch a))
		  (setq a* (remove 'T a*))   
		  (setq a* (remove '(true) a* :test 'equal)) 
		  (setq a* (remove '(empty) a* :test 'equal))
		  (setq a* (cond ((member 'nil (cdr a*)) nil)
				 ((member '(false) (cdr a*) :test 'equal) nil)
				 (t a*)))
		  t)
           (not (equal a* a)))
      (reduce2 a* parity ))

     ((and (eq 'or (car a))
	   (progn (setq a* (crunch a))
		  (setq a* (remove 'nil a*))   
		  (setq a* (remove '(false) a* :test 'equal))
		  (setq a* (remove '(empty) a* :test 'equal))
		  (setq a* (cond ((member 't (cdr a*)) t)
				 ((member '(true) (cdr a*) :test 'equal) t)
				 (t a*)))
		  t)
	   (not (equal a* a)))
      (reduce2 a* parity ))

     ((MEMBER (CAR A) '(& OR))
      (setq s (car a))
      (cond ((setq a* (assoc-r s (cdr a))) ;see replace.lisp for assoc-r
	     ;;(& a b (& c d) e) goes to (& a b c d e), etc.  a* = (& c d)
	     (setq a** (remove-one a* a)) ;a** = (& a b e)
	     (setq a** (cons s (append (cdr a*) (cdr a**))))
	     (reduce2 a** parity ))
	    (t (CONS s (MAPCAR #'(lambda(Q) (REDUCE2 Q PARITY )) (CDR A)))) ))

     ((EQ 'SUM (CAR A))
      (COND ((EQUAL (SETQ NN (CADDDR A)) 0) 0)
	    ((ATOM NN) A)
	    ((AND (EQ '+ (CAR NN)) (EQUAL 1 (CADDR NN)))
	     (LIST '+
		   (SUBST= (CADR NN) NN A)
		   (SUBST= NN (CADR A) (CADDR (CDDR A))) ))
	    (T A)))

     ((AND (EQ '+ (CAR A)) (NOT (ATOM (CADR A)))(EQ '+ (CAADR A)))
      (REDUCE2 (CONS '+ (append (CDADR A) (CDDR A))) parity ))
     ((and (member (car a) '(+ * - /)) (every #'numberp (cdr a)))  (eval a))
     ((and (eq '+ (car A)) 
	   (not (equal (cdr a) (setq a* (reduce-alg (cdr a))))))
      (cond ((null a*) 0)
	    ((numberp a*) a*)
	    (t (reduce2 (cons '+ a*) parity))))

     ((and (eq '* (car A)) 
	   (not (equal (list A) (setq A* (reduce-alg (list A))))))
      (cond ((null a*) 0)
	    ((numberp a*) a*)
	    (t (reduce2 (cons '+ a*) parity))))

     ((MEMBER (CAR A) '(+ * - = SUM < <= || lim))
      (COND ((EQUAL A 
		    (SETQ A* 
			  (CONS (CAR A) 
				(MAPCAR #'(lambda (x) (reduce2 x parity )) (CDR A)))))
	     A)
	    (T (REDUCE2 A* parity ))))
     ((eq-car A 'e) 
      (cond ((and (not (atom (caddr A))) (member (caaddr A) '(res res-L))
		  (equal (caddr (caddr A)) 0))
	     (reduce2 (subst (cadr (caddr A)) (caddr A) A)))
	    (t A)))
     (T A)


)))

(defun reduce-el-split (A) ;Called by Eligible-split
 (let (a*)
  (cond 
   ((atom A) A)
   ((eq '& (car A))
    (cond ((null (cdr a)) nil)
          ((null (cddr a)) (cadr a))
          ((not (equal a (setq a* (remove 't a)))) (reduce-el-split a*))
          ((member nil a) nil)
          ((not (equal a (setq a* (mapcar #'reduce-el-split a))))
           (reduce-el-split a*))
          (t a)))
   ((eq 'or (car a))
    (cond ((null (cdr a)) t)
	  ((null (cddr a)) (cadr a))
	  ((not (equal a (setq a* (remove nil a))))(reduce-el-split a*))
	  ((member 't a) t)
          ((not (equal a (setq a* (mapcar #'reduce-el-split a))))
           (reduce-el-split a*))
	  (t a)))
   ((eq '-> (car a))
    (cond ((eq 't  (cadr a))  (reduce-el-split (caddr a)))
          ((eq nil (cadr a))  t)
          ((eq 't  (caddr a)) t)
          ((eq nil (caddr a)) (reduce-el-split (list 'not (cadr a))))
          ((not (equal a (setq a* (list '-> (reduce-el-split (cadr a))
                                            (reduce-el-split (caddr a))))))
           (reduce-el-split a*))
          (t a)))
   ((eq 'not (car a))                       
    (cond ((eq nil (cadr a)) t)
          ((eq 't  (cadr a)) nil)
          ((atom   (cadr a)) a)
          ((not (equal a (setq a* (list 'not (reduce-el-split (cadr a))))))
           (reduce-el-split a*))
          ((eq '<  (caadr a)) (list '<= (minus-on-reduce 0 (cadadr a))))
          ((eq '<= (caadr a)) (list '<  (minus-on-reduce 0 (cadadr a))))
          (t a)))
   (t a) )))


(DEFUN APPLY-ELLIPSES (THETA A)    ;USED-IN (A INSTANCE-ELLIPSES REDUCE)
  ;; This is for applying a substitution that has elliptical entries in it. 
  (declare (special THETA))
  (APPLY-SUBSTITUTION THETA (APPL-E A)))

(DEFUN APPL-E (A)		;CALLED BY APPLY-ELLIPSES (REDUCE)
  ;; This function - as a part of APPLY-ELLIPSES - takes care of applying the
  ;;  elliptical part of THETA to A 
  (declare (special theta))           ;Theta changes 
  (Let  (R* TEMP TEMP*)

   (COND ((NULL A) NIL)
	 ((ATOM A) A)
	 ((EQ (CADR A) 'REPEAT)
	  (COND ((SOME #'(LAMBDA (X)
				 (AND (OCCUR X (CAR A))
				      (CDR (ASSOC (TRIPLE X) THETA :test #'equal))))
		       '(= % -))
		 (SETQ R* (APPL-E (CAR A)))
		 (CONS R* (APPL-E A)))
		(T NIL)))
	 ((NULL THETA) A)
	 ((EQ 'T THETA) A)
	 ((and (member a  '((===) (%%%)) :test 'equal) 
	       (equal (list a) (sublis= theta a)))  nil)	;??  it is OK
	 ((MEMBER (CAR A) '(--- === %%% /// ))
	  (APPEND (SUBLIS= THETA (CAR A)) (APPL-E (CDR A))))
	 ((MEMBER (CAR A) '(= % -))
  	  (SETQ TEMP (ASSOC (TRIPLE (CAR A)) THETA :test #'equal))
	  (COND ((NULL (CDR TEMP)) NIL)
		(T (SETQ TEMP* (SUBST (CDDR TEMP) (CDR TEMP) TEMP :test 'equal))
		   (SETQ THETA (SUBST= TEMP* TEMP THETA))
		   (CONS (CADR TEMP) (APPL-E (CDR A))))))
	 (T 
	  (CONS (SUBLIS= THETA (CAR A)) (APPL-E (CDR A)))))))

(DEFUN ELLIPSE-PART (SYM X A A-ELLIPSE)  ;used by INSTANCE-ELLIPSES
  ;;This binds to the symbol SYM, the list of leading elements of A,
  ;;  which come before the element of A which matches X.
  ;;Thus if SYM = ---, X = (+ ---), and A = (a b (+ d e) c), then
  ;;  it binds --- to (a b) and returns that sub and the rest of A.
  ;;See the trace following INSTANCE-ELLIPSES.
  (Let (THETA U)
       (COND ((NULL A) NIL)
	     ((NULL SYM) NIL)
	     ((SETQ THETA (INSTANCE-ELLIPSES (CAR A) X))
	      (SETQ U (CONS SYM (REVERSE A-ELLIPSE)))
	      (SETQ THETA (COMPOSE (LIST U) THETA))
	      (LIST THETA (CDR A)))
	     (T (ELLIPSE-PART SYM X (CDR A) (CONS (CAR A) A-ELLIPSE))
		;Keep moving members from the head of A to the head of 
                ; A-ELLIPSE until (car A) matches x.
		))))

(defun ground-reduce (A R)
  ;;This Reduces the formula A by the Rules (pairs) in R
  (let (A*)
       (setq A* A) 
       (mapc #'(lambda (rl) (setq A* (subst (cadr rl) 
                                            (car rl) A* :test 'equal)))
             R)
       (cond ((equal A* A) A)
             (t (ground-reduce A* R)))))

(defun ground-CSOR (RULES &optional flag)
  ;;This generates the complete set of reductions from the set RULES.
  (let (RS)
    (cond ((and (Equal RULES (setq RS (G-CSOR RULES))) flag) RULES)
	  (t (ground-CSOR RS t)))))
         

(defun G-CSOR (RS)      ;called by Ground-CSOR
  (let (RS* R)
    (setq RS* RS)
    (mapc #'(lambda (rl) 		    
              (setq R (ground-reduce rl (cdr RS*)))
              (cond ((< (length (car R)) (length (cadr R))) 
                 (setq R (reverse R))))
              (setq RS* (Append (Cdr RS*) (list R))) )
          RS)
    RS*))

(DEFUN INSTANCE (A B &optional L)	;USED-IN (DEFINE REDUC)
  ;; IF A IS AN INSTANCE OF B UNDER SUBSTITUTION THETA THEN RETURN THETA
  ;; A AND B ARE FORMULAS OR TERMS
  (Let (THETA LAMBDA B* L*)
       (COND ((ATOM B)
	      (COND ((EQ B A) T)
		    ((NUMBERP B) NIL)
                    ((member B L) nil)
;;                  ((occur B A) NIL)          ;no problem with instance?
		    (T (LIST (CONS B A)))))
	     ((ATOM A) NIL)
	     ((NOT (EQ (CAR B) (CAR A)))
              (cond ((or (atom (car a)) (atom (car b))) nil)
;;                  ((not (equal (caar a) (caar b))) nil)
                    (t (instance (cons 'unique-symbol a) 
                                 (cons 'unique-symbol b)))))
                    ;;eg, ((+ (f) (g)) c)   ((+  f* g* ) x)
                    
	     ((NULL (CDR B)) (COND ((NULL (CDR A)) T)))
	     ((NULL (CDR A)) NIL)
	     ((SETQ THETA (INSTANCE (CADR A) (CADR B) L))
              (cond ((eq t theta) 
                     (instance (cons (car A) (cddr A)) 
                               (cons (car B) (cddr B)) L))
                    ((and (setq B* (apply-substitution theta B))
                          (setq L* (union-w L 
                           (mapcan #'(lambda (u) (cond ((atom (cdr u)) 
                                                        (list (cdr u)))))
                               theta))) 
                          nil))
                    ((setq lambda (instance A B* L*))
                     (compose theta lambda)))))))

 
(DEFUN INSTANCE-ELLIPSES (A B)  ;Used-in (ELLIPSE-PART I REDUC)
   ;; Called by REDUC
   ;; This is like INSTANCE except that B can be more complicated -- like the
   ;;   left side of a REDUCE rule.
   ;; --- is like the "three dots" in an elliptical expr, (similarly === %%%)
   ;; For example, if A = (* a b (+ d e) c) and B = (* --- (+ ===) %%%) then
   ;; (INSTANCE-ELLIPSES a b) binds the lists (A B), (D E) and (C), to 
   ;;    ---, ===, and %%%, respectively.  That is it returns the substitution 
   ;;    ((--- A B) (=== D E) (%%% C)).  See the trace below.

  (Let (W A-prime THETA SIGMA)
       (COND ((EQUAL A B) T)
	     ((ATOM B)
	      (COND ((NUMBERP B) NIL)
		    ((MEMBER B *CONSTANT-LIST*) NIL)
                    ((occur B A) NIL)
		    (T (LIST (CONS B A)))))          ;ordinary binding
	     ((AND (NULL A) (member b '((===) (%%%)) :test 'equal))
	      (list (list (car b) (car b)))) ;??
	     ((ATOM A) NIL)
	     ((EQUAL (CAR A) (CAR B))
	      (INSTANCE-ELLIPSES (CDR A) (CDR B)))
	     ((MEMBER (CAR B) '(--- === %%% /// ))
	      (COND ((NULL (CDR B)) (LIST (CONS (CAR B) A)))
		    ((SETQ W (ELLIPSE-PART (CAR B) (CADR B) A nil))
		     (SETQ THETA (CAR W))
		     (SETQ A-prime (CADR W))
		     (COND ((SETQ SIGMA
				  (INSTANCE-ELLIPSES A-prime 
						     (CDDR B)))
			    (COMPOSE THETA SIGMA))))))
             ((atom (car b)) nil)
	     ((SETQ THETA (INSTANCE-ELLIPSES (CAR A) (CAR B)))
	      (COND ((SETQ SIGMA
			   (INSTANCE-ELLIPSES
			    (APPLY-ELLIPSES THETA (CDR A))
			    (APPLY-ELLIPSES THETA (CDR B))))
		     (COMPOSE THETA SIGMA)))))))
 
#|  Tracing INSTANCE-ELLIPSES-
>a
(* A B (+ D E) C)
>b
(* --- (+ ===) %%%)

>(instance-ellipses a b)
  1> (INSTANCE-ELLIPSES (* A B (+ D E) C) (* --- (+ ===) %%%))
    2> (INSTANCE-ELLIPSES (A B (+ D E) C) (--- (+ ===) %%%))
      3> (ELLIPSE-PART --- (+ ===) (A B (+ D E) C) NIL)
        4> (INSTANCE-ELLIPSES A (+ ===))
        <4 (INSTANCE-ELLIPSES NIL)
        4> (ELLIPSE-PART --- (+ ===) (B (+ D E) C) (A))
          5> (INSTANCE-ELLIPSES B (+ ===))
          <5 (INSTANCE-ELLIPSES NIL)
          5> (ELLIPSE-PART --- (+ ===) ((+ D E) C) (B A))
            6> (INSTANCE-ELLIPSES (+ D E) (+ ===))
              7> (INSTANCE-ELLIPSES (D E) (===))
              <7 (INSTANCE-ELLIPSES ((=== D E)))
            <6 (INSTANCE-ELLIPSES ((=== D E)))
          <5 (ELLIPSE-PART (((--- A B) (=== D E)) (C)))
        <4 (ELLIPSE-PART (((--- A B) (=== D E)) (C)))
      <3 (ELLIPSE-PART (((--- A B) (=== D E)) (C)))
      3> (INSTANCE-ELLIPSES (C) (%%%))
      <3 (INSTANCE-ELLIPSES ((%%% C)))
    <2 (INSTANCE-ELLIPSES ((--- A B) (=== D E) (%%% C)))
  <1 (INSTANCE-ELLIPSES ((--- A B) (=== D E) (%%% C)))
((--- A B) (=== D E) (%%% C))
|#
 
(DEFUN NEG-ON (A) (REDUC (LIST '- A)))

(defun Numbers-negs-first (x y)		;Used by reduce-alg, in SORT
  ;;Put numbers (in order) first, then negations
  (cond ((numberp x) (cond ((numberp y) (<= x y)) (T T)))
        ((numberp y) nil)
        ((atom x) (cond ((atom y) T) ((eq '- (car y)) nil) (t t)))
        ((eq '- (car x)) 
	 (cond ((atom y) t) 
	       ((eq '- (car y)) (numbers-negs-first (cadr x) (cadr y)))
	       (t t)))
        ((atom y) t)
        ((eq '- (car y)) nil)
        ((eq '+ (car x)) t)
        ((eq '+ (car y)) nil)
        ((eq '* (car x)) t)
        ((eq '* (car y)) nil)
        (t t) ))

(defun numbers-in (a) 
 (let (W)
  (cond ((null a) (list nil nil))
        ((atom a) (cond ((numberp a) (list (list a) nil))
                        (t (list nil (list a)))))
        ((atom (car a)) (cond ((numberp (car a))
                               (setq w (numbers-in (cdr a)))
                               (list (cons (car a) (car w)) (cadr w)))
                              )))))

(DEFUN OCCUR (X Y)   ;USED-IN (APPL-E CALLED-IN CHOOSE NOT-FORMULA  REDUC)
  (COND ((EQUAL Y X))
	((ATOM Y) NIL)
	((OCCUR X (CAR Y)))
	((OCCUR X (CDR Y)))))
 
(defun prove= (A)			;Called from options
  ;;This tries to prove A (an equality theorem) by a stronger 
  ;;set of reduction rules than are used in REDUC
  ;;eg,A is (= (+ (* (N) (N)) (N) (* 2 (N)) 2) (+ (* (N) (N)) (N) (N) (N) 1 1))
  (format t "~%        +++++++++Try to Use  PROVE= to prove++++++++++ ~%~% ~s ~%~%" A)
  (ans* 'TRY--PROVE=)
  (let (A* A**)
    (cond ((atom A) (options))
	  ((not (eq '= (car a))) (options))
	  ((and (numberp (setq A* (minus-on-reduce (cadr A) (caddr A))))
                (zerop A*))
           (format t "~% Proved by REDUCE ~%~%")	   
 	   (ans* 'Proved-by-Reduce)  T)
          ((not (eq '+ (car A*))) (options))
	  ((null (setq A** (reduce-alg (cdr A*))))      ;Algebraically Reduce
           (format t "~% Proved by REDUCE-ALG ~%~%")	   
	   (ans* 'Proved-by-reduce-alg)  T)
          ((not (equal A** (cdr A*)))
           (prove= (list '= (cons '+ A**) 0)))
	  (t (options)) )))

(defun reduce-algebraically  (A)			;Called from options
  ;;This tries to reduce A (an inequality formuls) by a stronger 
  ;;set of reduction rules than are used in REDUC
  ;;eg, (< (+ (- (* (X1) (X1))) -1 (* (X1) (X1)))) is reduced to 
  ;;    (< -1)  (which means ( -1 < 0 ), which is true.)

  (let (A* A**)
    (cond ((atom A) A)
	  ((not (member (car a) '(< <=))) A)
          ((atom (cadr a)) A)
          ((not (eq '+ (car (setq A* (cadr A))))) A)

	  ((null (setq A** (reduce-alg (cdr A*))))
;	   (format t "~% ~s  REDUCED ALGEBRAICALLY to ~s ~%" A* 0)
	   (ans* (list A* 'Reduced-Algebraically-to 0))
	   (list (car A) 0))

          ((not (equal A** (cdr A*)))
           (setq A** (cond ((cdr A**) (cons '+ A**)) (T (car A**))))
;	   (format t "~% ~s  REDUCED ALGEBRAICALLY to ~% ~s ~%" A* A**)
	   (ans* (list A* 'Reduced-Algebraically-to A**))
           (reduce-algebraically (list (car A) A**)) )
	  (t A) )))

(defun reduce-alg (L)    ; Called by Prove=
  ;; eg, L = ((* (N) (N)) (N) (* 2 (N)) 2 (- (* (N) (N))) 
  ;;          (- (N)) (- (N)) (- (N)) -1 -1)
  ;; L should be thought as the argument list of an + expression,
  ;;   eg, for (+ a b c),  L = (a b c)

  (and (not *jr-light*) (format t "+"))  ;Print a + each time this is accessed

  (let (L* H A L**)

    (setq L* (sort (copy-tree L) 'Numbers-negs-first))  
    ;;Put numbers (in order) first, then negations
    ;; eg, L* = (-1 -1 2 (- (* (N) (N))) (- (N)) (- (N)) (- (N)) 
    ;;                   (* (N) (N)) (N) (* 2 (N)))

    (cond ((atom L*) L*)
          ((numberp (setq H (car L*)))
           (cond ((zerop H) (reduce-alg (cdr L*))) ;remove 0's
                 ((null (cdr L*)) L*)
                 ((numberp (cadr L*))	;L* = (2 3 c ...)                 
                  (reduce-alg (cons (+ H (cadr L*)) (cddr L*))))
                 (t (cons H (reduce-alg (cdr L*))))))
          
          ((atom H) (cons H (reduce-alg (cdr L*))))

          ((eq '- (car H))		;L* = ((- a) ....), H = (- a)
	   (cond ((numberp (cadr H))	;L* = ((- 3) ....), H = (- 3)
                  (reduce-alg (cons (- (cadr H)) (cdr L*)))) ;to (-3 ...)
                 ((member (setq A (cadr H)) (cdr L*) :test 'equal) 
					;L* = ((- a) ... a ...)
		  (reduce-alg (remove-one A (cdr L*))))
                 ((and (not (atom (cadr H))) (eq '+ (caadr H)))
                     ;;L* ((- (+ a b)) c d) goes to ((- a)(- b) c d)
                  (reduce-alg (append (mapcar #'(lambda (x) (list '- x)) 
                                              (cdadr H))
                                      (cdr L*))))
		 (t (cons H (reduce-alg (cdr L*)))) ))
	  ((eq '* (caar L*))		;L* = ((* a b ...) ...)
	   (setq A (sort (copy-tree (cdar L*)) 'Numbers-negs-first)) 
					;A = (a b ...)
	   (cond ((equal 0 (car A))	;L* = ((* 0 b ...) c ...), A=(0 c ...)
		  (reduce-alg (cdr L*)))
		 ((member (car A) '(1 2 3 4 5 6) :test 'equal)
					;L* = ((* 2 b ...) ...)  A=(2 b ...)
		  (setq L** (multiply-out A))
		  ;;L** = ((* b ...)(* b ...) ...)
		  (reduce-alg (append L** (cdr L*))))
                 ((and (not (atom (car a))) (eq '* (caar A)))
		  ;;A = ((* c d ...) e f ...)
                  (setq A (append (cdar A) (cdr A))) ; A = (c d ... e f ...)
		  (reduce-alg (cons (cons '* A) (cdr L*))))
		 (t (cons (car L*) (reduce-alg (cdr L*)))) ))
	  ((eq '+ (caar L*))		;L* = ((+ a b) c d), to (a b c d)
	   (reduce-alg (append (cdar L*) (cdr L*))))
	  (t (cons (car L*) (reduce-alg (cdr L*)))) )))
 
(defun multiply-out (A)  ;A has the form (2 b ...) or (3 b ...) etc
  ;;(2 b ...) goes to ((b ...) (b ...)) etc.
  (cond ((atom A) A)
        ((not (numberp (car A))) A)
        ((zerop (car A)) 0)            ;This should not happen
        ((equal 1 (car A)) (cond ((cddr A) (list (cons '* (cdr A))))
                                 ((cdr  A) (cdr A))  ;ie, (list (cadr A))
                                 (t A)))   ;This should not happen
        (t (cons (cond ((cddr a) (cons '* (cdr a)))
		       ((cdr  a) (cadr a))
		       (t a))		   ;This should not happen
                 (multiply-out (cons (1- (car a)) (cdr a))))) ))


(defun remove-one (x L)   ;remove the first x from L 
  (cond ((equal x L) nil)
        ((atom L) L)
        ((equal x (car L)) (cdr L))
        (t (cons (car L) (remove-one x (cdr L)))) ))

(SETQ *REDUCE-TABLE*    ;USED-IN (ADD-REDUCE REDUCE)
           '(((<= (- (* x x))) (true))
             ((< (* x x)) (false))
             ((ELEMENT X8 (SCSR X8))   (TRUE)) 
             ((ELEMENT T7 (SCSR X7))   (OR (ELEMENT T7 X7) (= T7 X7)))
             ((ELEMENT X-ANYTHING 0)   (FALSE))
             ((= (scsr X-ANYTHING) 0)   (FALSE))
             ((ELEMENT T7 (INTERSECTION A7 B7)) 
              (& (ELEMENT T7 A7) (ELEMENT T7 B7))) 
             ((ELEMENT T1 (UNION A7 B7)) 
              (OR (ELEMENT T1 A7) (ELEMENT T1 B7))) 
             ((ELEMENT T2 (SUBSETS A7))   (SUBSET T2 A7))
             ((SUBSET T3 (INTERSECTION A7 B7)) 
              (& (SUBSET T3 A7) (SUBSET T3 B7))) 
             ((INTERSECTION A7 A7)   A7) 
             ((UNION A7 A7)   A7) 
             ((INTERSECTION A7 (EMPTY-SET))   (EMPTY-SET)) 
             ((UNION A7 (EMPTY-SET))   A7) 
             ((SUBSET (EMPTY-SET) A7)    (TRUE)) 
             ((ELEMENT A7 (SINGLETON B7))   (= A7 B7)) 
             ((ELEMENT (CHOICE A7) A7)   (NOT (= A7 (EMPTY-SET))))
             ((REFINEMENT G7 G7)   (TRUE)) 
             ((REFINEMENT G7 (CLOSURE-F G7))   (TRUE)) 
             ((SUBSET A7 A7) (TRUE)) 
             ((SUBSET A7 (CLOSURE A7))   (TRUE)) 
             ((ELEMENT 0 (OMEGA)) (TRUE))
            ;; LOGICAL Reductions        
             ((OR A7 (NOT A7))   (TRUE)) 
             ((& A7 (NOT A7))   (FALSE)) 
             ((& (TRUE) A7)   A7) 
             ((& A7 (TRUE))   A7) 
              ((OR A7 (TRUE))   (TRUE)) 
             ((OR (TRUE) A7)   (TRUE)) 
             ((& A7 (FALSE))   (FALSE)) 
             ((& (FALSE) A7)   (FALSE)) 
             ((OR A7 (FALSE))   A7) 
             ((OR (FALSE) A7)   A7)
        ))
 
(SETQ *REDUCE-TABLE-ELLIPSES*    ;USED-IN (ADD-REDUCE REDUCE)
                  '(((& --- (TRUE) ===) (& --- ===)) 
                    ((OR --- (FALSE) ===)  (OR --- ===)) 
		    ((* --- 1 ===) (* --- ===)) 
                    ((+ --- 0 ===) (+ --- ===)) 
                    ((& --- (FALSE) ===) (FALSE)) 
                    ((OR --- (TRUE) ===) (TRUE))
                    ((* --- 0 ===) 0) 
                    ((* --- (+ ===) %%%) (+ (* --- = %%%) REPEAT))
                    ((- --- (+ ===) %%%) (+ (- =) REPEAT))
                    ((+ --- (+ ===) %%%) (+ --- === %%%))
         ))

(defun rlt () (print *reduce-light*))

(DEFUN TRIPLE (X)                  ;USED-IN (APPL-E)
  ;; USED BY THE ellipses FUNCTIONS LIKE INSTANCE-ELLIPSE 
  (COND ((EQ '- X) '---) ((EQ '= X) '===) ((EQ '% X) '%%%)))



(defun Test-Reduce-timing (e n) ;time to reduce e n times
 (time (prog () 
        top
         (if (< n 1) (progn (print 'finished) (return t))
                     (progn (reduc e t) (setq n (- n 1))
                             (go top)) )
)))

(defvar thm '(= (+ (* (N) (N)) (N) (* 2 (N)) 2) (+ (* (N) (N)) (N) (N) (N) 1 1)))

(defvar thm* '(= (+ (* 3 (N)) 2) (+ (N) (N) (N) 1 1)))

(defvar am '((* (N) (N)) (N) (* 2 (N)) 2 (- (* (N) (N))) 
(- (N)) (- (N)) (- (N)) -1 -1))

