
;;;*********************************************************************
;;;
;;;                  REPLACE.LISP    In KCL
;;;                         
;;;               W. W. Bledsoe     4 Dec 1987            6 April 1988
;;;
;;;             This is used in  IMPLY.LISP
;;;        
;;;            (This is used in my Analogy Effort)
;;;***********************************************************************
;;;The main function here is REPLACE-A which is used by REPLACE-TERM,
;;;  which is called in response to the option call REP in OPTIONS.
;;;  It is also called by the options APPLY-REDUCE, APPLY-LEMMA, 
;;;  APPLY-HYPOTHESIS via the function APPLY-RULE-OR-LEMMA.
;;;
;;;The function  REPLACE-A  is used to replace the term (or formula) e in c by 
;;; back-chaining on it, using either UNIFY or CHAIN or by another action 
;;; designated by AC. (The default case is UNIFY, but also the CHAIN
;;; case is detected automatically, as is the reduce-rule case,which uses =-->.
;;;
;;;It proceeds somewhat like IMPLY-RULES until it reaches e.
;;;Then it continues as in IMPLY-RULES to back-chain using UNIFY or CHAIN
;;; until e is replaced by e*, so that c becomes c*. 
;;;Once e is replaced in c by another subformula, resulting in c*,
;;; (using the mgu THETA), then the triple (c* THETA PATH) is returned, unless
;;; PATH is a member of reject-list, RJ.  PATH is the path from the top of H
;;;  down to the subformula of H which is matched with e.
;;;If c is unchanged then (c nil) is returned, the NIL indicating failure.
;;;For example, if h = (<= (+ (f (a)) (- (c))))),        Ex62
;;;                c = (<  (+ (f  x ) (- (d))))), and
;;;                e = (f x)),  
;;;  then  (replace-a h c e)  returns the value ((< (+ (c) (- (d))))  ((x a))),
;;;  where         c*= (< (+ (c) (- (d))))  and THETA =  ((x a)).
;;;
;;;Actually, REPLACE-A allows the parameter e to be a pattern for the term
;;;  that is being replaced.  For example, if we wish to replace the term
;;;  (P (f (a) (g (c))) ( h y)), we could use e = (P (f x y) z), or 
;;;  (P x y) if this uniquely defines the term that is to be replaced in c.
;;;
;;;The user can accept or reject a replacement for e, by the use of the
;;;  function EXAMINE.  See how EXAMINE is called by the options rule REP.
;;;**********************************************************************

(defun actions (h e action reject-list)    ;called by replac  NOT USED??
  ;;carry out the designated action (ACTION) of REPLACE
  ;;ACTION can take one of the values (UNIFY CHAIN DEFINE)
  (let (theta)
    (cond 
     ((eq 'unify action)
      (cond ((and (setq theta (unify h e))
		  (not (member (list t theta) reject-list)))
             (list t theta))
            (t (list e nil)) ))
     ((eq 'chain action) (chain h e reject-list))
     ;;DEFINE later
     )))
 
(defun apply-Forward-Lemma (L A pattern parity)    ;called by APPLY-LEMMA
  ;;apply the forward lemma L to A whenever A matches the pattern pattern.
  (let (theta)
    (cond ((and parity
                (or (not pattern) 
		    (instance A pattern)
		    (and (eq 'occur (car pattern))
			 (every #'(lambda (x) (occur x A)) (cdr pattern))))
                (setq theta (instance A (caddr L))))
           (apply-substitution theta (cadr L)))
          ((atom A) A)
          ((eq 'not (car A)) 
           (list 'not (apply-Forward-Lemma L (cadr A) pattern (not parity))))
          ((eq '-> (car A))
           (list '-> (apply-Forward-lemma L (cadr A) pattern (not parity))
                     (apply-Forward-lemma L (caddr A) pattern parity)))
          ((member (car A) '(& or < <=))
           (cons (car A) (mapcar #'(lambda (u) 
                 (apply-Forward-Lemma L u pattern parity)) (cdr A))))
          (t A) )))
 
(defun apply-Lemma ()
  ;; 7/7/88 If this is to be used it must be corrected to account for the fact 
  ;;  that H is now of the form  (use h-gu H= Hv), instead of a single 
  ;;  conjunction.  So we must call this (or apply-Lemma*) with each of these
  ;;  h's.

  (declare  (special h c tl anc lt yet))
  (let (L pattern h* c* B L* L**)

     (format t "        ===== APPLY a FORWARD LEMMA ===== ~%")
     (format t "~%Type in the FORWARD LEMMA that you wish to apply -    ")
     (setq L (read-h))
     (format t "~%The Forward Lemma you have chosen is ~%~% ~s   ~s ~%" L (eval L))
     (format t "~%Do you wish to give a pattern for applying the Lemma?
                  If so type the pattern, if not type the letter n ~%")
     (setq pattern (read-h))  
      ;;The pattern will be used to screen the formulas to which the lemma
      ;; is being applied.  Also it might an atom or list of atoms that
      ;; must occur in the formula to which the lemma is applied.
     (setq pattern (cond ((or (eq 'n pattern)(null pattern)) nil) 
                         ((atom pattern) (list 'occur pattern))
			 (t   pattern)))
     (ans* (list 'The 'Forward-Lemma L  'is 'being 'applied))
     (setq %goal-message% (list 'Apply-Forward-Lemma  L)) 
     (setq L (eval L))
     (setq B (new-bindings L))
     (setq L* (apply-substitution B L))
     (setq h* (apply-Forward-Lemma L* h pattern nil))
     (setq c* (apply-Forward-Lemma L* c pattern t))

     (cond ((and (equal c c*) (equal h h*)
                 (or (equal L (setq L** (binary-to-unary< L)))
		     (and (setq B (new-bindings L**))
			  (setq L* (apply-substitution B L**))
			  (setq h* (apply-Forward-Lemma L* h pattern nil))
			  (setq c* (apply-Forward-Lemma L* c pattern t))
			  (equal c c*) (equal h h*))))
            (print '(Forward-Lemma did not change H or C))
           (ans*  '(Forward-Lemma did not change H or C))
            (options))
           (t (setq c c*) (setq h h*) (options)))))
;;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(defun apply-reduce ()       ;;something is wrong here. 7/7/88
  ;;I believe that it is OK now.  3/27/89  Except that H has the
  ;;  the form (use H-gu Hv H=).

  (declare  (special h c tl anc lt yet))
  (let (R pattern h* c* B R*)

     (format t "        ===== APPLY a REDUCE RULE ===== ~%")
     (format t "~%Type in the RULE that you wish to apply -    ")
     (setq R (read-h))
     (format t "~%The Reduce Rule you have chosen is ~%~% ~s   ~s ~%" R (eval R))
     (format t "~%Do you wish to give a pattern for applying the rule?
                  If so type the pattern, if not type the letter n ~%")
     (setq pattern (read-h))
     (setq pattern (cond ((eq 'n pattern) nil) (t pattern)))
     (ans* (list 'The 'Reduce-rule R  'is 'being 'applied))
     (setq %goal-message% (list 'Reduce-rule R)) 
     (setq R (eval R))
     (setq B (new-bindings R))
     (setq R* (apply-substitution B R))

     (setq h* (apply-reduce-rule R* h pattern))
     (setq c* (apply-reduce-rule R* c pattern))

     (cond ((and (equal c c*) (equal h h*)) 
            (print '(Reduce-Rule did not change H or C))
            (ans*  '(Reduce-Rule did not change H or C))
            (options))
           (t (setq c c*) (setq h h*) (options)))))



;;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(defun apply-reduce-rule (R A pattern)    ;called by APPLY-REDUCE
  ;;apply the reduce rule R to A whenever A matches the pattern pattern.
  (let (theta)
    (cond ((and (or (not pattern) (instance A pattern))
                (setq theta (instance A (cadr R))))
           (apply-substitution theta (caddr R)))
          ((atom A) A)
          (t (cons (apply-reduce-rule R (car A) pattern)
                   (apply-reduce-rule R (cdr A) pattern))) )))

(defun apply-rule-or-lemma (action h c) ;called by OPTIONS 
  (if (eq 'original-goal %goal-message%) (write-goal-message) nil)
  (let (h* c* n w)
   (cond 
    ((member action '(apply-r apply-reduce)) ;Apply a Reduce rule
     (format t "        ===== APPLY a REDUCE RULE ===== ~%")
     (format t "~%Type in the RULE that you wish to apply -    ")
     (setq h* (read-h))
     (ans* (list 'The 'Reduce-rule h*  'is 'being 'applied))
     (setq %goal-message% (list 'Reduce-rule h*)) 
     (cond ((and (setq c* (match-in (cadr h*) c 'instance))
		 (cadr (setq w (replace-a h* c c*))))
	    (ans* (list 'apply-rule-rule 'succeded 'for h* 'with 'result w)))
	   (t (ans* (list 'apply-rule-rule 'failed 'for h*)))))

    ((member action '(apply-l apply-lemma))  ;apply a Lemma
     (format t "        ===== APPLY a LEMMA ===== ~%")
     (format t "~%Type in the LEMMA that you wish to apply -    ")
     (setq h* (read-h))
     (ans* (list 'The 'Lemma  h*  'is 'being 'applied))
    (setq %goal-message% (list 'Apply-lemma  h*)) 
     (cond ((and (setq c* (match-in (caddr h*) c 'instance))
		 (cadr (setq w (replace-a h* c c*))))
	    (ans* (list 'apply-lemma 'succeded 'for h* 'with 'result w)))
	   (t (ans* (list 'apply-lemma 'failed 'for h*)))))
     
    ((member action '(apply-h apply-hypothesis))  ;apply a hypothesis of h
     (format t "        ===== APPLY a HYPOTHESIS ===== ~%")
     (format t "The current Hypotheses of H are - ~%~% ~s ~%"
         (number&print h))
     (format t "~%Type in the HYPOTHESIS that you wish to apply -    ")
     (setq n (read-h))
     (setq h* (nth n h))
     (ans* (list 'Hypothesis 'Number n   h*  'is 'being 'applied))
     (setq %goal-message% (list 'Apply-hypothesis-number n  h*)) 
     (cond ((and (setq c* (match-in (caddr h*) c 'unify))
		 (cadr (setq w (replace-a h* c c*))))
	    (ans* (list 'apply-Hypothesis n h* 'succeded 'with 'result w)))
	   (t (ans* (list 'apply-Hypothesis n h* 'failed))))) ) w))

(DEFUN APPLY-substitution (THETA D)	;widely used
       ;; USED-IN 
       ;;    (AND-C APPLY APPLY-ELLIPSES BACK-CHAIN-O CONFLICT DEFINE INSTANCE OR-H
       ;;         PROVE-THE-PREMISE PROVE-THE-PREMISE* REDUC UNIFY-LISTS)

       ;; the substitution THETA is applied to the form D
       (COND ((EQ T THETA) D)
	     ((TWO-XS THETA NIL)
	      (PROG (X2)
		    (SETQ X2 (TWO-XS THETA))
 		    (RETURN (reduc (AND-ON (APPLY-SUBSTITUTION (CAR X2) D)
					    (APPLY-SUBSTITUTION (CADR X2) D))))))
	     (T (SUBLIS= THETA D))))

(defun assoc-r (a L)     ;Called by REPLACE-TERM, REDUCE2
  ;;Return the first member of L of the form (a ...) if there is one, else NIL
  ;;This is different from ASSOC because L may have atomic members
  (cond ((atom L) nil) 
        ((atom (car L)) (assoc-r a (cdr L)))
        ((equal a (caar L)) (car L))
        (t (assoc-r a (cdr L))) ))


(defun examine (ans sym)   ;called by OPTIONS 
  ;;This lets the user approve of an action.
  ;;If the user approves of the ans = (c* Theta), from the previous action,
  ;; then try to finish the proof by calling IMPLY, with c* for c, and compose
  ;; theta with the result.

 (declare  (special h c tl anc lt yet))
   (let (c* theta theta* tl* yett)
   (cond (ans 
	  (setq c* (car ans)) (setq theta (cadr ans))
          (cond (theta (format t "~% C is replaced by ~2% ~s ~2% in response to the command- ~s , ~% with a resulting Substitution of-  ~s ~2%" c* sym theta))
                (t (format t "~% The user command- ~s resulted in failure, i.e., (C NIL) was returned. ~2% So It returned to options. ~2%" sym)
                   (print (list 'In-Examine 'h h 'c c))
                   (return-from examine (options))))

	  (if (y-or-n-p "Do you accept the result?  Type-   ")
	      (progn (setq c c*) 
		     (setq yett 
			   (cond ((eq t theta) yet)
				 (t (mapcar #'(lambda (u) 
                                      (cons (compose (car u) theta) (cdr u)))
					    yet)))) 
                       ;;compose THETA with existing subs of yet.
                     (setq tl* (append tl (list sym)))
		     (cond ((setq theta* (imply h c tl* anc lt yett))
			    (compose theta theta*))
                           (t (print '(PROOF OF c* FAILED)) 
                              (setq c c*) (options))))
	      (progn (print 'REJECTED-BY-USER) (ans* 'REJECTED-BY-USER)
                   (options)) ))
         (t (print 'FAILED) (options)) )))

(defun flip-unit (L e)			;called by replac
  ;;this finds the first member A of L which contains e and returns
  ;;the pair (A L*) where L* is the remainder of L  

  (cond ((atom L) (cond ((eq L e) (L nil)) (t nil)))
	((member (car L) '(& or))
	 (some #'(lambda (A) (and (occur e A) (list A (remove2 A L)))) L))
        ((occur e L) (list L nil))
        (t nil)))

(defun insert (D* e-path theta &optional (parity t))    ;called by repla-a
  ;;E-path is a (reversed) path through C to e (or c* if e is inside an
  ;; inequality term (eg, (< a b))).  For example, if 
  ;; C = (-> H (& A B (-> H2 G))), and e = G, then e-path is 
  ;; ( G  (-> H2 G)  (& A B (-> H2 G))  (-> H (& A B (-> H2 G))) )
  ;;Insert replaces e (or c*) by D* in C by working back through the path.
  ;;Theta is applied when appropriate, depending on the predicate symbol of
  ;; D, and the parity.
  ;;Also T-REDUCE is applied to combine t and (FALSE) appropriately.
  (let (D E E* parity* theta*)
    (cond ((null (cdr e-path)) D*)
	  (t (setq D (car e-path))
	     (setq E (cadr e-path))
	     (setq E* (t-reduce (subst D* D E :test 'equal)))
	     (cond ((insert-parity E parity)
		    (setq theta* (restrict-theta theta E D))
		    (setq E* (apply-substitution theta* E*))))
	     (setq parity* (insert-carry-parity E D parity))
	     (insert E* (cdr e-path) theta parity*) ))))

(defun insert-parity (E parity)            ;called by INSERT
  (cond ((eq '& (car E)) parity)
	((member (car E) '(or -> not)) (not parity))
	(t parity)))

(defun insert-carry-parity (E D parity)    ;called by INSERT
  (cond ((atom E) nil)    ;alarm
	((member (car E) '(& or)) parity)
	((eq 'not (car E)) (not parity))
	((eq '-> (car E))
	 (cond ((occur D (cadr E)) (not parity))
	       (t parity)))))
  
(defun L-LE (sign1 sign2)      ;called by reduc
 (cond ((and (eq sign1 '<=) (eq sign2 '<)) '<) (t '<=)))

(defun match-in (A B unify-or-instance &optional term-reject-list occur-list)
  ;;find a match of A in B and return it if it is not in the term-reject-list,
  ;; provided that the atoms in OCCUR-LIST occur in it.

  (cond ((and (every #'(lambda (x) (occur x B)) occur-list)
	      (or (and (eq 'unify unify-or-instance) (unify a b))
		  (and (eq 'instance unify-or-instance) (instance b a)))
	      (not (member b term-reject-list :test 'equal)))
         b)      
        ((atom b) nil)
        ((match-in a (car b) unify-or-instance term-reject-list occur-list))
        ((match-in a (cdr b) unify-or-instance term-reject-list occur-list)) ))
  
(defun remove2 (x L)    ;this removes the first occurance of x from L
                        ;and condenses L if it has form (& a) etc
 (setq l (remove1 x L))
 (cond ((cddr L) L)
       ((cdr L) (cadr L))
       (t 'alarm)))

(defun remove-shields (h c) ;called by OPTIONS -- requires user input
  (if (eq 'original-goal %goal-message%) (write-goal-message) nil)
  (let (x)
   (format t " ===== REMOVE THE SHIELDING TERMS OF A VARIABLE of C  ===== ~%") 
   (format t "~%Type in the VARIABLE x for which you wish the shielding terms removed-    ")
   (setq x (read-h))
   (setq %goal-message% (list 'Remove-Shields-of x))
   (remove-shs h c x)))          ;remove each member of shields from c
  
(defun remove-shs (h c x)        ;called by remove-shields
  ;;It returns a list (C** theta*) if successful, else NIL,
  ;;where C** is the result of removing the shields of x in C

  (let (shield theta theta* c* c**)
    (cond ((setq shield (a-shielding-term x c))	;a shielding term of x in c
	   (cond ((and (setq theta (replace-a h c shield)) (cadr theta))
		  (setq c* (car theta)) (setq theta (cadr theta))
		  (cond ((setq theta* (remove-shs h c* x))
                         (setq c** (car theta*)) (setq theta* (cadr theta*))
                         (setq theta* (compose theta theta*))
                         (list c** theta*))
                        (t nil) ))        ;return NIL if failure
                 (t nil) ))
          (t (list c t)) ))) ;if there are no more shielding terms return (c t)
  
(defun remove-variable (h c) ;called by OPTIONS -- requires user input
  ;;This returns a list (C** theta) if successful, else NIL, 
  ;;where C** is the result of removing x from C, with substitution theta

  (if (eq 'original-goal %goal-message%) (write-goal-message) nil)

  (let (x theta c* c**)
   (format t " ======= REMOVE A VARIABLE x from C  ======== ~%") 
   (format t " ==(Using VE - after Shielding-Term removal)== ~%") 

   (format t "~%Type in the VARIABLE x which you wished removed-    ")
   (setq x (read-h))
   (setq %goal-message% (list 'Remove-the-variable x))
   (cond ((setq theta (remove-shs h c x))  ;remove the shielding terms of x
          (setq c* (car theta)) (setq theta (cadr theta))
          (setq c** (VE x c*))
          (list c** theta)) )))
  
 (defun replace-a (h c e &optional (par t) action rj)   
  ;;Used by Replace-TERM(from options), APPLY-RULE-OR-LEMMA, REMOVE-SHIELD
  ;;See explanation at the head of this file (replace.lisp)
  ;;rj is a reject-list of PATHs, from h down to the term in h matched with e 
  ;;par is the parity
  (let (h* c* e* e** rl)
;(format t "~%par= ~s ~%" par)
    (setq rl *reduce-light*)
;;    (setq  *reduce-light* t)
    (setq e** (gensym))
    (setq c* (subst= e** e c))  ;So that reduc does not mangle e
    (setq h* (binary-to-unary< h))
    (setq c* (binary-to-unary< c*))
    (setq e* (binary-to-unary< e))
    (setq c* (subst= e* e** c*))
;(format t "~%par= ~s ~%" par)
    (setq c* (replac h* c* e* par action rj (list h)))
    (setq  *reduce-light* rl)

    (list (reduc (car c*)) (cadr c*))))
  
(defun replac (h c e par &optional action rj path)   ;called by replace-a
  ;;rj is the reject-list.  par is the parity

  (let (theta c* c2 the the2 w h*)

   (cond 
    ((member path rj :test 'equal) (list c nil))
    ((equal c e)
     (cond 

      ((and (not (atom c)) (member (car c) '(< <=))
            (not (atom h)) (member (car h) '(< <=)) 
            (or (eq 'chain action) (not action)) par
	    (setq W (<chain h c)))
       (append (reverse (car W)) (list path)))

      ((and par (setq theta (unify h c)))       ;simple match
       (cond ((or (eq 'unify action) (not action))
                                 (list t theta path))  ;replace c by t
             ((eq 'chain action) (list 0 theta path))  ;replace c by 0 
             (t (list c nil)) ))

      ((not par) (setq h* (not-on h))
                 (replac h* c e t action rj (cons h* path)))

      ((atom h) (list t nil))
            
      ((eq '& (car h))
       (cond
        ((setq theta (some #'(lambda (u) 
              (and (setq the (replac u c e par action rj (cons u path)))
		   (cadr the) the)) 
                           (cdr h))) 
         theta)
        (t (list c nil)) ))

      ((eq '-> (car h)) 
       (cond
        ((and (setq theta (replac (caddr h) c e par action rj 
				  (cons (caddr h) path)))
              (cadr theta)		;back-chain
	      (setq c* (car theta)) (setq theta (cadr theta))
 	      (setq c* (and-on (apply-substitution theta (cadr h)) c*)))
         (list c* theta path))

	((and (setq theta (replac (caddr (setq h* (reverse-> h))) c e par 
				  action rj (cons (caddr h*) path)))
	                                          ;also try reversing h
	      (cadr theta)
	      (setq c* (car theta)) (setq theta (cadr theta))
 	      (setq c* (and-on (apply-substitution theta (cadr h*)) c*)))
         (list c* theta path)) 

        (t (list c nil)) ))

      ((eq '=--> (car h)) ;Note the dual actions for -> and =--> here,
       (cond              ; (cadr h) and (caddr h) are interchanged.
        ((and (setq theta (replac (cadr h) c e par action rj 
				  (cons (cadr h) path))) 
              (cadr theta)		;use a reduce rule
	      (setq c* (car theta)) (setq theta (cadr theta))
 	      (setq c* (and-on (apply-substitution theta (caddr h)) c*)))
         (list c* theta))
        (t (list c nil)) ))

      ((eq 'or (car h))
       (cond                               ;"or" back-chaining
        ((setq theta 
               (some #'(lambda (u)
                  (and (setq the (replac u c e par action rj (cons u path)))
		       (setq c* (car the)) 
                       (setq the (cadr the))
		       (setq c* (and-on (reduc (apply-substitution the 
					   (not-on (remove2 u h)))) c*))
		       (list c* the))) (cdr h)))
         theta)
        (t (list c nil)) ))
    
      ((eq 'not (car h))
       (cond ((equal h (setq h* (push-nots-in-ve h))) (list c nil))
	     (t (replac h* c e par action rj (cons h* path)))))

      (t (list c nil)) ))

     ;;Now, the (c not= e) case
     
    ((atom c) (list c nil))

    ((eq '& (car c)) 
     (cond 
      ((setq theta 
	     (some #'(lambda (u) 
		       (and (occur e u)
			    (setq the (replac h u e par action rj path))
			    (cadr the) 
			    (setq c* (subst= (car the) u c))
			    (setq c* (apply-substitution (cadr the) c*))
			    (list c* (cadr the)) )) 
		   (cdr c)))
       theta)
      (t (list c nil))))
 
    ((eq 'or (car c))     ;note this is the same as the (eq '& (car c)) case
     (cond 
      ((setq theta 
	     (some #'(lambda (u) 
		       (and (setq the (replac h u e par action rj path))
			    (cadr the) 
			    (setq c* (subst= (car the) u c))
                            (setq c* (apply-substitution (cadr the) c*))
			    (list c* (cadr the)) )) 
                   (cdr c)))
       theta)
      (t (list c nil))))
 
    ((eq '-> (car c))                   
     (cond ((occur e (caddr c))
	    (setq theta 
		  (replac h (caddr c) e par action rj path))
	    ;;theta = (c2 the2)
            (setq c2 (car theta)) (setq the2 (cadr theta))
            (setq c* (list '-> (cadr c) c2))
            (list c* the2) )
           ((and (occur e (cadr c)) (not (occur e (caddr c))))
	    ;;if e occurs in (cadr c), reverse c
	    (setq c* (reverse-> c))            ;(c1->c2) goes to (~c2->~c1)
            (setq w (replac h c* e par action rj path))
	    (cond (w (list (reverse-> (car w)) (cadr w)))))
           (t (list c nil)) ))

    ((member (car c) '(< <=))
     (setq action (cond (action) (t 'chain)))
     (cond 
      ((atom h) (list t nil))
      ((eq '& (car h))
       (cond
        ((setq theta (some #'(lambda (u) 
                (and (setq the (replac u c e par action rj (cons u path)))
		     (cadr the) the)) 
                           (cdr h))) 
         theta)
        (t (list c nil)) ))
      ((eq '-> (car h)) 
       (cond
        ((and (setq theta (replac (caddr h) c e par action rj 
				  (cons (caddr h) path))) 
              (cadr theta)		;back-chain
	      (setq c* (car theta)) (setq theta (cadr theta))
 	      (setq c* (and-on (apply-substitution theta (cadr h)) c*)))
         (list c* theta path))
	((replac (setq h* (reverse-> h)) c e par action rj (cons h* path))) 
	                                             ;also try reversing h
        (t (list c nil)) ))
      ((eq '=--> (car h))	;Note the dual actions for -> and =--> here,
       (cond			; (cadr h) and (caddr h) are interchanged.
        ((and (setq theta
		    (replac (cadr h) c e par action rj (cons (cadr h) path)))
              (cadr theta)		;use a reduce rule
	      (setq c* (car theta)) (setq theta (cadr theta))
 	      (setq c* (and-on (apply-substitution theta (caddr h)) c*)))
         (list c* theta))
        (t (list c nil)) ))
      ((eq 'or (car h))
       (cond				;"or" back-chaining
        ((setq theta 
               (some #'(lambda (u)
		 (and (setq the (replac u c e par action rj (cons u path)))
		      (cadr the)
		      (setq c* (car the)) (setq the (cadr the))
		      (setq c* (and-on (reduc (apply-substitution the 
				  (not-on (remove2 u h)))) c*))
		      (list c* the))) (cdr h)))
         theta)
        (t (list c nil)) ))
     
      ((member (car h) '(< <=))    ;Recall, that also (car c) is < or <=
       (cond ((or (eq 'chain action)(not action))
              (cond ((setq w (<chain-on h c e)) (list (cadr W) (car W) path))
		    (t (list c nil))))
             ((setq theta (unify h c))
              (print 'how-could-this-happen?)
              (list t theta path))
             (t (list c nil))))

      (t (list c nil)) ))
    ((eq 'not (car C))
     (cond ((cadr (setq c* (replac h (cadr C) e (not par) action rj)))
	    (list (not-on (car c*)) (cadr c*)))
	   (t (list c nil))))

    (t (list c nil)) )))

 (defun replacc-a (c e &optional (par t) rj)   
  ;;Used by REMOVE-SHIELD
  ;;rj is a reject-list of PATHs, from h down to the term in h matched with e 
  ;;par is the parity
  (let (c* e* e** rl)
    (setq rl *reduce-light*)
    (setq e** (gensym))
    (setq c* (subst= e** e c))  ;So that reduc does not mangle e
    (setq c* (binary-to-unary< c*))
    (setq e* (binary-to-unary< e))
    (setq c* (subst= e* e** c*))
    (setq c* (replacc c* nil e* par nil rj (list c)))
    (setq  *reduce-light* rl)

    (list (reduc (car c*)) (cadr c*))))

(defvar *e-path* nil)

(defun repla-a (c e &optional (par t) rj)   ;;Used by REMOVE-SHIELD
  ;;rj is a reject-list of PATHs, from h down to the term in h matched with e 
  ;;par is the parity
  ;;The idea is to replace the term e in C, by back-chaining or inequality-
  ;; chaining on other parts of c.  
  ;;If successful, it returns a list (c2 theta path), where c2 is a replacement
  ;; for e (or the atom holding e, if e is contained within an inequality atom)
  ;; theta is the corresponding substitution, and PATH is the path leading to
  ;; the formula that replaced e.
  ;;Otherwise, it returns nil.
  (let (c* e* e** c2 rl)
    (setq *e-path* nil)
    (setq rl *reduce-light*)
    (setq e** (gensym))
    (setq c* (subst= e** e c))  ;So that reduc does not mangle e
    (setq c* (binary-to-unary< c*))
    (setq e* (binary-to-unary< e))
    (setq c* (subst= e* e** c*))
    (setq c* (repla c* e* nil par rj (list c)))
    (setq  *reduce-light* rl)
    (cond ((not c*) nil)
	  ((eq 'e* (car c*)) nil)
	  (t (setq c2 (insert (car c*) *e-path* (cadr c*)))
	     (setq c2 (new-variables c2))
	     (cons c2 (cdr c*))) )))

(defun repla (D e e* par rj path)   ;Called by REPLA-A
  ;;See REPLA-A for explanation of parameters.  D is the current value of c.
  ;;e* is either e, or the inequality atom containing e.  
  ;;If e has been replaced, this  returns a list (c2 theta path), where c2 is 
  ;; the replacement for e (or for e*), theta is the corresponding substitution
  ;; and PATH is the path leading to the formula that replaced e.
  ;; ('e* parity e*) is returned if e has been found but not replaced, and
  ;; nil is returned if e is not found.

  (let (W D* D2 e2 theta u* e-par path* par* u2)
    (cond ((not e*)
      (cond ((equal D e) 
	     (setq *e-path* path)
	     (list 'e* par e))
	    ((atom D) nil)      ;alarm
	    
	    ((member (car D) '(& or))
	     (some #'(lambda (u) (and (occur e u)
				      (setq W (repla u e nil par rj 
						     (cons u path)))
				      (setq u* u)))
		   (cdr D))
	     (cond ((eq 'e* (car W))
		    (setq e-par (cadr W)) (setq e* (caddr W))
		    (setq par* (eq par e-par))
		    (setq D* (remove u* D :test 'equal))
		    (repla D* e e* par* rj path))
		   (t  W)))
	    
	    ((eq '-> (car D))
	     (cond ((occur e (setq u* (caddr D)))
		    (setq W (repla u* e nil par rj (cons u* path)))
		    (cond ((eq 'e* (car W))
			   (setq e-par (cadr W)) (setq e* (caddr W))
			   (setq par* (eq par e-par))
			   (repla (cadr D) e e* (not par*) rj path))
			  (t W)))
		   ((occur e (setq u* (cadr D)))
		    (setq W (repla u* e nil (not par) rj (cons u* path)))
		    (cond ((eq 'e* (car W))
			   (setq e-par (cadr W)) (setq e* (caddr W))
			   (setq par* (eq par e-par))
			   (repla (caddr D) e e* par* rj path))
			  (t W)))))
	    ((eq 'not (car D))
	     (setq u* (cadr D))
	     (setq W (repla u* e nil (not par) rj (cons u* path)))
	     (cond ((eq 'e* (car W))
		    (setq e-par (cadr W)) (setq e* (caddr W))
		    (setq par* (eq par e-par))
		    (repla D e e* par* rj path))
		   (t W)))

	    ((member (car D) '(< <=))
	     (setq *e-path* path)
	     (list 'e* par D))

	    (t nil) ))

;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;;   now e* is t   ie, e has been found.

	  ((member path rj :test 'equal) (list t nil))
	  ((atom D) (list 'e* par e*))

	  ((and (member (car D) '(< <=)) e* 
		(setq D* (cond (par (push-not-on D)) (t D)))
		(setq W (<chain-on D* e* e)))
	   (list (cadr W) (car W) path))
	  
	  ((and (not par) (setq theta (unify D e*)))
	   (list t theta path))

	  ((eq '& (car D))
	   (cond ((some #'(lambda (u) 
                           (and (setq W (repla u e e* par rj (cons u path)))
				(cadr W)))
			(cdr D))
		  W)
		 (t (list 'e* par e*))))

	  ((eq 'or (car D))
	   (cond ((some #'(lambda (u)		;"or" back-chaining
	    	     (and (setq W (repla u e e* par rj (cons u path)))
			  (not (eq 'e* (setq e2 (car W))))
			  (setq theta (cadr W))
			  (setq D2 (and-on e2 (reduc (apply-substitution theta 
				      (push-not-on (remove2 u D))))))
			  (setq path* (caddr W)) ))
			(cdr D))
		  (list D2 theta path*))
		 (t (list 'e* par e*)) ))

	  ((eq '-> (car D))
	   (cond ((and (setq u2 (caddr D)) 
		       (setq W (repla u2 e e* par rj (cons u2 path)))
		       (not (eq 'e* (setq e2 (car W))))
		       (setq theta (cadr W))   ;back-chain
		       (setq D2 (imp-on (apply-substitution theta (cadr D)) 
					e2)))
		  (list D2 theta (cons u2 path)))
		 ((and (setq u2 (caddr (setq D* (reverse-> D))))
		       (setq W (repla u2 e e* par rj (cons u2 path)))
		       (not (eq 'e* (setq e2 (car W))))
		       (setq theta (cadr W))   ;back-chain
		       (setq D2 (imp-on (apply-substitution theta (cadr D*)) 
					e2)))
		  (list D2 theta (list u2 path)))    	;also try reversing D
		 (t (list 'e* par e*)) ))

	  ((eq 'not (car D))
	   (cond ((not (equal D (setq D* (push-nots-in-ve D))))
		  (repla D* e e* par rj (cons D* path)))
		 (t (repla (cadr D) e e* (not par) rj (cons (cadr D) path)))))

	  (t (list 'e* par e*)) )))

(defun replacc (c c* e par e-lt rj path) ;called by     NOT USED 21 June 89
  (let (W W* e-par par* c2 c3 c4 theta)
    (cond ((not e-par)
	   (cond ((eq 'e-lt (car (setq W (repl c c* par rj path))))
		  (setq e-par (cadr W)) (setq c* (caddr W))
		  (setq par* (and par e-par))
		  (replacc c c* e par* t rj path))
		 (t W)))
	  ;e-lt is now t
	  ((member path rj :test 'equal) (list c nil))
          ((atom c) (list c nil))
	  ((and (member (car c) '(< <=)) c* (setq W (<chain-on c c* e)))
	   (list (cadr W) (car W) path))
	  ((and par (setq theta (unify c e))) (list t theta path))
          ((eq '& (car c))
	   (cond ((setq W (some #'(lambda (u) 
                    (and (setq W* (replacc u c* e par t rj (cons u path)))
			 (cadr W*) W*))
				(cdr c)))
		  W)
		 (t (list c nil))))
	  ((eq 'or (car c))
	   (cond ((setq W (some #'(lambda (u)		;"or" back-chaining
	    	     (and (setq W* (replac u c* e par t rj (cons u path)))
			  (setq c4 (car W*)) (setq theta (cadr W*))
			  (setq c4 (and-on (reduc (apply-substitution theta 
				      (not-on (remove2 u c)))) c4))
			  (list c4 theta)))
				(cdr c)))
		  W)
		 (t (list c nil)) ))
	  ((eq '-> (car c)) 
	   (cond ((and (setq c3 (caddr c)) 
		       (setq W (replac c3 c* e par t rj (cons c3 path)))
		       (setq c4 (car W)) (setq theta (cadr W))   ;back-chain
		       (setq c4 (and-on (apply-substitution theta 
							    (cadr c)) c4)))
		  (list c4 theta path))
		 ((and (setq c3 (caddr (setq c2 (reverse-> c))))
		       (setq W (replac c3 c* e par t rj (cons c3 path)))
		       (setq c4 (car W)) (setq theta (cadr W))   ;back-chain
		       (setq c4 (and-on (apply-substitution theta 
							    (cadr c2)) c4)))
		  (list c4 theta path))    	;also try reversing c
		 (t (list c nil)) ))

	  ((eq 'not (car c))
	   (cond ((equal c (setq c2 (push-nots-in-ve c))) (list c nil))
		 (t (replac c2 c* e par t rj (cons c2 path)))))

	  (t (list c nil)) )))

(defun repl (c e par rj path)      ;Called by replacc
  (cond ((equal c e) (list 'e-lt par nil))
	((atom c) (list c nil))
	((eq '& (car c))  (some #'(lambda (u) (replacc u nil e par nil rj
						       (cons u path)))
				(cdr c)))
	((eq 'or (car c)) (some #'(lambda (u) (replacc u nil e par nil rj
						       (cons u path)))
				(cdr c)))
	((eq '-> (car c))
	 (or (replacc (caddr c) nil e   par     nil rj (cons (caddr c) path))
	     (replacc (cadr c)  nil e (not par) nil rj (cons (cadr  c) path))))

	((eq 'not (car c))
	 (replacc (cadr c) nil e (not par) nil rj (cons (cadr c) path)))

	((and (member (car c) '(< <=)) (occur e c))
	 (list 'e-lt par c))

	(t (list c nil)) ))

(defun replace-term (h c &optional term-pattern term-reject-list)
   ;;called by OPTIONS -- requires user input.

  (let (rlt term action reject-list w occur-list)
   (setq rlt *reduce-light*)        ;Save the rlt
;;   (setq *reduce-light* t)          ;Turn on REDUCE

   (format t "        ===== REPLACE A TERM ===== ~%")
   (cond 
    (term-pattern (format t "~%        TRY THE NEXT MATCH ~%"))
    (t 
     (format t "~%Type in the TERM that you wish replaced in C - 
       You may type in a more general pattern that uniquely defines it -
       Please Type your Term-pattern:   ")
     (setq term-pattern (read-h)) ))
   ;;Term-pattern may be a pattern such as (|| z) which will be used by 
   ;; MATCH-IN to fetch the TERM from C which is to be replaced. Also 
   ;; TERM-PATTERN may contain a member expression of the form(occur x y ...), 
   ;; where the atoms x y etc, must occur in the TERM that is fetched.  
   ;; eg, (|| z (occur F)), would fetch  (|| (+ ((F) (XS D)) (- (L)))) 
   ;; but not (|| (+ (XS D) (- (A))))

   (cond ((setq occur-list (assoc-r 'occur term-pattern))
          (setq term-pattern (remove occur-list term-pattern :test 'equal))
          (setq occur-list (cdr occur-list))))

   (setq term (match-in term-pattern c 'instance term-reject-list occur-list))
   (cond ((not term)
          (format t "~% Your term-pattern did not determine a term in C ~%")
          (if (y-or-n-p "Do you wish to try again?  Type-  ")
	      (return-from replace-term (replace-term h c))
	    (progn (format t "~% User opted to return to OPTIONS ~%")
		   (return-from replace-term (options))))))
   
   (format t "~%You have requested the replacement of the term-pattern ~% ~s 
   which determined the term ~% ~s  of C ~%" term-pattern term)
   (if (y-or-n-p "Is this OK?  Type-  ") t
     (if (y-or-n-p "Shall we seek another match in C of the same pattern? Type- ")
	      (return-from replace-term 
                  (replace-term h c term-pattern (cons term term-reject-list)))

       (if (y-or-n-p "Do you wish to try again?  Type-  ")
	   (return-from replace-term (replace-term h c))
	 (progn (format t "~% User opted to return to OPTIONS ~%")
		(return-from replace-term (options))))))

   (ans* (list 'Replace-term 'is 'being 'tried 'for term))
   (setq %goal-message% (list 'Remove-the-term term))

;;   (if (y-or-n-p "Do you wish to give a ACTION? Type-  ")
;;     (setq action (read-h)) (setq action nil))
;; (if (y-or-n-p "Do you wish to give a REJECT-LIST? Type-  ")
;;     (setq reject-list (cons (read-h) reject-list)) nil)

   (cond ((cadr (setq w (replace-a h c term action reject-list)))
          (ans* (list 'Replace-term 'succeded 'for term 'with 'result w)))
         (t (ans* (list 'Replace-term 'failed 'for term))))
   (setq *reduce-light* rlt)        ;Restore the rlt
   w))

(defun restrict-theta (theta E D)   ;called by INSERT
  (cond ((eq t theta) t)
	(t (setq E (subst 'D% D E :test 'equal))
	   (mapcan #'(lambda (u) (cond ((occur (car u) E) (list u)))) 
		   theta))))

(defun reverse-> (A) ;called by replac.  A has the form (-> P Q).
  (list '-> (push-not-on (caddr A)) (push-not-on (cadr A))))

(defun t-reduce (D)     ;called by INSERT
  (cond ((atom D) D)
	((eq '& (car D)) 
	 (cond ((member '(false) D :test 'equal) '(false))
	       (t (clear-& (remove t D)))))
	((eq 'or (car D))
	 (cond ((member t D) t)
	       (t (clear-or (remove '(false) D :test 'equal)))))
	((eq '-> (car D))
	 (cond ((eq t (cadr D)) (caddr D))
	       ((equal '(false) (cadr D)) t)
	       ((eq t (caddr D)) t)
	       ((equal '(false) (caddr D)) '(false))
	       (t D)))
	((eq 'not (car D))
	 (cond ((eq t (cadr D)) '(false))
	       ((equal '(false) (cadr D)) t)
	       (t D)))
	(t D) ))
 
;;Examples

(defvar r1 '(-> (<= (+ (f (a)) (- (c)))) (<  (+ (f  x ) (- (d))))))

(defvar r2 '(-> (& (<= (+ (f (a)) (- (c))))
		   (<  (+ (- (f  x ) (d)))))
		(< (+ (d) (- (c))))))

(defvar h '(<= (+ (f (a)) (- (c)))))

(defvar c '(<  (+ (f  x ) (- (d)))))

(defvar e '(f x))
