;;;               Sb1.lisp

;;;  From Larry Hines    28 July 1989
(require "SLOOP")
;;;(in-package "SLOOP" :use '(LISP))
;;;
;;;(defmacro loop (&body body) (parse-loop body))
;;;    is assumed to have been defined.
;;;

(provide "sb1")

(defmacro set-diff? (form)
  `(and (consp ,form) (eq 'set-diff (car ,form))))
(defmacro intersection? (form) 
  `(and (consp ,form) (eq 'intersection (car ,form))))
(defmacro union? (form) 
  `(and (consp ,form) (eq 'union (car ,form))))
(defmacro complement? (form) 
  `(and (consp ,form) (eq 'complement (car ,form))))
(defmacro singleton? (form) 
  `(and (consp ,form) (eq 'singleton (car ,form))))
(defmacro nonempty? (lit) `(and (consp ,lit) (eq '[= (car ,lit))))
(defmacro element? (lit) `(and (consp ,lit) (eq '[= (car ,lit))))
(defmacro subset=? (lit) `(and (consp ,lit) (eq '[= (car ,lit))))
(defmacro strict? (sign) `(equal '[ ,sign))
(defmacro sign (lit) `(car ,lit))
(defmacro left-side (lit) `(cadr ,lit))
(defmacro right-side (lit) `(caddr ,lit))
(defmacro make-literal (sign left-side right-side)
  `(list ,sign ,left-side ,right-side))
(defmacro cl-number (clause) `(car ,clause))
(defmacro cl-type (clause) `(cadr ,clause))
(defmacro sis? (type) `(eq 'sis ,type))
(defmacro input? (type) `(eq 'input ,type))
(defmacro subsumed? (clause) `(eq 'subsumed (cl-type ,clause)))
(defmacro literals (clause) `(caddr ,clause))
(defmacro make-clause (number type clause) `(list ,number ,type ,clause))
(defmacro variable? (term) `(atom ,term))
(defmacro function-args (function-term) `(cdr ,function-term))
(defun non-ground? (term)
  (or (variable? term) (some #'non-ground? (function-args term))))

(defvar sb-example
  '(-> (& ([= (singleton (L1)) (c12)) ([= (singleton (not* (L1))) (c21)))
       ([= (union (union (set-diff (c12) (singleton (L1)))
			 (set-diff (c21) (singleton (not* (L1)))))
		  (singleton (L)))
	   (union (set-diff (union (c12) (singleton (L))) (singleton (L1)))
		  (set-diff (c21) (singleton (not* (L1))))))))

(defvar sb-ex1
  '(-> ([= (intersection (singleton (L)) (singleton (L1))) 0)
       ([= (union (union (set-diff (c12) (singleton (L1)))
			 (set-diff (c21) (singleton (not* (L1)))))
		  (singleton (L)))
	   (union (set-diff (union (c12) (singleton (L))) (singleton (L1)))
		  (set-diff (c21) (singleton (not* (L1))))))))

(defvar sb-ex1=
  '(-> ([= (intersection (singleton (L)) (singleton (L1))) 0)
       (= (union (union (set-diff (c12) (singleton (L1)))
			(set-diff (c21) (singleton (not* (L1)))))
		 (singleton (L)))
	  (union (set-diff (union (c12) (singleton (L))) (singleton (L1)))
		 (set-diff (c21) (singleton (not* (L1))))))))

(defvar sb-ex2
  '(some c (-> ([= (intersection (singleton (L)) (singleton (L1))) 0)
	       ([= (union (union (set-diff (c12) (singleton (L1)))
				 (set-diff (c21) (singleton (not* (L1)))))
			  (singleton (L)))
		   (union (set-diff c (singleton (L1)))
			  (set-diff (c21) (singleton (not* (L1)))))))))

(defvar sb-ex2=
  '(some c (-> ([= (intersection (singleton (L)) (singleton (L1))) 0)
	       (= (union (union (set-diff (c12) (singleton (L1)))
				(set-diff (c21) (singleton (not* (L1)))))
			 (singleton (L)))
		  (union (set-diff c (singleton (L1)))
			 (set-diff (c21) (singleton (not* (L1)))))))))

(defvar sb-ex3e  '(-> ([= (A) (B)) ([= (A) (B))))
(defvar sb-ex3u  '(-> ([= (A) (B)) ([= (A) (UNION (B) (C)))))
(defvar sb-ex3i  '(-> ([= (A) (B)) ([= (intersection (A) (C)) (B))))
(defvar sb-ex4e  '(-> ([  (A) (B)) ([= (A) (B))))
(defvar sb-ex4s  '(-> ([  (A) (B)) ([  (A) (B))))
(defvar sb-ex4eu '(-> ([  (A) (B)) ([= (A) (UNION (B) (C)))))
(defvar sb-ex4su '(-> ([  (A) (B)) ([  (A) (UNION (B) (C)))))
(defvar sb-ex4ei '(-> ([  (A) (B)) ([= (intersection (A) (C)) (B))))
(defvar sb-ex4si '(-> ([  (A) (B)) ([  (intersection (A) (C)) (B))))
(defvar sb-ex5e  '(-> (& ([= (A) (C)) ([= (C) (B))) ([= (A) (B))))
(defvar sb-ex5u  '(-> (& ([= (A) (C)) ([= (C) (B)))
		      ([= (A) (UNION (B) (C)))))
(defvar sb-ex5i  '(-> (& ([= (A) (C)) ([= (C) (B)))
		      ([= (intersection (A) (C)) (B))))
(defvar sb-ex6e  '(-> (& ([  (A) (B)) ([= (C) (B))) ([= (A) (B))))
(defvar sb-ex6s  '(-> (& ([  (A) (B)) ([= (C) (B))) ([  (A) (B))))
(defvar sb-ex6eu '(-> (& ([  (A) (B)) ([= (C) (B)))
		      ([= (A) (UNION (B) (C)))))
(defvar sb-ex6su '(-> (& ([  (A) (B)) ([= (C) (B)))
		      ([  (A) (UNION (B) (C)))))
(defvar sb-ex6ei '(-> (& ([  (A) (B)) ([= (C) (B)))
		      ([= (intersection (A) (C)) (B))))
(defvar sb-ex6si '(-> (& ([  (A) (B)) ([= (C) (B)))
		      ([  (intersection (A) (C)) (B))))
(defvar sb-ex7e  '(-> (& ([= (A) (B)) ([  (C) (B))) ([= (A) (B))))
(defvar sb-ex7eu '(-> (& ([= (A) (B)) ([  (C) (B)))
		      ([= (A) (UNION (B) (C)))))
(defvar sb-ex7ei '(-> (& ([= (A) (B)) ([  (C) (B)))
		      ([= (intersection (A) (C)) (B))))
(defvar sb-ex8e  '(-> (& ([  (A) (B)) ([  (C) (B))) ([= (A) (B))))
(defvar sb-ex8s  '(-> (& ([  (A) (B)) ([  (C) (B))) ([  (A) (B))))
(defvar sb-ex8eu '(-> (& ([  (A) (B)) ([  (C) (B)))
		      ([= (A) (UNION (B) (C)))))
(defvar sb-ex8su '(-> (& ([  (A) (B)) ([  (C) (B)))
		      ([  (A) (UNION (B) (C)))))
(defvar sb-ex8ei '(-> (& ([  (A) (B)) ([  (C) (B)))
		      ([= (intersection (A) (C)) (B))))
(defvar sb-ex8si '(-> (& ([  (A) (B)) ([  (C) (B)))
		      ([  (intersection (A) (C)) (B))))

(defvar grd-sb-examples
  '(sb-ex1 sb-ex1= sb-ex3e sb-ex3u sb-ex3i
	   sb-ex4e sb-ex4s sb-ex4eu sb-ex4su sb-ex4ei sb-ex4si
	   sb-ex5e sb-ex5u sb-ex5i
	   sb-ex6e sb-ex6s sb-ex6eu sb-ex6su sb-ex6ei sb-ex6si
	   sb-ex7e sb-ex7eu sb-ex7ei
	   sb-ex8e sb-ex8s sb-ex8eu sb-ex8su sb-ex8ei sb-ex8si))


(defun negate-prep&cnf-formula (formula)
  (subsume-within-set
    (loop for c in (clausify (negate&prep-formula formula) nil nil nil)
	  unless (some #'(lambda (l) (not (subset=? l))) c)
	  collect (loop for l in c for ll = (init-literal l)
			unless (false-literal ll) collect ll))))

(defun init-literal (lit)
  (if (subset=? lit)
      (let ((l (cond ((intersection? (cadr lit)) 
		      (if (member 0 (cdadr lit)) (list 0)
			(remove '(universe) (delete-duplicates
					      (cdadr lit) :test #'equal)
				 :test #'equal)))
		     ((equal (cadr lit) '(universe)) nil)
		     (t (list (cadr lit)))))
	    (r (cond ((union? (caddr lit)) 
		      (if (member '(universe) (cdaddr lit) :test #'equal)
			  '((universe))
			(remove 0 (delete-duplicates
				    (cdaddr lit) :test #'equal))))
		     ((eq (caddr lit) 0) nil)
		     (t (list (caddr lit))))))
	(cond ((and (consp (car r)) (equal '(universe) (car r)))
	       (make-literal
		 '[= (loop for ll in l if (singleton? ll) collect ll)
		 r))
	      ((and (consp (car l)) (eq 0 (car l)))
	       ;; The literal is true, so we just return 0[=0.
	       (make-literal '[= l l))
	      (t (make-literal '[= l r))))
    lit))

;;;(defvar universe-expr '(universe))
;;;(defmacro universe () universe-expr)
(defun subsume-within-set (cs) 
  (and cs (let ((cs1 (subsume-within-set 
		       (not-subsumed-by (cdr cs) (car cs)))))
	    (if (subsumed-by-oneof (car cs) cs1)
		cs1 (cons (car cs) cs1)))))

(defun subsumed-by-oneof (clause set-of-subsumers)
  "Returns nonNIL if CLAUSE is subsumed by a clause in SET-OF-SUBSUMERS."
  (declare (special verbose))
  (loop for c2 in set-of-subsumers thereis (subsumed-by clause c2)))
(defun not-subsumed-by (cl-set subsumer)
  "Returns clauses not subsumed by SUBSUMER."
  (declare (special verbose))
  (loop for c1 in cl-set 
	unless (subsumed-by c1 subsumer) collect c1))
(defun subsumed-by (clause subsumer) 
  " Returns T if CLAUSE is subsumed by SUBSUMER."
  (every #'(lambda (lit) (lit-subsumed-by lit clause)) subsumer))
(defun lit-subsumed-by (s-lit clause)
  (some #'(lambda (lit2) 
	    (and (subsetp (left-side s-lit) (left-side lit2) :test #'equal)
		 (subsetp (right-side s-lit) (right-side lit2) :test #'equal)))
	clause))


(defun false-literal (lit)
  "Returns nonNIL iff the literal represents either that the non-empty
universe or a sinleton set is a subset or equal to the empty set.  The 
universe is assumed to be non-empty iff *NONEMPTY-UNIVERSE* is nonNIL."
  (declare (special *nonempty-universe*))
  (and (null (right-side lit)) ;; The rightside is empty.
       (let ((l (left-side lit))) 
	 ;; Either the universe is a subset or equal to empty set
	 (or ;;(and *nonempty-universe* (null l))
	     (and (null (cdr l)) (singleton? (car l))
		  (not (non-ground? (car l))))))))

