;;;             sb2.lisp

;;;  From Larry Hines    28 July 1989
(require "SLOOP")
;;;(in-package "SLOOP" :use '(LISP))
;;;
;;;(defmacro loop (&body body) (parse-loop body))
;;;    is assumed to have been defined.
;;;
;;;
(require "sb1")
(provide "sb2")

(defun negate&prep-formula (form)
  (let ((conjuncts (prep-not-formula (remove-set-diffs form))))
    (if (null (cdr conjuncts)) (car conjuncts)
      (cons '& conjuncts))))

(defun remove-set-diffs (form)
  (cond ((atom form) form)
	((member (car form) '(some all))
	 (list (car form) (cadr form) (remove-set-diffs (caddr form))))
	((member (car form) '(not & or -> <-> [= [ = union intersection))
	 (cons (car form) (mapcar #'remove-set-diffs (cdr form))))
	((eq (car form) 'set-diff) 
	 `(intersection ,(cadr form) (complement ,(caddr form))))
	(t form)))

(defun prep-formula (form)
  (cond ((atom form) (list form))
	((member (car form) '(some all))
	 `((,(car form) ,(cadr form) (& .,(prep-formula (caddr form))))))
	((eq '-> (car form))
	 (cross-fn-over 'or (list (prep-not-formula (cadr form))
				  (prep-formula (caddr form)))))
	((eq '<-> (car form))
	 (append (prep-formula `(-> .,(cdr form)))
		 (prep-formula `(-> ,(caddr form) ,(cadr form)))))
	((eq 'or (car form))
	 (cross-fn-over 'or (mapcar #'prep-formula (cdr form))))
	((eq '& (car form)) (mapcan #'prep-formula (cdr form)))
	((eq 'not (car form)) (prep-not-formula (cadr form)))
	((eq '[ (car form)) (prep-sb (cadr form) (caddr form)))
	((eq '= (car form)) (prep= (cadr form) (caddr form)))
	((eq '[= (car form)) (prep-sb= (cadr form) (caddr form)))
	((eq 'element (car form)) (prep-sb= (list 'singleton (cadr form))
					    (caddr form)))
	((eq 'nonempty (car form)) (prep-nonempty (cadr form)))
	(t (list form))))
	
(defun prep-not-formula (form)
  (cond ((atom form) `((not ,form)))
	((eq 'not (car form)) (prep-formula (cadr form)))
	((eq 'some (car form))
	 `((all ,(cadr form) (& .,(prep-not-formula (caddr form))))))
	((eq 'all (car form))
	 `((some ,(cadr form) (& .,(prep-not-formula (caddr form))))))
	((eq '-> (car form)) (append (prep-formula (cadr form))
				     (prep-not-formula (caddr form))))
	((eq '<-> (car form)) 
	 (append (cross-fn-over 'or (mapcar #'prep-formula (cdr form)))
		 (cross-fn-over 'or (mapcar #'prep-not-formula (cdr form)))))
	((eq '& (car form))
	 (cross-fn-over 'or (mapcar #'prep-not-formula (cdr form))))
	((eq 'or (car form)) (mapcan #'prep-not-formula (cdr form)))
	((eq '[ (car form)) (prep-not-sb (cadr form) (caddr form)))
	((eq '= (car form)) (prep-not= (cadr form) (caddr form)))
	((eq '[= (car form)) (prep-not-sb= (cadr form) (caddr form)))
	((eq 'element (car form)) 
	 (prep-sb= `(intersection (singleton ,(cadr form)) ,(caddr form))
		   0))
	((eq 'nonempty (car form)) (prep-sb= (cadr form) 0))
	(t `((not ,form)))))

(defun prep-not-sb (left right) " A /[  B   >>>> A /[= B  v  B [= A."
  (cross-fn-over 'or (list (prep-not-sb= left right)
			   (prep-sb= right left))))
(defun prep-sb (left right) " A[B   >>>>  A [= B  &  B /[= A."
  (append (prep-sb= left right) (prep-not-sb= right left)))
(defun prep-not= (left right) " A/=B   >>>> A n B \[= A u B."
  (let ((var (gentemp)))
    `((some ,var (& .,(append (prep-sb=-*-union `(singleton ,var) 
						(list left right))
			      (prep-sb=-intersect 
				`((singleton ,var) ,left ,right) 0)))))))
(defun prep= (left right) " A=B     >>>> A[=B & B[=A."
  (append (prep-sb= left right) (prep-sb= right left)))
(defun prep-not-sb= (left right) " A /[= B   >>>> {x} [= A n -B."
  (let ((var (gentemp)))
    `((some ,var (& ,@(prep-sb= `(singleton ,var) left)
		    .,(prep-sb=-intersect `((singleton ,var) ,right) 0))))))

(defun prep-nonempty (term)
  (let ((var (gentemp)))
    `((some ,var (& .,(prep-sb= `(singleton ,var) term))))))

(defun prep-sb= (left right) 
  "Returns a list of conjuncts which are in canonical form and represents
the LEFT is a subset of or equal to RIGHT."
  (cond ((not (consp right)) (prep-sb=-left left right))
	((eq 'union (car right)) (prep-sb=-*-union left (cdr right)))
	((eq 'intersection (car right))
	 (prep-sb=-*-intersect left (cdr right)))
	((eq 'complement (car right)) 
	 (prep-sb=-intersect (list (cadr right) left) 0))
	(t (prep-sb=-left left right))))
  

(defun prep-sb=-*-union (left cdr-right)
  "Returns a list of conjuncts which are in canonical form and represents
the LEFT is a subset of or equal to union applied to CDR-RIGHT."
  (loop with comps = nil and inters = nil and rest = nil
	for r = (pop cdr-right)
	if (complement? r) do (push (cadr r) comps)
	;; A [= B u -C      >>>> A n C [= B.
	else if (union? r) do (setq cdr-right (append (cdr r) cdr-right))
	;; A [= B u (C u D) >>>> A [= B u C u D.
	else if (intersection? r) do (push (cdr r) inters)
	;; A [= B u (C n D) >>>> A [= (B u C) n (B u D).
	else if (not (eq 0 r)) do (push r rest)
	unless cdr-right return
	(cond (inters (if comps (prep-sb=-*-intersect 
				  `(intersection left .,comps)
				  (cross-intersection inters rest))
			(prep-sb=-*-intersect
			  left (cross-intersection inters rest))))
	      (comps (prep-sb=-intersect (cons left comps)
		       (if (cdr rest) `(union .,rest) (car rest))))
	      (t (prep-sb=-left 
		   left (if (cdr rest) `(union .,rest) (car rest)))))))

(defun prep-sb=-*-intersect (left cdr-right)
  "Returns a list of conjuncts which are in canonical form and represents
the LEFT is a subset of or equal to intersection applied to CDR-RIGHT."
  ;; A [= B n C       >>>> A[=B & A[=C.
  (mapcan #'(lambda (r) (prep-sb= left r)) cdr-right))

;;; Right has been fully canonicalized.
(defun prep-sb=-left (left right) 
  "Returns a list of conjuncts which are in canonical form and represents
the LEFT is a subset of or equal to RIGHT.  RIGHT is assumed in canonical
form."
  (cond ((not (consp left)) `(([= ,left ,right)))
	((eq 'union (car left)) (prep-sb=-union (cdr left) right))
	((eq 'intersection (car left)) (prep-sb=-intersect (cdr left) right))
	((eq 'complement (car left))
	 (prep-sb=-*-union '(universe) (list (cadr left) right)))
	(t `(([= ,left ,right)))))

;;; Right has been fully canonicalized.
(defun prep-sb=-intersect (cdr-left right)
  "Returns a list of conjuncts which are in canonical form and represents
intersection applied to CDR-LEFT is a subset of or equal to RIGHT.  RIGHT 
is assumed in canonical form."
  (loop with comps = nil and unions = nil and rest = nil
	for l = (pop cdr-left)
	if (complement? l) do (push (cadr l) comps)
	;; A n -B [= C      >>>> A [= B u C.
	else if (intersection? l) do (setq cdr-left (append (cdr l) cdr-left))
	;; A n (B n C) [= D >>>> A n B n C [= D.
	else if (union? l) do (push (cdr l) unions)
	;; A n (B u C) [= D >>>> (A n B) u (A n C) [= D.
	else do (push l rest)
	unless cdr-left return
	(cond (comps (if unions (prep-sb=-*-union
				  `(union .,(cross-union unions rest))
				  (cons right comps))
		       (prep-sb=-*-union (if (cdr rest) `(intersection .,rest)
					   (car rest))
				  (cons right comps))))
	      (unions (prep-sb=-union (cross-union unions rest) right))
	      (t `(([= ,(if (cdr rest) `(intersection .,rest) (car rest))
		       ,right))))))


(defun cross-fn-over (fn nonempty-sets &optional set)
  "Returns the cross-product of each of the sets and in addition to set.
FN is consed onto each member of the cross-product."
  (mapcar #'(lambda (x) (cons fn x)) 
	  ;; If (NULL REST) then don't confuse CROSS-PRODUCT by adding NIL.
	  (CROSS-PRODUCT (IF set (CONS set nonempty-sets) nonempty-sets))))

(defun cross-intersection (intersections rest)
  "Returns the cross-product of each of the intersections and the rest."
  ;; If (NULL REST) then don't confuse CROSS-PRODUCT by adding NIL.
  (cross-fn-over 'union (if rest (cons rest intersections) intersections)))

(defun cross-union (unions rest)
  "Return the cross-product of each of the unions and the rest."
  ;; If (NULL REST) then don't confuse CROSS-PRODUCT by adding NIL.
  (cross-fn-over 'intersection (if rest (cons rest unions) unions)))

(defun cross-product (listof-nonempty-lists)
  (and listof-nonempty-lists
       (if (null (cdr listof-nonempty-lists))
	   (mapcar #'list (car listof-nonempty-lists))
	 (let ((cprod (cross-product (cdr listof-nonempty-lists))))
	   (loop for elt in (car listof-nonempty-lists)
		 append (loop for set in cprod collect (cons elt set)))))))

(defun prep-sb=-union (cdr-left right)
  "Returns a list of conjuncts which are in canonical form and represents
union applied to CDR-LEFT is a subset of or equal to RIGHT.  RIGHT is assumed 
in canonical form."
  ;; A u B [= C       >>>> A[=C & B[=C.
  (mapcan #'(lambda (l) (prep-sb= l right)) cdr-left))

(defun term-lessp (t1 t2) (eq t (term-compare t1 t2)))
(defun term-compare (t1 t2)
  (cond ((atom t1) (cond ((consp t2) t)
			 ((eq t1 t2) '=)
			 (t (numberp (string-lessp t1 t2)))))
	((atom t2) nil)
	(t (loop for (t11 . rest1) on t1 for (t22 . rest2) on t2 
		 for comp = (term-compare t11 t22)
		 unless (eq '= comp) return comp
		 finally (return (cond (rest1 nil) (rest2 t) (t '=)))))))


