
;;;*********************************************************************
;;;
;;;                  UNIFY.LISP      In KCL
;;;                                  19 Aug 87
;;;               W. W. Bledsoe      12 Feb 1987    3 Mar 87
;;;                                                 21 Jan 88
;;;
;;;                 Called by  IMPLY.LISP
;;;
;;;
;;;***********************************************************************

(DEFUN OCCUR-UNIFY (X A)   ;used-in Unify2
  (COND ((ATOM A) (EQ X A))
	((CDR A) (SOME #'(LAMBDA (Y) (OCCUR-UNIFY X Y)) (CDR A)))))

(defun sublis= (theta A)  ;apply the substitution theta to A
   ;;So that (sublis= '( ((x) y)) '(P x))     correctly gives   (P x).
   ;;Whereas (subst   '( ((x) y)) '(P x) :test 'equal) incorrectly gives (P y).
     (mapc #'(lambda (u) (setq A (subst= (cdr u) (car u) A))) theta)
     A)

(DEFUN TWO-XS (THE &optional (THE* nil)) ;USED-IN (APPLY-substitution 
                                         ;COMPOSE CONFLICT)
 ;; Example:  If Theta = ((x a) (x b) (y d) (x c)) then
 ;; (TWO-XS Theta ) = ( ((x a) (y d) (x c))  ((x b) (y d) (x c)) )
 ;; If the substitution Theta has at least two conflicting entries,
 ;; then these two are split into two separate substitutions

 (IF (NULL THE) NIL
  (Let ((temp (ASSOC (CAAR THE) (CDR THE) :test #'equal)))
    (COND (TEMP 
           (LIST (APPEND THE* (REMOVE TEMP THE)) (APPEND THE* (CDR THE))))
          (T (TWO-XS (CDR THE) (CONS (CAR THE) THE*))))))) 

(DEFUN UNIFY (A B)   ;USED-IN (BACK-CHAIN BACK-CHAIN* BACK-CHAIN-O 
                     ;         IMPLY-RULES MATCH TEST UNIFY-LISTS)
  ;; THE MAJOR PART OF UNIFY IS FOUND IN Unify2 WHICH IS CALLED HERE , THE
  ;;  REST OF THIS IS CONCERNED WITH &exclude& 
  ;; IT RETURNS THE MGU OF a AND b IF THERE IS ONE 
  ;; HOWEVER IF ANY OF THE UNITS IN THIS MGU ARE IN &exclude& THEN THE
  ;;  SUBSTITUTION IS REJECTED 
  ;;;   8/19/87  for the  YET  version of the prover, we have eliminated
  ;;;   the need for &exclude% so it is suppressed here 
  ;;;   See UNIFYi81887.lisp for the other version.
      (UNIFY2 A B))

(DEFUN UNIFY2 (A B)                         ;USED-IN (UNIFY UNIFY3)
  (COND ((EQUAL A B) T)
	((and (ATOM A) (not (numberp A)) (not (member A '([]))))
	 (COND ((atom B) (list (cons B A))) ((OCCUR-UNIFY A B) NIL)
	       (A (LIST (CONS A B)))))
	((and (ATOM B) (not (numberp B)) (not (member A '([]))))
	 (COND ((OCCUR-UNIFY B A) NIL) (B (LIST (CONS B A)))))
	((OR (ATOM A) (ATOM B)) NIL)
        ((NOT (EQUAL (CAR A) (CAR B))) NIL)
	((member (car A) '(ALL SOME PRODUCT SUM))   ;ignore quantified variable
	 (unify3 (cddr A) (subst (cadr A) (cadr B) (cddr B) :test 'equal)))
        ((eq '|| (car A))  ;;Equate |A| and |-A|
         (or (unify (cadr A) (cadr b)) 
             (unify (cadr A) (reduc* (list '- (cadr B))))))
        ((eq '+ (car A))   ;;Allow asso commu of + with mandatory head-matching
         (cond ((not (equal (length (cdr A)) (length (cdr B)))) nil)
               (t (or (unify3 (cdr A) (cdr B)) (unify+ (cdr A) (cdr B))))))
	(T (UNIFY3 (CDR A) (CDR B) NIL))))
 
(DEFUN UNIFY3 (L L* &optional THETA)		;USED-IN (UNIFY2)
  (LET (THE)
       (COND ((NOT (EQUAL (LENGTH L) (LENGTH L*))) NIL)
	     ((NULL L) (COND (L* NIL) (THETA) (T T)))
	     ((EQUAL (CAR L) (CAR L*)) (UNIFY3 (CDR L) (CDR L*) THETA))
	     ((SETQ THE (UNIFY2 (CAR L) (CAR L*)))
              (cond ((eq t the)     ;a call to UNIFY+ might have made THE t
		     (UNIFY3 (CDR L) (CDR L*) THETA))
                    (t (SETQ THETA (UNION-W (SUBLIS THE THETA) THE))
		       (UNIFY3 (SUBLIS THE (CDR L)) 
			       (SUBLIS THE (CDR L*)) THETA)))))))

(defun unify-tp (H C &optional tp)   ;Called by Back-chain, tp = triple
  ;;This is ordinary unification if tp is nil
  ;; Else, it tries to carry out the PAIRS operation
  ;; tp is a triplet from the *pairs-structure* (actually a doubleton)
  ;;Note that C is not used.  However, C has already been matched with
  ;; the other unit in the triple from *paris-structure* before this call
  ;; and H2 is returned to be matched later.
  (let (theta H2)

    (cond ((not tp) (unify H C))
          ((setq theta (unify H (car tp)))
	   (setq H2 (apply-substitution theta (cadr tp)))
           (list theta H2))
          (t nil))))

(defun unify+ (L L*)   ;Called by UNIFY2, the + has been suppressed
  (let (W sigma)
    (cond ((null L) t)
          ((every 'atom L)  (unify3 L L*))  ;if L is all variables use UNIFY3
          ((every 'atom L*) (unify3 L L*))  ;if L*is all variables use UNIFY3
          ((setq W (unify-a-term L L*));Headers must match, W = (theta L2 L*2) 
           (cond ((setq sigma (unify+ (cadr W) (caddr W)))
                  (compose (car W) sigma))
                 (t nil)))
          (t nil))))

(defun unify-a-term (L L* &optional (L2 L) (L*2 L*))  ;Called by Unify+
  ;;This unifies a term tm of L with a term tm* of L*, with the requirement
  ;; that the heads of tm and tm* are equal.
  ;;And returns the triple W = (theta L2 L*2), where theta is the unifying
  ;; substitution, and L2 and L*2 are the result of removing one instance of
  ;; (tm theta) from (L theta) and (L* theta).
  (let (tm tm* theta)
    (cond ((or (null L) (null L*)) nil)
          ((or (and (not (atom (setq tm (car L))))
                    (setq tm* (assoc-r (car tm) L*))
		    (setq theta (unify tm tm*)))
	       (and (not (atom (setq tm* (car L*))))
		    (setq tm (assoc-r (car tm*) L))
		    (setq theta (unify tm tm*))) )
	   (setq L2  (apply-substitution theta (remove-one tm L2)))
	   (setq L*2 (apply-substitution theta (remove-one tm* L*2)))
	   (list theta L2 l*2))
	  (t (unify-a-term (cdr L) (cdr L*) L2 L*2)) )))

(DEFUN UNION-W (A B)			;a version of UNION

       (COND ((NULL B) A)
	     ((MEMBER (CAR B) A :test #'equal) (UNION-W A (CDR B)))
	     (T (UNION-W (APPEND A (LIST (CAR B))) (CDR B)))))

