
;;                   high-low2.lisp                           2Sept94

;;        Renamed from high-low.lisp  12 Oct 94
;;        See high-low.lisp for most recent version

#|
Basic Idea:  We wish to handle hard fol theorems, using both 
a "high" level routine and a "low" level one.
The low one (otter for now) handle most fol theorems very
well but has trouble when (1) there are too many hypotheses,
(2) when the theorem is just a construction of "big hunks" that
a person sees easily, but the "low" does not, (3) when there
is a large KB of lemmas given from which it has to choose
(and extention of (1) and (2)).  

See FOL-BC.OTTER.LISP for the first step on this mission.  
See experiments-hl.lisp, for the experimentating with HIGH-LOW

But we will add to that some basic notions:
(1) we will separate the KB of lemmas from the other hypotheses,
(2) use an "image" for each lemma to help focus on the one that
is needed for the theorem being proved, (3) and evaluator, to
help determine the value of the selected lemmas (or subgoals).
The evaluator might use (abstraction&otter) and/or abduction.  

Loading:  

(load "imply.lisp")(ls)(lu)(lp)(le)                               ;(lplns)
(lres)(lpi)(lpci)(lop)(lsg)(lpbr)(lppc)(lpc)(lpbe)(lpLe)

(load "fol-bc.lisp")
(load "fol-bc-otter.lisp") (load "ground-resolution.lisp")
(load "experiments-hl.lisp")
(load "high-low.lisp")
|#

(defun lf  () (load "fol-bc.lisp"))
(defun lfo () (load "fol-bc-otter.lisp"))
(defun lhl () (load "high-low.lisp"))
(defun g* (&optional a) (setq *grind-version* (cond (a t) (t nil))))

(defvar %theorem-variables% nil)
(defvar *n* nil)
(defvar *V* nil)
(defvar *repeats* nil)
(defvar *repeats-limit* nil)
(setq   *repeats-limit* 40)
(defvar *lemmas* nil)
(defvar *MINIMAL-IMAGE-MATCH* nil)
(defvar *predicate-list* 
 '(some all or & not clsd-rec clsd-int open finite <<= <= << < e subset))
(defvar *grind-version* nil)
(defvar *grind*-version* nil)
(defvar *matrix-version* nil)
(setq *gv* *grind*-version*)
(defvar *LSS* nil)
(defvar *theta* nil)

(defun prove-HL (Thm &optional H Lemmas copies Lm) ;Main Call  2Sept94
  (let (Thm* thm** thm*** H* W lemmas* number-of-lemmas pv nl W)
    (setq %theorem-variables% nil)
    (setq *n* 20)
    (setq *repeats* 0)
    (setq *MINIMAL-IMAGE-MATCH* 1)
    (cond ((eq t Lemmas) (setq Lemmas nil)))   ;4 Oct 94
    (setq thm* (imp-on H thm))
    (cond ((and lemmas (eq-car lemmas 'exs-from-to)) 
	   (setq lemmas (exs-from-to (cadr lemmas) (caddr lemmas)
				     (cadddr lemmas) (fifth lemmas)))))
    (and copies (setq lemmas (union-n lemmas copies)))
    (and lm (setq lemmas (append (remove lm lemmas :test 'equal) (list lm))))
    (setq number-of-lemmas (length lemmas)) (prnt number-of-lemmas)
    (setq lemmas* (mapcar #'(lambda (u) (cond ((atom u) (eval u)) (t u)))
			  lemmas))
    (setq lemmas* (mapcar #'replace-multiple-quantifiers lemmas*))

    (setq thm** (imp-on (andify lemmas*) thm*))
    (setq thm** (sublis '((< . <<) (<= . <<=)) thm**))
;;    (setq thm** (reducef (unique-ize-quan-vars thm**)))   ;8 Oct 94
    (setq thm** (unique-ize-quan-vars thm**))               ;8 Oct 94
    (setq *V* (unifiable-variables Thm**))
    (setq pv (predicate-symbols thm**))    ;15Sept94
    (setq pv (set-difference pv *predicate-list*))
    (setq *predicate-list* (append pv *predicate-list*))
    (cond (lemmas (setq thm* (caddr thm**))
		  (setq *lemmas* (un-andify (cadr thm**))))
	  (t (setq thm* thm**) (setq *lemmas* nil)))
    (cond ((eq-car Thm* '->) 
	   (setq H* (cadr Thm*)) (setq Thm* (caddr Thm*))))
    (cond ((eq-car Thm* '->) 
	   (setq H* (and-on (cadr Thm*) H*)) (setq Thm* (caddr Thm*))))
    (setq H* (forward-chain-on H*))                        ;8 Oct 94
    (setq *lemmas* (subst '<< '< (subst '<<= '<= *lemmas*)))
    (cond (*grind*-version* (setq *grind-version* t)))
    (cond (*matrix-version* (setq *grind*-version* t)
			    (setq *grind-version* t)))

    ;;%%%%%%%%%%%%% Which Version of HL? %%%%%%%%%%%%%%
    (cond (*matrix-version* 
	   (cond ((eq '[] (setq W (HLm H* thm*))) '[])
		 ((eq t (cadr W)) '[])))
          (*grind-version*
	   (setq *LSS* (mapcar #'(lambda (Lm) (clausify-gr lm t))
			       *lemmas*))
	   (HLg H* Thm*))
	  (t (setq *lemmas* (mapcar #'(lambda (u) (list u (image u))) 
				    *lemmas*))
	   (HL H* Thm*))) ))

(defun HL (H C &optional reject (repeat 1))                  ;2Sept94
(prnt 'HL)
  (let (image W P path* theta theta* reject* H* C* ans)
    (cond ((and (member repeat '(1 2)) 
		(setq ans (time (otter-prove C
		    :inference-rules '(binary_res para_into para_from)
		    :time 2 :factor t :base-axioms H)))    ;Proved by otter
		(eq-car ans t)) 
	   (return-from HL t)))
;;$$ Evaluate here
    (cond ((< *repeats-limit* (setq *repeats* (1+ *repeats*))) 
	   (format t "~%Failed because of two many repeats~%") 
	   (return-from HL nil)))
    (setq image (image C))

    (cond ((and (member repeat '(1 2)) 
		(setq W (bcc-HL H C image reject))) ;W=(theta P path H*)
	   (setq theta (car W)) (setq P (cadr W)) (setq path* (caddr W))
	   (cond ((eq t P) W)
		 ((eq t (setq W (HL H P reject 2))) t)
		 (t (setq theta* (compose (car W) theta))  ;W=(theta P path H*)
		    (cons theta* (cdr W)))))
	  ((and (setq reject* (cons path* reject))
		(setq W (HL H C reject* 3))) W)
	  ((and (member repeat '(3)) 
		(setq ans (time (otter-prove C
		     :inference-rules '(binary_res para_into para_from)
		     :time 3600 :factor t :base-axioms H)))
		(eq ans t)) ans)
	  (t nil) )))

(defun HLg (H C &optional reject (n 0))                       ;22Sept94
(prnt 'HLg)
  (let (ans W W2 theta P path* theta* reject* HS CS CS*)
    (cond ((< 3 (setq n (1+ n))) (format t "~%Too Deep in HLg~%") 
                       	         (return-from HLg nil))
	  ((and (setq ans (time (otter-prove C
	               :inference-rules '(binary_res para_into para_from)
		       :time 2 :factor t :base-axioms H))) 
		(eq-car ans t))                          ;Proved by otter
	   (return-from HLg '[])))
    (cond ((and (setq HS (clausify-gr H t))
		(setq CS (clausify-gr C))
		(eq '[] (setq CS* (grind HS CS))))
	   (cond ((setq W (bcc-HL H C t reject)) ;W=(theta P path H*)
		  (setq theta (car W)) (setq P (cadr W)) (setq path* (caddr W))
		  (cond ((eq t P) W)
			((eq t (setq W2 (HLg H P reject n))) '[])
			((eq '[] (setq W2 (HLg H P reject n))) '[])
			(W2 (setq theta* (compose (car W2) theta)) 
                                                           ;W=(theta P path H*)
			    (cons theta* (cdr W2)))
			((and W (not W2) (setq reject* (cons path* reject))
			      (setq W (HLg H C reject* n))) W)))
		 ;;% Lemmas needed >= 1,2
		 ((or (HLgl H C HS CS* reject n 1)
		      (HLgl H C HS CS* reject n 2))
		  '[])
		 ((and (setq ans (time (otter-prove C
			    :inference-rules '(binary_res para_into para_from)
			    :time 1000 :factor t :base-axioms H))) 
		       (eq-car ans t))	;Proved by otter
		  '[])
		 (t nil)))

	  ;;% Lemmas needed >= 1,2
	  ((or (HLgl H C HS CS* reject n 1)
	       (HLgl H C HS CS* reject n 2))
	   '[])
	  ((and (setq ans (time (otter-prove C
			    :inference-rules '(binary_res para_into para_from)
			    :time 1000 :factor t :base-axioms H))) 
		       (eq-car ans t))	;Proved by otter
	   '[])
	  (t nil) )))

(defun HLm (H C &optional reject (n 0))                       ;26Sept94
(prnt 'HLm)
  (let (ans W W2 theta theta* theta** thetaf P path* reject* HS CS CS*)
    (cond ((< 10 (setq n (1+ n))) (format t "~%Too Deep in HLm~%") 
                       	         (return-from HLm nil))
	  ((and (setq ans (time (otter-prove C  
	               :inference-rules '(binary_res para_into para_from)
		       :time 50 :factor t :base-axioms H))) 
		(eq-car ans t))                          ;Proved by otter
	   (return-from HLm '[])))
    (setq HS (clausify-matrix H t)) 
    (setq CS (clausify-matrix C))
    (prnt 1)
    (setq *theta* nil)
    (cond ((and (eq '[] (setq CS* (grind-matrix HS CS)))
		(or (setq theta** *theta*) t))
	   (setq *repeats* 0)
	   (setq C (demodu H C))	;4 Oct 94
	   (cond ((setq W (bcc-HL H C t reject)) ;W=(theta P path H*)
		  (setq theta (car W)) (setq P (cadr W)) (setq path* (caddr W))
		  (cond ((eq t P) W)
			((eq t (setq W2 (HLm H P reject n))) '[])
			((eq '[] W2) '[])
			(W2 (setq theta* (compose (car W2) theta)) 
					;W=(theta P path H*)
			    (cons theta* (cdr W2)))
			((and W (not W2) (setq reject* (cons path* reject))
			      (setq W (HLm H C reject* n))) W)))
#|
		 ((and (or (format t "~%Warning. Grind=[] but Bcc-HL failed~%")
			   (break) t)
		       (setq theta2 (double-units theta**))
		       (setq thetaf (block-matrix HS CS theta2))
		       (setq HS (apply-sub thetaf HS))
		       (setq CS (apply-sub thetaf CS)) nil))
|#
                 ((and *lemmas* (HLml2 H C HS CS reject n 'm[])) '[]) ;28Sept94

		 ((and (setq ans (time (otter-prove C
			    :inference-rules '(binary_res para_into para_from)
			    :time 1000 :factor t :base-axioms H))) 
		       (eq-car ans t))	;Proved by otter
		  '[])
		 (t nil)))

          ((and *lemmas* (HLml2 H C HS CS reject n)) '[]) ;28Sept94

	  ((and (setq ans (time (otter-prove C
			  :inference-rules '(binary_res para_into para_from)
			  :time 1000 :factor t :base-axioms H))) 
		(eq-car ans t))		;Proved by otter
	   '[])
	  (t nil) )))

(defun HLgl (H C HS CS reject n depth-of-lemmas)  ;22Sept94  Called by HLg
  (let (LSS1 LSS2 LSS2* SS H* SS m)
    (cond ((equal 1 depth-of-lemmas) 
	   (setq *repeats* 0)
	   (cond ((and (setq LSS1 (mapcan #'(lambda (LS Lm) 
                              (setq SS (union-w HS LS))
                              (cond ((eq '[] (grind SS CS)) (list Lm))))
			      *LSS* *lemmas*))
		       ;;later, order LSS1 according to functions, etc
		       (some #'(lambda (lm) (and (setq H* (and-on lm H))
					  (eq '[] (HLg H* C reject n))))
                                                            ;change reject??
			     LSS1))
		  '[])
		 (t nil)))

	  ((equal 2 depth-of-lemmas) 
	   (setq *repeats* 0)
	   (setq LSS2 (mapcan #'(lambda (LS Lm) 
                (setq SS (union-w HS LS))
                (cond ((eq '[] (setq CS* (grind SS CS))) nil)  ;failed already
		      ((< (setq m (measure CS CS*)) 10) nil)   ;10??
		      (t (list (list m Lm)))))
			      *LSS* *lemmas*))
	   (setq LSS2 (sort LSS2 'HLgl-sort-function))
	   (setq LSS2* (mapcar #'cadr LSS2))
	   (cond ((some #'(lambda (lm) (and (setq H* (and-on lm H))
					  (eq '[] (HLg H* C reject n))))
                                                            ;change reject??
			     LSS2*)
		  '[])
		 (t nil))) )))

(defun HLml (H C HS CS reject n depth-of-lemmas)  ;26Sept94  Called by HLm
  (let (LSS1 LSS2 LSS2* SS H* SS m)
    (cond ((equal 1 depth-of-lemmas) 
	   (setq *repeats* 0)
	   (cond ((and (setq LSS1 (mapcan #'(lambda (LS Lm) 
                              (setq SS (and-on LS HS))
			      (or (prnt 2) t)
                              (cond ((eq-car (grind-matrix SS CS) '[]) 
				     (prnt '[] Lm)
				     (list Lm))))
			      *LSS* *lemmas*))
		       (setq LSS1 (sort-HLml LSS1 C))         ;27Sept94
		       (some #'(lambda (lm) (and (setq H* (and-on lm H))
					  (eq '[] (HLm H* C reject n))))
                                                            ;change reject??
			     LSS1))
		  '[])
		 (t nil)))

	  ((equal 2 depth-of-lemmas) 
	   (setq *repeats* 0)
	   (setq LSS2 (mapcan #'(lambda (LS Lm) 
                (setq SS (and-on LS HS))
(or (prnt 3) t)
                (cond ((eq-car (setq CS* (grind-matrix SS CS)) '[]) nil)  
		      ((< (setq m (measure-matrix (and-on CS SS) CS*)) 1) nil)
		      (t (list (list m Lm)))))
			      *LSS* *lemmas*))
	   (setq LSS2 (sort LSS2 'HLgl-sort-function))  ;not defined yet
	   (setq LSS2* (mapcar #'cadr LSS2))
	   (cond ((some #'(lambda (lm) (and (setq H* (and-on lm H))
					  (eq '[] (HLm H* C reject n))))
                                                            ;change reject??
			     LSS2*)
		  '[])
		 (t nil))) )))

(defun HLml2 (H C HS CS reject n &optional m[])   ;28Sept94  Called by HLml
  (let (lemmas LS SS LSS2 LSS2* CS* H* m nn ans lm*)
(prnt 'HLml2)
    (setq lemmas (sort-HLml *lemmas* C))
    (setq lemmas (first-n 10 lemmas))                  ;10?
    (cond ((some #'(lambda (lm) 
                     (and (or (prnt 2) t)
			  (or (and m[] (or (prnt m[]) t))
			      (and (setq LS (clausify-matrix lm t))
				   (setq SS (and-on LS HS))
				   (eq '[] (grind-matrix SS CS))
				   (or (prnt lm '[]) t)))
			  (setq H* (and-on lm H))
			  (eq '[] (HLm H* C reject n))))
		 lemmas)
	   '[])
	  ;;%%%%%%% NOTE %%%%  1Oct94
          ((and (< (length *lemmas*) 4)
		(or (setq H* (and-on (andify *lemmas*) H)) t)
		(or (prnt '4) t)
		(eq '[] (HLm H* C reject n)))
	   '[])

	  ;;%%%%%%% NOTE %%%%  10Oct94  This OUTLAWS the next option
          ((and	(or (setq H* (and-on (andify *lemmas*) H)) t)
		(or (prnt '5) t)
		(eq '[] (HLm H* C reject n)))
	   '[])

	  (t (setq LSS2 (mapcan #'(lambda (Lm) 
		(or (prnt 3) t)
		(setq LS (clausify-matrix lm t))
                (setq SS (and-on LS HS))
                (cond ((eq '[] (setq nn (grind-matrix SS CS))) nil)
		      ((< (setq m (measure-matrix (and-on CS SS) nn)) 1) nil)
                       ;;This has to be changed  28,29Sept94
		      (t (prnt Lm nn) (list (list m Lm))) ))
				lemmas))
	     (setq LSS2 (sort LSS2 'HLml-sort-function)) 
	     (setq LSS2* (mapcar #'cadr LSS2))
	     (cond ((some #'(lambda (lm)
                      (and (setq H* (and-on lm H))
			   (eq '[] (HLm H* C reject n))   ;change reject??
			   (setq *lemmas* 
				 (append (remove lm *lemmas* :test 'equal)
					 (list lm)))  ;put lm last  1Oct94
			   ))
			  LSS2*)
		    '[])
		   (t nil))) )))
	  
(defun bcc-HL (H C image reject &optional path VC)  ;2Sept94
  ;;  Called by HL
(prnt 'bcc-HL)
  (let (path-bch path-h path-hh path-c1 W W2 c# theta theta2 P P2  H* C* CL 
		 reject* xs x Cii image-i)

    (cond ((< *repeats-limit* (setq *repeats* (1+ *repeats*))) 
	   (format t "~%Failed because of too many repeats~%") 
	   (return-from bcc-HL nil))
	  ((member path reject :test 'equal) (prnt path)
	   (return-from bcc-HL nil)))

    (setq path-bch (cons 'bch-HL  path))
    (cond ((member path-bch reject :test 'equal) (Prnt path-bch) nil)
                                                          ;skip next line
	  ((not (setq W (bch-HL H C image reject path-bch VC))) 
	   (setq reject (cons path-bch reject))))         ;reject next time
    (cond (W W)                                        ;W= (theta P path-h H*)
	  
	  ((atom C) nil)
	  
	  ((eq-car C '&)
	   (setq c# 0)
	   (cond ((some #'(lambda (Ci) 
                            (and (setq c# (1+ c#))
				 (setq path-h (append (list c# '&C) path))
				 (setq image-i (image Ci))
				 (setq W(bcc-HL H Ci image-i reject path-h VC))
				 (setq Cii Ci)))
			(cdr C))                        ;W= (theta P path-h H*)
		  (setq theta (car W)) (setq P (cadr W)) 
		  (setq path-hh (caddr W)) (setq H* (cadddr W))
		  (setq CL (remove Cii (cdr C) :test 'equal))
		  (setq CL (mapcar #'(lambda (u) (apply-substitution theta u))
				   CL))
		  (setq CL (set-difference CL (cdr H*) :test 'equal))
		  (cond ((null CL) W)                          ;bc'ed all of C
			((setq W2 (bcc-HL H (andify CL) reject path-h VC))
			 (setq theta2 (compose (car W2) theta))   ;rest of C
			 (setq P2 (and-on P (cadr W2)))
			 (cons theta2 (cons P2 (cddr W2))))

			(t (setq reject* (cons path-hh reject))
			   (bcc-HL H C reject* path VC))))
		 (t nil)))

          ((member (setq path-c1 
			 (cons (maknam (list (cond ((atom C) C) (t (car C)))
					     'C)) path))
		   reject :test 'equal) (prnt 'c-name) nil)

	  ((and (eq-car C '->) 
		(setq image* (image (caddr C)))
		(setq W (bcc-HL (and-on (cadr C) H) (caddr C) image* reject 
			      (cons '->C path) VC)))
	   (setq P2 (imp-on (cadr c) (cadr W)))
	   (put-in P2 2 W))
			
	  ((eq-car C 'or) 
	   (or (bcc-HL (and-on (not-fcnf (rest-&or C)) H) (cadr C)
		       (image (cadr C)) reject (cons 'orC1 path) VC)
	       (bcc-HL (and-on (not-fcnf (cadr C)) H) (rest-&or C)
		       (image (rest-&or C)) reject (cons 'orC2 path) VC)))

	  ((eq-car C 'not) nil)                           ;later

	  ((eq-car C 'some)
	   (bcc-HL H (caddr C) (image (caddr C)) reject 
		     (cons 'SomeC path) (cons (cadr C) VC)))

	  ((eq-car C 'all)
	   (setq xs (make-skolem-expresn (setq x (cadr C)) VC))   ;(xs ...)
	   (setq C* (subst xs x (caddr C) :test 'equal))
	   (bcc-HL H C* (image C*) reject (cons 'Call path) VC))

	  )))

(defun bch-HL (H C image reject path VC)  ;3Sept94  Called by bcc-HL
  (let (W)
;;(prnt 'bch-HL)
    (cond ((member path reject :test 'equal) (Prnt path) 
	   (return-from bch-HL nil)))

    (cond ((setq W (bch H C image reject (cons 'bch path) VC)) W)
	  ((null *lemmas*) nil)
	  (*grind-version* nil)   ;22Sept94
	  ((setq W (bcL H C image reject (cons 'bcL path) VC)) W) 

	  )))

(defun bch (H C image reject path VC &optional VH (P t) H*) 
  ;3Sept94 Called by bch-HL
;;(prnt 'bch)
  (let (path-h1 path-u theta P* H** h# path-h W xs x (V (union-w VC VH)) )

    (cond ((member path reject :test 'equal) (Prnt path) 
	   (return-from bch nil)))

    (setq path-u (cons 'u  path))
    (cond ((member path-u reject :test 'equal) (Prnt path-u) nil)
                                                              ;skip next line
	  ((not (setq theta (unify-v H C V))) 
	   (setq reject (cons path-u reject))))               ;reject next time

    (cond (theta (setq P*  (apply-sub-hb theta P))
		 (setq H** (apply-sub-hb theta H*))
		 (list theta P* path-u H**))

	  ((atom H) nil)

	  ((eq-car H '&)
	   (setq h# 0)
	   (cond ((some #'(lambda (Hi) 
                           (and (setq h# (1+ h#))
				(setq path-h (append (list h# '&H) path))
				(setq W (bch Hi C image 
					     reject path-h VC VH P H))))
			(cdr H))                    ;W= (theta P path-h H*)
		  W)
		 (t nil)))	  

	  ((member (setq path-h1 (cons (maknam (list (cond ((atom H) H) 
							   (t (car H))) 'H)) 
				       path))
		   reject :test 'equal) (prnt 'h-name) nil)

	  ((eq-car H '->) 
	   (or (bch (caddr H) C image reject (cons '->H1 path) VC VH 
		     (and-on (cadr H) P) H*)
	       (bch (not-fcnf (cadr H)) C image reject (cons '->H2 path) VC VH 
		     (and-on (not-fcnf (caddr H)) P) H*)))

	  ((eq-car H 'or) 
	   (or (bch (cadr H) C image reject (cons 'orH1 path) VC VH
		     (and-on (not-fcnf (rest-&or H)) P) H*)
	       (bch (rest-&or H) C image reject (cons 'orH2 path) VC VH
		     (and-on (not-fcnf (cadr H)) P) H*)))

	  ((eq-car H 'all)
	   (bch (caddr H) C image reject (cons 'allH path) VC 
		 (cons (cadr H) VH) P H*))

	  ((eq-car H 'some)
	   (setq xs (make-skolem-expresn (setq x (cadr H)) VH))   ;(xs ...)
	   (setq H** (subst xs x (caddr H) :test 'equal))
	   (bch H** C image reject (cons 'someH path) VC VH P H*))

	  ((eq-car H 'not) nil)                            ;later

	  (t nil) )))

(defun bcL (H C image reject path VC &optional VH P H*) 
  ;;3Sept94 Called by bch-HL.  Not called if *lemmas* is nil
  ;;                                or if *grind-version* is t
  (let (images lemmas*)
(prnt 'bcL) 
    ;;NOTE: each lm of *lemmas* has the form (lemma image)
    (setq images (mapcar #'(lambda (lm) 
                   (list (car lm) (match-images image (cadr lm))))
			 *lemmas*))
    (setq images  (mapcan #'(lambda (lm)   ;prune the poor ones
                    (cond ((< *minimal-image-match* (cadr lm)) (list lm))))
			  images))
    (setq lemmas* (sort images #'(lambda (x y) (< (cadr x) (cadr y)))))
    (setq lemmas* (mapcar #'car lemmas*))
    (bch (andify lemmas*) C image reject (cons 'bch path) VC)))

(defun apply-sub-hb (theta C) ;10Aug94  Called by PR, BC
  (let (C*)
    (cond ((eq t theta) C)
	  ((atom C) (sublis theta C))
	  ((member (car C) '(all some))
	   (setq C* (apply-sub-hb theta (caddr C)))
	   (cond ((equal (cadr C) (sublis theta (cadr C)))
		  (subst C* (caddr C) C :test 'equal))
		 (t C*)))
	  (t (cons (car C) (mapcar #'(lambda (u) (apply-sub-hb theta u))
				   (cdr C)))) )))

(defun before (x y L)  ;14Sept94  Is x before y in list L?
  (<= (length (member y L)) (length (member x L))))

(defun block-matrix (HS CS theta &optional theta2)   ;30Sept94
  (let (HS* CS* u x y)
    (cond ((null theta) 
	   (format t "~%A Problem in BLOCK-MATRIX~%") (break) nil)
	  ((and (setq u (car theta)) (setq x (car u)) (setq y (cdr u))
		(setq theta2 (cons u theta2))
		(setq HS* (subst y x HS)) (setq CS* (subst y x CS)) nil))
	  ((eq '[] (grind-matrix HS* CS*))
	   (block-matrix HS* CS* (cdr theta) theta2))
	  (t theta2))))

(defun combine-thetas (addrs) ;10Aug94  Called by PR, BC
  (let (theta)
    (setq theta t)
    (mapc #'(lambda (u) (setq theta (compose (car u) theta))) addrs)))

(defun crunch-hl (L) ;SEE FOLLOWING .widely used. This version does not recurse
  ;;removes duplication from a list
  (let (L*)
    (setq L* L)
    (mapcan #'(lambda (x) (setq L* (cdr L*))
               (cond ((member x L* :test 'equal) nil)
		     (t (list x)))) 
	    L)))

(defun crunch-hl (L)   ; 11 Oct 94 Widely used. This version does not recurse
  ;;removes duplication from a list
  (let (L*)
    (setq L* nil)
    (mapc #'(lambda (x) (cond ((member x L* :test 'equal) nil)
			      (t (setq L* (cons x L*)))))
	  L)
    (reverse L*)))

(defun crunch-inside (A)  ;11 Oct 94
  ;;removes duplications from (& ...) and (or ...) inside A
  (cond ((atom A) A)
	((eq-car A '&)  (andify (crunch-hl (mapcar #'crunch-inside (cdr A)))))
	((eq-car A 'or) (orify  (crunch-hl (mapcar #'crunch-inside (cdr A)))))
	(t (cons (crunch-inside (car A))
		 (crunch-inside (cdr A))))))

(defun dnf (A)   ;14Sept94  Called by STRUCTURE-HL
  ;;Use CNF
  ;;Ie, (dnf '(& (or A B) C)) =  (or (& A C) (& B C))
    (setq A (switch '& 'or A))   ;switch & for or in A
    (setq A (cnf A))
    (setq A (switch '& 'or A)))

(defun double-units (theta)   ;30Sept94
  (let (u2)
    (cond ((null (cdr theta)) nil)
	  ((some #'(lambda (u) (and (eq (caar theta) (car u))
				    (setq u2 u)))
		 (cdr theta))
	   (setq theta* (remove u2 (cdr theta) :test 'equal))
	   (cons (car theta) (double-units theta*)))
	(t (double-units (cdr theta))) )))

(defun equal* (A B)  ;23Sept94
  ;;Same as EQUAL except that SOME and ALL are forced to have common variables,
  ;; and x=v if v is in *V*, for any x.  *V* is universal.
  ;; *V* is the set of all unifiable variables for the theorem being proved.
  (cond ((equal A B) t)
	((member A *V*) (setq *nm* (1+ *nm*)) 
	 (setq *theta* (cons (cons A B) *theta*)) t)
	((member B *V*) (setq *nm* (1+ *nm*)) 
	 (setq *theta* (cons (cons B A) *theta*)) t)
	((atom A) nil)
	((atom B) nil)
	((not (eq (car A) (car B))) nil)
	((member (car A) '(SOME ALL))
	 (equal* (caddr A) (subst (cadr A) (cadr B) (caddr B))))
	(t (every #'(lambda (u v) (equal* u v)) (cdr A) (cdr B))) ))

(defun equal-sa (A B) ;8 Oct 94   Called by FCOH
  (cond ((equal A B) t)
	((or (atom A) (atom B)) nil)
	((not (eq (car A) (car B))) nil)
	((and (member (car A) '(some all)) 
	      (equal A (subst (cadr A) (cadr B) B :test 'equal))) t)
	((every #'(lambda (x y) (equal-sa x y)) (cdr A) (cdr B)) t)
	(t nil)))

(defun exs-from-to (m n &optional j k)  ;13Sept94  Called by prove-HL
  ;;Return the list of formulas, exm to exn from file example-theorems.lisp
  (let (ex exx)
    (cond ((or (null n) (< n m))
	   (cond (j (exs-from-to j k)) (t nil)))
	  (t (setq ex (maknam (list 'ex m)))
	     (setq exx (eval ex))
	     (cond (exx (cons exx (exs-from-to (1+ m) n j k)))
		   (t (exs-from-to (1+ m) n j k)))) )))

(defun eval-pf (Lpf Cpf)   ;27Sept94  Called by SORT-HLml
  ;;Lpf = (Lp Lf), Cpf= (Cp Cf)
  (+ (length (set-intersection (car  Lpf) (car  Cpf)))
     (length (set-intersection (cadr Lpf) (cadr Cpf)))))

(defun forward-chain-on (H)  ;8 Oct 94  Called by Prove-HL
  ;;eg, (forward-chain-on '(& A (-> A (& B C)))) = (A (& B C))
  ;;    (forward-chain-on '(& A (-> (& D A) (& B C)))) = (A (& B C))
  (let (H*)
    (cond ((not (eq-car H '&)) H)
	  ((equal H (setq H* (andify (fcoh (cdr H))))) H)
	  (t (forward-chain-on H*)) )))

(defun fcoh (HL)              ;8 Oct 94  Called by FORWARD-CHAIN-ON
  (let (HL1 CV)
    (setq HL1 HL)
    (mapc #'(lambda (u) 
             (setq HL1 (mapcar #'(lambda (v) 
		   (cond ((eq-car v '->)
			  (cond ((equal-sa u (cadr v)) (caddr v))
				((and (eq-car (cadr v) '&)
				      (member u (cadr v) :test 'equal-sa))
				 (setq CV (remove u (cdadr v) :test 'equal-sa))
				 (imp-on (andify CV) (caddr v)))
				(t v)))
			 (t v)))
			       HL1)))
	  HL)
    HL1))
	  
(defun function-symbols (A)  ;27Sept94 Called by PROVE-HL
  (let (pv fv)
    (setq pv (predicate-symbols A))
    (setq fv (crunch-hl (fvi A)))
    (set-difference (set-difference fv '(-> & or not)) pv)))

(defun fvi (A)           ;27Sept94  Called by FUNCTION-SYMBOLS
  (cond ((atom A) nil)
	((member (car A) '(some all)) (fvi (caddr A)))
	(t (cons (car A) (mapcan #'fvi (cdr A)))) ))

(defun HLml-sort-function (x y) x)  ;12Sept94  This has to be changed

(defun image (A)    ;14Sept94  Called by prove-HL, etc
  ;;We use only STRUCTURE-HL now, and add later, the number of atoms,
  ;; the non-logical predicates, the functions, etc.
  (cond (*grind-version* t)    ;22Sept94
	(t (structure-HL A))))

(defun make-skolem-expresn (x V)  ;12Aug94
  (cond (V (cons (new-symbol (maknam  (list x 's))) V))
	(t x)))

(defun match-images (Im1 Im2)  ;3Sept94  Called by bcL
  ;;Im1 and Im2 are images of two formulas.  This compares them.
  ;;Im1 is the image of a goal, C. Im2 is the image of a "lemma".
  ;;0 means nothing in common.  The better the match the higher the output.
  ;;14Sept94  We are using only STRUCTURE-HL.  
  ;;20Sept94  If they are eaual, then 10.
  ;;          If Im1 is equal to a prime sub-structure of Im2, then 7
  (cond ((equal Im1 Im2) 10) 
	((and (eq-car Im2 'or) (member Im1 (cdr Im2) :test 'equal)) 7)
	(t 0)))

(defun measure (CS CS*)      ;22Sept94  Called by HLgl
  ;;CS and CS* are sets of clauses, 
  ;;This returns the (difference in the number of literals)/(total number)
  (let (ncs ncs*)
    (setq ncs 0) (setq ncs* 0) 
    (mapc #'(lambda (C) (setq ncs  (+ ncs  (length C)))) CS)
    (mapc #'(lambda (C) (setq ncs* (+ ncs* (length C)))) CS*)
    (* 10 (/ (- ncs ncs*) ncs)) ))

(defun measure-matrix (S1 n) ;28Sept94
  ;;S1 is a set of literals and
  ;;  is the number of literals removed by grind-matrix
  ;;This returns (n/nl)*10 where nl is the number of literals in S1
  (let (nl)
    (setq nl (length (literals-in S1)))
    (* 10 (/ n nl)) ))

(defun not-fcnf (A)  ;23Aug94  Called by PC, BCC-HL, BCH-HL
  (reducef (not-fcn A)))

(defun or-start-in (H) ;24Aug94  Called by PR
  ;;eg, (or-start-in '(or A B)) = t
  ;;    (or-start-in '(-> P (or A B))) = t
  (let () 
    (cond ((eq-car H 'or) t)
	  ((atom H) nil)
	  ((eq-car H '->) (or-start-in (caddr H)))
	  ((member (car H) '(all some)) (or-start-in (caddr H)))
	  ((eq-car H '&) (some #'or-start-in (cdr H)))
	  (t nil) )))

(defun predicate-symbols (A)  ;15Sept94 Called by PROVE-HL
  (crunch-hl (psy A)))   

(defun psy (A)   ;15Sept94  Called by PREDICATE-SYMBOLS
  (cond ((atom A) (list A))
	((member (car A) '(some all)) (psy (caddr A)))
	((member (car A) '(-> & or not)) (mapcan #'psy (cdr A)))
	(t (list (car A))) ))

(defun put-in (P n W)  ;29aug94   Called by BCC-HL
  (cond ((atom W) W)
	((= n 1) (cons P (cdr W)))
	(t (cons (car W) (put-in P (1- n) (cdr W))))))

(defun q-subst (Nu A AA) ;2Aug94   Just for tracing
  (q2-subst Hu A AA))

(defun q2-subst (Nu A AA) ;2Aug94  
  (let (theta)
    (cond ((setq theta (instance AA A))
	   (setq (cond ((eq t theta) t)
		       (t (mapcar #'(lambda (u) (cons (cdr u) (car u)))
				  theta))))
	   (apply-sub theta Hu))
	  ((atom AA) AA)
	  ((member (car A) '(-> & or not all some))
	   (cons (car A) (mapcar #'(lambda (u) (q2-subst Nu A u))))) )))

(defun reducef (A)  ;23Aug94  Called by PROVE
  (let (A*)
    (cond ((equal A (setq A* (redf A))) A)
	  (t (reducef A*)))))

(defun redf (A)     ;23Aug94  Called by REDUCEF
  (let ()           ;need parity?
    (cond ((atom A) A)
	  ((and (eq-car A 'not) (eq-car (cadr A) 'not))   ;(not (not A)) ->> A
	   (cadadr A))
	  ((and (eq-car A 'not) (eq-car (cadr A) 'all))   ;(not (All x B))
	   (list 'some (cadadr A) (not-fcnf (caddr (cadr A)))))  ;= (some x ~B)
	  ((and (eq-car A 'not) (eq-car (cadr A) 'some))  ;(not (Some x B))
	   (list 'all (cadadr A)  (not-fcnf (caddr (cadr A)))))  ;= (all x ~B)
	  ((and (eq-car A 'not) (eq-car (cadr A) '->))
           (list '& (cadr A) (not-fcnf (caddr (cadr A)))))
	  ((and (eq-car A 'not) (eq-car (cadr A) '&))
	   (cons 'or (mapcar #'not-fcnf (cdadr A))))
	  ((and (eq-car A 'not) (eq-car (cadr A) 'or))
	   (cons '& (mapcar #'not-fcnf (cdadr A))))
	  ((member (car A) '(some all))
	   (list (car A) (cadr A) (redf (caddr A))))
	  ((member (car A) '(& or -> not))
	   (cons (car A) (mapcar #'redf (cdr A))))
	  (t A))))

(defun reduce&or (A)  ;15Sept94
  (let (u2 A2)
    (cond ((atom A) A)
	  ((and (member (car A) '(& or))   ;(& a (& b c) d) ->> (& a b c d)
		(some #'(lambda (u) (and (eq-car u (car A)) (setq u2 u)))
		      (cdr A)))
	   (setq A2 (cons (car A) (mapcan #'(lambda (u) 
                            (cond ((equal u u2) (cdr u)) (t (list u))))
					  (cdr A))))
	   (reduce&or A2))
	  ((member (car A) '(-> & or not some all))
	   (cond ((equal A (setq A2 (cons (car A) 
					  (mapcar #'reduce&or (cdr A)))))
		  A)
		 (t (reduce&or A2))))
	  (t A))))

(defun reject-last-unify (path)  ;12Aug94
  ;;path starts reversed.  We will change the first (u theta) to (u reject)
  ;; and reverse
  (cond ((null path) nil)
	((eq-car (car path) 'u) (reverse (cons '(u reject) (cdr path))))
	(t (reject-last-unify (cdr path)))))

(defun rest-&or (A)  ;12Aug94  
  ;;eg, (rest-&or '(& a b c)) = (& b c), (rest-&or '(& a b)) = b.
  (cond ((or (not (cddr A)) (not (member (car A) '(& or))))
	 (format t "~%Failed in rest-&or~%") (break) nil)
	((null (cdddr A)) (caddr A))
	(t (cons (car A) (cddr A))) ))

(defun sko-free-vars (A)
  (let ((V (free-variables-in A)))
    (mapc #'(lambda (x) (setq A (subst (list x) x A :test 'equal))) V)
    A))

(defun sort-HLml (LSS C)  ;27Sept94  Called by HLml
  ;;Sort the member (lemmas) in LSS according to their closeness to C
  (let (pC fC Lpf)
    (setq pC (predicate-symbols C))
    (setq fC (function-symbols  C))
    (setq Lpf (mapcar #'(lambda (lm) (list lm (predicate-symbols lm)
					      (function-symbols  lm)))
		      LSS))
    (setq Lpf (mapcar #'(lambda (u) 
                  (list (car u) (eval-pf (cdr u) (list pC fC))))
		      Lpf))
    (sort Lpf #'(lambda (x y) (< (cadr x) (cadr y))))
    (reverse (mapcar #'car Lpf)) ))

;;(defun sort-HLml (LSS C) LSS)   ;not used

(defun structure-HL (A)   ;13Sept94  A part of IMAGE  (see 12Sep94 notes)
  ;;eg, (structure-HL '(-> (& (all x (-> (P x) (Q x))) (P b)) (Q b)))
  ;;    =                     (OR (SOME (& (NOT Q) P)) (NOT P) Q)
  (let ()           
    (setq A (convert-implication-signs A))
    (setq A (push-nots-in A))
    (setq A (dnf A))  ;overall and within ALL and SOME
    (struc (reduce&or A))))

(defun struc (A)  ;13Sept94  Called by STRUCTURE
  (let () 
    (cond ((atom A) A)
	  ((member (car A) '(all some))
;;	   (list (car A) (struc (caddr A)))    ;20Sept94  Try the next line
           (struc (caddr A))                   ;20Sept94  Instead of last line
           )
	  ((eq-car A 'not) (list 'not (struc (cadr A))))
	  ((member (car A) '(or &))
	   (cons (car A) (struc-sort (mapcar #'struc (cdr A)))))
	  (t (car A)) )))

(defun struc-sort (L)  ;13Sept94  Called by STRUC
  ;;Why does this not work?   Use the next.
  (sort L 'struc-sort-function))

(defun struc-sort (L)  ;16Sept94  Called by STRUC
  (let (L*) 
    (cond ((null (cdr L)) L)
	  ((and (setq L* (struc-sort (cdr L))) nil))   
	  ((equal (car L) (struc-sort-function (car L) (car L*)))
	   (cons (car L) L*))
	  (t (cons (car L*) (struc-sort (cons (car L) (cdr L*))))) )))

(defun struc-sort-function (x y)  ;16Sept94  Called by STURC-SORT
  (cond ((equal x y) x)
	((atom x)
	 (cond ((atom y) 
		(cond ((before x y *predicate-list*) x) 
		      (t y)))
	       (t y)))
	((atom y) x)
	((equal (car x) (car y))
	 (cond ((equal (cdr x) (struc-sort-function (cdr x) (cdr y))) x)
	       (t y)))
	((equal (car x) (struc-sort-function (car x) (car y))) x)
	(t y) ))
 
(defun switch (x y A)   ;14Sept94 Switch x for y in A
  (setq A (subst 'x* y A))
  (setq A (subst y x A))
  (setq A (subst x 'x* A)) )

(defun un-andify (A) 
  (cond ((eq-car A '&) (cdr A)) (t (list A))))

(defun un-sko-free-vars (A)
  (cond ((atom A) A)
	((and (atom (car A)) (null (cdr A))) (car A))
	(t (cons (car A) (mapcar #'un-sko-free-vars (cdr A))))))

(defun unify-v (A B &optional V)  ;10Aug94  Called by BCH-HL
  (let ()
    (cond ((equal A B) t)
	  ((and (member A *V*) (not (occur A B)) (equal-types A B))
	   (list (cons A B)))                    ;7Nov94
	  ((and (member B *V*) (not (occur B A)) (equal-types A B))
	   (list (cons B A)))                    ;7Nov94
	  ((or (atom A) (atom B)) nil)
	  ((not (eq (car A) (car B))) nil)
	  ((member (car A) '(& or some all)) (unify-v2 (cdr A) (cdr B) V t))
	  (t (unify-v2 (cdr A) (cdr B) V)) )))

(defun unify-v2 (L1 L2 V &optional commute)  ;22Aug94  Called by UNIFY-V
  (let (the1 the2)
    (cond ((not (equal (length L1) (length L2))) nil)
	  ((null L1) t)
	  ((not commute)
	   (cond ((setq the1 (unify-v  (car L1) (car L2) V))
		  (setq L1* (apply-substitution the1 (cdr L1)))
		  (setq L2* (apply-substitution the1 (cdr L2)))
		  (cond ((setq the2 (unify-v2 L1* L2* V))
			 (compose the1 the2))))))
	  (commute
	   (cond ((some #'(lambda (u) (and (setq the1 (unify-v (car L1) u V))
					   (setq u* u)))
			L2)
		  (setq L1* (apply-substitution the1 (cdr L1)))
		  (setq L2* (remove u* L2 :test 'equal))
		  (setq L2* (apply-substitution the1 L2*))
		  (cond ((setq the2 (unify-v2 L1* L2* V t))
			 (compose the1 the2))))))
	  (t nil) )))

(defun unifiable-variables (thm &optional (par t)) ;29aug94 called byPROVE 
  (crunch-hl (un-var thm par))) 

(defun un-var (thm &optional (par t)) 
  ;29aug94 Called by  UNIFIABLE-VARIABLES
  (let ()
    (cond ((atom thm) nil) 	 
	  ((or (and (eq-car thm 'some) par) 	 
	       (and (eq-car thm 'all) (not par))) 	 
	   (cons (cadr thm) (un-var (caddr thm) par)))
	  ((member (car thm) '(all some)) (un-var (caddr thm) par))
	  ((eq-car thm 'not) (un-var (cadr thm) (not par)))
	  ((eq-car thm '->) (union-w (un-var (cadr thm) (not par))
				     (un-var (caddr thm) par)))
	  ((member (car thm) '(& or))
	   (mapcan #'(lambda (u) (un-var u par)) (cdr thm))) )))

(defun union-n (L n) ;20Sept94  unionize n copies of L
  (cond ((<= n 0) nil)
	(t (append L (union-n L (1- n))))))

;;------------------------------------------------------------------

(defun ee (L) 
  (let (W A A*)
    (cond ((null L) nil)
;;	  ((and (or (prnt L) t) (break) nil))
	  ((and (setq A (car L)) (setq A* (eval A)) (prnt A) nil))
	  ((setq W (proveo A*))
	   (prnt W)
	   (ee (cdr L)))
	  (t (prnt 'no)
	     (ee (cdr L))))))
(defun tt (L)                  ;OTTER prove LL in 28Sec
  (let (W A A*)
    (cond ((null L) nil)
;;	  ((and (or (prnt L) t) (break) nil))
	  ((and (setq A (car L)) (setq A* (eval A)) (prnt A) nil))
	  ((setq W (of A*))    ;otter-prove
	   (prnt W)
	   (tt (cdr L)))
	  (t (prnt 'no)
	     (tt (cdr L))))))

(defun rr (L) 
  (let (W A A*)
    (cond ((null L) nil)
;;	  ((and (or (prnt L) t) (break) nil))
	  ((and (setq A (car L)) (setq A* (eval A)) (prnt A) nil))
	  ((setq W (resolution A*))
	   (prnt W)
	   (rr (cdr L)))
	  (t (prnt 'no)
	     (rr (cdr L))))))
(setq LL 
'(EX1 EX2 EX3 EX4 EX6 EX7 EX7* EX9 EX9* EX9** Ex9***
  ex10 ex10f ex11 ex12 ex12* ex13 ex14 ex15 ex16 ex17 ex17* ex17**
  ex18 ex19 ex19* ex20bc ex20A
  ex24 ex24* ex25 ex26 ex27 ex28))

;;---------------------------------------------------------------------

(defun tro  () (trace HL bcc-HL bch-HL bch bcL otter-prove))
(defun trro () (trace HL bcc-HL        otter-prove))
