
                     IMPLY Documentation             ATP85A
                        Woody Bledsoe             21 February 1989


                        Introduction

This paper attempts to document the Theorem Proving Program known as 
IMPLY, which has been used by Bledsoe and his students off and on
for a number years at the University of Texas.  It has changed 
gradually over the years so no existing documentation is adequate.

The Program we refer to is an INTERACTIVE Prover, which is used to
prove theorems in mathematics.  The user can employ a user interface,
which offers a number of Options to assist with the proof, or to help 
follow a computer generated  proof as it evolves.  This prover is not
designed for speed.  But it hopefully serves as a convenient tool 
for user interaction, for building components for more powerful provers.

The basic prover was described by Bledsoe and Tyson [BT75,78].  This
extends the version described earlier by Bledsoe and Bruell [BB73].
It is also briefly described in Bledsoe's 1977 paper on Non-Resolution
Theorem Proving, [B77].  Another version of IMPLY, which omits the 
interactive component, is described in [B83].  Another Natural Deduction
style prover is used by Brock, Cooper, and Pierce, in the research on
Analogy in proof discovery [BCP87].

This document will concentrate on the current version, an interactive
version, which is written in Kyoto Common Lisp, and presently running
on SUN work stations, under UNIX.  (Though it in no way depends on UNIX).

-------------------------------------------------------------------------
DISCLAIMER  

We do not promise that this prover works completely as intended.  Things
that have worked properly for years tend to get flaky as we add new
things, as we have -- and do regularly.  Hopefully, any new errors will be 
corrected soon. Please let us know if you have questions or observe errors.

       bledsoe@cs.utexas.edu      512-471-9568

Also, it is not our aim to produce the most elegant LISP code, but rather 
correct and readable code.  We believe that the future of ATP rests on
IDEAS, not clever coding.  (Although we continue to urge extensive
exercising of provers on substantial examples.)

A revision is being prepared that improves the coding but not the basic 
concepts. 
-------------------------------------------------------------------------
For those in my ATP Group, from any UT CS computer (on CS network), do

        CD ~bledsoe/imply      
-------------------------------------------------------------------------

To read the tape onto SUN Computer Workstation:
        Tar -xvf /dev/rst0

                      Loading the Program

The program is located in the following files:
         
             IMPLY.LISP                  And other not really used in 
             PROVE1.LISP                  the Examples shown below:
             PROVE2.LISP)
             OPTIONS1.LISP                   REPLACE.LISP
             OPTIONS2.LISP                   INFIX.LISP
             EXAMPLE-theorems.lisp           PLANS.LISP
             UNIFY.LISP                      PROOFS.LISP
             REDUCE.LISP                     ANALOGY.LISP
             SKOLEM.LISP                     LEMMAS.LISP
             VE.LISP                           others
             SYM.LISP
             REPLACE.LISP
    
One can load these files into a KCL environment by the following commands:

            (Load "imply.lisp")
            (load-lisp-all)           ;Defined in IMPLY.LISP.

That would enable the non-compiled version.  For a compiled version do

            (Load "imply.lisp")
            (compile-all)             ;Defined in IMPLY.LISP
            (load-all)                ;Defined in IMPLY.LISP.
 
Of course, they can be loaded and/or compiled individually
-----------------------------------------------------------------------

                     Description of Files

IMPLY.LISP is rather short, just a shell to initialize parameters, etc.
It need not be compiled.

EXAMPLE-THEOREMS.LISP contains a number of example theorems (and some
non-theorems) on which the prover can be exercised.  These are numbered
Ex1, Ex2, etc.  Simply call 

             (prove ex1)   (or (pr ex1))
             (prove ex2), etc

to see how it goes.  These examples force the prover through a variety of 
actions, thereby showing how the various features work.  We will shortly
list the interactive commands ("options") available to the user, and
explain each one.  These examples are tutorial in nature, and not meant
to be a list of "challenge" theorems for other provers.
        
PROVE1.LISP and PROVE2.LISP make up the basic proving engine (originally
combined into PROVE.LISP, but broken up for ease of compiling).
It contains the main routines:

             PROVE, IMPLY, IMPLY-RULES
and many subsidiary ones.  There is a great deal of documentation with
each of the files and functions.  
 
                    The Main Functions

PROVE (via the routing PROVE-executor) accepts a theorem, Skolemizes it,
and sends it to IMPLY for proof.  Before doing so, it sets a number of
parameters and structures to help communicate with the user.  For
example, a PROOF-LIST keep a list of the various steps taken in the
proof, and types them out at the end of the proof in a easy to understand
way. (This is now suppressed: one can look to the workstation screen for a
record of a proving session.)  Also run-time calculations are made.  

The notation used by the prover is prefix (e.g., we write (-> H C) instead 
of (H -> C)), but some ability to convert infix to prefix is available.
It uses n-ary conjunctions, disjunctions, etc.  So we write 

     (& a b c), (or c d e), (+ x y z w), (* s t v), etc.
  
IMPLY is the main routine of the prover.  But it sends to IMPLY-RULES for the
bulk of the work.  And it also works closely with OPTIONS to enable the user
interface.  IMPLY has 6 parameters:  H C TL ANCESTRY LT YET.  H and C are
its main parameters:

	H  is the current hypothesis, and
        C  is the current subgoal, that we are trying to prove.
        
Also

        TL ("theorem-label") is an encoding which shows the user where 
            he "is" in the proof.  See examples later.

        ANCESTRY contains information of previous sub-goals, enabling the 
            prover to accept certain subgoals as true and to reject others.  
            See IMPLY-RULES.  Its initial value is NIL.

        LT is a control parameter, used mainly by the user interface.
          
        YET contains the subgoals that must still be proved before the
            complete proof is finished.  (The addition of the parameter YET
            represents a variation from the earlier versions of IMPLY, as
            described in ATP17B).

When IMPLY is called,

             (IMPLY H C ANCESTRY LT YET),

it returns a substitution THETA.  If THETA is NIL then the sub-goal is
false (or the prover is unable to prove it), otherwise 

             (H -> C) THETA

is tautologically valid.  (I.e., if the substitution is applied to the
variables of H and C, then C follows from H).
  
Such a substitution, returned from IMPLY or one of the other routines,
may be passed back to another function which called it, or might be
combined with other substitutions, that result from other 
subgoals.  A function COMPOSE is used to combine substitutions.  In our 
system, the substitution  T  represent the identity  
substitution, where no variables are bound.  This allows us to reserve
the value  NIL  for FAILURE throughout the prover. 
     
IMPLY-RULES is a set of rules that are applied to the parameters H, C, 
etc, manipulating them, calling other functions -- including IMPLY --
until a successful substitution or failure is reached.  

Central to this is UNIFICATION.  The call (UNIFY A B) returns the mgu
of A and B if there is one, else NIL.  For example, if H is (P x)
and C is (P (a)), then IMPLY-RULES will return the substitution ((x a)).
(We use the convention that skolem variables, those that can accept bindings
-- like the x in the example --, are represented by lisp atoms (non-numeric), 
and skolem constants are parenthesized.  Of course, such a constant can be 
thought of as a 0-ary functional expressions.)

Another simple rule of IMPLY-RULES is MATCH, where a conjunct of H is
unified with C.  Other more complicated rules involve BACK-CHAINING,
Substitution-of-equals, "Promoting", etc.  The PROMOTE rule causes a 
premise in the conclusion to be moved to the Hypothesis:  e.g., 

      (IMPLY 'h '(-> h2 c1) ...)

is converted to     

      (IMPLY '(& h h2) 'c1  ...) .

 
Another important rule is
AND-SPLITTING, whereby a conjunctive goal (& A B ...) is divided into a 
number of sub-goals: A, B, ..., which are proved separately.  Of course, 
the associated substitutions will have to be compatible, and are combined.  
A similar rule, CASE-SPLITTING, applies when the hypothesis, H, is a 
disjunction, e.g., H = (or H1 H2).  Then the conclusion C is proved twice, 
using first H = H1 and then H = H2.  (See list of rules in [BT75,78,B77,B83]).


 
                   The User Interface

The user interface is enabled by an interaction between IMPLY and OPTIONS.
The main feature is  **-IMPLY-STOP-**.  The program often stops 
during the computation, at a point in the computation called the 
"IMPLY STOP".  At that point it prints out the values of TL and C,
then prints **-IMPLY-STOP-**, and halts for instructions from the user.  For 
example, it might print

         (1 . P-> . [3] . 2 . R . [2] .)
         (P X (A))

         **-IMPLY-STOP-**

Which means that TL (the theorem label) is (1 . P-> . [3] . 2 . R . [2] .), 
and the current subgoal, C, is  (P X (A)).  The user now can type one of the 
many atoms listed below (see also the function OPTIONS) to elicit various 
actions.
  
TL represents a cryptic history of what has happened so far, on the current 
branch of the proof tree.  In the above example, 

         1      Indicates the main theorem itself

         P->    Indicates a promotion (See above)

         [3]    Indicates a backchaining on the 3rd hypothesis of H

         2      Indicates that the SECOND conjunct is being proved

         R      Indicates that REDUCE was used to change either C or H

         [2]    Indicates a backchaining on the 2nd hypothesis of H

         .      The periods represent "Back-up points".  See B and BACK below.
                These periods are automatically inserted after the major
                actions of the prover, and can also be user-inserted elsewhere
                (see B-POINT below). 

This not only offers a easy way for the user to see where he is in the proof,
but is also used by the prover at times to direct the search.  For example,
the prover can be made to not back-chain on the same hypothesis more than
n times on the same branch of the proof tree.

The prover might also add the word  "PROVED"  or  "FAILED".  E.g.,

         (1 . P-> . [3] . 2 . R . [2] .)
         (P X (A))
         PROVED - 

        **-IMPLY-STOP-**

which indicates that the current sub-goal (P X (A)) has been proved.  
 (At this point the User could type  "X"  to find out what substitution 
was used in the proof of that sub-goal.  See below)
 
After  **-IMPLY-STOP-**  is printed and the prover has stopped, 
the user has a number of options that allow him to direct the
computation.  Some of these are for printing values of parameters, 
others regulate the amount of printed output, still others allow
him to reorient the prover's direction in proving the theorem.  See
the function, OPTIONS, for some further documentation.  Of course,
one of the easiest ways to learn about these options is to use them 
in proofs.

We now list all of the options (or most of them) with a short explanation
of their functions.  And then show some examples of the prover in action.
  
                        OPTIONS LIST

OPTION            ACTION
------      --------------------------------------------

<CR>        Proceed.  By pressing the Carriage Return Key, the computation
            proceeds in a normal manner.  This is the most used option.

QED         Finish the proof of the current sub-goal, C, without stopping.
            Under this option it will NOT stop at **-IMPLY-STOP-**, but
            will proceed instead until that sub-goal is proved, or until
            it has failed on it.

QEDN        This is exactly like QED, except it proceed to the end of the
 (or N)     WHOLE THEOREM, not just the sub-goal C, before stopping.

DETAIL      This directs the prover to break up the proof of C into 
            two conjuncts (if there are two) and to report on each.
            See [BB73] and the remarks below.
 
A           Return T (TRUE) for the current sub-goal.  This lets the
            user force the prover to consider it true, whether or not
            it is.  This is very useful for debugging a proof.  
            (The prover keeps a list of such assumptions and 
            reminds the user at the end that the "proof" may
            not be valid. It also keeps a record of lemmas that are used
            -- and not proved -- under the USE option (below)).  
            (Note: Currently, the "keeps a record" feature is not implemented.)

F           Return NIL (FALSE) for the current sub-goal.  Similar to A.

B           Back-up in the proof to the last . in TL (the theorem label).
            By the CATCH and THROW mechanisms the prover is required to
            reprove the portion of the proof represented in TL after the
            last period.  (If TL ends in a . then it backs up to the 
            previous one).  One can use B several times to back up 
            further, or use  BACK.

BACK        This causes the prover to back-up n levels as commanded by
            (prompted) user input.
 
B-POINT     This inserts a back-up point, ., in the TL.

E           Break to the LISP level.  For user interaction with the code.
            (Type :R to proceed,  or :Q to return to the top level of LISP,etc)
       
PUT         This allows the user to make instantiations for variables.  He is
            prompted for his input.
  
USE         This allows the user to add a lemma to the hypothesis (without
            proof).  The user input is prompted.  As mentioned under A,
            the prover keep track of all unproved lemmas that are added in
            this way.

CLAIM       This exactly like USE except that the claim-lemma, that the user
            gives, is first proved before it is used.

LEMMA       Same as CLAIM.

INDUCTION   This lets the user use proof by induction.  He is prompted for the
            Induction Variable and the Induction Hypothesis, but can let
            the prover select these automatically if he chooses.  In this 
            case he can reject the provers selections (or even reject his
            own) if he chooses, and enter them himself. Ex's 30,31,33,36,37.
            See the (partial) proofs below of Ex's 30, 36, 37.  (If the user
            wants to use a more general Induction Hypothesis, than C, he can
            ease his typing load by using the characters "&H" and "&C"
            to represent the evaluation of H and C.)

I           Same as INDUCTION.

INEQ        This calls the inequality prover.  For access to the techniques,
            such as VE (variable elimination) for proving general inequality
            theorems.

VE-LIGHT    This enables the function, INEQUALITY-PROVE, so the VE, etc, can
            occur in the normal sequence.

NOT-VE-LIGHT  This disables INEQUALITY-PROVE.
 
TIME        This prints the amount of time spent so far in the proof.

HH          This prints H (the hypothesis), numbering the conjuncts. I
            prints the (non-empty) components of H separately: 
            Hu, H-gu(ground units), H=(equality units), Hv.  Eg., 

	**-IMPLY-STOP-**
	hh                %% The User types this hh, and it responds:

        H-Ground-units:

        1 :  (FOOD (APPLES))

        2 :  (FOOD (CHICKEN))

        3 :  (EATS (BILL) (PEANUTS))

        4 :  (ALIVE (BILL))

        Hv:

        1 :  (-> (FOOD X) (LIKES (JOHN) X))

        2 :  (-> (& (EATS Y X1) (NOT (KILLED-BY Y X1))) (FOOD X1))

        3 :  (-> (EATS (BILL) X2) (EATS (SUE) X2))

        4 :  (-> (ALIVE Y1) (NOT (KILLED-BY Y1 X3)))
 
C           This prints C (the sub-goal), numbering the conjuncts. Like HH.

THM         This Prints (H --------> C), numbering the conjuncts.

TL          This Prints TL.

LT          This Prints LT.

YET         This Prints YET.
 
X           This Prints Bigx, the substitution that has just been returned
            for the proof of C, the current sub-goal.

^^          Return to top level of LISP.

R or R*     Reduce C.  (Using the Rewritter, REDUC).
RH          Reduce H.

no-R        Turn-on the *global-no-reduce-light*, so absolutely no 
            REDUCing is possible.  (Turn it back off with off-no-R).
off-no-R    See above.

R-alg       Reduce C Algebraically.  A more powerful reducer for 
            algebraic expressions only.  Use in conjunction with R.
RA          Same as R-alg.

RAH         Reduce H Algebraically.

ADD-REDUCE  Lets the user add reduce rules to the Reduce-table. (He can 
            also do this using the LISP editor).
DEL-REDUCE  Lets the user delete a reduce rule from the Reduce-table.
PRT         Print the Reduce Table.

DC          This opens the definition of the sub-goal C, or of its 
            main predicate symbol.

D           This lets the user open the definition of other terms in
            H or C.

SHOW-DEF    Show the definition of the (Prompted) term.

ADD-DEF     Lets the user add definitions to the Definition-table. (He can 
            also do this using the LISP editor).
DEL-DEF     Lets the user delete a definition from the Definition-table.
PDT         Print the Definition Table.
     
MOVE->      Promote.
  
(H n1 n2 n3 ...)   Reorder the hypotheses of H in this order (putting 
                   any whose number is missing at the end).  E.g., if H is 
                   (& A B C D E), then the user command (H 3 2 5) resets 
                   H to (& C B E A D).
                   Actually H has four components: Hu, H-gu, H=, Hv. 
                   This action is carried out on  Hv  only.  A similar
                   command is used for H-gu (eg, (H-gu 7 3) would place
                   reorder H-gu, putting hypotheses 7 and 3 in first
                   and second place).

(n1 n2 n3...)      Same as (H n1 n2 n3 ...).

(HD n1 n2 n3 ...)  Same as (H n1 n2 n3 ...), except the unreferenced 
                   hypotheses are deleted.  H-guD is similar.

(C n1 n2 n3 ...)   Similar to (H n1 n2 n3 ...).


------------------------------------------------------------------
There are other options not shown here such as: LET, N0-OP, CASES, PROMOTE, 
APPLY, SUB=, SPLIT-H, SPLIT-C, VE-LIGHT, REPLACE, REMOVE-SHIELDS, REMOVE-
VARIABLE, APPLY-REDUCE, APPLY-LEMMA, APPLY-HYPOTHESIS, ELIGIBLES,
ELIMINATE-VARIABLE, ACTION.  One should consult the code if interested.

These are not used in the examples given below, but are used in various
other examples related to research projects.

-------------------------------------------------------------------


                FURTHER COMMENTS ON OPTIONS, etc
 
DETAIL

DETAIL does not now work exactly as it was originally planned. (See [BB73]).
This is because of our change in the IMPLY function, giving it a new parameter,
YET.  But it still acts approximately right, and serves the following useful 
purpose:

	when a subgoal C fails, a call to DETAIL directs the prover to 
        work down it proof tree until it finds its first conjunct (& A B),
        -- in any --, and to then report on A and B separately (without 
	stopping to report the outcome of other subgoals in the process).

        In this way the user can concentrate on the part that has gone wrong,
        and not wander all over the place looking for the problem.  
 
However, one can get a similar effect (as DETAIL) by using QED and CR, but
it requires more work on the part of the user.

INEQUALITY PROVER

Earlier versions of IMPLY.LISP emphasized inequality theorems such as Ex47,

        (-> (< A B) (SOME X (& (< A X) (< X B)))),

the limit theorems of calculus, and theorems of intermediate analysis.[B84].

An important part of this was our GROUND INEQUALITY PROVER, based on the
Sup-Inf technique.  But after we discovered the usefulness of VE (Variable
Elimination) [BH80], we switched to a Resolution format, which is much more
convenient for VE.  We have now built the VE methodology into this natural
deduction style prover (thus the need for the YET parameter of IMPLY), but
have not yet incorporated the Inequality Ground Prover.  This can be done
quite easily and probably will be done shortly.  In the meantime, we use
a user-prompted INEQUALITY GROUND PROVER, whereby the user is asked if 
a pure inequality ground theorem is true.  If he answers "yes", then it
is remembered as TRUE by the prover for future proofs. (Caution: it believes
the user!!).
 
REDUCE

The reduce unit (Rewrite Rules) has also been a central part of our provers
in the past, but the one in this prover is rather meager.  One can build
up the reduce rules, and thereby make it more powerful, but using the
command ADD-REDUCE (and DEL-REDUCE for deleting them) or by using the LISP 
editor.  The option PRT can be used to easily print the current table of 
REDUCE rules.  But even then the reducer is rather slow.  We expect to 
incorporate a much stronger and faster version soon.
 
VARIABLE-SUPPLY

This holds a fairly large list of symbols such A B C D etc., which are
used when a new symbol is needed.  A GEN-SYM approach is probably better,but it
tends to face the user with ugly symbols in the proofs and lessens his
ability to easily interact.
 
SYMBOL-LIST

This lists all predicate and function symbols used in any proof supplied to
the prover.  So one should add to this list (in the file  SYM.LISP) if
other symbols are used.  This is clearly a bad arrangement.  We hope to soon
correct the situation so that any symbol that occurs in a proper position of a 
formula is alright.
 
GENERALIZED SUBSTITUTIONS

In the proofs of some theorems, certain variable are bound to more than one
term.  Even though there are ways to dynamically rename variable in such a
way that multiple bindings are not needed, we find that awkward and prefer
a generalized substitution approach [BT79], and use that here.  Thus in 
the proof of the theorem, Ex9,

          (-> (ALL X (P X)) (& (P A) (P B)))

we obtain the "answer", THETA = ((X A) (X B)).  This offers no difficulty as
long as that substitution is not required to be applied to another
formula.  But when it is, one must act appropriately.  For example, in the
proof of Ex9**,

   (-> (& (Q A) (Q B)) (SOME X (& (-> (P X) (& (P A) (P B))) (Q X))))

after skolemization and preliminary steps, the prover is given the 
conjunction,
 
      (&  (-> (P X) (& (P (A)) (P (B))))  (Q x))

to be proved.  The first conjunct yields the substitution, 
THETA = ((X A) (X B)).  (As did Ex9).  When that is substituted into the 
second conjunct, (Q x), the subgoal,

        (& (Q (A)) (Q (B))),

is obtained, and proved, yielding a final substitution,  ((X A) (X B)), for
the theorem.


References.     

[BB73] Bledsoe, W.W. and Peter Bruell, A Man-machine Theorem Proving System,
JAR 5 (1974) 51-72.

[BT75,78] Bledsoe, W.W. and Mabry Tyson, The UT Interactive Theorem Prover,
Univ of Texas Math Dep Memo, ATP 17, May 1975, ATP 17A, June 1978.

[B77] Bledsoe, W.W., Non-Resolution Theorem Proving, JAR 9 (1977) 1-35.

[B83] Bledsoe, W.W., The UT Natural Deduction Prover, Univ of Texas Math
Dep Memo, ATP 17B, Apr 1983.

[B84] Bledsoe, W.W., Some Automatic Proofs in Analysis, in Automated Theorem
Proving: After 25 years, W.Bledsoe and D. Loveland (Eds), Amer Math Soc,
Contemporary Mathematics Series 19, 1984, 89-118.

[BH80] Bledsoe, W.W. and L. Hines, Variable elimination and chaining in 
a resolution based prover for inequalities, CADE-80, 70-87, 1980.

[BCP87] Brock, B., S. Cooper and W. Pierce, Some Experiments with Analogy in
Proof Discovery:  A Natural Deduction Approach, MCC Tech Report ACA-AI-274-87,
August, 1987.
 
[Sim84] Simon, D., A Linear Time Algorithm for a subcase of Second-order
Instantiation, 7th Conf on Automated Deduction, Napa, Ca, May 1984.

[BT79] Bledsoe, W.W. and Mabry Tyson, Generalized Substitutions, CADE-4.

[Bl71] Bledsoe, W.W., Splitting and Reduction Heuristics in Automatic
Theorem Proving, Artificial Intelligence 2 (1971), 55-77.
-------------------------------------------------------------------------

                   Some example Proofs

We suggest that the user try several of the examples, Ex1, Ex2, etc.
The following subset is suggested as an interesting group:

  Ex1,2,3,4,8,9,9*,9***,21,22,23,24,30(see below),30(Induction)(See below),
    32(See below),33,35,36,37,42(see below),47(see below),

Now we give "photos" of the proofs (or parts) of some example theorems.
We repeat, that these example theorems are not offered as "Challenge"
problems, but are here only to check out various parts of the prover.

They are however very useful for the person unfamiliar with the prover,
who is finding out how it works.  Some are not theorems:  some need more
hypotheses (use USE or LEMMA);  some require additional REDUCE rules; 
and some require INDUCTION;  and some are theorem that IMPLY cannot
prove (e.g., Ex27).  (It would be a simple matter to force IMPLY to
prove theorems like Ex27, but that would require destroying the basic
structure of the theorem,-- like clausifying it --  and that is not in 
the spirit of Natural Deduction as we see it).

Annotation is indicated by the symbol %%.  @@@@@@ indicated that a number
of steps have been ommitted from this presentation.
  
Ex1**********************
>(prove ex1)                          %% User input

 Theorem Being Proved: EX1 

(-> A A)

Theorem being proved -- Skolemized:  EX1 
		 
 H:                                   %% This gives the hypothesis H,
H-Ground-units:                       %%  by component. In this case the
(A)                                   %%  only non-NIL component is H-gu

C:  (A)                               %% This gives the conclusion C.

(1 .)                                 %% This is the Theorem Label, TL
(A)                                   %% This is the conclusion(again),
______________                        %%  repeated at each IMPLY-STOP
  **-IMPLY-STOP-**                    %% The prover waits at IMPLY-STOP,
Proceed-ed                            %%  for user input. Here the user issues
                                      %%  a CR(carriage return),& it proceeded.
C was found in the Hypothesis         
(1 .)
(A)
PROVED-                               %% Here it reports that the goal has
______________                        %%  been proved and awaits further
  **-IMPLY-STOP-**                    %%  user input
Proceed-ed 

 PROVED: EX1 

 With MGU THETA: T 
                        Returned for the Theorem
(TIME-USED 24 MSEC)

 
Ex3*************************
>(pr ex3)

 Theorem Being Proved: EX3 

(-> (& (P A) (ALL X (-> (P X) (Q X)))) (Q A))

Theorem being proved -- Skolemized:  EX3 
		 
 H:                                %% In this case H contains 
H-Ground-units:
(P (A))                            %%  ground unit, (P (A))

Hv:
(-> (P X) (Q X))                   %%  and one non-ground hypothesis

C:  (Q (A)) 


(1 .)
(Q (A))
______________
  **-IMPLY-STOP-**                 %% Here the user issued a CR.
Proceed-ed                         %%  (Similarly below, where "Proceed-ed"
                                   %%  is shown)
 Back-Chain on Hyp: 1 of HV with MGU: ((X A)) 

(1 . [1] .)
(P (A))
______________
  **-IMPLY-STOP-**  
Proceed-ed 

C was found in the Hypothesis
(1 . [1] .)
(P (A))
PROVED-
______________
  **-IMPLY-STOP-**  
Proceed-ed 
(1 .)
(Q (A))
PROVED-
______________
  **-IMPLY-STOP-**  
Proceed-ed 

 PROVED: EX3 

 With MGU THETA: ((X A)) 
                        Returned for the Theorem
(TIME-USED 317 MSEC)



Ex3 using QED********************************
>(pr ex3)

 Theorem Being Proved: EX3 

(-> (& (P A) (ALL X (-> (P X) (Q X)))) (Q A))

Theorem being proved -- Skolemized:  EX3 
		 
 H:  
H-Ground-units:
(P (A))

Hv:
(-> (P X) (Q X))

C:  (Q (A)) 

(1 .)
(Q (A))
______________
  **-IMPLY-STOP-**  qed          %% User input.

(QED BEING TRIED)....
 Back-Chain on Hyp: 1 of HV with MGU: ((X A)) 

C was found in the Hypothesis    %% Note: the prover did not stop (at
(1 . QED)                        %%  IMPLY-STOP) until the subgoal, where
(Q (A))                          %%  QED was called, was completed proved
PROVED-BY-QED
((X A))
______________
  **-IMPLY-STOP-**               %% Now it stops and awaits a CR to proceed.
Proceed-ed 
(1 .)
(Q (A))
PROVED-
  @@@@@@

 PROVED: EX3 

 With MGU THETA: ((X A)) 
                        Returned for the Theorem
(TIME-USED 304 MSEC)

%%     If QEDN is used, instead of QED, then it proceeds with the proof ALL the
        way to the end, including the proof summary, without stopping. 
        See next.

Ex3 using QEDN********************************

>(pr ex3)

 Theorem Being Proved: EX3 

(-> (& (P A) (ALL X (-> (P X) (Q X)))) (Q A))

Theorem being proved -- Skolemized:  EX3 
		 
 H:  
H-Ground-units:
(P (A))

Hv:
(-> (P X) (Q X))

C:  (Q (A)) 

(1 .)
(Q (A))
______________
  **-IMPLY-STOP-**  qedn                   %% Note user input

(QEDN BEING TRIED)....
 Back-Chain on Hyp: 1 of HV with MGU: ((X A)) 

C was found in the Hypothesis
(1 . QED)
(Q (A))
PROVED-BY-QED
((X A))(1 .)
(Q (A))
PROVED-

 PROVED: EX3 

 With MGU THETA: ((X A)) 
                        Returned for the Theorem
(TIME-USED 304 MSEC)

Ex9**  *********************************************************
>(pr ex9**)                  %% Note that generalized substitutions are
                             %%  used in this example, wherein the variable
 Theorem Being Proved: EX9**     x is bound to more than one term.

(-> (& (Q A) (Q B)) (SOME X (& (-> (P X) (& (P A) (P B))) (Q X))))

Theorem being proved -- Skolemized:  EX9** 

 H:  
H-Ground-units:
1 :  (Q (A))
2 :  (Q (B))

C:  (& (-> (P X) (& (P (A)) (P (B)))) (Q X)) 

(1 .)
(& (-> (P X) (& (P (A)) (P (B)))) (Q X))
______________
  **-IMPLY-STOP-**  
Proceed-ed 

 SPLIT-C: NEW-C:   (-> (P X) (& (P (A)) (P (B)))) 

(1 . 1 .)
(-> (P X) (& (P (A)) (P (B))))
______________
  **-IMPLY-STOP-**  qed

(QED BEING TRIED)
 SPLIT-C: NEW-C:   (P (A)) 
 Match C and H with MGU: ((X A)) 

 SPLIT-C: NEW-C:   (P (B)) 
 Match C and H with MGU: ((X B)) 
                                   %% Here the generalized substition 
                                   %%  ((X A) (X B)) is returned and
(1 . 2 .)                          %%  applied to (Q X) getting the new
(& (Q (A)) (Q (B)))                %%  subgoal (& (Q (A)) (Q (B)))
______________
  **-IMPLY-STOP-**  qedn

(QEDN BEING TRIED)
 Match C and H with MGU: T 
(1 . 2 . QED)
(& (Q (A)) (Q (B)))
PROVED-BY-QED
 @@@@@@
 @@@@@@
 PROVED: EX9** 

 With MGU THETA: ((X A) (X B)) 
                        Returned for the Theorem
(TIME-USED 545 MSEC)

Ex9***  *********************************************************
>(pr ex9***)               %% Note that a GENERALIZED substitution is
                           %%  returned, wherein x is bound to three terms.
 Theorem Being Proved: EX9*** 

(-> (ALL X (P X)) (& (P A) (P B) (P C)))

Theorem being proved -- Skolemized:  EX9*** 
		 
 H:  
Hv:
(P X)

C:  (& (P (A)) (P (B)) (P (C))) 

(1 .)
(& (P (A)) (P (B)) (P (C)))
______________
  **-IMPLY-STOP-**  qedn

(QEDN BEING TRIED).....
 SPLIT-C: NEW-C:   (P (A)) 
 Match C and H with MGU: ((X A)) 

 SPLIT-C: NEW-C:   (P (B)) 
 Match C and H with MGU: ((X B)) 

 SPLIT-C: NEW-C:   (P (C))                           %%
 Match C and H with MGU: ((X C)) 

(1 . QED)
(& (P (A)) (P (B)) (P (C)))
PROVED-BY-QED
 @@@@@@
 PROVED: EX9*** 

 With MGU THETA: ((X A) (X C) (X B)) 
                        Returned for the Theorem
(TIME-USED 653 MSEC)


Ex32*********************************************************

  For this theorem we will use the OPTIONS:  PUT and REORDER H (hypotheses).
  (It is also very convenient at times to use REORDER C,  backup (B or BACK),
    H (number and show the hypotheses) C, etc).

  This is a theorem in topology (Paracompactness).  You do not need to
     know the following semantics:
   (OPEN F) Means that F is a family of Open sets (in a topology)
   (RF F G) Means that F is a refinement of G. (Each member of G is
              is a subset of a member of F)
   (LF F)   Means that F is a locally finite family.  (Each point of
              the space has a neighborhood which intersects at most finitely
               many member of F).
   ____________________

   In the following, %% indicated user annotation.  Most of the proof steps
   are left out left out (as indicated by "@@@@@@").

>(pr ex32)                        
                                  
 Theorem Being Proved: EX32 

(-> (& (ALL J (CLOSED (CLSR J))) (ALL K (-> (LF K) (LF (CLSR K))))
       (ALL M (-> (COVER M) (COVER (CLSR M))))
       (ALL P (ALL Q (ALL R (-> (& (REF P Q) (REF Q R)) (REF P R)))))
       (ALL U (ALL V (-> (REF U V) (REF (CLSR U) (CLSR V)))))
       (ALL F
            (-> (& (OPEN F) (COVER F))
                (SOME G (& (OPEN G) (COVER G) (REF (CLSR G) F)))))
       (ALL F*
            (-> (& (OPEN F*) (COVER F*))
                (SOME G* (& (OPEN G*) (COVER G*) (REF G* F*) (LF G*))))))
    (ALL FF
         (-> (& (OPEN FF) (COVER FF))
             (SOME H (& (CLOSED H) (COVER H) (REF H FF) (LF H))))))
   ...
______________
  **-IMPLY-STOP-**  hh             %% The user asks to see the Hypotheses
                                   %%  which are separated into 4 components,
H-Ground-units:                    %%  Hu, H-gu, H=, and Hv.  Only H-gu and
                                   %%  and Hv are non-empty in this example.
1 :  (OPEN (FF))                   

2 :  (COVER (FF))

Hv:

1 :  (CLOSED (CLSR J))

2 :  (-> (LF K) (LF (CLSR K)))

3 :  (-> (COVER M) (COVER (CLSR M)))

4 :  (-> (& (REF P Q) (REF Q R)) (REF P R))

5 :  (-> (REF U V) (REF (CLSR U) (CLSR V)))

6 :  (-> (& (OPEN F) (COVER F))
         (& (OPEN (GS F)) (COVER (GS F)) (REF (CLSR (GS F)) F)))

7 :  (-> (& (OPEN F*) (COVER F*))
         (& (OPEN (G*S F*)) (COVER (G*S F*)) (REF (G*S F*) F*)
            (LF (G*S F*))))

(1 . P-> .)
(& (CLOSED H) (COVER H) (REF H (FF)) (LF H))                %% Current Goal
______________
  **-IMPLY-STOP-**  put                                     %% User Command

Substitute for what Variable?:  h                           %% User Response

PUT What, for that Variable:  (clsr (g*s (gs (ff))))        %%  "     "

====MANUAL SUBSTITUTION====
The result of PUT is:

(NIL (& (OPEN (FF)) (COVER (FF))) NIL
     (& (CLOSED (CLSR J)) (-> (LF K) (LF (CLSR K)))
        (-> (COVER M) (COVER (CLSR M)))                     %% The Components 
        (-> (& (REF P Q) (REF Q R)) (REF P R))                  of H
        (-> (REF U V) (REF (CLSR U) (CLSR V)))
        (-> (& (OPEN F) (COVER F))
            (& (OPEN (GS F)) (COVER (GS F)) (REF (CLSR (GS F)) F)))
        (-> (& (OPEN F*) (COVER F*))
            (& (OPEN (G*S F*)) (COVER (G*S F*)) (REF (G*S F*) F*)
               (LF (G*S F*))))))
 ------>
(& (CLOSED (CLSR (G*S (GS (FF))))) (COVER (CLSR (G*S (GS (FF)))))   
   (REF (CLSR (G*S (GS (FF)))) (FF)) (LF (CLSR (G*S (GS (FF))))))       

 %% Note that the variable H has been instantiated with 
     (clsr (g*s (gs (ff)))).

IS THIS OK?   Type:    (Y or N) y

(1 . P-> . PUT .)
(& (CLOSED (CLSR (G*S (GS (FF))))) (COVER (CLSR (G*S (GS (FF)))))    %% Current
   (REF (CLSR (G*S (GS (FF)))) (FF)) (LF (CLSR (G*S (GS (FF))))))        Goal,C
 @@@@@@
(1 . P-> . PUT . 1 .)
(CLOSED (CLSR (G*S (GS (FF)))))
 @@@@@@

PROVED-
 @@@@@@

(1 . P-> . PUT . 2 . 1 .)
(COVER (CLSR (G*S (GS (FF)))))
 @@@@@@

 Back-Chain on Hyp: 3 of HV with MGU: ((M G*S (GS (FF)))) 
(1 . P-> . PUT . 2 . 1 . [3] .)      %% Back-chains on Hyp 3
(COVER (G*S (GS (FF))))              %% Proved, using Hyps 7,6
 @@@@@@

(1 . P-> . PUT . 2 . 2 . 1 .)
(REF (CLSR (G*S (GS (FF)))) (FF))
______________
  **-IMPLY-STOP-**  (Hv 5 7)         %% This reorders the hypotheses in Hv
                                     %%  putting numbers 5 and 7 first
****-REORDER THE HYPOTHESES OF H-****

1 :  (-> (REF U V) (REF (CLSR U) (CLSR V)))
2 :  (-> (& (OPEN F*) (COVER F*))
         (& (OPEN (G*S F*)) (COVER (G*S F*)) (REF (G*S F*) F*)
            (LF (G*S F*))))
3 :  (CLOSED (CLSR J))
4 :  (-> (LF K) (LF (CLSR K)))
5 :  (-> (COVER M) (COVER (CLSR M)))
6 :  (-> (& (REF P Q) (REF Q R)) (REF P R))
7 :  (-> (& (OPEN F) (COVER F))
         (& (OPEN (GS F)) (COVER (GS F)) (REF (CLSR (GS F)) F)))

(1 . P-> . PUT . 2 . 2 . 1 .)
(REF (CLSR (G*S (GS (FF)))) (FF))
______________
  **-IMPLY-STOP-**  
 Back-Chain on Hyp: 6 of HV with MGU: ((P CLSR (G*S (GS (FF)))) (R FF)) 

(1 . P-> . PUT . 2 . 2 . 1 . [6] .)
(& (REF (CLSR (G*S (GS (FF)))) Q) (REF Q (FF)))
 @@@@@@
PROVED-                              %% Proved using Hyps: 1,2,7
______________
  **-IMPLY-STOP-**  
Proceed-ed                           %% This just echo's the fact that
                                     %%  the user pressed the CR key.
(1 . P-> . PUT . 2 . 2 . 1 . [6] . 2 .)
(REF (CLSR (GS (FF))) (FF))
______________
  **-IMPLY-STOP-**  (Hv 7)           %% This reorders Hv, putting 7 first.
      %% Now (REF (CLSR (GS (FF))) (FF)) follows from it, and other hyps.
 @@@@@@

(1 . P-> . PUT . 2 . 2 . 2 .)
(LF (CLSR (G*S (GS (FF)))))          %% Follows from Hyp 2,7,6, etc.
 @@@@@@
      %% Now it returns T (true) through its various levels.
 @@@@@@

 PROVED: EX32 
 With MGU THETA: ((H CLSR (G*S (GS (FF)))) (J G*S (GS (FF)))
                  (M G*S (GS (FF))) (F* GS (FF)) (F FF)
                  (P CLSR (G*S (GS (FF)))) (R FF) (U G*S (GS (FF)))
                  (Q CLSR (GS (FF))) (V GS (FF)) (K G*S (GS (FF)))) 
                        Returned for the Theorem
(TIME-USED 4339 MSEC)  T

%%   This shows a computing time of about 4 secconds.  No real attempt is
       being made here to optimize that.  It would be even longer, except 
       that we prevented the repreated use of REDUCE by setting the *global-
       no-reduce-light* to T.  In many examples, such as this, no reduction is
       is needed.  

%%   Actually the prover is capable of proving Ex32 by itself, but it might
       take an inordinate amount of time, because only a DEPTH CUTOFF mechanism
       prevents it from recursing infinitely when a "bad" hypothesis is 
       encountered (like the transitivity of REF, Hyp 6 above).

%%   But one can easily obtain a proof (as we did) by using the PUT option, and
       by properly ordering (and reordering) the hypotheses and conclusions, 
       using the commands (h ...) and (c ...). 

Ex30**********************************************************
>(pr ex30)                      %% For this example we use INDUCTION

 Theorem Being Proved: EX30 

(ALL X
     (ALL T
          (-> (& (ELEMENT X (OMEGA)) (ELEMENT T X))
              (ELEMENT T (OMEGA)))))

Theorem being proved -- Skolemized:  EX30 
		 
 H:  
H-Ground-units:
1 :  (ELEMENT (X) (OMEGA))
2 :  (ELEMENT (T) (X))

C:  (ELEMENT (T) (OMEGA)) 

(1 .)
(ELEMENT (T) (OMEGA))
______________
  **-IMPLY-STOP-**  i                      %% User input ("i" for induction).

     ====== INDUCTION BEING TRIED ======

Please Type the INDUCTION VARIABLE, N: (You may Type NIL or CR -- if you wish to let it find the induction variable)  
Type N:                                    %% The user types  CR  here.

Please Type the INDUCTION HYPOTHESIS, IH:
       
(You may Type NIL or CR -- if you wish to use  C  for the Ind. Hyp.)
       
Type IH:                                   %% The user types  CR  here.
Ordinary INDUCTION ON: 
 (X)
USING INDUCTION HYPOTHESIS:
  (-> (& (ELEMENT (T) (X))) (ELEMENT (T) (OMEGA)))

C1:  (-> (& (ELEMENT (T) 0)) (ELEMENT (T) (OMEGA)))     %% Base Case.


C2:                                                     %% Induction Step case.
(-> (& (ELEMENT (X) (OMEGA))
       (-> (& (ELEMENT (T) (X))) (ELEMENT (T) (OMEGA))))
    (-> (& (ELEMENT (T) (SCSR (X)))) (ELEMENT (T) (OMEGA))))

IS THIS OK?   Type:    (Y or N) y                       %% User response

(1 . INDUCTION-1 .)
(-> (& (ELEMENT (T) 0)) (ELEMENT (T) (OMEGA)))          %% Base Case Proof
______________
  **-IMPLY-STOP-**  
Proceed-ed 
....
(1 . INDUCTION-1 . R .)
(-> (ELEMENT (T) 0) (ELEMENT (T) (OMEGA)))
______________
  **-IMPLY-STOP-**  
Proceed-ed 
.....                                       %% The dots indicate that Reduction
(1 . INDUCTION-1 . R . R .)                 %%  has taken place.  The R's in 
(-> (FALSE) (ELEMENT (T) (OMEGA)))          %%  TL indicate the same.
______________
  **-IMPLY-STOP-**  
Proceed-ed 
.
(1 . INDUCTION-1 . R . R . R .)
(TRUE)
______________
  **-IMPLY-STOP-**  
Proceed-ed 
(1 . INDUCTION-1 . R . R . R .)
(TRUE)
PROVED-
 @@@@@@
 @@@@@@
 $$$$$-Induction-1 Proved with MGU-&&&: T

(1 . INDUCTION-2 .)                         %% Induction Step Case Proof.
(-> (& (ELEMENT (X) (OMEGA))
       (-> (& (ELEMENT (T) (X))) (ELEMENT (T) (OMEGA))))
    (-> (& (ELEMENT (T) (SCSR (X)))) (ELEMENT (T) (OMEGA))))
______________
  **-IMPLY-STOP-**  
 @@@@@@

(1 . INDUCTION-2 . R . R .)
(-> (& (ELEMENT (X) (OMEGA))
       (-> (ELEMENT (T) (X)) (ELEMENT (T) (OMEGA))))
    (-> (OR (ELEMENT (T) (X)) (= (T) (X))) (ELEMENT (T) (OMEGA))))    %% C
     %% Note that (element (t) (scsr (x))) has been reduced to
     %%           (or (element (t) (x)) (= (t) (x))).
______________
  **-IMPLY-STOP-**  
 @@@@@@
(1 . INDUCTION-2 . R . R . P-> .)
(-> (OR (ELEMENT (T) (X)) (= (T) (X))) (ELEMENT (T) (OMEGA)))         %% C
______________
  **-IMPLY-STOP-**  hh

H-Ground-units:                %% Note that it has promoted two hypotheses
1 :  (ELEMENT (X) (OMEGA))     %%  into H-gu
2 :  (-> (ELEMENT (T) (X)) (ELEMENT (T) (OMEGA)))   

(1 . INDUCTION-2 . R . R . P-> .)
(-> (OR (ELEMENT (T) (X)) (= (T) (X))) (ELEMENT (T) (OMEGA)))         %% C
______________
  **-IMPLY-STOP-**  
 @@@@@@
(1 . INDUCTION-2 . R . R . P-> . P-> .)
(ELEMENT (T) (OMEGA))                                                 %% C
______________
  **-IMPLY-STOP-**  hh

H-Ground-units:                         
(OR (& (ELEMENT (T) (X)) (ELEMENT (X) (OMEGA))
       (-> (ELEMENT (T) (X)) (ELEMENT (T) (OMEGA))))
    (& (= (T) (X)) (ELEMENT (X) (OMEGA))
       (-> (ELEMENT (T) (X)) (ELEMENT (T) (OMEGA)))))
          %% Note that when the hypothesis (OR (ELEMENT (T) (X)) (= (T) (X)))
          %%  was promoted to H-gu, that the OR was pushed to the outside.
          %% It will now SPLIT on that OR.
(1 . INDUCTION-2 . R . R . P-> . P-> .)
(ELEMENT (T) (OMEGA))
______________
  **-IMPLY-STOP-**  
Proceed-ed 

 SPLIT-H:NEW-H: 
H-Ground-units:

1 :  (ELEMENT (T) (X))

2 :  (ELEMENT (X) (OMEGA))

3 :  (-> (ELEMENT (T) (X)) (ELEMENT (T) (OMEGA)))

(1 . INDUCTION-2 . R . R . P-> . P-> . 1 .)
(ELEMENT (T) (OMEGA))                                               %% C
______________
  **-IMPLY-STOP-**  
 @@@@@@

 Back-Chain on Hyp: 3 of H-GROUND-UNIT-S with MGU: T 

(1 . INDUCTION-2 . R . R . P-> . P-> . 1 . [3] .)
(ELEMENT (T) (X))

   %% Easily proved using Hyp 1.
 @@@@@@
                                     %% Now for the other half of the OR-SPLIT

(1 . INDUCTION-2 . R . R . P-> . P-> . 2 .)

(ELEMENT (T) (OMEGA))
______________
  **-IMPLY-STOP-**  hh

H-Ground-units:

1 :  (ELEMENT (X) (OMEGA))

2 :  (-> (ELEMENT (T) (X)) (ELEMENT (T) (OMEGA)))

H=:

(= (T) (X))                          %% In this case we have this equality unit

(1 . INDUCTION-2 . R . R . P-> . P-> . 2 .)

(ELEMENT (T) (OMEGA))
______________
  **-IMPLY-STOP-**  
Proceed-ed 

Proved by the GROUND-UNIT-PROVER     %% note
(1 . INDUCTION-2 . R . R . P-> . P-> . 2 .)
(ELEMENT (T) (OMEGA))
PROVED-
 @@@@@@
 @@@@@@
 PROVED: EX30 
 With MGU THETA: T 
                        Returned for the Theorem
(TIME-USED 2879 MSEC)

----------------------------------------------------------------------------
Ex36**********************************************************
>(pr ex36)

 Theorem Being Proved: EX36 

(-> (ELEMENT N OMEGA) (= (* 2 (SUM I 1 N I)) (* N (+ N 1))))

Theorem being proved -- Skolemized:  EX36 
		 
 H:  
H-Ground-units:
(ELEMENT (N) (OMEGA))

C:  (= (* 2 (SUM (I) 1 (N) (I))) (* (N) (+ (N) 1))) 

(1 .)
(= (* 2 (SUM (I) 1 (N) (I))) (* (N) (+ (N) 1)))
______________
  **-IMPLY-STOP-**  i

     ====== INDUCTION BEING TRIED ======

Please Type the INDUCTION VARIABLE, N: (You may Type NIL or CR -- if you wish to let it find the induction variable)  
Type N:  

Please Type the INDUCTION HYPOTHESIS, IH:
       
(You may Type NIL or CR -- if you wish to use  C  for the Ind. Hyp.)
       
Type IH:  
Ordinary INDUCTION ON: 
 (N)
USING INDUCTION HYPOTHESIS:
  (= (* 2 (SUM (I) 1 (N) (I))) (* (N) (+ (N) 1)))

C1:  (= (* 2 (SUM (I) 1 0 (I))) (* 0 (+ 0 1)))

C2:
(-> (& (ELEMENT (N) (OMEGA))
       (= (* 2 (SUM (I) 1 (N) (I))) (* (N) (+ (N) 1))))
    (= (* 2 (SUM (I) 1 (+ (N) 1) (I))) (* (+ (N) 1) (+ (+ (N) 1) 1))))

IS THIS OK?   Type:    (Y or N) y

(1 . INDUCTION-1 .)
(= (* 2 (SUM (I) 1 0 (I))) (* 0 (+ 0 1)))
 @@@@@@

PROVED-
 $$$$$-Induction-1 Proved with MGU-&&&: T

(1 . INDUCTION-2 .)
(-> (& (ELEMENT (N) (OMEGA))
       (= (* 2 (SUM (I) 1 (N) (I))) (* (N) (+ (N) 1))))
    (= (* 2 (SUM (I) 1 (+ (N) 1) (I))) (* (+ (N) 1) (+ (+ (N) 1) 1))))
______________
  **-IMPLY-STOP-**  
 @@@@@@

(1 . INDUCTION-2 . R .)
(-> (& (ELEMENT (N) (OMEGA))
       (= (* 2 (SUM (I) 1 (N) (I))) (+ (* (N) (N)) (N))))
    (= (+ (* 2 (SUM (I) 1 (N) (I))) (* 2 (N)) 2)
       (+ (* (N) (N)) (N) (N) (N) 1 1)))
 @@@@@@

(1 . INDUCTION-2 . R . P-> .)
(= (+ (* 2 (SUM (I) 1 (N) (I))) (* 2 (N)) 2)
   (+ (* (N) (N)) (N) (N) (N) 1 1))
______________
  **-IMPLY-STOP-**  hh

H-Ground-units:
(ELEMENT (N) (OMEGA))

H=:
(= (* 2 (SUM (I) 1 (N) (I))) (+ (* (N) (N)) (N)))

(1 . INDUCTION-2 . R . P-> .)
(= (+ (* 2 (SUM (I) 1 (N) (I))) (* 2 (N)) 2)
   (+ (* (N) (N)) (N) (N) (N) 1 1))
______________
  **-IMPLY-STOP-**  

    %% At this point when all fails, it tries PROVE=, and succeeds.

        +++++++++Try to Use  PROVE= to prove++++++++++ 

 (= (+ (* 2 (SUM (I) 1 (N) (I))) (* 2 (N)) 2)
    (+ (* (N) (N)) (N) (N) (N) 1 1)) 

        +++++++++Try to Use  PROVE= to prove++++++++++ 

 (= (+ (- (* (N) (N))) (- (N)) (- (N)) (- (N)) (N) (N)
       (SUM (I) 1 (N) (I)) (SUM (I) 1 (N) (I)))
    0) 

 @@@@@@

 $$$$$-Induction-2 Proved with MGU-&&&: T

 PROVED: EX36 

 With MGU THETA: T 
                        Returned for the Theorem
(TIME-USED 4609 MSEC)

%% Note: The use of  prove= (when all else fails) does indeed succeed in
proving the theorem.  If one were to use IMPLY for this kind of theorem
then a stronger REDUCE-SIMPLIFIER should be installed.  We feel that that
is a rather routine implementation task.

-- ---------------------------------------------------------------------------
Ex37*********************************************************************
>(pr ex37)

 Theorem Being Proved: EX37 

(-> (ELEMENT N OMEGA)
    (= (* 6 (SUM I A N (* I I))) (* N (+ N 1) (+ (* 2 N) 1))))

Theorem being proved -- Skolemized:  EX37 
		 
 H:  
H-Ground-units:
(ELEMENT (N) (OMEGA))

C:  (= (* 6 (SUM (I) (A) (N) (* (I) (I))))
       (* (N) (+ (N) 1) (+ (* 2 (N)) 1))) 

(1 .)
(= (* 6 (SUM (I) (A) (N) (* (I) (I))))
   (* (N) (+ (N) 1) (+ (* 2 (N)) 1)))
______________
  **-IMPLY-STOP-**  i

     ====== INDUCTION BEING TRIED ======
 @@@@@@

Ordinary INDUCTION ON: 
 (N)
USING INDUCTION HYPOTHESIS:
  (= (* 6 (SUM (I) (A) (N) (* (I) (I))))
     (* (N) (+ (N) 1) (+ (* 2 (N)) 1)))

C1:  (= (* 6 (SUM (I) (A) 0 (* (I) (I)))) (* 0 (+ 0 1) (+ (* 2 0) 1)))

C2:
(-> (& (ELEMENT (N) (OMEGA))
       (= (* 6 (SUM (I) (A) (N) (* (I) (I))))
          (* (N) (+ (N) 1) (+ (* 2 (N)) 1))))
    (= (* 6 (SUM (I) (A) (+ (N) 1) (* (I) (I))))
       (* (+ (N) 1) (+ (+ (N) 1) 1) (+ (* 2 (+ (N) 1)) 1))))

(1 . INDUCTION-1 .)
(= (* 6 (SUM (I) (A) 0 (* (I) (I)))) (* 0 (+ 0 1) (+ (* 2 0) 1)))
 @@@@@@

PROVED-
 $$$$$-Induction-1 Proved with MGU-&&&: T

(1 . INDUCTION-2 .)
(-> (& (ELEMENT (N) (OMEGA))
       (= (* 6 (SUM (I) (A) (N) (* (I) (I))))
          (* (N) (+ (N) 1) (+ (* 2 (N)) 1))))
    (= (* 6 (SUM (I) (A) (+ (N) 1) (* (I) (I))))
       (* (+ (N) 1) (+ (+ (N) 1) 1) (+ (* 2 (+ (N) 1)) 1))))
______________
  **-IMPLY-STOP-**  
Proceed-ed 
......................................... etc
(1 . INDUCTION-2 . R .)
(-> (& (ELEMENT (N) (OMEGA))
       (= (* 6 (SUM (I) (A) (N) (* (I) (I))))
          (+ (* (N) (N) (* 2 (N))) (* (N) (N)) (* (N) (* 2 (N))) (N))))
    (= (+ (* 6 (SUM (I) (A) (N) (* (I) (I)))) (* 6 (* (N) (N)))
          (* 6 (N)) (* 6 (N)) 6)
       (+ (* (N) (N) (* 2 (N))) (* (N) (N) 2) (* (N) (N))
          (* (N) (* 2 (N))) (* (N) 2) (N) (* (N) (* 2 (N))) (* (N) 2)
          (N) (* (N) (* 2 (N))) (* (N) 2) (N) (* 2 (N)) 2 1 (* 2 (N)) 2
          1)))

 %%  Note:  PROVE= is not strong enough to prove this in reasonable time.
            A fine algebraic simplifier routinely finishes this kind of proof.
---------------------------------------------------------------------

Ex42*************************************************************
>(pr ex42)                     C%%  See [Bl71], pp. 64-66.

(SET-= (OMEGA) (UNION-OF (OMEGA)))

 @@@@@@

(1 . D .)
(& (SUBSET (OMEGA) (UNION-OF (OMEGA)))
   (SUBSET (UNION-OF (OMEGA)) (OMEGA)))

 %%  The first Conjunct is proved by employing the Option, USE, 
         using the lemma:  Ex42-lemma1,
     and the second Conjunct is proved by induction.
 ----------------------------------------------------------------------

Ex47. ***************************************************************

>(pr ex47)

 Theorem Being Proved: EX47 

(-> (< A B) (SOME X (& (< A X) (< X B))))

++++++++++++++++++Theorem being proved -- Skolemized:  EX47 
		 
 H:  
H-Ground-units:
(< (+ (A) (- (B))))

C:  (& (< (+ (A) (- X))) (< (+ X (- (B))))) 

(1 .)
(& (< (+ (A) (- X))) (< (+ X (- (B)))))
______________
  **-IMPLY-STOP-**  ineq        %% Note: the user command, "ineq", asking for 
                                    the Inequality prover to be used.
.................
(1 . VE X .)

(-> (< (+ (A) (- (B)))) (< (+ (A) (- (B)))))
______________                  %% Note that this new subgoal is free of "x",
  **-IMPLY-STOP-**                  which has been eliminated by VE (variable
                                    elimination).
 @@@@@@

 PROVED: EX47 
 With MGU THETA: T   Returned for the Theorem
(TIME-USED 481 MSEC)
-------------------------------------------------------------------------

-----------------------------------------------------------------------------

Ex48 ************************************************************************
>(pr ex48)                           %% Like ex47

 Theorem Being Proved: EX48 

(-> (& (< C B) (< A C)) (SOME X (& (< A X) (< X B))))

++++++++++++++++++++++++Theorem being proved -- Skolemized:  EX48 
 H:  
H-Ground-units:
1 :  (< (+ (C) (- (B))))
2 :  (< (+ (A) (- (C))))

C:  (& (< (+ (A) (- X))) (< (+ X (- (B))))) 
(1 .)
(& (< (+ (A) (- X))) (< (+ X (- (B)))))
______________
  **-IMPLY-STOP-**  ineq
.................
(1 . VE X .)
(-> (& (< (+ (C) (- (B)))) (< (+ (A) (- (C))))) (< (+ (A) (- (B)))))
______________
  **-IMPLY-STOP-**  ineq
     +++++ Ground Inequality Prover +++++

           REQUIRES USER INPUT         

(A =
   (-> (& (< (+ (C) (- (B)))) (< (+ (A) (- (C))))) (< (+ (A) (- (B))))))
Is this a Theorem?  (Y or N) y                  %% User response
(1 . VE X .)
(-> (& (< (+ (C) (- (B)))) (< (+ (A) (- (C))))) (< (+ (A) (- (B)))))

 PROVED: EX48 
 With MGU THETA: T   Returned for the Theorem
  
  %% At present IMPLY does not have a ground inequality prover (like
  %%  the SUP-INF prover) so such simple theorems as 

  (-> (& (< (+ (C) (- (B)))) (< (+ (A) (- (C))))) (< (+ (A) (- (B)))))

  %%  are done by asking the user.  That will be added shortly.
