
;;;***************************************************************************
;;;
;;;                       P-IMPLY.LISP                              22 June 93
;;;                    (Potential Imply)
;;;
;;; (load "imply.lisp") (load-lisp-all) (load "p-imply.lisp")
;;;
;;;This version of IMPLY, P-imply,
;;;    "potential Back-chaining" (see below)
;;;     A precondition prover, P-PC, working backwards from the "bottom"
;;;     Q-prove, a prover for finishing a particular goal (OTTER?)
;;;
;;;Potential Back-chaining, P-BACK-CHAIN, tries to BC (back-chain) on a 
;;;  conclusion "clump" C, by first computing the "characteristics" of C, 
;;;  just enough to allow a UNIQUE selection of a lemma or hypothesis on which
;;;  to back-chain.  Then BC on that lemma, recalling P-IMPLY to 
;;;  prove-the-premise.  
;;;
;;;Start by a call to (P-Prove sko-thm) or (P-Prove nil thm) which processes it
;;;  (skolemize if necessary) and calls p-imply.
;;;

(defun lpi () (load "p-imply.lisp"))
(defvar *print* t)
(setq *reduce-light* nil)

(defun p-prove (sko-thm &optional thm) ;21June93
  (let ()
    (setq sko-thm (cond (sko-thm sko-thm)                          ;22Jun93 sko
		    (t (skolemize thm))))
    (p-imply nil sko-thm '(1) sko-thm t)))

(defun p-imply (H C tl &optional (THM C) first)  ;22June93  
  ;;Called by P-PROVE
  ;;All formulas are in skolemized-in-place form
  ;;tl is the theorem label. If ~FIRST=nil then it is the "first" call.
  ;;THM is the original value of C (the whole theorem).
  (let (C*  (thm* (list '-> H C)) theta tl* n tll)
    (and *print* (prog2 (print-tlC tl C) (format t "~%")))
    (cond (nil first                   ;22June93  For now it is skipping this 
	   (cond ((setq theta (Q-prove thm*)) (prove-rest theta H C tl THM))
		 ;;Can it Quick-prove thm*?
		 ((equal C (setq C* (p-pc C)))
		  ;;process C with the pre-condition prover
		  (format t "~%P-PC did not change C~%") 
		  (p-imply H C tl THM nil))
		 ((eq '[] C*) (prove-rest t H C tl THM))
                                              ;should we have THETA here?
		 (t (p-imply H C* tl THM t))))

	  ((eq t C) t)
	  ((atom C) nil)                  ;maybe (eval C)??

	  ((setq theta (p-back-chain H C tl THM)) theta) 

	  ((eq-car C '->) 
	   (setq tl* (append tl '(P->)))
	   (p-imply (and-on (cadr C) H) (caddr C) tl* THM))

	  ((eq-car C '&)
	   (setq tl* (new-tl tl '&))
	   (p-imply H (cadr C) tl* THM))

	  ((eq-car H 'or)
	   (setq tl* (new-tl tl 'or))
	   (p-imply H (cadr C) tl* THM))

;;	  ((eq-car C 'or)  (later))
;;	  ((eq-car C 'not) (later))
	  )))


(defun apply-sub (theta D)  
  ;;The substitution THETA is applied to the form D
   (cond ((or (eq t theta) (null theta)) D)
	 (t (sublis theta D))))

(defun bu-pc (A)      ;3July93  Bottoms-up precondition Prover
  ;;This takes the formula A and tries to abduct it,using a bottoms-up approach
  ;; returning a list W= (A* theta)
  ;;Eg, (bu-pc '(-> (P) (& (P) (-> (Q) (& (R) (Q))))))
  ;;         = ((-> (P) (-> (Q) (R))) t)
  ;;Eg, (bu-pc Q from GCLR example)=
  (let (W W* theta WL A* thL cnf)
    (cond ((atom A) (list A nil))                    ;Should we eval A?
	  ((eq-car A '->)                 ;A= (-> H C)
	   ;;This assumes that H and C have no common (instantiable) variables
	   (setq W (bu-pc (caddr A)))
	   (setq A* (imp-on (cadr A) (car W)))
	   (setq W* (bu-pc* nil A*))          
	   (setq theta (cond ((cadr W) (compose (cadr W) (cadr W*)))
			     (t (cadr W*))))      ;concern about conflicts??
	   (list (car W*) theta))

	  ((eq-car A '&)
	   (setq WL (mapcar #'(lambda (u) (list u (bu-pc u))) (cdr A)))
	   (setq A* (andify (mapcar #'car WL)))
	   (setq thL (mapcar #'cadadr WL))
	   (setq cnf (conflict thL))
	   (setq theta (cond (cnf t) (t (compose-all thL))))
	   (setq A* (cond (cnf A) (t A*)))
	   (setq W (bu-pc* nil A*))
	   (list (car W) (compose theta (cadr W))))
	  (t (bu-pc* nil A)) )))

(defun bu-pc* (H C &optional (tl '(pc 1)))    ;26June93  Called by BU-PC
  (let ((VH (variables-in H)) (VC (variables-in C)) 
	(tlc '(pcc)) (tlh '(pch)) headc headh C* H* thm* pc W Hi*)
    (cond ((p-imply H C tl (imp-on H C)) (list t t))
	  ((atom C) (list (imp-on H C) nil))              ;;(eval C) first ??
	  ((eq-car C '->) (bu-pc* (and-on (cadr C) H) (caddr C)))
	  ((eq-car H 'or) (setq W (bu-or-h  H C VH tlh))
                          (list (imp-on H (car W)) (cadr W)))
	  ((eq-car C '&)  (setq W (bu-and-c H C VC tlc))
	                  (list (imp-on H (car W)) (cadr W)))
	  ((and (eq-car H '&)            ;H= (& ... (or ...) ...)
		(some #'(lambda (Hi) (and (eq-car Hi 'or)(setq Hi* Hi)))
		      (cdr H)))
	   (bu-pc* (or-outt H Hi*) C tl))
	  ((eq-car C 'or) 
	   (cond ((some #'(lambda (u) (and (setq W (bu-pc* H u tl)) (cadr W)))
			(cdr C)) W)
		 (t (list (imp-on H C) nil))))
	  ((setq W (define= C H))
	   (setq C* (car W)) (setq P (cadr W))
	   (cond ((and (setq H* (define=-some H C W))
		       (setq pc (bu-pc (imp-on H* C*)))
		       (setq thm* (imp-on H* C*))
		       (not (equal& thm* (car pc)))) pc)
		 ((and (setq pc (bu-pc (imp-on H C*)))
		       (setq thm* (imp-on H C*))
		       (not (equal& thm* (car pc)))) pc)
		 (t (list (imp-on H C) nil))))

	  ((setq H* (define=-some H C nil))               ;5July93 This clause
	   (cond ((and (setq pc (bu-pc (imp-on H* C)))    ; is explosive
		       (setq thm* (imp-on H* C))
		       (not (equal& thm* (car pc)))) pc)
		 (t (list (imp-on H C) nil))))

	  (t (list (imp-on H C) nil)))))

(defun bu-and-c (H C VC tlc)       ;3July93  Called by BU-PC*
  ;;C has the form (& ...)
  (let (n uths tln)
    (setq n 0)
    (setq uths (mapcar #'(lambda (u) (setq n (1+ n)) 
			             (setq tln (append tlc (list n)))
				     (list u (bu-pc* H u tln)))
		       (cdr C)))
    (un-conflict uths VC '&)))

(defun bu-or-h (H C VC tlh)       ;3July93  Called by BU-PC*
  ;;H has the form (or ...). or (& ... (or ...) ...) if Hi*=(or ...), not nil
  (let (n uths tln)
    (setq n 0)
    (setq uths (mapcar #'(lambda (u) (setq n (1+ n)) 
			             (setq tln (append tlh (list n)))
				     (list u (bu-pc* u C tln)))
		       (cdr H)))
    (un-conflict uths VC 'or)))

(defun andify (L)   ;3July93  
  ;;Make the list L= (a b c ...) into a conjucntion (& a b c ...)
  (setq L (remove t L))
  (cond ((null L) t)
	((member nil L) nil)
	((null (cdr L)) (car L))
	(t (cons '& L))))

(defun cdr-& (H) (cond ((eq-car H '&) (cdr H)) (t (list H))))  ;29June93
	  
(defun CH-conflicts (uths VCH)  ;26June93  Called by BU-PC*
  ;;eg, (setq uths '(((P x z) ((x a) (y b) (z c))) 
  ;;                 ((Q x z) ((x a) (y d) (z e))) ))
  ;; (CH-conflicts uths '(x z)) = unths.  If e=c, then nil.
  (let (thi thj x y)
    (setq C-conflicts (mapcan #'(lambda (uth1)                 ;uth1=  (u  thi)
       (cond ((and 
;;	       (setq thi (cadr uth1))
	       (setq thi (cadadrr uth1))                       ;3July93
	       (not (eq t thi))
		   (some #'(lambda (xt) 
		     (and (member (setq x (car xt)) VCH)
			  (some #'(lambda (uth2)               ;uth2= (u2 thj)
			    (and
;;			     (setq thj (cadr uth2))
			     (setq thj (cadadrr uth2))         ;3July93
			     (not (eq t thj))
				 (some #'(lambda (ys)
				   (and (member (setq y (car ys)) VCH)
					(eq x y) 
					(not (equal (cdr xt) (cdr ys)))))
				       thj)))
				uths)))
			 thi))
	      (list uth1))))
			      uths)) ))

(defun characteristics (H C V)  ;21Jun93 Called by P-BACK-CHAIN
#|  (setq H  '(ALL S
	 (-> (& (CLAUSES-L S) (FINITE S) (NOT (E [] S)) (< 0 (EXL S)))
	     (SOME C (& (E C S) (CLAUSE-L C) (NU C)
		        (SOME L (& (E L C) (LITERAL-L L)
			           (ALL C* (ALL L*
					   (->
					    (& (E C* S) (NU C*) (E L* C*))
					    (<= (I L*) (I L))))))))))))
    (setq C '(SOME C (& (E C (S)) (NU C) (SOME L (& (E L C)
		    (ALL C* (ALL L* (->  (& (E C* (S)) (NU C*) (E L* C*))
					 (<= (I L*) (I L))))))))))
    (characteristics H C nil) =
|#
  (let (HH syms hsyms)
    (setq HH (cond ((eq-car H '&) (cadr H)) (t (list H))))
    (setq syms (symbols-of C))
     ;;Now, we are just using the symbols of C.  Later more.
    (mapcan #'(lambda (hyp) (cond ((and (setq hsyms (symbols-of hyp))
					(subset-p syms hsyms H C V))
				   (list (list syms hsyms)))))
	    H)))

(defun compose-all (ths)  ;26June93  Called by BU-PC*
  ;;Compose the non-nil substitutions in ths
  (let (th*)
    (setq th* t)
    (mapc #'(lambda (th) (cond (th (setq th* (compose th* th)))))
	  ths)
    th*))

(defun conflict (L)  ;3July93  Called by BU-PC  
  ;;Does the list L of substitutions have conflicting entries?
  ;;eg, (conflict '(((x a) (y b) (z c)) t ((x a) (y d) (s a)) t )) = t
  ;;    (conflict '(((x a) (y b) (z c)) t ((x a) (y b) (s a)) t )) = nil
  (some #'(lambda (th) (and (not (eq t th)) (some #'(lambda (xt)
    (some #'(lambda (th*) (and (not (eq t th*)) (some #'(lambda (ys)
              (and (eq (car xt) (car ys)) (not (equal (cdr xt) (cdr ys)))))
						      th*)))
	  L))
						  th)))
	L))




(defun define= (C H &optional (par t))          ;29June93    Called by BU-PC*
  ;;ie, (define= '(R) '(& (yy) (= xx (R)))) = (xx =)
  ;;    (define= '(e (t) (merge-low A)) nil) = ( (defn of C) C)

  (let (xx W def-ent B*)
    (cond ((some #'(lambda (Hi)                           ;Hi= (= C xx)
                    (and (eq-car Hi '=) (member C (cdr Hi) :test 'equal)
			 (or (and (equal C (cadr Hi)) (setq xx (caddr Hi)))
			     (and (equal C (caddr Hi)) (setq xx (cadr Hi))))))
			(cdr-& H))
	   (list xx '=))
	  ((setq def-ent (definition C par)) 
	   (list def-ent C))
	  ((member (car C) '(e element))                          ;C= (e x B)
	   (cond ((setq W (define= (caddr C) H par))
		  (setq B* (car W))
		  (cond ((eq-car (cdr W) '=)
			 (list (list (car C) (cadr C) B*) '=))
			(t (list (list (car C) (cadr C) B*) C))))
		 (t nil)))
	  (t nil) )))

(defun define=-some (H C W)    ;29June93   Called by BU-PC*
  (let ((cdrh (cdr-& H)) Wh Hi* H*)
    (cond ((some #'(lambda (Hi) 
              (and (setq Wh (define= Hi H nil)) (setq Hi* Hi)
		   (or (not W) 
		       (and (eq-car (cdr W) '=) (eq-car (cdr Wh) '=))
		       (and (not (eq-car (cadr W) '=)) (similar Hi C)))))
		 cdrh)
	   (setq H* (reduce-& (subst (car Wh) Hi* H :test 'equal)))  ;5July93
	   H*) )))

(defun definition (C par)            ;29 June93  called by DEFINE=
  (let (theta u99 C* PP Ps Pf* Vu (V (variables-in C)))
    (cond ((some #'(lambda (u)                     ;u= (C' xx yy PP PP)
	      (and (setq theta (instance C (car u)))
		   (or (setq Vu (crunch (variables-in u))) t)
		   (setq u99 u)
		   (or (mapc #'(lambda (x) 
                      (setq x99 (pack (append (explode x) '(9 9))))
		      (setq u99 (subst x99 x u99)))
			 Vu) t)
		   (setq theta (instance C (car u99)))))
		 definition-table)
	   (setq C* (cond (par (cadr u99)) ((caddr u99)) (t (cadr u99))))
	   (setq C* (apply-sub theta C*))
	   (setq PP (cond (par (cadddr u99)) (t (fifth u99))))
	   (mapc #'(lambda (P) 
		 (cond ((setq Ps (find-all P C*))    ;Ps= ((P x y) ...)
                        (mapc #'(lambda (pf)         ;Pf=  (P x y)
			            (setq Pf* (append Pf V)) ;Pf*= (P x y u v) 
				    (setq C* (subst Pf* pf C* :test 'equal)))
			      Ps))))
		 PP)
	   C*)
	  (t nil))))

(setq definition-table '(
 ((e x (sng y)) (= x y))
 ((e x (U A B)) (or (e x A) (e x B)))
 ((e x (~ A B)) (& (e x A) (not (e x B))))
 ((e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2))))
  (or (e x (U (~ C1 (sng L1)) (~ C2 (sng L2)))) (= x L)))

 ((e x (merge-low C))
  (& (e x C) (-> (& (e (L_) C) (= (lit (L_)) (lit x)))
		 (<= (i x) (i (L_)))))
  (& (e x C) (-> (& (e L_ C) (= (lit L_) (lit x)))
		 (<= (i x) (i L_))))
  nil (L_))

))

(defun eqs-car (A L) ;26June93  Called by BU-PC
  ;;If A is not an atom and (car A) is in L, return t, else nil
  (and (not (atom A)) (member (car A) L :test 'equal)))

(defun equal& (A B)    ;3July93
  ;;Allows reordering of the conjuncts
  (cond ((equal A B) t)
	((or (atom A) (atom B)) nil)
	((not (equal (car A) (car B))) nil)
	((or (eq '& (car A)) (eq 'or (car A)))
	 (and (subset& (cdr A) (cdr B)) (subset& (cdr B) (cdr A))))
	((member (car A) '(-> not))
	 (every #'(lambda (u v) (equal& u v)) (cdr A) (cdr B)))
	(t nil)))

(defun subset& (L1 L2) ;3July93   Called by EQUAL&
  (every #'(lambda (A) (some #'(lambda (B) (equal& A B)) L2)) L1))
 
(defun explode (A) 
  (map 'list #'(lambda (char) (intern (string char)))
       (symbol-name a)))

(defun find-all (P C)    ;29June93   Called by DEFINITION
  ;;Return the list of all formulas of the form (P ...) in C.
  (let ()
    (cond ((atom C) nil)
	  ((eq-car C P) (list C))
	  (t (union-w (find-all P (car C))
		      (find-all P (cdr C)))) )))

(defun first-integer-or-o (tll) ;23June93  Called by PROVE-REST
  (cond ((null tll) nil)
	((or (integerp (car tll)) (member (car tll) '(o1 o2 o3 o4 o5)))
	 (car tll))
	(t (first-integer-or-o (cdr tll)))))

(defun imp-on (H C) 
  (cond ((null H) C)
	((eq t H) C)
	((eq t C) t)
	((eq nil C) nil)
	(t (list '-> H C))))

(defun imp-on* (A B)    ;called by PC               27May92
  (cond ((eq-car B '->) (imp-on* (and-on A (cadr B)) (caddr B)))
	(t (imp-on A B))))

(defun new-tl (tl action) ;23June93   Called by P-IMPLY
  (let ((tll (reverse tl)) n nl n*)
    (cond ((eq '& action)
	   (setq tl* (cond ((integerp (setq n (car tll)))
			    (reverse (cons (1+ n) (cdr tll))))
			   (t (append tl '(1))))))
	  ((eq 'or action)
	   (setq tl* (cond ((integerp (setq n (car tll)))
			    (append tl '(o1)))
			   ((eq 'o1 (car tll))
			    (reverse (cons 'o2 (cdr tll))))
			   ((eq 'o2 (car tll))
			    (reverse (cons 'o3 (cdr tll))))
			   ((eq 'o3 (car tll))
			    (reverse (cons 'o4 (cdr tll))))
			   (t (append tl '(o1))))))
	  (t tl))))

(defun non-nils (L)  ;26June93  Called by BU-PC*
  ;;The non-nil members of L
  (crunch (mapcan #'(lambda (u) (cond (u (list u)))) L)))

(defun orify (L)   ;5July93  
  ;;Make the list L= (a b c ...) into a disjucntion (or a b c ...)
  (setq L (remove nil L))
  (cond ((null L) nil)
	((member t L) t)
	((null (cdr L)) (car L))
	(t (cons 'or L))))

(defun or-outt (H Hi)  ;5July93  Called by BU-PC*
  ;;eg, (or-outt '(& A (or B C D) E F) '(or B C D))  where Hi= (or B C D)
  ;; =  (or (& A B E F) (& A C E F) (& A D E F))
   (cons 'or (mapcar #'(lambda (u) (subst u Hi H :test 'equal)) (cdr Hi))))

(DEFUN PACK (L)

;  PACK returns a symbol interned in USER whose print name
;  consists of the characters that are printed by PRINCing
;  the members of L with *PRINT-BASE* 10.

  (LET ((*PRINT-PRETTY* NIL)
	(*PRINT-BASE* 10)
	(*PRINT-RADIX* NIL)
	(*PRINT-LEVEL* NIL)
	(*PRINT-LENGTH* NIL)
        (*PRINT-CASE* :UPCASE))
    (INTERN (FORMAT NIL "~{~A~}" L)
            'USER)))

(defun p-back-chain (H C tl THM &optional (C* t))
  ;;22Jun93  Called by P-IMPLY
  (let ((sC (structure-size C)) (sH (structure-size H)) 
	 theta lambda C** B* H**)
    (cond ((atom H) nil)
	  ((and (setq theta (unify-> H C))
		(setq lambda (pprove-the-premise theta H C tl THM C*)))
	   (compose theta lambda))

	  ((< (+ sH 1) sC) nil)  ;H must be as "large" as C

	  ((eq-car H '->)
	   (setq C** (and-on (cadr H) C*))
	   (p-back-chain (caddr H) C (append tl '(P->)) THM C**))

	  ((eq-car H '&)
	   (cond ((some #'(lambda (H*) 
                    (setq theta (p-back-chain H* C tl THM C*)))
			(cdr H))
		  theta)))
      
	  ((eq-car H 'or)
	   (cond ((some #'(lambda (H*) 
              (setq B* (delete-one H* H))
	      (setq C** (and-on (not-on B*) C*))
	      (setq theta (p-back-chain H* C tl THM C**)))
			(cdr H))
		  theta)))

	  
	  ((and (eq '-> (car H)) (EQ '-> (caaddr H)))
	   ;;eg, H = (-> H1 (-> H2 C1))
	   (setq H** (list '-> (and-on (cadr H) (cadr (caddr H)))
			   (caddr (caddr H))))
	   ;;eg, H** = (-> (& H1 H2) C1)
	   (back-chain H** C tl THM C*))
	  )))

(defun print-tlc (tl C)  ;3July93   Called by P-IMPLY
  (let ((n (length tl)) tn (s* '********************))
    (setq tn (- (* n 2) 2))
    (format t "~%") (prsp tn) (format t "~s" tl) 
    (format t "~%") (prsp tn) (format t "~s" s*)
    (format t "~%") (prsp tn) (format t "~s~%" C)))

(defun prsp (n)  ;Print n spaces
  (cond ((< n 1) (format t ""))
	((< n 2) (format t "  "))
	((< n 3) (format t "   "))
	((< n 4) (format t "    "))
	((< n 5) (format t "     "))
	((< n 6) (format t "      "))
	((< n 7) (format t "       "))
	((< n 8) (format t "        "))
	((< n 9) (format t "         "))
	((< n 10)(format t "          ")) 
	((< n 11)(format t "           ")) 
	((< n 12)(format t "            ")) 
	((< n 13)(format t "             ")) 
	((< n 14)(format t "              ")) 
	((< n 15)(format t "               ")) 
	((< n 16)(format t "                ")) 
	((< n 17)(format t "                 ")) 
	((< n 18)(format t "                  ")) 
	((< n 19)(format t "                   ")) 
	((< n 20)(format t "                    ")) 
	(t       (format t "                     ")) ))

(defun prove-rest (theta H C tl THM)  ;22Jun93  
  ;;Called by P-IMPLY, PPROVE-THE-PREMISE
  ;;The idea is to look at the original theorem being proved, THM,
  ;;  replace the part proved, C, by t, and apply THETA at the
  ;;  appropriate places in THM, and call p-imply to prove the
  ;;  resulting THM.
  (let (THM* (tll (reverse tl)) n th2)
    (setq n (first-integer-or-o tll))
    (setq THM* (cond ((integerp n) 
		      (setq thm2 (subst t C THM))
		      (put-tnil '& t theta thm2))
		     ((member n '(o1 o2 o3 o4 o5))
		      (setq thm2 (subst nil H THM))
		      (put-tnil 'or nil theta thm2))
		     (t (prnt 'ALARM 'prove-rest))))
    (setq THM* (reduce-tnil THM*))
    (p-imply nil THM* tl THM* t)))

(defun pprove-the-premise (theta H C tl THM C*)
  (let (lambda sigma)
    (setq C* (apply-sub theta C*))
    (cond ((and (setq lambda (p-imply H C* tl (imp-on H C*) t)) 
		(or (setq lambda (compose theta lambda)) t)
		(setq sigma (prove-rest lambda H C tl THM))
		(compose theta sigma)))
	  (t nil))))

(defun put-tnil (action tnil theta THM)  ;23June93 Called by PROVE-REST
  (cond ((equal tnil THM) tnil)
	((atom THM) THM)
	((eq-car THM action)
	 (cond ((not (occur tnil THM)) THM)
	       ((Some #'(lambda (u) (and (occur tnil u)(not (occur action u))))
		      (cdr THM))
		(apply-sub theta THM))
	       (t (cons (car THM) (mapcar #'(lambda (u)
                                             (put-tnil action tnil theta u))
				       (cdr THM))))))
	(t (cons (car THM) (mapcar #'(lambda (u)
                                             (put-tnil action tnil theta u))
				       (cdr THM))) )))

(defun reduce-tnil (A)  ;22June93  Called by PROVE-REST
  ;;The goal here is to evaluate those subformulas involving t or nil
  ;;Eg, (reduce-tnil '(&  (P x) Q  t  (R y))) = (&  (P x) Q (R y))
  ;;    (reduce-tnil '(or (P x) Q nil (R y))) = (or (P x) Q (R y))
  (let (A*)
    (cond ((equal A (setq A* (reduce-tn2 A))) A)
	  (t (reduce-tnil A*)))))

(defun reduce-tn2 (A)  ;22June93  Called by REDUCE-TNIL
  (let (A*)
    (cond ((eq A t) t)
	  ((atom A) A)
	  ((eq-car A '&) 
	   (cond ((equal A (setq A* (and-off A t)))
		  (cons '& (mapcar #'reduce-tn2 (cdr A))))
		 (t (reduce-tn2 A*))))

	  ((eq-car A 'or) 
	   (cond ((equal A (setq A* (or-off A t)))
		  (cons 'or (mapcar #'reduce-tn2 (cdr A))))
		 (t (reduce-tn2 A*))))
	  ((eq-car A '->)
	   (setq A* (list '->  (reduce-tn2 (cadr A))
			       (reduce-tn2 (caddr A))))
	   (cond ((eq t (caddr A*)) t)
		 ((eq nil (cadr A*)) t)
		 ((eq nil (caddr A*)) (list 'not (cadr A*)))
		 ((eq t (cadr A*)) (caddr A*))
		 (t A*)))
	  ((eq-car A 'not) (list 'not (reduce-tn2 (cadr A))))
	  (t A))))

(defun reduce-& (H)  ;5July93  Remove double &'s in H
  ;;eg, (reduce-& '(& A (& B C) D)) = (& A B C D)
  (cond ((eq-car H '&) (cons '& (crunch (reduce-&2 (cdr H))))) (t H)))

(defun reduce-&2 (L) ;5July93  Called by REDUCE-&
  (cond ((null L) nil)
	((eq-car (car L) '&)                      ;L= ((& A B C) D E)
	 (append (cdar L) (reduce-&2 (cdr L))))
	(t (cons (car L)  (reduce-&2 (cdr L)))) ))

(defun similar (H C)   ;29June93   Called by DEFINE=-SOME
  ;;eg (similar '(e (t) A) '(e (t) B)) = T
  ;;   (similar '(e (t) A) '(e (s) B)) = nil
  (let()
    (cond ((equal H C) t)
	  ((atom H) nil)
	  ((atom C) nil)
	  ((not (eq (car H) (car C))) nil)
	  ((some #'(lambda (u v) (equal u v)) (cdr H) (cdr C)) t)
           ;;Seems a little severe here.  Needs to be worked on.
	  (t nil))))

(defun structure-size (C) ;21Jun93  Called by UNIFY->
  (let ()
    (cond ((atom C) 1)
	  ((member (car C) '(& or not ->))
	   (1+ (apply '+ (mapcar 'structure-size (cdr C)))))
	  (t 1))))

(defun subset-p (syms hsyms H C V &optional par) ;21Jun93 C. by CHARACTERISTICS
  ;;each symbol f of syms should either be in hsyms or be in 
  ;; a "variable position" of hsyms.  
  ;;Eg, (subset-p '(g h) '(h) '(P y (f x)) '(P y (f (g z))) '(x))
  (let (s1 s2)
    (cond ((and (not par) (eq-car H 'all))
	   (subset-p syms hsyms (caddr H) C (cons (cadr H) V) par))
	  ((and par (eq-car H 'some))
	   (subset-p syms hsyms (caddr H) C (cons (cadr H) V) par))
	  ((eq-car H '->)
	   (or (subset-p syms hsyms (cadr H)  C V (not par))
	       (subset-p syms hsyms (caddr H) C V par)))
	  (t (setq s1 (mapcan #'(lambda (s)	;Those in hysms
			 (cond ((member s hsyms :test 'equal) (list s))))
		     syms))
	     (setq s2 (set-difference syms s1))	;Those not in hsyms
	     (cond ((null s2) t)
		   ((every #'(lambda (s) (var-position s H C V par)) s2) t)
		   (t nil))))))

(defun symbols-of (C)			;21Jun93 Called by CHARACTERISTICS
  ;;eg, (symbols-of '(SOME C (& (E C (S)) (NU C) (SOME L (& (E L C)
  ;;		    (ALL C* (ALL L* (->  (& (E C* (S)) (NU C*) (E L* C*))
  ;;					 (<= (I L*) (I L))))))))))
  ;; =  (NU E <= I)
  (crunch (syms C)))

(defun syms (C)				;Called by SYMBOLS-OF
  (cond ((atom C) nil)
	((member (car C) '(all some)) (syms (caddr C))) ;not enough??
	((member (car C) '(& some not ->))
	 (mapcan #'syms (cdr C)))
	;;	((null (cdr C)) nil)
	(t (cons (car C) (mapcan #'syms (cdr C))))))

(defun un-conflict (WL VCH &or)    ;3July93    Called by BU-AND-C, BU-OR-H
  ;;WL is a list of elments, W= (u (u* theta)), 
  ;; where u* is the pc of u with substitution theta.
  ;;If &or=&, want to conjoin, for each W,
  ;; t if theta=t, 
  ;; u if theta is nil, or if theta conflicts with another theta in WL.
  ;;ie,    
  ;;WLn=  those u's not proved
  ;;WLn2= those u's proved, but with a conflict (so assume not proved)
  ;;WLP = those u*'s for u is pc'ed to u* (with theta)
  ;;conjoin (WLn U WLn2 U WLP), andify to A*, return (A* theta)

  (let (WLn WLt WLn2 WLP WLth th thm)                    ;th=theta
    (mapc #'(lambda (W) (setq th (cadadr W))             ;W= (u (u* th))
                (cond ((not th) (setq WLn (cons (car W) WLn)))
		      ((eq t th) (setq WLt (cons (caadr W) WLt)))
		      (t (setq WLth (cons W WLth)))))
	  WL)
    (cond ((and (not WLn) (not WLth))
	   (setq A* (cond ((eq '& &or) (andify WLt)) (t (orify WLt))))
	   (return-from un-conflict (list A* t)))
	  ((equal& (cons &or WLn) (setq thm (cons &or (mapcar #'car WL))))
	   (return-from un-conflict (list thm nil))))
    (setq CH-conflicts (CH-conflicts WLth VCH))
    (setq WLth* (set-difference WLth CH-conflicts))
    (setq theta (compose-all (mapcar #'cadadr WLth*)))
    (setq WLn2 (mapcar #'car CH-conflicts))
    (setq WLP (mapcar #'caadr WLth*))
    (setq WL* (append WLn WLn2 WLP))
    (setq A* (cond ((eq '& &or) (andify WL*)) (t (orify WL*))))
    (list A* theta) ))

(defun unify-> (H C)   ;21Jun93  
  ;;This is high level unification, allowing also for H to have
  ;; conjuncts that C does not, etc.
  ;;This assumes that H and C are skolemized (in place)
  ;;Eg, 
  (let ((sC (structure-size C)) (sH (structure-size H)) theta sigma C1 H1)
    (cond ((setq theta (unify H C)) theta)
	  ((or (atom H) (atom C)) nil)
;;	  ((< (+ sH 1) sC) nil)  ;H must be as "large" as C
	  ((and (eq-car C '&) (eq-car H '&))
	   (unify-L (cdr H) (cdr C)))
	  ((and (not (eq-car C '&)) (eq-car H '&))
	   (unify-L (cdr H) (list C)))
	  ((and (eq-car C '->) (eq-car H '->))
	   (cond ((setq theta (unify-> (caddr H) (caddr C)))
		  (setq C1 (apply-sub theta (cadr C)))
		  (setq H1 (apply-sub theta (cadr H)))
		  (cond ((setq sigma (unify-> C1 H1))
			 (compose theta sigma))))))
;;	  ((or not etc later))
	  (t nil))))

(defun unify-L (LH LC)  ;21Jun93  Called by UNIFY->
  (let (theta sigma u* LH* LC*)
    (cond ((null LC) t)
	  ((some #'(lambda (u) 
            (and (setq theta (unify-> u (car LC)))
		 (or (setq LH* (apply-sub theta (remove u LH :test 'equal))) t)
		 (or (setq LC* (apply-sub theta (cdr LC))) t)
		 (setq sigma (unify-L LH* LC*))))
		 LH)
	   (compose theta sigma))
	  (t nil))))

(defun var-position (s H C V &optional par)	;21Jun93 Called by SUBSET-P
  ;;Is the symbol s in a variable position of H?
  ;;Ie, Is s part of a subformula (f (s ..)) in C for which there is
  ;;    a subformula (f x) of H with x in V?
  ;;Eg, (var-position 'g '(P y (f x)) '(P y (f (g z))) '(x)) = t
  (let ()
    (cond ((not (occur s C)) nil)
	  ((atom C) nil)
	  ((member (car C) '(all some)) (var-position s H (caddr C) V par))
	  ;;Should we put (cadr C) in V here?? In one of these cases?
	  ((member (car C) '(& or not ->))
	   (some #'(lambda (C*) (var-position s H C* V par)) (cdr C)))
	  ((or (and par (eq-car H 'all)) (and (not par) (eq-car H 'some)))
	   (var-position s (cadr H) C (cons (cadr H) V) par))
	  ((atom H) nil)
	  ((member (car H) '(all some)) (var-position s (caddr H) C V par))
	  ((eq-car H '&)
	   (some #'(lambda (H*) (var-position s H* C V par)) (cdr H)))
	  ((instance-p s H C V) t)
	  (t nil))))

(defun instance-p (s H C V)		;21Jun93  Called by VAR-POSITION
  ;;Eg, (instance-p 'g  '(P a (f b z)) '(P a (f b (g c d))) '(z)) = t
  (let ()
    (cond ((equal C H) t)
	  ((and (eq-car C s) (member H V)) t)	;C= (s ...)
	  ((atom C) nil)
	  ((atom H) nil)
	  ((not (eq (car C) (car H))) nil)
	  (t (every #'(lambda (x y) (instance-p s x y V)) H C)))))

;;--------------------------------------------------------------
;;@
(load "strive-1-1/load-latest.lsp")
(si::chdir "struve")
(or (find-package "PVR")
    (make-package "PVR" :use
		  '(SLOOP LISP user)))
(import '(user::sb-prove) "pvr")
(load "struve-pvr.lsp")
(si::chdir "../")
(defun convert-for-struve (thm)               ;27Jul93
  ;;convert each atom (P x ... y) to (nonempty (ff x ... y))
  (cond ((atom thm)  thm)
	((null (cdr thm)) thm)
	(t (case (car thm)
		 ((some or not)
		  (cons (car thm) (mapcar #'convert-for-struve (cdr thm))))
		 ((& and user::& user::and)
		  (cons 'user::& (mapcar #'convert-for-struve (cdr thm))))
		 ((-> user::->)
		  (cons 'user::-> (mapcar #'convert-for-struve (cdr thm))))
		 ((-> <-> iff user::<->)
		  (cons 'user::<-> (mapcar #'convert-for-struve (cdr thm))))
		 ((all user::all)
		  (cons 'user::all (mapcar #'convert-for-struve (cdr thm))))
		 (([ proper-subset subset* user::[)
		  (cons `user::[  (mapcar #'convert-for-struve1 (cdr thm))))
		 (([= subset user::[=)
		  (cons `user::[= (mapcar #'convert-for-struve1 (cdr thm))))
		 (=  (cons `user::=  (mapcar #'convert-for-struve1 (cdr thm))))
		 ((element e user::element)
		  (cons `user::element (mapcar #'convert-for-struve1 
					       (cdr thm))))
		 ((nonempty user::nonempty)
		  (cons 'user::nonempty (mapcar #'convert-for-struve1
						(cdr thm))))
		 (t (list 'user::nonempty (convert-for-struve1 thm)))))))

(defun convert-for-struve1 (expr)	;27Jul93
  (cond ((atom expr) expr)
	((null (cdr expr)) expr)
	(t (case (car expr)
		 ((singleton sng user::singleton)
		  (cons 'user::singleton (mapcar #'convert-for-struve1 
						 (cdr expr))))
		 ((intersection user::intersection) 
		  (cons 'user::intersection (mapcar #'convert-for-struve1 
						    (cdr expr))))
		 ((union u user::union)
		  (cons 'user::union (mapcar #'convert-for-struve1
					     (cdr expr))))
		 ((~ set-difference set-diff user::set-diff)
		  (cons 'user::set-diff (mapcar #'convert-for-struve1
						(cdr expr))))
		 (t (cons (car expr) (mapcar #'convert-for-struve1 (cdr expr)))
		    )))))

(setq pvr::non-input-cl-penalty 1)

(defun struve-prove (formula &optional (x 40) (y 40)) 
  (or (pvr::sb-prove (convert-for-struve formula) x y)
      (pvr::sb-prove (convert-for-struve `(some larry ,formula)) x y)))

(defun finish-by-struve (CL)              ;27Jul93
  ;;This version was patterned after the 15Jan93 version of
  ;; FINISH-BY-STRIVE in SET-VAR.LISP
  ;; 
  ;;Try to find [] by the STR+VE-subset prover.  But first,
  ;; Remove ALL parents of any Rule1 or Rule11 literals, (not yet)
  ;;    and all parents of *set-theory-definitional-clauses*
  ;; Collapse ALL set-of literals
  ;; 29oct92: For collapsing, use an ELIGIBLE-FOR-COLLAPSE (set-of ...)
  ;;   to replace all A's in CL. 
  ;; Subsume all possible  (for now we will assume that this is already done)
  (let (S6 S7 S8 S9 S10 CL2 CL4 Thm RR H Thm* CL* a-value V SxS) 
    (prnt-%%)
    (format t "~%Finish-by-Struve CL= ~s~%" (mapcar #'car CL))          
    (setq S6 (mapcan #'(lambda (C) 
			  (cond ((equal (caddr C) '(input)) (list C)))) *CL*))
    (setq S7 (set-difference S6 *set-theory-definitional-clauses*))
    (setq S8 (remove (C 0) S7 :test 'equal))
     ;;S8 is the input clauses except for the set-theory definitional ones.
    (format t "~%S8 ~s~%" (mapcar #'car S8)) 
    (setq S9 (mapcan #'(lambda (C) 
		  (cond ((or (eq-car (caddr C) 'r1) (eq-car (caddr C) 'r11)
			     (equal '(0 1) (caddr (caddr C))))
			 (list C)))) *CL*))
     ;;S9 consists of R1 and R1.1 resolvents, and collapsed clauses
    (setq S10 (sort (union-w S8 S9) 'clause-number<))
    (setq CL2 (remove-r1-parents-cr=cr S10))
    (setq CL2 (remove-parent-of-collapsed CL2))
    (cond ((setq a-value (select-an-a-value-u CL2))               ;6Jan93
	   (prnt 'a-value-u)
	   (setq CL4 (collapse-with-value CL2 *A* A-value t)))    ;13Jan93
	  ((setq a-value (select-an-a-value CL2))                 ;24Nov92
	   (prnt 'a-value-not-u)
	   (setq CL4 (collapse-with-value CL2 *A* A-value)))      ;13Jan93
	  ((not (occur 'set-of CL2)) 
	   (setq CL4 CL2))
	  (t (return-from finish-by-struve nil)))                 ;6Jan93
    (prnt A-value)
    (format t "~%CL4= ~s~%" (mapcar #'car CL4))
    (setq CL4 (collapse-with-value CL4 *A* A-value t))
    (and A-value (setq CL4 (subst (list *A*) A-value CL4 :test 'equal)));4Feb93
    (setq CL4 (red* CL4))
    (setq SxS (split-s (mapcar #'cadr CL4)))                      ;15Jan93
;;%%%%%%4Feb93 The following is for today only, until Larry fixes Str+ve
    (setq SxS (mapcan #'(lambda (S) 
      (cond ((some #'(lambda (C) (equal C '((NOT (<<= R (B1)))))) S) nil)
	    (t (list S))))
		      SxS))                
    (cond ((every #'(lambda (S) 
      (prnt S) 
      (setq Thm (negate-&-requantify S))
      (setq Thm (rewrite-subset-set-of thm))        ;5oct92
      (setq thm* (remove-constant-parentheses Thm))
      (setq thm* (universally-quantify thm*))       ;12oct92
      (setq thm* (subst '<= '<<= thm*))      
      (setq Thm* (unique-ize-quan-vars Thm*))
      (setq Thm* (subst 'z+ 'z% thm*))
      (setq Thm* (convert-for-struve Thm*))         ;27Jul93,20Nov92
      (setq RR (pvr::sb-prove thm* 40 4000)) ;27Jul93
      (cond ((equal ':TIME-LIMIT RR)                                  ;4Feb93
	     (format t "~%:TIME-LIMIT~%") nil)
	    ((equal ':FAILED RR)
	     (format t "~%:FAILED~%") nil)
	    (RR '[])
	    (t nil))) SxS)
	   '[])
	  (t nil)) ))

(push 'pvr::subset pvr::prover-features)
(push 'pvr::inequality pvr::prover-features)
;;(load "struve/struve-convert")

(pvr::sb-prove (convert-for-struve '(e (x) (sng (x)))) 40 40)
;;@
(defun tr () (trace p-prove p-imply p-back-chain pprove-the-premise 
                    prove-rest))
(defun trb () (trace bu-pc bu-pc*))

(setq thm1 '(-> (& (P a) (P b) (Q b)) (some x (& (P x) (Q x)))))
(defun p1 () (time (p-prove nil thm1)))

(setq thm2 '(-> (all x (P x)) (& (P a) (P b))))
(defun p2 () (time (p-prove nil thm2)))

(setq thm3 '(-> (& (Q a) (Q b)) (some x (& (-> (P x) (& (P a) (P b))) (Q x)))))
(defun p3 () (time (p-prove nil thm3)))

(setq thm4 '(-> (P) (& (P) (-> (Q) (& (R) (Q))))))
(defun b4 () (bu-pc thm4))

(setq thm5 '(-> (e x (U (~ C1 (sng L1)) (~ C2 (sng L2))))
		(e x (U (~ C1 (sng L1)) (~ C2 (sng L2))))))
(defun b5 () (bu-pc thm5))

(setq thm6 '(-> (e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2)))) 
		(e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2))))))
(defun b6 () (bu-pc thm6))

(setq thm7
 '(-> (merge-low (e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2))))) 
      (merge-low (e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2)))))))
(defun b7 () (bu-pc thm7))

(setq fml8 '(-> (& (= (A) (U (~ C1 (sng L1)) (~ C2 (sng L2))))
		   (= (B) (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2)))))
		(-> (e (t) (merge-low (A))) (e (t) (merge-low (B))))))
(setq fml8 '(-> (& (= (A) (U (~ (C1) (sng (L1))) (~ (C2) (sng (L2)))))
		   (= (B) (U (~ (C1) (sng (L1))) 
			     (~ (U (C2) (sng L)) (sng (L2))))))
		(-> (e (t) (merge-low (A))) (e (t) (merge-low (B))))))
(setq fml8 '(-> (e (t) (merge-low (U (~ (C1) (sng (L1))) (~ (C2) (sng (L2))))))
		(e (t) (merge-low (U (~ (C1) (sng (L1))) 
				     (~ (U (C2) (sng (L))) (sng (L2))))))))

(defun b8 () (time (bu-pc fml8)))

