
;;;***************************************************************************
;;;
;;;                       PCI.LISP                              7 July 93
;;;         (INSTANCE version of pre-condition prover)
;;;
#|
    (load "imply.lisp") (load-lisp-all) (load "p-imply.lisp")
    (load "examples-pbrL.lisp") (load "hq-cq")(ls)(load "pci.lisp")
|#
;;;
;;;This version of Prec-condition prover used INSTANCE instead of UNIFY
;;; assumming that the goal is always "ground", and only the hypothesis
;;; has variables to be instantiated.  
;;;

;;; WWB: continue or start a rule only if there is HOPE.  How?
;;; WWB: the test for (equal C* C) below is not realistic,
;;;      they are equal in essence many times when changed.  How to fix?

(defun lpci () (load "pci.lisp"))
(defvar *print* t)
(defvar *reduce-light* nil)
(defvar *reduce* nil)
(defvar *plan* nil)
(defvar *plan** nil)
(defvar *C-list* nil)
(defvar %theorem-variables% nil)
(defun tri() (trace (pci :entry `(pci ,(cadr  si::arglist))) ;prints only thm
		    (bci :entry `(bci ,(cadr si::arglist))) ;prints only C
		    ))

(defun pci-start (H C &optional H* C*)  ;14July93
  (let ()
    (cond ((eq t H) (return-from pci-start (pci-start nil C H* C*))))
    (cond ((eq t C) (return-from pci-start t)))
    (setq %theorem-variables% nil)
    (setq H* (cond (H* H*) (t (skolemize H nil nil))))
    (setq C* (cond (C* C*) (t (skolemize C nil  t ))))
;;(setq lh (length H*)) (setq lc (length C*)) (prnt lh lc) (break)
    (pci H* C*)))

(defun pci (H C &optional (bc 2))  ;NEW VERSION 10Aug93
  ;;10Aug93: Now pci always returns a pair, either (t t) or (t A1) or (nil A1).
  ;;The first t means that there was a successful change. The second t means
  ;; the proof was completed, the second A1 means the returned change in A.
  ;;bc is a parameter used by the function bci.  We allows recalling of
  ;; bci only twice, bc keeps track of that.
  ;;If (bc < 1), then this can only return (t t) or (nil A).
  ;;eg, (pci '(& (P) (-> (P) (Q)) (-> (Q) (R))) '(R)) = t
 (let (P (A (imp-on* H C)) C* E W WW HC H* Hi*)
  ;;Normally we do not used a *plan*. 
  (cond (*plan* (setq P (car *plan*)) (setq *plan* (cdr *plan*))))
  (cond
   ((eq P 'halt) (format t "~%Halted by User~%") (break) (pci H C bc))
   ((eq P 'tt) '(t t))
   ((eq P 'ttt) (setq *plan** *plan*) (setq *plan* nil)
    (cond ((equal '(t t) (pci H C bc)) (setq *plan* *plan**) '(t t)) 
	  (t '(nil A))))
   ((eq P 'end) '(nil A))
   ((and 1 (eq t C)) '(t t))
   ((and 2 H (instance C H)) (format t "~%Instance H~%") '(t t))
   ((and 3 (eq-car H '&) (some #'(lambda (Hi) (instance C Hi)) (cdr H))) 
     (format t "~%Instance Hi~%") '(t t))
   ((and 4 H (or (not P)(test P 'bci)) (eq-car (setq W (bci H C t H bc)) t)) W)
   ((and 5 (eq-car H 'or) (test P 'or-h))
    (setq C* (andify (mapcar #'(lambda (Hi) (imp-on Hi C)) (cdr H)))) 
    (format t "~%Or-H")
    (cond ((eq-car (setq W (pci nil C* bc)) t) W) (t (list nil A))))
   ((and 6 (eq-car H '&) (test P 'or-hi)
	 (some #'(lambda (Hi) (and (eq-car Hi 'or) (setq Hi* Hi)))
	       (cdr H))) (format t "~%Or-Hi")
    (cond ((eq-car (setq W (pci nil (or-outi H Hi* C) bc)) t) W) 
	  (t (list nil A))))
   ((and 7 (eq-car C '&) (test P 'and-c) (or (format t "~%And-C") t))
    (setq WW (mapcar #'(lambda (Ci) (format t "~%And-C Next") (pci H Ci bc))
		     (cdr C)))
    (cond ((every #'(lambda (W) (equal W '(t t))) WW) '(t t))
	  ((< bc 1) (list nil A))
	  ((every #'(lambda (W) (eq nil (car W))) WW) (list nil A))
	  (t (list t (andify (mapcar #'cadr WW)))) ))

   ((and 8 (eq-car C '->) (test P 'P->)) (format t "~%P->")
    (cond ((eq-car (setq W (pci (and-on (cadr C) H) (caddr C) bc)) t) W)
	  (t (list nil A))))

   ((and 9 (eq-car C 'or) (test P 'or-c)) (format t "~%Or-C") (setq WW nil)
    (cond ((some #'(lambda (Ci) (and (or (format t "~%OR-C Next") t)
				     (equal '(t t) (setq W (pci H Ci bc)))
				     (setq WW (append WW (list W)))))
		 (cdr C)) '(t t))
	  ((< bc 1) (list nil A))
	  ((every #'(lambda (W) (eq nil (car W))) WW) (list nil A))
	  (t (list t (orify (mapcar #'cadr WW)))) ))

   ((and 10 (or (not P) (test P 'sub=)) (setq E (=s-in (cdr H) H C))) 
    (format t "~%Sub=") (setq HC (sub= E H C))
    (cond ((eq-car (setq W (pci (car HC) (cadr HC) bc)) t) W) 
	  (t (list nil A))))
   ((and 11 *reduce* (not (equal C (setq C* (reduci C)))))
    (cond ((eq-car (setq W (pci H C* bc)) t) W) (t (list nil A))))
   ((and 11 *reduce* (not (equal H (setq H* (reduci H))))) 
    (cond ((eq-car (setq W (pci H* C bc)) t) W) (t (list nil A))))
   ((and 12 (test P 'defn-ch) (not (logic-in C)) (setq HC (defn-CH H C)) 
	    (not (equal C (cadr HC))))                      ;HC= (H* C*)
    (format t "~%Def-CH")
    (cond ((eq-car (setq W (pci (car HC) (cadr HC) bc)) t) W)
	  (t (list nil A))))
   ((and 13 (test P 'defn-deep) (not (logic-in C)) (setq HC (defn-deep-HC H C))
	    (not (equal A (imp-on (car HC) (cadr HC)))))
    (format t "~%Defn-Deep")
    (cond ((eq-car (setq W (pci (car HC) (cadr HC) bc)) t) W)
	  (t (list nil A))))
   (t (list nil A)) )))

(defun and-off* (A L) (and-off A (cons '& L)))

(defun andify-nil (L) (cond ((eq t L) nil) (t (andify L)))) ;12Aug93

(defun apply-def (entry C par)            ;called by DEFN-HC
  (let (theta C* PP Ps (V (variables-in C)) Pf*)
    (cond ((not (setq theta (instance C (car entry))))
	   (return-from apply-def nil)))
    (setq C* (cond (par (cadr entry)) ((caddr entry)) (t (cadr entry))))
    (setq C* (apply-sub theta C*))
    (setq PP (cond (par (fourth entry)) (t (fifth entry))))
    (mapc #'(lambda (P) (cond ((setq Ps (find-all P C*))     ;Ps= ((P x y) ...)
			       (mapc #'(lambda (pf)          ;Pf=  (P x y)
                                 (setq Pf* (append pf V))    ;Pfx =(P x y u v)
				 (setq C* (subst Pf* pf C* :test 'equal)))
				    Ps))))
	  PP)
    C*))



(defun bci (H C C* HH bc)         ;10Aug93  NEW VERSION
  ;;back-chain, using INSTANCE instead of UNIFY
  ;;10Aug93  This version works with the new version of PCI which
  ;; returns either (t t) or (t A1) or (nil A1)
  (let ((SH (structure-size H)) (SC (structure-size C)) (A (imp-on H C)) 
	theta H* B* nB* W WW)
    (cond ((and (setq theta (instance* C H C* HH))
		(or (format t "~%From BCI") t)
		(eq-car (setq W (pci HH (apply-sub theta c*) (1- bc))) t)) W)
	  ((< (+ SH 1) SC) (list nil A))           ;H must be as "large" as C
	  ((eq-car H '->) (setq C* (and-on (cadr H) C*)) 
	   (cond ((eq-car (setq W (bci (caddr H) C C* HH bc)) t) W)
		 (t (list nil A))))
	  ((eq-car H '&) (setq WW nil)              ;10Aug93 This is like OR-C
	   (cond ((some #'(lambda (Hi) 
                              (and 
;;			       (equal '(t t) (setq W (bci Hi C C* HH bc)))
			       (eq-car (setq W (bci Hi C C* HH bc)) t)
			       (setq WW (append WW (list W)))))
			(cdr H)) W)
		 ((every #'(lambda (W) (eq nil (car W))) WW) (list nil A))
;;		 (t (list t (orify (mapcar #'cadr WW))))
		 (t (setq WW (mapcan #'(lambda (W) (cond ((eq-car W t) 
							  (list W))))
				     WW))
		    (list t (orify (mapcar #'cadr WW))))))

	  ((eq-car H 'or) (setq WW nil)
	   (cond ((some #'(lambda (Hi) 
                    (and (or (setq B* (delete-one Hi H)) t)
			 (or (setq nB* (cond ((eq-car B* 'not) (cadr B*))
					     (t (list 'not B*)))) t)
			 (setq C* (and-on nB* C*))
			 (equal '(t t) (setq W (bci Hi C C* HH bc)))
			 (setq WW (append WW (list W)))))
			(cdr H)) '(t t))
		 ((every #'(lambda (W) (eq nil (car W))) WW) (list nil A))
;;		 (t (list t (orify (mapcar #'cadr WW)))) 
		 (t (setq WW (mapcan #'(lambda (W) (cond ((eq-car W t) 
							  (list W))))
				     WW))
		    (list t (orify (mapcar #'cadr WW))))))

	  ((and (eq-car H '->) (eq-car (caddr h) '->))
           ;;H= (-> (H1 (-> H2 C1))) ~> (-> (& H1 H2) C1)
	   (setq H* (list '-> (and-on (cadr H) (cadr (caddr H)))
			      (caddr (caddr H))))
	   (cond ((eq-car (setq W (bci H* C C* HH bc)) t) W) (t (list nil A))))
	  (t (list nil A)) )))

(defun defn-ch (H C)       ;called by PCI
  (let (entry H* C* Hi2 Hi*)
    ;;The defintion-table is ordered so that the preferred ones are first
    (cond ((setq entry (first-in definition-table C))
           ;;entry= (C% D1 D2) 
	   (setq C* (apply-def entry C t))   ;t is parity
	   (cond ((setq H* (apply-def entry H nil))
		  (list H* C*))
		 ((not (eq-car H '&)) nil)
		 ((some #'(lambda (Hi) (setq Hi2 Hi)
			               (setq Hi* (apply-def entry Hi nil)))
			(cdr H))
		  (setq H* (reduce-& (subst Hi* Hi2 H :test 'equal)))
		  (list H* C*))
		 (t (list H C*))))
	  (t nil) )))

(defun defn-deep-hc (H C)  ;called by PCI
  (let (entry fmls fH fC C2* H2* C* H* C3 H3)
    (setq fmls (union-w (setq fH (formulas-of H)) 
			(setq fC (formulas-of C))))
    (cond ((some #'(lambda (fml) 
     (and (setq entry (first-in definition-table fml))
	  (some #'(lambda (C2) (setq C2* (apply-def entry (setq C3 C2) t))) fC)
	  (some #'(lambda (H2) (setq H2* (apply-def entry (setq H3 H2) t))) fH)
	  (setq C* (reduce-& (subst C2* C3 C :test 'equal)))
	  (setq H* (reduce-& (subst H2* H3 H :test 'equal))) ))
		 fmls)
	   (list H* C*))
	  ((some #'(lambda (fml) 
     (and (setq entry (first-in definition-table fml))
	  (some #'(lambda (C2) (setq C2* (apply-def entry (setq C3 C2) t))) fC)
	  (setq C* (reduce-& (subst C2* C3 C :test 'equal))) ))
		 fmls)
	   (list H C*))
	  ((some #'(lambda (fml) 
     (and (setq entry (first-in definition-table fml))
	  (some #'(lambda (H2) (setq H2* (apply-def entry (setq H3 H2) t))) fH)
	  (setq H* (reduce-& (subst H2* H3 H :test 'equal))) ))
		 fmls)
	   (list H* C))
	  (t nil) )))

(defun pci-echo (H C H2 C2 &optional bc)   ;9Aug93   ANALOGY IN ABDUCTION
  ;;The idea here is to use PCI on BOTH  H,C  and  H2,C2.
  ;;PCI already can prove H,C; Use that as a guide to prove H2,C2 (on the fly).
  ;;At each step of PCI:
  ;; (1) Perform the step for H,C,  getting the step# and P
  ;; (2) (a) If P=P2 (the P for H2,C2), then perform step# on H2,C2.
  ;; (3) (b) If not, then provide code (at each such step#) telling
  ;;                 it how to proceed. Eg, if P2 is (& ... P ...)
  ;;         then try to prove as much of P2 as possible (including P)
  ;;         and return the unproved as the pre-condition.  
 (let ( (A (imp-on* H C)) (A2 (imp-on* H2 C2)) C* E W H* Hi*
	    A*2 C*2 H*2 Hi*2 u2 C*1 L2 C2* E2 W2)
  (cond 
   ((or (atom H) (atom C) (atom H2) (atom C2)) nil)
     ;;9Aug93  This is an addition (to the structure of PCI)
   ((and (not (eq (car C) (car C2))) (eq-car C2 '&) 
	 (some #'(lambda (u) (and (similar-echo u C) (setq u2 u)) (cdr C2)))
    (setq C*1 (pci-echo H C H2 u2))
    (setq L2 (remove u2 (cdr C2) :test 'equal))
    (setq C2* (andify (mapcar #'(lambda (u) (pci H2 u bc)) L2)))
    (and-on C*1 C2*)))

   ((and 1 (eq t C))         (pci H2 C2))  ;no help from the guide
   ((and 2 H (instance C H)) (pci H2 C2))  ;no help from the guide

   ((and 3 (eq-car H  '&) (some #'(lambda (Hi) (instance C  Hi)) (cdr H))) 
    (cond ((and 
           (eq-car H2 '&) (some #'(lambda (Hi) (instance C2 Hi)) (cdr H2))) t)
	  (t (I am not sure yet)))) ;maybe learn which Hi was used by H,C
                                    ;Or do something like a SOME-ECHO ??

   ((and 4 H   (eq t (bci H C)))
    (cond ((and 
	    H2 (eq t (bci H2 C2))) t)
	  (t (bci-echo H C H2 C2))))

   ((and 5 (eq-car H 'or))           
    (setq C* (andify (mapcar #'(lambda (Hi) (imp-on Hi C)) (cdr H)))) 
    (format t "~%Or-H")
    (cond ((eq-car H2 'or)
	   (setq C*2 (andify (mapcar #'(lambda (Hi) (imp-on Hi C2))
				     (cdr H2)))) 
	   (format t "~%Or-H2")
	   (pci-echo nil C* nil C*2 bc))
	  (t (I am not sure yet))))

   ((and 6 (eq-car H '&) 
	 (some #'(lambda (Hi) (and (eq-car Hi 'or) (setq Hi* Hi)))
	       (cdr H))) 
    (format t "~%Or-Hi")
    (cond ((and (eq-car H2 '&)
		(some #'(lambda (Hi) (and (eq-car Hi 'or) (setq Hi*2 Hi)))
		      (cdr H2))
		(similar-echo Hi*2 Hi2))           ;We need to think about this
	   (pci-echo nil (or-outi H Hi* C) (or-outi H2 Hi*2 C2) bc))
	  (t (I am not sure yet))))

   ((and 7 (eq-car C '&) (or (format t "~%And-C") t)
	 (cond ((and (eq-car C2 '&) (or (format t "~%And-C2") t))
		(cond (bc (every #'(lambda (Ci Cj) 
                            (eq t (pci-echo H Ci H2 Cj bc))) (cdr C) (cdr C2)))
		      (t (and (setq C*2 (andify (mapcar #'(lambda (Ci Cj) 
				       (format t "~%And-C Next") 
				       (pci-echo H Ci H2 Cj bc))
						   (cdr C) (cdr C2))))
			      (not (equal A2 (setq A*2 (imp-on* H2 C*2)))))
			 A*2)))
	       (t (I am not sure yet)))))

   ((and 8 (eq-car C '->)) (format t "~%P->")
    (cond ((eq-car C2 '->)
	   (pci-echo (and-on (cadr C)   H) (caddr C)
                     (and-on (cadr C2) H2) (caddr C2)  bc))
	  (t (I am not sure yet))))

   ((and 9 (eq-car C 'or)) (format t "~%Or-C")
    (cond ((eq-car C2 'or) (format t "~%Or-C2")
	   (cond ((some #'(lambda (Ci Cj) (eq t (pci H Ci H2 Cj bc)))
			(cdr C) (cdr C2)) t)
		 (bc nil)
		 ((some #'(lambda (Ci Cj) 
		     (not (equal (imp-on H2 Cj) 
				 (setq C*2 (pci-echo H Ci H2 Cj bc)))))
			(cdr C) (cdr C2)) C*2)   ;13July93 WWB equal hard to cp
		 (t (imp-on H2 C2))))
	  (t (I am not sure yet))))
	   

   ((and 10 (setq E (=s-in (cdr H) H C))) (format t "~%Sub=")
    (cond ((setq E2 (=s-in (cdr H2) H2 C2))
	   (setq W (sub= E H C)) (setq W2 (sub= E2 H2 C2))
	   (pci-echo (car W) (cadr W) (car W2) (cadr W2)  bc))
	  (t (I am not sure yet))))

   ((and 11 *reduce* (not (equal C (setq C* (reduci C)))))
    (cond ((not (equal C2 (setq C*2 (reduci C2))))
	   (pci-echo H C* H2 C*2 bc))
	  (t (I am not sure yet))))
   ((and 11 *reduce* (not (equal H (setq H* (reduci H)))))
    (cond ((not (equal H2 (setq H*2 (reduci H2))))
	   (pci-echo H* C H*2 C2 bc))
	  (t (I am not sure yet))))

   ((and 12 (not (logic-in C)) (setq W (defn-CH H C))     ;W= (H* C*)
	 (not (equal C (cadr W))))
    (format t "~%Def-CH")
    (cond ((and (not (logic-in C2)) (setq W2 (defn-CH H2 C2))
	   (not (equal C2 (cadr W2))))
	   (pci-echo (car W) (cadr W) (car W2) (cadr W2) bc))
	  (t (I am not sure yet))))

   ((and 13 (not (logic-in C)) (setq W (defn-deep-HC H C))
	 (not (equal A (imp-on (car W) (cadr W))))) 
    (format t "~%Defn-Deep")
    (cond ((and (not (logic-in C2)) (setq W2 (defn-deep-HC H2 C2))
		(not (equal A2 (imp-on (car W2) (cadr W2)))))
	   (pci-echo (car W) (cadr W) (car W2) (cadr W2) bc))
	  (t (I am not sure yet))))
    
	  


   (bc nil)
   (t A) )))

  

(defun =s-in (L &optional H C)                   ;called by PCI, INSTANCE*
  ;;The list of (= x t) units in L for which x occurs elsewhere in H or C
  ;;If H And C are nil, then the list of all (= x t) units in L
  (let (H2)
    (mapcan #'(lambda (u) (setq H2 (remove u H :test 'equal))
        (cond ((and (eq-car u '=)
		    (or (and (not H2) (not C))
			(occur (cadr u) H2) (occur (cadr u) C)))
	       (list u)))) L)))

(defun =s-in2 (L H C)    ;not used                   ;called by instance* 
  ;;The list of (= x t) units in L for which t occurs elsewhere in H or C
  (let (H2)
    (crunch (mapcan #'(lambda (u) (setq H2 (remove u H :test 'equal))
        (cond ((and (eq-car u '=) 
		    (or (occur (caddr u) H2) (occur (caddr u) C)
			(occur (cadr u) H2) (occur (cadr u) C)))
	       (list u)))) L))))

(defun first-in (L C)            ;Called by DEFN-HC DEFN-DEEP-HC
  ;;Each member u of L has the form u= (C' xx yy PPx PPy)
  ;;where xx is the defn of C' with parity t, yy with parity nil,
  ;;PPx is the list of predicates in xx, PPy is the list of preds in yy
  ;;This finds the first entry u in L for which C is an instance of C'
  ;;And we select new variables for the entry. And return (C' xx yy).
  (let (u* Vu x99 u99)
    (cond ((some #'(lambda (u) (and (instance C (car u)) (setq u* u))) L)
	   (setq Vu (variables-in u*))
	   (setq u99 u*)
	   (mapc #'(lambda (x) (setq x99 (pack (append (explode x) '(9 9))))
		               (setq u99 (subst x99 x u99)))
		 Vu)
	   u99)
	  (t nil) )))

(defun formulas-of (C)   ;called by DEFN-DEEP-HC
  ;;eg, (formulas-of '(& (E x (A)) (E (T) (B)) 
  ;;                     (-> (& (E y (D)) (= (f x) (f (t))) 
  ;;				(<= (I (T)) (I y))))))
  ;; = ((E x (A)) (E (T) (B)) (E y (D)) (= (f x) (f (t))) (<= (I (T)) (I y)))
  (let ()
    (cond ((atom C) nil)
	  ((null (cdr C)) nil)
	  ((member (car C) '(& or -> not) :test 'equal)
	   (crunch (mapcan #'formulas-of (cdr C))))
	  (t (list C)))))

(defun instance* (C H C* HH) ;9Aug93  Called by BCI
  (let (H* E W)
    (cond ((instance C H))              ;if instance works, return it
	  ((and (setq H* (and-on C* HH))  ;else sub= and try again
		(setq E (=s-in (cdr H*)))
		(setq W (sub== E H C)))
	   (instance (cadr W) (car W)))
	  (t nil))))

(defun instance (C H)  ;9Aug93 Called by instance*
  (let () 
    (cond ((equal C H) t)
	  ((atom H) 
	   (cond ((and H (not (occur C H)) (not (numberp H)))
		  (list (cons H C)))
		 (t nil)))
	  ((atom C) 
	   (cond ((and (member C *C-list* :test 'equal)
		       (not (occur H C)) (not (numberp C)))
		  (list (cons C H)))
		 (t nil)))
	  ((not (eq (car H) (car C))) nil)
	  ((not (equal (length H) (length C))) nil)
	  (t (instance2 (cdr C) (cdr H))) )))

(defun instance2 (L1 L2)
  (let (theta sigma)
    (cond ((null L1) (null L2))
	  ((null L2) nil)
	  ((not (setq theta (instance (car L1) (car L2)))) nil)
	  ((not (setq sigma (instance2 (apply-sub theta (cdr L1))
				       (apply-sub theta (cdr L2))))) nil)
	  (t (compose theta sigma)) )))

(defun logic-in (A) (some #'(lambda (s) (occur s A)) '(& or -> not)))

(defun or-outi (H Hi C)  ;13July93  Called by PCI
  ;;eg, (or-outi '(& A (or B C D) E F) '(or B C D) '(CC)) 
  ;; =  (& (-> (& A B E F) (CC)) (-> (& A C E F) (CC)) (-> (& A D E F) (CC)))
  (let (L)
    (setq L (mapcar #'(lambda (u) (subst u Hi H :test 'equal)) (cdr Hi)))
     ;;L= ((& A B E F) (& A C E F) (& A D E F))
    (andify (mapcar #'(lambda (u) (imp-on u C))  ;(-> (& A B E F) (CC))
		    L)) ))

(defun quantify-set (A)   ;14July93
  ;;eg, (quantify-set '(-> (all (x y) (P x y)) (all z (Q z))))
  ;;           =       (-> (all x (all y (P x y))) (all z (Q z)))
  (let (C)
    (cond ((atom A) A)
	  ((member (car A) '(all some))
	   (setq C (quantify-set (caddr A)))
	   (cond ((atom (cadr A)) (list (car A) (cadr A) C))
		 (t (quan-set (car A) (cadr A) C))))
	  ((member (car A) '(& or not ->))
	   (cons (car A) (mapcar #'quantify-set (cdr A))))
	  (t A))))

(defun quan-set (P L C) ;14July93  Called by QUANTIFY-SET
  (cond ((null L) C)
	((null (cdr L)) (list P (car L) C))  ;L= (x)
	(t (list P (car L) (quan-set P (cdr L) C)))))

(defun rest-of (A) 
  ;;eg, (rest-of '(& a b c)) = (& b c), (rest-of '(& a b)) = b
  (cond ((cdddr A) (cons (car A) (cddr A)))  ;(& a b c ...) -> (& b c ...)
	((cddr  A) (cadr A))                 ;(& a b)       -> b
        (t nil)))

(defun similar-echo (A B)  ;needs to be done   9Aug93
  (prnt A B) t)

(defun sub= (E H C)                    ;called by PCI
  ;;apply the equality units u= (= x t) in E to H and C
  ;;eg, (sub= '((= x (f t)) (= y (g s)) (= z (c))) '(P x) '(Q x z))
  ;;       =   ((P (F T)) (Q (F T) (C)))
  (let (H2 E* H* (C* C) x tm u*)
    (setq E* (mapcan #'(lambda (u) (setq u* (list (car u) (caddr u) (cadr u)))
              (setq H2 (remove u  H  :test 'equal))
              (setq H2 (remove u* H2 :test 'equal))
              (cond ((or (occur (cadr u) H2) (occur (cadr u) C)) (list u))))
		     E))
    (setq H* (and-off* H E*))
    (mapc #'(lambda (u) (setq x (cadr u)) (setq tm (caddr u))
                        (setq H* (subst-w tm x H*))
                        (setq C* (subst-w tm x C*)))
	  E*)
    (list H* C*)))

(defun sub== (E H C)  ;9Aug93  Called by INSTANCE*   not used
  ;;repetitely sub=
  ;;eg, (sub== '((= x (f y)) (= y (g z)) (= z d)) '(H x y z) '(C x b z))
  ;;         =  ((h (f (g d)) (g d) d) (c (f (g d)) b d))
  (let (W)
    (cond ((equal (list H C) (setq W (sub= E H C))) W)  ;no change
	  (t (sub== E (car W) (cadr W))))))

(defun sub== (E H C)  ;9Aug93  Called by INSTANCE*
  ;;repetitely sub=
  ;;eg, (sub== '((= x (f y)) (= y (g z)) (= z d)) '(H x y z) '(C x b z))
  ;;         =  ((h (f (g d)) (g d) d) (c (f (g d)) b d))
  ;;eg, (sub== '((= x a) (= y b) (= z (f z x)) (= u (g z x)) (= v (h z y)))
  ;;           '(P x y z u v) '(Q x y z u v)) 
  ;;           ((P A B (F Z A) (G (F Z A) A) (H (F Z A) B))
  ;;            (Q A B (F Z A) (G (F Z A) A) (H (F Z A) B)))
  (let (E1 E2 E3 E4 (H2 H) (C2 C))
    (cond ((null E) (list H C))
	  ((equal E (setq E1 (mapcan #'(lambda (u)  ;in terms of self
                                (cond ((occur (cadr u) (caddr u))(list u))))
				     E)))
	   (mapc #'(lambda (u) (setq H2 (subst-w (caddr u) (cadr u) H2))
		               (setq C2 (subst-w (caddr u) (cadr u) C2)))
		 E)
	   (list H2 C2))

	  ((and (setq E2 (set-difference E E1)) nil))    ;not in terms of self
	  ((setq E3 (mapcan #'(lambda (u)           ;u= (x s), v= (y t) 
            (cond ((every #'(lambda (v) (not (occur (cadr v) (caddr u))))
			  E2)     ;E1 is those u's of E for which no y is in s
		   (list u))))
			    E2))

	   (setq E4 (union (set-difference E2 E3 :test 'equal) E1))
	   (mapc #'(lambda (u)  
                     (setq E4 (subst-w (caddr u) (cadr u) E4))
                     (setq H2 (subst-w (caddr u) (cadr u) H2))
                     (setq C2 (subst-w (caddr u) (cadr u) C2)))
		 E3)
	   (sub== E4 H2 C2))
	  ;;Still concerned about "tied-to"  Handle later
	  (t (format t "~%Alarm in sub==~%") (break)))))

(defun subst-w (x y A)  ;14July93 
  ;;To prevent    (subst '(x) '(y) '(P (y) y) :test 'equal) = (P (x) x) ???
  ;;Now correct (subst-w '(x) '(y) '(P (y) y))              = (P (x) y). !!
  (cond ((equal y A) x)
	((atom A) A)
	(t (cons (subst-w x y (car A)) 
		 (mapcar #'(lambda (u) (subst-w x y u)) (cdr A))))))

(defun test (P P*) (cond ((not P) t) (t (eq P P*))))

;;--------------------------------------------------------------------

(defun tr () (trace pci))

(setq thm2 '(-> (all x (P x)) (& (P a) (P b))))
(defun p2 () (time (pci nil thm2)))

(setq thm5 '(-> (e x (U (~ C1 (sng L1)) (~ C2 (sng L2))))
		(e x (U (~ C1 (sng L1)) (~ C2 (sng L2))))))
(defun p5 () (pci nil thm5))

(setq thm6 '(-> (e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2)))) 
		(e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2))))))
(defun p6 () (pci nil thm6))

(setq thm7
 '(-> (merge-low (e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2))))) 
      (merge-low (e x (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2)))))))
(defun p7 () (pci nil thm7))

(setq fml8 '(-> (& (= (A) (U (~ C1 (sng L1)) (~ C2 (sng L2))))
		   (= (B) (U (~ C1 (sng L1)) (~ (U C2 (sng L)) (sng L2)))))
		(-> (e (t) (merge-low (A))) (e (t) (merge-low (B))))))
(setq fml8 '(-> (& (= (A) (U (~ (C1) (sng (L1))) (~ (C2) (sng (L2)))))
		   (= (B) (U (~ (C1) (sng (L1))) 
			     (~ (U (C2) (sng L)) (sng (L2))))))
		(-> (e (t) (merge-low (A))) (e (t) (merge-low (B))))))
(setq fml8 '(-> (e (t) (merge-low (U (~ (C1) (sng (L1))) (~ (C2) (sng (L2))))))
		(e (t) (merge-low (U (~ (C1) (sng (L1))) 
				     (~ (U (C2) (sng (L))) (sng (L2))))))))

(setq plan8 '(p-> defn-ch and-c tt P-> defn-deep or-hi and-c P-> tt
              P-> defn-deep defn-deep or-hi and-c P-> tt
              P-> defn-deep sub= end)) 
(defun p8 () (setq *plan* plan8) (time (pci nil fml8))) ;see p8-bcq

(defun trs () (trace (pci :entry `(pci ,(cadr  si::arglist))) ;prints only C
		     (pci-start :entry `(pci ,(cadr  si::arglist))) 
;;		     (sub= :entry `sub=) ;prints no arguments
;;		     (bci :entry `(bci ,(cadr  si::arglist))) ;prints only C
;;		     (bci :entry `bci) ;prints no arguments
		     ))   ;short printing version

(setq thm1 '(all b (-> (all x (P x)) (some y (& (P b) (P y))))))
(defun p1 () (time (pci-start thm1)))

(setq thm2 '(all b (-> (& (P b) (Q b)) (some y (& (Q y) (P y))))))
(setq c-list2 '(y))
(defun p2 () (setq *c-list* c-list2) (time (pci-start thm2)))

#| 2Aug93  We temporarily excluded the next 11 lines
;;See file atp.116 for values of HP, Lemmas1*, and C**
;;See file hq-cq for the skolemized values Hq* Cq*
(setq lemsq (andify (mapcar 'eval Lemmas1*)))
(setq Hq (quantify-set (and-on HP lemsq)))
(setq Cq C**)
(setq c-listq '(C9 L10))
(defun pq1 () (setq *plan* plan-q) (time (pci-start Hq Cq)))
(defun pq () (setq *plan* plan-q) (setq *C-list* c-listq)
             (time (pci-start nil nil Hq* Cq*)))
(setq plan-q '(and-c P-> and-c P-> P-> halt defn-ch halt 
               and-c halt tt P-> defn-deep halt))
|#

#| 11Aug93  Save these old versions for awhile
(defun pci (H C &optional (bc 2))  ;OLD VERSION
  ;;If (bc < 1), then this can only return t or nil, used by BCI
  ;;Else it returns a Q for which  (Q & H --> C).
  ;;eg, (pci '(& (P) (-> (P) (Q)) (-> (Q) (R))) '(R)) = t
 (let (P (A (imp-on* H C)) A* C* E W H* Hi*)
  (cond (*plan* (setq P (car *plan*)) (setq *plan* (cdr *plan*))))
  (cond
   ((eq P 'halt) (format t "~%Halted by User~%") (break) (pci H C bc))
   ((eq P 'tt) t)
   ((eq P 'ttt) (setq *plan** *plan*) (setq *plan* nil)
    (cond ((eq t (pci H C bc)) (setq *plan* *plan**) t) (t nil)))
   ((eq P 'end) A)

   ((and 1 (eq t C)) t)
   ((and 2 H (instance C H)) (format t "~%Instance H~%") t)
   ((and 3 (eq-car H '&) (some #'(lambda (Hi) (instance C Hi)) (cdr H))) 
     (format t "~%Instance Hi~%") t)
;; ((and *struve* (eq t (struve-prove (imp-on H C) 12 12))) t)
   ((and 4 H (or (not P) (test P 'bci)) (eq t (bci H C t H bc))) t)
   ((and 5 (eq-car H 'or) (test P 'or-h))
    (setq C* (andify (mapcar #'(lambda (Hi) (imp-on Hi C)) (cdr H)))) 
    (format t "~%Or-H")
    (pci nil C* bc))
   ((and 6 (eq-car H '&) (test P 'or-hi)
	 (some #'(lambda (Hi) (and (eq-car Hi 'or) (setq Hi* Hi)))
	       (cdr H))) (format t "~%Or-Hi")
    (pci nil (or-outi H Hi* C) bc))
   ((and 7 (eq-car C '&) (test P 'and-c) (or (format t "~%And-C") t)
	 (cond ((< bc 1) (every #'(lambda (Ci) (eq t (pci H Ci bc))) (cdr C)))
	       ((and (setq C* (andify (mapcar #'(lambda (Ci) 
				  (format t "~%And-C Next") (pci H Ci bc))
					      (cdr C))))
		     (not (equal A (setq A* (imp-on* H C*))))))))
    A*)
   ((and 8 (eq-car C '->) (test P 'P->)) (format t "~%P->")
    (pci (and-on (cadr C) H) (caddr C) bc))
   ((and 9 (eq-car C 'or) (test P 'or-c)) (format t "~%Or-C")
    (cond ((some #'(lambda (Ci) (eq t (pci H Ci bc))) (cdr C)) t)
	  ((< bc 1) nil)
	  ((some #'(lambda (Ci) 
		     (not (equal (imp-on H Ci) (setq C* (pci H Ci bc)))))
		 (cdr C)) C*)       ;13July93 WWB equal is hard to compute here
	  (t (imp-on H C))))
   ((and 10 (or (not P) (test P 'sub=)) (setq E (=s-in (cdr H) H C))) (format t "~%Sub=")
    (setq W (sub= E H C))
    (pci (car W) (cadr W) bc))
   ((and 11 *reduce* (not (equal C (setq C* (reduci C))))) (pci H C* bc))
   ((and 11 *reduce* (not (equal H (setq H* (reduci H))))) (pci H* C bc))
   ((and 12 (test P 'defn-ch) (not (logic-in C)) (setq W (defn-CH H C)) 
                                                              ;W= (H* C*)
	 (not (equal C (cadr W)))) (format t "~%Def-CH")
    (pci (car W) (cadr W) bc))
   ((and 13 (test P 'defn-deep) (not (logic-in C)) (setq W (defn-deep-HC H C))
	 (not (equal A (imp-on (car W) (cadr W))))) (format t "~%Defn-Deep")
    (pci (car W) (cadr W) bc))
   ((< bc 1) nil)
   (t A) )))
(defun bci (H C C* HH bc)             ;OLD VERSION
  ;;back-chain, using INSTANCE instead of UNIFY
  (let ((SH (structure-size H)) (SC (structure-size C)) (A (imp-on H C)) 
	theta H* B* nB*)
    (cond ((and (setq theta (instance* C H C* HH))
		(or (format t "~%From BCI") t)
		(eq t (pci HH (apply-sub theta c*) (1- bc)))) t)
	  ((< (+ SH 1) SC) (list nil A))           ;H must be as "large" as C
	  ((eq-car H '->) (setq C* (and-on (cadr H) C*)) 
	                  (bci (caddr H) C C* HH bc))
	  ((eq-car H '&)
	   (cond ((some #'(lambda (Hi) (bci Hi C C* HH bc)) (cdr H)) t)))  
	  ((eq-car H 'or)
	   (cond ((some #'(lambda (Hi) (setq B* (delete-one Hi H))
			               (setq nB* (cond ((eq-car B* 'not) 
							(cadr B*))
						       (t (list 'not B*))))
			               (setq C* (and-on nB* C*))
				       (bci Hi C C* HH bc)) 
			(cdr H)))))
	  ((and (eq-car H '->) (eq-car (caddr h) '->))
           ;;H= (-> (H1 (-> H2 C1))) ~> (-> (& H1 H2) C1)
	   (setq H* (list '-> (and-on (cadr H) (cadr (caddr H)))
			      (caddr (caddr H))))
	   (bci H* C C* HH bc))
	  (t nil) )))
|#
