
;;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;;;                 res-skip.lisp                           15Mar93
;;;
;;;This is to help determine the nature of Lemma1 in the proof of
;;; the completeness of SET-VAR.  
;;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(defun lss () (load "res-skip.lisp"))

  ;;This next function, RES-SKIP, is ordinary ground resolution when Skip=nil.
  ;;If Skip is not nil, then the literals of Skip are NOT resolved upon.  
  ;;S is an unsatisfiable set of ground clauses and Skip is a collection
  ;; of ground atoms.
  ;;This derives the set of all resolvents from S except that the
  ;; atoms of Skip are not resolved upon.  It returns all such resolvents
  ;; which contain members of Skip or their negations.  Subsumption is done.
  ;;eg,  if S= '(P ~P), Skip= (P), then (res-skip S Skip) = S.
  ;;and, if S= '((A B) (A (NOT B)) ((NOT A) B) ((NOT A) (NOT B))), Skip= '(A), 
  ;;then, (res-skip S Skip) = ((P) ((NOT P)))
  ;;Theorem: If S is insat then (res-skip S Skip) is unsat (for any Skip).

(defun res-skip (S &optional Skip)  ;15Mar93
  (let (S2 RR RR2)
    (setq S2 S)
    (setq RR (mapcan #'(lambda (C1) (setq S2 (cdr S2)) (mapcan #'(lambda (C2)
               (cond ((equal C1 C2) nil)
		     ((eq '[] (setq RR2 (resolve-skip C1 C2 Skip)))
		      (return-from res-skip '[]))
		     (RR2 RR2)))
							       S2))
		     S))
    (setq RR (subsume-out-g  (crunch (append S RR))))
    (cond ((equal RR '(nil)) '[])
	  ((not (subset S RR)) (res-skip RR skip))
	  (RR RR)
	  (t 'ALARM) )))

(defun resolve-skip (C1 C2 Skip)  ;15Mar93  Called by RES-SKIP
  ;;eg, (resolve-skip '(L A (Not B)) ((not A) C)) = ((L (not B) C))
  (let (RR R)
    (setq RR (mapcan #'(lambda (L1) (mapcan #'(lambda (L2)
              (cond ((or (member L1 Skip) (member L2 Skip)) nil)
		    ((not (equal L1 (not-fcn L2))) nil)
		    ((eq '[] (setq R (crunch (append(remove L1 C1 :test 'equal)
						(remove L2 C2 :test 'equal)))))
		     '[])
		    ((tautology R) nil)
		    (t (list R))))
					    C2))
		     C1)) ))

(defun skip-test (A P n) ;16Mar93
  ;;Generate A partial full-set S on the literal of the list A, 
  ;; verify that is is unsatisfiable,
  ;; generate S*= (res-skip S P), S**= (res-skip S* (rest of P))
  ;; and verify that (res-skip S**) = [].
  ;;ie, Theorem:  (The above)
  (let (S RR S* S** P*)
    (setq S (full-set-partial A n))
    (cond ((not (eq '[] (setq RR (res-skip S))))      ;verify that S is unsat
	   (format t "~%S is not unsatisfiable~%")
	   (return-from skip-test nil))) 
    (cond ((eq '[] (setq S* (res-skip S P))) (return-from skip-test '[])))
    (setq P* (set-difference A P))
    (setq S* (mapcan #'(lambda (C) (cond ((occur-set P C) (list C)))) S*))
;;The next command is for trial only  16Mar93
    (setq S* (mapcan #'(lambda (C) (cond ((not (occur-set P* C)) 
					  (list C)))) S*))
    (setq S** (res-skip S* P*))
;;    (prnt S RR S* P* S**)
    (cond ((eq '[] S**) '[])
	  (t (res-skip S**))) ))

(defun crunch (L) 
  (cond ((null (cdr L)) L)
	((member (car L) (cdr L) :test 'equal) (crunch (cdr L)))
	(t (cons (car L) (crunch (cdr L))))))

(defun eq-car (A x) (and (not (atom A)) (eq (car A) x)))

(defun full-set (A) ;16Mar93 
 ;;This generates the full-set of clauses on the ground literals in list A
 ;;eg, (full-set '(A B)) = ((A B) (A (NOT B)) ((NOT A) B) ((NOT A) (NOT B)))
 (let (S)
   (cond ((null A) (list nil))
	 (t (setq S (full-set (cdr A)))
	    (mapcan #'(lambda (C) (list (cons (car A) C)
					(cons (list 'not (car A)) C)))
		    S)) )))

(defun full->partial (S n) ;S is a set of ground clauses, 1<=n<=10 16Mar93
  ;;This shortens each clause C of S, by removing a literal of L of C,
  ;; if the (random-number 10) is greater than n.  1<=n<=10.
  (let (S* C*)
    (setq S* (mapcar #'(lambda (C) (setq C* (mapcan #'(lambda (L)
       (cond ((<= (random-number 10) n) (list L))))
					   C))
	   (cond (C* C*) (t C)))
		     S))
    (subsume-out-g (crunch S*))))

(defun full-set-partial (A n) ;16Mar93 
  ;;Generate a full-set on the list of literals A, and shorten the clauses.
  (let (S)
    (setq S (full-set A))
    (setq S (full->partial S n))))

(defun not-fcn (A) (cond ((eq-car A 'not) (cadr A)) (t (list 'not A))))

(defun occur (x A) 
  (cond ((equal x A) t)
	((atom A) nil)
	(t (or (occur x (car A)) (occur x (cdr A))))))

(defun occur-set (V A)              ;2Feb93  used in INFECTED
  (some #'(lambda (x) (occur x A)) V))

;;  From Bob Boyer (The Knuth Algorithm, he says)
(DEFPARAMETER *RANDOM-SEED* 0)
(DEFUN RANDOM-NUMBER (N)
  (LET ((LINEAR 267) (MULTIPLIER 317) (MODULUS 4096))
    (REM (SETQ *RANDOM-SEED*
               (REM (+ LINEAR (* MULTIPLIER
                                 *RANDOM-SEED*))
                    MODULUS))
         N)))

(defun subset (A B) (subsetp A B :test 'equal))

(defun subsume-out-g (S &optional (S* S))  ;16Mar93  Called by RES-SKIP
  (let ()
    (mapc #'(lambda (C*) 
       (cond ((member C* S :test 'equal) ;Was C* already removed?
	      (setq S (mapcan #'(lambda (C) (cond ((equal C* C) (list C))
						  ((subset C* C) nil)
						  (t (list C)))) S))) ))
	  S*)
    S))

(defun tautologies-out (S) ;16Mar93
  (mapcan #'(lambda (C) (cond ((tautology C) nil) (t (list C)))) S))
(defun tautology (C) ;16Mar93
  (cond ((null C) nil)
	((member (not-fcn (car C)) (cdr C) :test 'equal) t)
	(t (tautology (cdr C))) ))

(defun variables-in (S) ;16Mar93 S is a set of ground clauses
  (cond ((null S) nil)
	((eq 'not S) nil)
	((atom S) (list S))
	(t (crunch (append (variables-in (car S))
			   (variables-in (cdr S))))) ))

;;---------------------------------------------------------------------
;;HB experiment using res-skip
;;Idea: Take the ground proof of HB, and "skip" the ground literals
;; that are derived directly from (t e A) for ~(t e A).
;; Then are we left with a set S' for which resolving ONLY on the
;; ground literals are derived directly from (t e A) for ~(t e A).

(setq S '(
 ((<= a0 b0))                                                 ;1
 ((not (element B3 G3)) (open B3))                            ;2.1
 ((not (element B1 G0)) (open B1))                            ;2.2
 ((subset (nt a0 b0) (union G0)))                             ;3
 ((not (subset B1 G0)) (not (finite (sng B1)))                ;4.1
  (not (subset (nt a0 b0) (union (sng B0)))))
 ((not (subset (U Ha4 (sng B3)) G0)) (not (finite (U Ha4 (sng B3))))
  (not (subset (nt a0 b0) (union (U Ha4 (sng B3))))))         ;4.2
 
 ((not (<= a0 b0)) (not (subset (sgn B1) G0)) (not (finite (sng B1)))   ;11.1
                   (not (subset (nt a0 a0) (Union (sng B1)))) (<= a0 L0))
 ((not (<= b2 b0)) (not (subset (U Ha0 (sng B3))) G0)                   ;11.2
                   (not (finite (union (U Ha0 (sng B3)))))
                   (not (subset (nt a0 b2) (union (U Ha0 (sng B3)))))
                                                               (<= b2 L0))
 ((not (<= b3 b0)) (not (subset (sgn B1) G0)) (not (finite (sng B1))) ;11.3
                   (not (subset (nt a0 b3) (Union (sng B1))))  (<= b3 L0))

 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;13.1
  (not (subset (nt a0 a0) (union (sng B1)))) (<= Zb0 b0)       (<= L0 b0))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;13.2
  (not (subset (nt a0 a0) (union (sng B1)))) (<= Za2 b0)       (<= L0 a2))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;13.3
  (not (subset (nt a0 a0) (union (sng B1)))) (<= Za4 b0)       (<= L0 a4))
  
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;14.1
  (not (subset (nt a0 a0) (union (sng B1)))) (subset Hb0 G0)    (<= L0 b0))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;14.2
  (not (subset (nt a0 a0) (union (sng B1)))) (subset Ha2 G0)    (<= L0 a2))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;14.3
  (not (subset (nt a0 a0) (union (sng B1)))) (subset Ha4 G0)    (<= L0 a4))
  
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;15.1
  (not (subset (nt a0 a0) (union (sng B1)))) (finite Hb0)       (<= L0 b0))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;15.2
  (not (subset (nt a0 a0) (union (sng B1)))) (finite Ha2)       (<= L0 a2))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;15.3
  (not (subset (nt a0 a0) (union (sng B1)))) (finite Ha4)       (<= L0 a4))

 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;16.1
  (not (subset (nt a0 a0) (union (sng B1)))) (subset (nt a0 Zb0) (union Hb0))
                                                                (<= L0 b0))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;16.2
  (not (subset (nt a0 a0) (union (sng B1)))) (subset (nt a0 Za2) (union Ha2))
                                                                (<= L0 a2))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;16.3
  (not (subset (nt a0 a0) (union (sng B1)))) (subset (nt a0 Za4) (union Ha4))
                                                                (<= L0 a4))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;17.1
  (not (subset (nt a0 b0) (union (sng B1)))) (not (<= Zb0 b0))  (<= L0 b0))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;17.2
  (not (subset (nt a0 b0) (union (sng B1)))) (<= Za2 a2)       (<= L0 a2))
 ((not (<= a0 b0)) (not (subset (sng B1) G0)) (not (finite (sgn B1)))   ;17.3
  (not (subset (nt a0 b0) (union (sng B1)))) (<= Za4 a4)       (<= L0 a4))
  
 ((not (<= a0 b0)) (= a0 b0) (< a0 b0))                                 ;L0

 ((not (<= a0 b0)) (not (subset (nt a0 b0) (union G0))) (element B1 G0));L1.1
 ((not (<= a0 b0)) (not (subset (nt a0 b0) (union G0))) (element a0 B1));L1.2
 ((not (<= a0 b0)) (not (subset (nt a0 b0) (union G0))) (element B2 G0));L1.3
 ((not (<= a0 b0)) (not (subset (nt a0 b0) (union G0))) (element b0 B2));L1.4

 ((not (element B1 G0)) (subset (sng B1) G0))                            ;L2.1
 ((not (element B2 G0)) (subset (sng B2) G0))                            ;L2.2
 ((not (subset (nt a0 b3) B1)) (subset (nt a0 b3) (union (sng B1))))     ;L2*.3

 ((not (element a0 B1)) (subset (nt a0 a0) (union (sng B1))))            ;L3.1
 ((not (element b0 B2)) (subset (nt b0 b0) (union (sng B2))))            ;L3.2

 ((finite (sng (B1))))                                                   ;L4.1
 ((finite (sng (B2))))                                                   ;L4.2

 ((not (<= a0 L0)) (not (<= L0 b0)) (element L0 (nt a0 b0)))             ;L5
 ((not (element L0 (nt a0 b0))) (not (subset (nt a0 b0) (union G0)))
                                (element L0 (union G0)))                 ;L6

 ((not (element L0 (union G0))) (element B3 G0))                         ;L7.1
 ((not (element L0 (union G0))) (element L0 B3))                         ;L7.1

 ((<= L0 a0) (<= b0 L0) (not (element L0 B3)) (not (open B3)) (<= a0 a2)) ;L8.1
 ((<= L0 a0) (<= b0 L0) (not (element L0 B3)) (not (open B3)) (<  a2 L0)) ;L8.2
 ((<= L0 a0) (<= b0 L0) (not (element L0 B3)) (not (open B3)) (<  L0 b2)) ;L8.3
 ((<= L0 a0) (<= b0 L0) (not (element L0 B3)) (not (open B3)) (<= b2 b0)) ;L8.4
 ((<= L0 a0) (<= b0 L0) (not (element L0 B3)) (not (open B3)) 
                                                 (subset (nt a2 b2) B3))  ;L8.5

 ((not (subset (nt a0 Za2) (union Ha2))) (not (subset (nt a2 b2) B3))     ;L9.1
  (not (<= a2 Za2))  (subset (nt a0 b2) (union (U Ha2 (sng B3)))))         
 ((not (subset (nt a0 Za4) (union Ha4))) (not (subset (nt a4 L0) B3))     ;L9.2
  (not (<= a4 Za4))  (subset (nt a0 L0) (union (U Ha4 (sng B3)))))         

 ((< L0 b0) (not (element L0 B1)) (not (open B1)) (<= a3 L0))           ;L10.1
 ((< L0 b0) (not (element L0 B1)) (not (open B1)) (<  L0 b3))           ;L10.2
 ((< L0 b0) (not (element L0 B1)) (not (open B1)) (subset (nt a3 b3) B1));L10.3
 ((< L0 b0) (not (element L0 B1)) (not (open B1)) (<= b3 b0))           ;L10.4

 ((not (subset (nt a3 b3) B1)) (not (<= a3 a0)) (subset (nt a0 b3) B1)) ;L11

 ((not (finite G0 )) (finite (U G0  (sng (B3)))))                       ;L12.11
 ((not (finite Ha2)) (finite (U Ha2 (sng (B3)))))                       ;L12.13
 ((not (subset Ha2 G0)) (not (element B3 G0)) (subset (U Ha2 (sng B3)) G0))
                                                                        ;L12.2

 ((not (subset Ha2 G0)) (not (element B3 G0)) (subset (U Ha2 (sng B3)) G0))
                                                                        ;L13.1
 ((not (subset Ha4 G0)) (not (element B3 G0)) (subset (U Ha4 (sng B3)) G0))
                                                                        ;L13.2

 ((not (element L0 B3)) (not (open B3)) (< a4 L0))                      ;L14.1
 ((not (element L0 B3)) (not (open B3)) (subset (nt a4 L0) B3))         ;L14.2

 ((not (element L0 B4)) (not (open B4)) (not (<= a0 L0)) (<= a0 a4))    ;L14*.1
 ((not (element L0 B4)) (not (open B4)) (not (<= a0 L0)) (<  a0 L0))    ;L14*.2
 ((not (element L0 B4)) (not (open B4)) (not (<= a0 L0))                ;L14*.3
                                               (subset (nt a4 b0) B4))
))
(defun st5 (n) (time (skip-test '(a b c d e) '(a b c) n)))
(defun st4 (n) (time (skip-test '(a b c d) '(a b) n)))
