
;;;*********************************************************************
;;;
;;;                      RESOLUTION.LISP    
;;;
;;;               W. W. Bledsoe     24 April 1989   26 May 1989
;;;
;;;  This is an implementation of ordinary resolution.
;;;
;;;  A call to (RESOLUTION formula) is used to prove FORMULA.
;;;  For clausal input, use  (RESOLUTION nil Clause-set),
;;;    and for Set-of-support use (RESOLUTION nil Clause-set SOS).
;;;
;;;  Some of these functions are taken from the file:  SIMPLER.LISP
;;;  (load "imply.lisp") (load "simpler.lisp")(load "skolem.lisp") 
;;;  (load "resolution.lisp")
;;;
;;;  24DEC91: The changes for using the file PBR.LISP, which are 
;;;    invoked when the flag *pbr-flag* is non-nill
;;;    Put *pbr-flag* to nil for ordinary resolution proving
;;;
;;;  This requires the file  skolem.lisp
;;;
;;;  See RES.LISP for a simpler (and faster?) version or resolution.
;;;**********************************************************************
(defun lres () (load "resolution.lisp"))
(defvar *break* nil)
(defvar *clauses* nil)
(defvar *Ancestry-clauses* nil)
(defvar *output* t)
(cond  (*pbr-flag* (setq *output* nil)))                  ;24Dec91
(defvar *subsume* nil)
(defvar *factors* nil)
(defvar *rv* nil)       ;resolution verbose
;(load "f.print")

(defun resolution (formula &optional clause-set SOS (rnl 20))   ;Top level call
 (prog (n round-number round-number-limit cs resolvents Base-cs apply-cs
	  Base-cs-old Base-cs* C1 C2 L1 L2 U C1* C2* N1 N2 C1-V C2-V 
          theta R R* factors rb pairs-light)

  (setq round-number 1)
  (setq round-number-limit rnl)

  (setq cs (cond (clause-set clause-set) 
		 (t (clausify (list 'not formula)))           ;10oct92
;;		 (t (clausify formula))                       ;12oct92
		 ))

  (cond (SOS (setq cs (set-difference cs SOS)))
	(t   (setq SOS cs) (setq cs nil)))

  (cond ((or (and cs (numberp (caar cs))) (and SOS (numberp (caar SOS))))
         (setq SOS (mapcar #'(lambda(C) (list (car C) (cadr C) nil)) SOS))
         (setq cs (mapcar #'(lambda(C) (list (car C) (cadr C) nil)) cs))
	 (setq n 60))     ;Clauses are already numbered
        (t          ;Number each clause and give them a parentage (nil).
	 (setq n 0)			;for numbering clauses 
	 (setq SOS (mapcar #'(lambda (C) (list (setq n (1+ n)) C nil)) SOS))
	 (setq cs  (mapcar #'(lambda (C) (list (setq n (1+ n)) C nil)) cs))))
;(format t "~%SOS= ~%~s~%cs= ~%~s~%"SOS cs)
   
 (setq resolvents SOS)
 (setq Base-cs-old nil)
 (setq Base-cs      cs)
(prnt cl)  (number-print cl)                                           ;12oct92
(format t "~%Above top1: n= ~s  round-number= ~s  round-limit= ~s ~% resolvents=~% ~s ~% cs=~% ~s~%" n round-number round-number-limit resolvents cs)

top1
 (cond ((or (cond ((> round-number round-number-limit) 
		   (cond ((not *rv*) (return-from resolution nil)))
		   (format t "~%Round-limit Exceeded. After ~s Rounds~%" 
			   round-number-limit) t))
            (cond ((null resolvents)
		   (format t "~%No New Resolvents Generated during Round ~s~%" 
			   round-number) t)))

;	(format t "~%*pbr-flag*= ~s~%" *pbr-flag*)
	(cond ((not *pbr-flag*)              ;pbr-flag is set in PBR.LISP
	       (format t "~%User Options:  
        Type a number as a new round limit and computation will continue
        Type the letter  P  to print the existing clause set
        Type nil to fail
        ~% Please Type your response: ")

               (cond ((numberp (setq U (read)))
		      (setq round-number-limit U) (go top1-1))
		     ((eq 'P U) 
		      (setq *clauses* (append resolvents Base-cs))
		      (setq *clauses* (sort *clauses* 'clause-number<))
;;		      (and *v* (number-print-h *clauses*))
		      (return nil))
		     (t (return nil))))
	      (t (setq *clauses* (append resolvents Base-cs))   ;added 24Dec91
		 (return (mapcar #'cadr *clauses*))))           ;added 24Dec91
     ))

top1-1
 (setq apply-cs resolvents)                ;Clauses to be "applied to" 
 (setq Base-cs-old Base-cs)                ; the Base-cs clauses
 (setq Base-cs (append resolvents Base-cs))
 (setq resolvents nil)

;(format t "~%Above top2: n= ~s  round-number= ~s  round-limit= ~s ~% apply-cs=~% ~s ~% Base-cs-old=~% ~s~%Base-cs=~% ~s~%" n round-number round-number-limit apply-cs Base-cs-old Base-cs)

top2
 (cond ((null apply-cs) 
	(and *v* (format t "~%Finished Round ~s.  ~s Resolvents Generated.~%" 
           Round-number (length resolvents)));*v* is verbose lt (script-prover)
	(setq round-number (1+ round-number)) (go top1)))
 (cond ((eq 'Subsumed (car (last (setq C1 (car apply-cs)))))
        (setq apply-cs (cdr apply-cs)) (go top2)))
 (setq C1* (cadr C1)) (setq n1 (car C1))
 (setq Base-cs* (append (cdr apply-cs) Base-cs-old))

;(format t "~%Above top3: C1= ~s  C1*= ~s  n1= ~s  ~%Base-cs* ~%~s~%" C1 C1* n1 Base-cs*)

top3
 (cond ((null Base-cs*) (setq apply-cs (cdr apply-cs)) (go top2)))
 (cond ((eq 'Subsumed (car (last (setq C2 (car Base-cs*)))))
        (setq Base-cs* (cdr Base-cs*)) (go top3)))
 (setq C2* (cadr C2)) (setq n2 (car C2))
 (setq C1-V C1*)                                  ;Reset
  
;(format t "~%Above top4:~% C1= ~s~% C1*= ~s~% C2= ~s~% C2*= ~s~% n1= ~s n2= ~s  ~%"  C1 C1* C2 C2* n1 n2)

top4
 (cond ((null C1-V) (setq Base-cs* (cdr Base-cs*)) (go top3)))
 (setq L1 (car C1-V))
 (setq pairs-list                                              ;12Jan90
       (cond ((and %pairs-light% (eq-car L1 'not) (not (atom (cadr L1)))) 
	      (assoc (caadr L1) *pairs-structure*))                
	     (t nil)))
 (setq C2-V C2*)

top5
 (cond ((null C2-V) (setq C1-V (cdr C1-V)) (go top4)))
 (setq L2 (car C2-V))

 (cond (*break* (format t "~%Break after top5~%") (break)))

 (cond ((not (or (and (setq theta (unify-r L1 (not-formula-r L2)))
		      (or (setq L1p nil) t))                          ;12Jan90
#|		 (and pairs-list (setq L1p (pairs-s L1 L2 pairs-list)) ;12Jan90
		      (setq L1p (car L1p))   ;12Jan90 we are just using one
		      (setq theta (car L1p)) ; of the pair answers for now
		      (setq L1p (cdr L1p))
;(format t "~%After Pairs-s Base-cs= ~s~%" Base-cs)
		      )                      ;curtail this for now  13oct92
|#
		 ))
	(setq C2-V (cdr C2-V)) (go top5))
       
       ((and (setq R (union-w (remove L1 C1* :test 'equal)
			      (remove L2 C2* :test 'equal))) nil))
       ((and (setq R (union-w L1p R)) nil))                           ;12Jan90

       ((null R)
	(setq R (list (setq n (1+ n)) '[] (list n1 n2)))
	(setq resolvents (cons R resolvents))
	(setq *clauses* (append resolvents Base-cs))
	(setq *clauses* (sort *clauses* 'clause-number<))
        (setq *Ancestry-clauses* 
	      (Ancestry-clauses *clauses* (car (last *clauses*))))
;;	(cond (*v* (format t "~%Success:  [] found~%")    ;deleted 10oct92
;;		   (number-print-h *Ancestry-clauses*)))  ;deleted 10oct92
;;       (cond (*output* (number-print-h *clauses*)))     ;deleted 10oct92
;;	(return t)
	(format t "~% [] obtained~%")                     ;24Dec91
	(return '[])                                      ;24Dec91
	)
       (t (setq R (crunch (apply-sub theta R)))))
                 ;;ordinary resolvent
;;(prnt C1 C2 L1 L2 R)                                    ;13oct92
;;(format t "~%Resolvent= ~s ~%" R)

 (cond ((tautology R) ;(format t "~%R is a tautology ~s ~%" R) 
	(setq C2-V (cdr C2-V)) (go top5))
;;     ((> (delay R) 9) ;Discard R if its delay number is too large
;;      (setq C2-V (cdr C2-V)) (go top5))  ;24Dec91 omit this for now
       )

;;;%%%% TEST FOR UNIT-SUBSUMPTION
 (cond (*unit-subsumption*
	(setq rb (append resolvents Base-cs))
	(cond ((not (unit-subsumption-r R rb t)) 
	       ;;R is unit-subsumed by an earlier clause
	       (setq C2-V (cdr C2-V)) (go top5))
		 
	      ((null (cdr R))		;R is a unit
	       (setq resolvents 
		     (unit-subsumption-r R resolvents ))
	       (setq Base-cs
		     (unit-subsumption-r R Base-cs))))))

 (cond ((setq factors (factors R))
	(cond ((setq R* (factor-subsumes R factors))
	       ;;Replace R by its factor R* if R* subsumes R
	       (setq factors (cons R* (factors R*))))

	      (t (setq factors (cons R factors)))))
					;{R}U{factors of R}
       (t (setq factors (list R))))

  ;;Remove factors of R that are subsumed by previous clauses 
 (cond ((not *subsume*) nil)                   ;Dont try subsumption
       ((setq factors (mapcan #'(lambda (C*) 
         (cond ((some #'(lambda (CC) (cond ((eq 'subsumed (car (last CC))) nil)
					   ((subsume-c (cadr CC) C*) t)
					   (t nil)))
		      (append resolvents Base-cs)) nil)
	       (t (list C*)))) factors)))

       (t (setq C2-V (cdr C2-V)) (go top5)))

  ;;Deactivate any previous clause which is subsumed by a factors of R 
  (cond ((not *subsume*) nil)
        (t (mapc #'(lambda (C*) (setq resolvents  
	     (mapcar #'(lambda (CC) 
	        (cond ((eq 'subsumed (car (last CC))) CC)
		      ((subsume-c C* (cadr CC)) 
                       (append CC (list 'subsumed)))
		      (t CC))) resolvents)))
		 factors)
	   (mapc #'(lambda (C*) (setq Base-cs  
	     (mapcar #'(lambda (CC) 
		(cond ((eq 'subsumed (car (last CC))) CC)
		      ((subsume-c C* (cadr CC)) 
                       (append CC (list 'subsumed)))
		      (t CC))) Base-cs)))
		 factors)))

  (mapc #'(lambda (R) (setq R (list (setq n (1+ n)) R (list n1 n2)))
	    (and *v* (format t "~%Round# ~s  R=~%~s~%" round-number R))
(prnt R)

	    (setq resolvents (cons R resolvents)))
	factors)

  (setq C2-V (cdr C2-V)) (go top5)))
;;;End of the RESOLUTION function.
;;;-------------------------------------------------------------------------

(DEFUN APPLY-sub (theta D)  
  ;; the substitution THETA is applied to the form D
  (COND ((or (EQ T THETA) (null theta)) D)
	(T (SUBLIS THETA D))))

(defun clause-number< (A B) 
  (cond ((and (car A) (car B)) (< (car A) (car B)))))

;; CLAUSIFY
(defun clausify (A &optional sko par)
  ;;This converts the FOL fomula A into ordinary clauses.
    (cond (sko nil)                              ;already skolemized
	  (par (setq A (skolemize A nil t)))     ;Skolemize in place
	  (t   (setq A (skolemize A nil nil))))  ;Skolemize in place
    (cond (par (setq A (not-formula-r A))))      ;If par=t, negate A
    (setq A (convert-implication-signs A))
    (setq A (push-nots-in A))
    (setq A (CNF A))     ;At this point A is a set of ordinary clauses.
    (setq A (remove-&&-or-signs A)))

(defun remove-&&-or-signs (A) ;called by g-clausify
  ;;A is assumed to be in CNF, the output is a set of clauses
  (cond ((atom A) (list (list A)))
        ((eq '& (car A))
         (mapcan #'(lambda (u) (remove-&&-or-signs u)) (cdr A)))
        ((eq 'or (car A)) (list (cdr A)))
        (t (list (list A)))))

(defun convert-implication-signs (A) 
  ;;Ie, (-> P Q) goes to (or (not P) Q)
  (cond ((atom A) A)
        ((eq '-> (car A)) 
         (list 'or (list 'not (convert-implication-signs (cadr A)))
                   (convert-implication-signs (caddr A))))
        ((eq 'not (car A)) (list 'not (convert-implication-signs (cadr A))))

        ((member (car A) '(& or))
         (cons (car A) (mapcar #'convert-implication-signs (cdr A))))

	((member (car A) '(all some))
	 (list (car A) (cadr A) (convert-implication-signs (caddr A))))

        (t A)))

(defun pairs-s (L1 H pairs-list)   ;called by RESOLUTION  12Jan90
  ;;pairs-list came from *pairs-structure*
  (let (L1p C theta sigma A)
    (setq C (cadr L1))             ;L1 had the form (NOT C)
    (setq L1p (mapcan #'(lambda (u)
     (cond ((and (setq theta (instance H (car u)))
		 (setq A (apply-substitution theta (cadr u)))
		 (setq sigma (instance C A)))
	    (setq theta (compose theta sigma))
	    (setq A (apply-substitution theta (caddr u)))
	    (list (list theta (list 'not A))) )))
		      (cadr pairs-list)))

    L1p))

(defun push-nots-in (A) 
  ;;Assume that A has no -> signs.
  ;;Ie, (not (& A (or B C))) goes to  (or (not A) (& (not B) (not C)))
  (cond ((atom A) A)
        ((eq 'not (car A))
         (cond ((atom (cadr A)) A)
               ((eq 'not (caadr A))       ;(not (not A)) goes to A
		(push-nots-in (cadadr A)))  

               ((eq '& (caadr A))
                (cons 'or (mapcar #'(lambda (x) 
                            (push-nots-in (list 'not x))) (cdadr A))))

               ((eq 'or (caadr A))
                (cons '& (mapcar #'(lambda (x) 
                            (push-nots-in (list 'not x))) (cdadr A))))
	       
	       ((eq 'all  (caadr A))
		(list 'some (cadr (cadr A))
		      (push-nots-in (list 'not (caddr (cadr A))))))

	       ((eq 'some (caadr A))
		(list 'all  (cadr (cadr A))
		      (push-nots-in (list 'not (caddr (cadr A))))))

	       ((eq '<  (caadr A))
		(list '<= (caddr (cadr A)) (cadr (cadr A))))

	       ((eq '<= (caadr A))
		(list '<  (caddr (cadr A)) (cadr (cadr A))))

               (t A)))
        ((member (car A) '(& or)) 
         (cons (car A) (mapcar #'push-nots-in (cdr A))))
	((member (car A) '(all some))
	 (list (car A) (cadr A) (push-nots-in (caddr A))))

        (t A)) )

(defun cnf (A)      ;put A in Conjunctive Normal Form.
  ;;Assume that A has no -> signs, and has nots pushed in.
  ;;Ie, (or (& A B) C)  goes to  (& (or A C) (or B C))
  (let (AL)
    (cond ((atom A) A)
          ((and (member (car A) '(& or)) 
                (setq A (cons (car A) (mapcar #'cnf (cdr A)))) nil))
          ((eq '& (car A)) 
           (setq AL (cnf-& (cdr A)))      ;remove the other &'s
           (cons '& AL))
;;((or (prnt A) (break)))
          ((eq 'or (car A)) 
           (cond ((occur '& (cdr A))      ;permute &'s and or's and recurse
                  (cnf-or-& A))
                 (t (setq AL (cnf-or (cdr A)))     ;remove the other or's
                    (cons 'or AL))))
	  ((member (car A) '(all some))
	   (list (car A) (cadr A) (cnf (caddr A))))
          (t A))))

(defun cnf-& (AL)   ;called by cnf
  ;;remove the top level &'s (flatten)
    (cond ((null AL) nil)
          ((atom (car AL)) (cons (car AL) (cnf-& (cdr AL))))
          ((eq '& (caar AL))
           (cnf-& (append (cdar AL) (cdr AL))))
          (t (cons (car AL) (cnf-& (cdr AL))))))

(defun cnf-or (AL)   ;called by cnf
  ;;remove the top level or's (flatten)
    (cond ((null AL) nil)
          ((atom (car AL)) (cons (car AL) (cnf-or (cdr AL))))
          ((eq 'or (caar AL))
           (cnf-or (append (cdar AL) (cdr AL))))
          (t (cons (car AL) (cnf-or (cdr AL))))))
          
(defun cnf-or-& (A &optional AF)   ;called by cnf
  ;;A starts with the form (or A B ...)
  (cond ((null A) AF)
        ((atom (car A)) (cnf-or-& (cdr A) (append AF (list (car A)))))
        ((eq '& (caar A))
         (setq A (mapcar #'(lambda (x) (append AF (cons x (cdr A)))) (cdar A)))
         (setq A (cons '& A))
         (cnf A))
        (t (cnf-or-& (cdr A) (append AF (list (car A)))))))

(defun factors (R)   ;Called by RESOLUTION
  ;;This finds the factors  of R, if any, unless *factors* is NIL
 (let (theta)
   (cond ((not *factors*) nil)
	 (t (crunch (mapcan #'(lambda (S) (cond
	         ((and (cdr S) (setq theta (variant-all S))) 
		  ;;note, only non-unit S's
		  (list (crunch (apply-sub theta R)))) )) (subsets R)))))))

(defun factor-subsumes (R Facs &optional R*) ;called by resL
  ;;Return the smallest member R* of Facs which is a subset of R, if there
  ;; is one, else return nil. (Unless *factors* is nil)
  (cond ((not *factors*) nil)
	((some #'(lambda (R2) (cond ((subset R2 R) (setq R* R2)))) facs)
         (factor-subsumes R* (remove R* facs :test 'equal) R*))
        (t R*)))

(defun free-variables-in (A)        
  ;;The same as variables-in, except that here we omit quantified vars
  (crunch (f-vars-in A)))
 
(defun f-vars-in (A)                ;used-in FREE-VARIABLES-IN
  (cond ((numberp A) nil)
        ((null A) nil)
        ((member A '([])) nil)
	((atom A) (list A))
	((member (car A) 
                '(all some set-of union-of-sets intersection-of-sets lambda
		      product sum))
	 (f-vars-in (subst= (list (cadr A)) (cadr A) (caddr A))))
	((atom (car A))
	 (cond ((null (cdr A)) nil) (t (mapcan 'f-vars-in (cdr A)))))
	(t (mapcan 'f-vars-in A))))

(DEFUN COMPOSE-R (THETA SIGMA)  ;widely used  
  ;; this returns the composition of the two substitutions THETA & SIGMA
    
  (COND ((or (null theta) (null sigma)) nil)
        ((EQ T THETA) SIGMA)
        ((EQ T SIGMA) THETA)
        (T (SETQ THETA (MAPCAR #'(lambda(U)
				   (CONS (CAR U) (SUBLIS= SIGMA (CDR U))))
			       THETA))
	   (SETQ THETA (UNION-W THETA SIGMA))
	   (cond ((MAPCAN #'(lambda(U)
		       (COND ((EQUAL (CAR U) (CDR U)) NIL) (T (LIST U))))
		   THETA))
		 (t t)))))
    
(DEFUN CRUNCH (L)   ;widely used  
  ;;removes duplication from a list
  (COND ((NULL L) L)
	((MEMBER (CAR L) (CDR L) :test #'equal) (CRUNCH (CDR L)))
	(T (CONS (CAR L) (CRUNCH (CDR L))))))

;;See the next version
(DEFUN INSTANCE (A B &optional L)	;used-in VARIANT, etc etc
  ;; If A is an instance of B this returns the substitution THETA for 
  ;; which A = B THETA,  else nil
  ;; This is the traditional INSTANCE.  See another version in Reduce.lisp
  ;;   that handles more complicated terms.
  (Let (THETA LAMBDA B* L*)
       (COND ((ATOM B)
	      (COND ((EQ B A) T)
		    ((NUMBERP B) NIL)
                    ((member B L) nil)
;                   ((occur B A) NIL)              ;not needed for instance?
		    (T (LIST (CONS B A)))))
	     ((ATOM A) NIL)
	     ((NOT (EQ (CAR B) (CAR A))) nil)
	     ((NULL (CDR B)) (COND ((NULL (CDR A)) T)))
	     ((NULL (CDR A)) NIL)
	     ((SETQ THETA (INSTANCE (CADR A) (CADR B) L))
              (cond ((eq t theta) 
                     (instance (cons (car A) (cddr A)) 
                               (cons (car B) (cddr B)) L))
                    ((and (setq B* (apply-sub theta B))
                          (setq L* (union-w L 
                            (mapcan #'(lambda (u) (cond ((atom (cdr u)) 
							 (list (cdr u)))))
				    theta)))
			  nil))
       			 ;;L* is the list of atoms y, for (cons x y) in theta
		    ((setq lambda (instance A B* L*))
                     (compose-r theta lambda)))))))

(DEFUN INSTANCE (A B &optional L)	;used-in VARIANT, etc etc  21Sept91
  ;; If A is an instance of B this returns the substitution THETA for 
  ;; which A = B THETA,  else nil
  ;; This is the traditional INSTANCE.  See another version in Reduce.lisp
  ;;   that handles more complicated terms.
  (Let (THETA LAMBDA B* A* L*)
       (COND ((ATOM B)
	      (COND ((EQ B A) (list (cons B B)))
		    ((NUMBERP B) NIL)
                    ((member B L) nil)
;                   ((occur B A) NIL)              ;not needed for instance?
		    (T (LIST (CONS B A)))))
	     ((ATOM A) NIL)
	     ((NOT (EQ (CAR B) (CAR A))) nil)
	     ((NULL (CDR B)) (COND ((NULL (CDR A)) (list (cons 'new 'new)))))
	     ((NULL (CDR A)) NIL)
	     ((SETQ THETA (INSTANCE (CADR A) (CADR B) L))
	      (setq B* (apply-sub theta (cons (car B) (cddr B))))
	      (setq A* (apply-sub theta (cons (car A) (cddr A))))
	      (setq L* (cond ((eq t theta) nil)
			     (t (union-w L (mapcar #'car theta)))))
               ;;the list of first members of THETA
	      (cond ((setq lambda (instance A* B* L*))
                     (compose-r theta lambda)))))))

(defun not-formula-r (A)  ;Called by RESOLUTION, RESOLVENTS, SUBSUME-C
  ;;Eg, (not B) goes to B, and B goes to (not B).
  (cond ((atom A) (list 'not A))
        ((eq 'not (car A)) (cadr A))
        (t (list 'not A))))

(defun resolvents (L1 L2)   ;called by SUBSUME2
  ;;L1 is a set of clauses, and L2 is a set of leterals.
  (let (theta R)
    (mapcan #'(lambda (L*) (mapcan #'(lambda (C) (mapcan #'(lambda (L) 
        (cond ((setq theta (unify-r L (not-formula-r L*)))
               (setq R 
		     (crunch (apply-sub theta (remove L C :test 'equal))))
               (list R)))) C)) L1)) L2)))

(defun subst= (a b c)  ;replace a by b in c
   ;;So that (subst= '(y) '(x) '(P x))                correctly gives   (P x).
   ;;Whereas (subst  '(y) '(x) '(P x) :test 'equal) incorrectly gives   (P y).
   (cond ((equal b c) a)
         ((atom c) c)
         (t (mapcar #'(lambda (x) (subst= a b x)) c))))

(defun sublis= (theta A)  ;apply the substitution theta to A
   ;;So that (sublis= '( ((x) y)) '(P x))     correctly gives   (P x).
   ;;Whereas (subst   '( ((x) y)) '(P x) :test 'equal) incorrectly gives (P y).
     (mapc #'(lambda (u) (setq A (subst= (cdr u) (car u) A))) theta)
     A)

(defun subsets (L)   ;Called by FACTORS.  The ordinary subsets of L.
  (let (S) 
    (cond ((null (cdr L)) (list nil L)) ; (a) -->  (nil (a))
          (t (setq S (subsets (cdr L)))
             (append S (mapcar #'(lambda (S2) (cons (car L) S2)) S))))))

;;;SUBSUME
(defun subsume-c (A B)   
  ;;Return t if A subsumes B. (Including if A=B).
  ;;Ala Chang&Lee, p.95 (See also the file SUBSUME.LISP).
  (let (v)
;;  (cond ((variant A B) (return-from subsume-c t)))  ;I had nil here before??
    (setq V (free-variables-in B))
    (mapc #'(lambda (x) 
               (setq B (subst (list (gensym)) x B :test 'equal))) V)
    (setq B (mapcar #'not-formula-r B))
    (subsume2 (list A) B) ))

(defun subsume2 (L1 L2)                   ;called by SUBSUME-C
  (let (L3)
    (cond ((member nil L1) T)
          ((setq L3 (resolvents L1 L2)) (subsume2 L3 L2)))))

(defun subsume-out (S &optional (S* S))     ;called by PR in PBR.LISP
  ;;Remove from S any clause subsumed by another clause in S.
  (let ()
    (cond ((null S) S*)
	  ((some #'(lambda (C) (and (not (equal C (car S)))
				    (subsume-c C (car S))))
		 S*)
	   (subsume-out (cdr S) (remove (car S) S* :test 'equal)))
	  (t (subsume-out (cdr S) S*)) )))

(defun unit-subsumption-r (C cs &optional flag)
  ;;called by RESOLUTION.  Assume the *unit-subsumption* is t
  ;;If FLAG is NIL we do backward unit subsumption, else forward.
  ;;If FLAG in NIL this returns the subset of clauses in cs not subsumed by C,
  ;;   if C is a unit, cs otherwise.
  ;;If FLAG is not NIL it returns nil if C is subsumed by a unit clause in cs,
  ;;   C otherwise.

  (cond (flag                        ;forward subsumption case
	 (cond ((some #'(lambda (CC) (and (null (cdadr CC)) 
					  (unit-subsume (caadr CC) C)))
		      cs) nil)
	       (t t)))
	((cdr C) cs)                   ;If C is not a unit clause, return nil
	(t (mapcan #'(lambda (CC) 
              (cond ((eq 'subsumed (car (last CC))) (list CC));already subsumd
	            ((unit-subsume (car C) (cadr CC)) 
                     (list (append CC (list 'subsumed))))
		     ;;mark subsumed clauses
		    (t (list CC)))) cs )) ))

(defun unit-subsume (L C)   ;called by UNIT-SUBSUMPTION
  ;;L is a literal and C is a clause.
  ;;If L subsumes some literal in C, this returns t.
  (some #'(lambda (L*) (instance L* L)) C))

(defun tautology (C)  ;called by RESOLUTION
  ;;Does C have literals of the form  A  and  ~A?
  (let (A) 
    (cond ((null (cdr C)) nil)
          ((atom (setq A (car C)))
           (cond ((member (list 'not A) (cdr C) :test 'equal))
                 (t (tautology (cdr C)))))
          ((eq 'not (car A)) (cond ((member (cadr A) (cdr C) :test 'equal))
                                   (t (tautology (cdr C)))))
          ((member (list 'not A) (cdr C) :test 'equal))
          (t (tautology (cdr C))))))

(defun variant (A B) (and (instance A B) (instance B A)))

(defun variant-all (S) ;FACTORS
  ;;Find a common variant-unifier of members of S.
  (let (theta lambda)
    (cond ((null (cdr S)) T)
          ((null (cddr S)) (variant (car S) (cadr S)))
          ((not (setq theta (variant-all (cdr S)))) nil)
          ((setq lambda (variant-all (crunch (apply-sub theta S))))
           (compose-r theta lambda)))))

(DEFUN UNIFY-R (A B)                         ;Regular Unify
  (let (ta tb)
    (COND ((EQUAL A B) T)
	  ((and (ATOM A) (not (numberp A)))
	   (COND ((or (OCCUR-UNIFY A B)
		      (and (setq ta (type-h A)) (setq tb (type-h B)) 
			   (not (equal ta tb)))) nil)
		 ;;This typing will have to be handled carefully, but can
		 ;; done automatically, by preprocessing the clauses being
		 ;; resolved
		 (t (LIST (CONS A B)))))
	  ((and (ATOM B) (not (numberp B)))
	   (COND ((or (OCCUR-UNIFY B A)
		      (and (setq ta (type-h A)) (setq tb (type-h B)) 
			   (not (equal ta tb)))) nil)
		 (t (LIST (CONS B A)))))
	  ((OR (ATOM A) (ATOM B)) NIL)
	  ((NOT (EQUAL (CAR A) (CAR B))) NIL)
	  (T (UNIFY-sets (CDR A) (CDR B) NIL)))))
 
(DEFUN UNIFY-sets (L L* &optional THETA) ;Called by UNIFY-R
  (LET (THE)
    (COND ((NOT (EQUAL (LENGTH L) (LENGTH L*))) NIL)
	  ((NULL L) (COND (L* NIL) (THETA) (T T)))
	  ((EQUAL (CAR L) (CAR L*)) (UNIFY-sets (CDR L) (CDR L*) THETA))
	  ((SETQ THE (UNIFY-R (CAR L) (CAR L*)))
	   (SETQ THETA (UNION-W (SUBLIS THE THETA) THE))
	   (UNIFY-sets (SUBLIS THE (CDR L)) 
		       (SUBLIS THE (CDR L*)) THETA)))))

(DEFUN OCCUR-UNIFY (X A)   ;used-in UNIFY-R
  (COND ((ATOM A) (EQ X A))
	((CDR A) (SOME #'(LAMBDA (Y) (OCCUR-UNIFY X Y)) (CDR A)))))

(defun type-h (A)    ;Called by Unify-R
  (let (ta)
    (cond ((atom A) 
	   (cond ((member A '(D D1 D2)) 'D)
		 ((member A '(E E0))    'E)))
	  ((type-h (car A)))
	  ((member (car A) '(half)) (type-h (cadr A)))
	  ((and (member (car A) '(min max)) 
		(equal (type-h (cadr A)) (setq ta (type-h (caddr A)))))
	   ta)
	  (t nil))))


(DEFUN UNION-W (A B)			;a version of UNION

       (COND ((NULL B) A)
	     ((MEMBER (CAR B) A :test #'equal) (UNION-W A (CDR B)))
	     (T (UNION-W (APPEND A (LIST (CAR B))) (CDR B)))))

(defun ancestry-clauses (cl C)
 ;;Cl is a list of numbered clauses, C is a clause of cl
 ;;Each clause of cl has the form   (n Clause (parent1 parent2))
 ;;L is a partial list of integers
 ;;This returns a sublist of cl which contains only those clauses
 ;; supporting []
  (sort (crunch (anc-cls cl C))	'clause-number<))

(defun anc-cls (cl C) ;called by ANCESTRY-CLAUSES
 (let (P C1 C2)
   (cond ((not (setq P (caddr C))) (list C))
         (t (setq C1 (assoc (car  P) cl))
	    (setq C2 (assoc (cadr P) cl))
	    (append (list C C1 C2) 
		    (append (anc-cls cl C1) (anc-cls cl C2)))))))

(defun anc-cls-h (cl C) ;called by ANCESTRY-CLAUSES
  ;;This is the hyper-resolution version of anc-cls
  ;; Here each clause has the form (n pos neg parentage)
 (let (P S)
   (cond ((not (setq P (cadddr C))) (list C))  ;parentage
         (t (setq S (mapcar #'(lambda (n) (assoc n cl)) (cdr p)))
	    (append (cons C S)
                    (mapcan #'(lambda (C*) (anc-cls-h cl (cadddr C*))) S))))))

(defun assoc3 (k L) ;assoc on Caddr instead of car
  (cond ((null L) nil)
        ((equal k (caddar L)) (car L))
	(t (assoc3 k (cdr L)))))


(defvar prob00 nil)
(setq prob00
 '(((NOT (A)) (NOT (B)) (D))
   ((A) (E))
   ((B) (D))
   ((H))
   ((NOT (D)) (NOT (H)))
   ((NOT (E)))))


(defun subsume-list (L) 
;;This function returns the set of clauses in L not subsumed by another 
;; clause in L
  (let (L* L**)
    (setq L** L)
    
    (mapc #'(lambda (C) (cond ((and (member C L** :test 'equal)
				    (setq L* (remove C L** :test 'equal))
                 (some #'(lambda (CC) (subsume-c CC C)) L*))
			       (setq L** L*)) ))
	  L)
    L**))
			      

(setq L '( ((<= 0 (D1 (MIN X Y))) (NOT (<= 0 X)) (NOT (<= 0 (MIN X Y))))
 ((<= 0 (D1 (MIN Y Y))) (NOT (<= 0 Y)) (NOT (<= 0 (MIN Y Y))))
 ((<= 0 (D1 (MIN X X))) (NOT (<= 0 X)) (NOT (<= 0 (MIN X X))))))


;;7Dec92 
(setq CL '(
 ((R (a)))
 ((H (F z)))
 ((M (a)))
 ((not (R x)) (not (M x)) (not (H (f y))) (L x y))
 ((H (g x)) (M x))
 ((not (L x (j x))) (Q x))
 ((not (R x)) (not (L x (g x))) (M x))
 ((not (Q (a)))) ))

(defun r () (time (RESOLUTION nil CL)))
