 
;;;*********************************************************************
;;;
;;;                      SET-VAR.Lisp
;;;                   STR+VE-use Version
;;;                  Split Clauses Version
;;;              Generate-Guns&Targets Version
;;;
;;;               W. W. Bledsoe         2Feb93
#|
To Load, zap the following 7 lines, and run them under *LISP*

 (import '(user::set-diff user::[ user::[= user::element user::<-> user::->
	  		  user::& user::all user::nonempty user::singleton
			  user::s*) "SLOOP")
 (load "imply.lisp") (load-set-var-all)

 (si::chdir "/users/woody/strive-1-1")
 (load "load-latest") (load "load-strive.lsp") (user::load-strive) 
 (si::chdir "/users/woody")
               
|#
;;; 
;;;   This is a prover which implements the SET-VAR system
;;;    described in ATP 106.  It uses the file (prover) hand.set-var.
;;;
;;;   It uses STR+VE to "finish" a proof. Eg, (user::prove in::imv 40 400)
;;;    or, (run-examples ineq-examples :start-wt 40 max-time-seconds 40)
;;;
;;;   This version implements another version of the generate and 
;;;   search strategy, emphasizing the PARTNERS role.  17Nov92
;;;
;;;**********************************************************************

(defun lsv () (load "set-var.lisp"))               ; Load strive.

(defvar *CL* nil)                                  ; Global variables
(defvar *CLL* nil)
(defvar *collapser* '((0 ((subset D% D%)) (input))))
(setq   *collapser* '((0 ((subset D% D%)) (input))))
(defvar *A* nil)
(defvar *th-clauses* nil)
(defvar *R-clauses* nil)
(defun  *pairs* nil)
(defun  *H* nil)
(defun  *theorem* nil)
(defvar *n* 1)
(defvar *set-theory-definitional-lemmas* nil)
(defvar *set-theory-definitional-clauses* nil)
(defvar *set-theory-definitional-clauses** nil)
(defvar *guns-find-lemmas* nil)
(defvar *guns-find-clauses* nil)
(defvar *guns-find-clauses** nil)
(defvar *A-value* nil)

(defun set-var (theorem &optional hypothesis clause-list A)     ;the top call
  ;;This is the MAIN call.  
  ;;If hypothesis is non-nil, the real theorem is (hypothesis -> theorem).
  ;;If clause-list is non-nil, it has the form 
  ;; ((conclusion clauses) (hypothsis clauses)), all numbered (n (clause)input)
  (let (CL CLt CLh (n 0) m sdl gfl sdc gfc lmc RR)   
    (setq *n* 1)
    (setq *A* t)
    (setq *A-value* nil)
    (setq %theorem-variables% (append 
	      (set-difference-w (all-variables-in theorem)
				(bound-variables-in theorem)) '(%x%)))
    ;;STEP1: Clausify the Theorem
    (cond (clause-list (setq CL (append (car clause-list) (cadr clause-list))))
	  (hypothesis
	   (setq CLt (clausify theorem))      ;clausify is defined below
	   (setq CLt (binary->unary< CLt))     
	   (setq CLt (mapcar #'(lambda (c) (setq c (factor-subsume c))
				           (list (setq n (1+ n)) c '(INPUT)))
			     CLt))
	   (setq CLh (clausify (list 'not hypothesis)))
	   (setq CLh (binary->unary< CLh))  
	   (setq CLh (mapcar #'(lambda (c) (setq c (factor-subsume c))
				           (list (setq n (1+ n)) c '(INPUT)))
			     CLh))
	   (setq CL (append CLt CLh)) )
	  (t (setq CL (clausify theorem))       ;CLAUSIFY is in this file
	     (setq CL (binary->unary< CL))     
	     (setq CL (mapcar #'(lambda (c) (setq c (factor-subsume c))
				            (list (setq n (1+ n)) c '(INPUT)))
			       CL))
	     (setq Clh CL) (setq CLt CL) ))
    ;;STEP2: Clausify the *set-theory-definitional-lemmas*  or adjust numbering
    (cond (*set-theory-definitional-clauses*   
	   (setq m (+ (- n (caar *set-theory-definitional-clauses*)) 1))
            ;;adjust the numbers of these clauses
	   (setq *set-theory-definitional-clauses*
		 (mapcar #'(lambda (C) (cons (+ (car C) m) (cdr C))) 
			 *set-theory-definitional-clauses*))
	   (setq *guns-find-clauses*
		 (mapcar #'(lambda (C) (cons (+ (car C) m) (cdr C))) 
			 *guns-find-clauses*))
	   (setq n (+ n m)))

	  (t (setq sdl *set-theory-definitional-lemmas*) ;reclausify and number
	     (setq gfl *guns-find-lemmas*)
	     (setq sdc nil) (setq gfc nil)
	     (mapc #'(lambda (lm) 
                   (prnt lm)
		   (setq lmc (clausify (list 'not (eval lm))))
		   (setq lmc (mapcar #'(lambda (c) 
				(setq c (factor-subsume c))       ;23oct92
                                (list (setq n (1+ n)) c '(INPUT)))
				     lmc))
		   (setq sdc (append sdc lmc)) ;add numbered clauses to sdc&gfc
		   (cond ((member lm gfl) (setq gfc (append gfc lmc))))) 
		   sdl)
	     (setq *set-theory-definitional-clauses* sdc)
	     (setq *guns-find-clauses* gfc)))

    (setq *CL* (append CL *set-theory-definitional-clauses*))
    (setq *pairs* (find-pairs *CL*))   
    (setq *CL* (append *collapser* *CL*))  
     ;;*CL* holds ALL Clauses and resolvents throughout the proof
    (setq *th-clauses* (cond (clause-list (car clause-list))
			     (hypothesis CL)
			     (t CL)))
    (setq *A* (cond (A A) (t (set-var-variable *th-clauses* CL))))
     ;;Find the set-variable A (unless given) and call it *A* (global var)
    (prnt *A*)
    (cond ((eq '[] (setq RR (sv CL)))
	   '[])
	  (t (format t "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx")
	     (number-print *CL*))
	  )))

(defun sv (CL)                 ;called by SET-VAR  2oct92
  (let (CL2 CL* CL**)

    (cond ((eq '[] (setq CL* (Red-subset CL))) '[])          ;5oct92
	  ((eq '[] (setq CL** (a-guns-sv11 CL*))) '[])
	  ((eq '[] (setq CL2 (A-guns-sv CL**))) '[])         ;17Nov92
	  (t nil))))

(defun A-guns-sv (CL &optional guns targets)        ;17Nov92  Called by SV
  (let (RR RR2 CL2 CL2* CL3 targets2)
    (setq *R-clauses* nil) 
    (and (not guns) (setq guns (new-guns CL)))
    (cond ((eq '[] guns) (return-from A-guns-sv '[])))
    (setq CL (union-w CL *R-clauses*))             
    (and guns (not targets) (setq targets (find-targets guns CL)))
    (cond ((eq '[] targets) (return-from A-guns-sv '[])))
    (cond ((not targets) (return-from A-guns-sv (finish-by-strive CL))))
    (setq targets (subsume-targets targets))
    (setq RR (fire targets))
    (cond ((eq '[] RR) '[])
	  (RR (setq RR (set-difference RR CL))           ;20Nov92
	       (setq CL2 (subsume-out RR CL))
	       (setq CL2 (union-w CL2 RR))
	       (setq CL2 (remove-ancesters CL2))            ;28Nov92
	       (setq CL2* (remove-descendants-cr=cr CL2))    ;18Jan93
	       (format t "~%~s~%remove-descendants-cr=cr~%~s" 
		       (mapcar #'car CL2) (mapcar #'car CL2*))
	       (setq CL2 CL2*)
	       (cond ((setq RR2 (finish-by-strive CL2)))
		     (t (A-guns-sv CL2))))
	  ((eq '[] (setq targets2 (unblock targets CL))) '[])
            ;;To unblock, do one round of resolution, restructure targets
	  (targets2 (A-guns-sv CL guns targets2))
	  (t nil) )))

(defun A-guns-sv11 (CL)        ;16Sep92 Called by SV
  (let (guns11 RR11 RR22)
    (cond ((setq guns11 (sv-guns CL 'r11))
	   (cond ((eq '[] (setq RR11 (fire11 guns11 CL))) '[])     ;replaces
		 ((eq '[] (setq RR22 (collapse1 RR11 t))) '[])     ;replaces
		 (t RR22) ))
	  (t CL) )))

(defun add-LA (LA0 R A parity &optional (theta t))       ;called by RULE1, etc
   ;;1Sep92 Changed from hand.set-var version
   ;;combine LA0 with any other Set-Of-Literal in R
   ;;LA0 has the form  (not (subset A (set-of z% (P z%))))
   ;;eg, If LA0 = (not (subset A (set-of z% (P z%))))
   ;;    and  R = ((not (subset A (set-of z% (Q z%)))) D)
   ;;   then (add-LA LA0 R) = ((not (subset A (set-of z% (& (P z%)(Q z%))))) D)
   ;;1sep92 also case La0= (not (subset (set-of z% (P z%)) A)), parity nil

   (let (B LA1 LA2 P1 P2 R2 sigma)
     (setq R (remove LA0 R :test 'equal))
     (cond ((some #'(lambda (LA) 
		      (and (eq-car LA 'not) (eq-car (setq B (cadr LA)) 'subset)
			   (or (and parity (eq (cadr B) A)
				    (eq-car (caddr B) 'set-of)
				    (eq (cadr (caddr B)) 'z%))
			       (and (not parity) (eq (caddr B) A)
				    (eq-car (cadr B) 'set-of)
				    (eq (cadr (cadr B)) 'z%)))
			   (setq LA1 LA)))
		  R)
	    ;;LA1 has the form  (not (subset A (set-of z% (Q z%))))
	    ;;              or  (not (subset (set-of z% (Q z%)) A))
	    (cond (parity
		   (setq P1 (caddr (caddr (cadr LA0))))	;(P z%)
		   (setq P2 (caddr (caddr (cadr LA1))))	;(Q z%)
		   (setq P1 (and-on-a P2 P1))             
		   (setq sigma (cadr P1)) ;see hand.set-var
		   (setq P1 (car P1))	;(& (P z%) (Q z%))
		   (setq LA2 (list 'not (list 'subset A 
					      (list 'set-of 'z% P1))))
		   (setq R2 (remove LA1 R :test 'equal))
		   (add-LA LA2 R2 A (compose sigma theta)))
		  (t 
		   (setq P1 (caddr (caddr (caddr LA0)))) ;(P z%)
		   (setq P2 (caddr (caddr (caddr LA1)))) ;(Q z%)
		   (setq P1 (and-on-a P2 P1))             
		   (setq sigma (cadr P1))
		   (setq P1 (car P1))	;(& (P z%) (Q z%))
		   (setq LA2 (list 'not (list 'subset 
					      (list 'set-of 'z% P1) A )))
		   (setq R2 (remove LA1 R :test 'equal))
		   (add-LA LA2 R2 A parity (compose sigma theta)))))

	   (t (apply-sub theta (cons LA0 R))))))

(defun ancesters (CL &optional (first t))  ;28Nov92  Called by REMOVE-ANCESTERS
  ;;This returns the list of clause #'s of all ancesters of members of CL
  (let (parents grand-ancesters LC anc CL*)
    (setq LC (mapcar #'car CL))         ;the clause #'s of CL
    (setq parents (mapcan #'(lambda (C) 
                 (cond ((equal '(input) (setq anc (caddr C))) nil)
					;anc=(action (m i) (n j))
		       ((atom (caddr anc)) (list (caadr anc)))
		       (t (list (caadr anc) (car (caddr anc))))))
			  CL))
    (cond ((not parents) (cond (first nil) (t LC)))
	  (t (setq parents (sort (crunch parents) '<))  ;the clause #'s
	     (setq CL* (mapcar #'C parents))         ;the corresponding clauses
	     (setq grand-ancesters (ancesters CL* nil))
	     (sort (crunch (append grand-ancesters parents)) '<)) )))

(defun ancestry-clauses-sv (cl C)
 ;;Cl is a list of numbered clauses, C is a clause of cl
 ;;Each clause of cl has the form   (n Clause (action (n i) (m j)))
 ;;                             or  (n Clause (action (n i) single))
 ;;19Aug92 This is different from the resolution.lisp version
 ;; in that there the form  was    (n Clause n m)
 ;;L is a partial list of integers
 ;;This returns a sublist of cl which contains only those clauses
 ;; supporting []
  (sort (crunch (anc-cls-sv cl C)) 'clause-number<))

(defun anc-cls-sv (cl C) ;called by ANCESTRY-CLAUSES-sv
 (let (P C1 C2 n m)
;;(prnt cl C) (break)
   (cond ((equal '(input) (setq P (caddr C))) (list C)) 
         (t (setq C1 (assoc (caadr P) cl))   ;P=(action (n i) (m j))
	    (cond ((member (caddr P) '(single)) ;P=(action (n i) single)
		   (append (list C C1) (anc-cls-sv CL C1)))
		  (t (setq C2 (assoc (caaddr P) cl))  ;Clause m
		     (append (list C C1 C2) 
			     (append (anc-cls-sv cl C1)
				     (anc-cls-sv cl C2)))))))))
(defun ancestry-multiple (CL CL2) 
  (let (CLL)
    (setq CLL (mapcan #'(lambda (C) (ancestry-clauses-sv CL C)) CL2))
    (setq CLL (sort CLL 'cl-sf))))

(defun C (n) (assoc n *CL*))
(defun CS (L) (mapcar #'C L))

(defun clause-sort-function (A B) (< (car A) (car B)))
(defun Cl-sf (A B) (clause-sort-function A B))

(defun change-variables (V thm)              ;24Nov92
  ;;rename the variables of V that appear in thm
  ;;ed, (change-variables (P x y z w) '(x z)) = (P x1 y z1 w)
  (let (V2 thm* x*)
    (cond ((setq V2 (variables-in thm))
	   (setq V2 (set-intersection V2 V))
	   (setq thm* thm)
	   (mapc #'(lambda (x) 
              (setq %theorem-variables% (cons x %theorem-variables%))
	      (setq x* (new-symbol x))
	      (setq thm* (subst x* x thm*)))
		 V2)
	   thm*)
	  (t thm))))

(defun clausify (A)
  ;;This converts the FOL fomula A into ordinary clauses.
;;(format t "~%clausify in set-var-r.lisp~%")
    (setq A (skolemize A nil t))  ;Skolemize in place
    (setq A (not-formula-r A))     
    (setq A (convert-implication-signs A))
    (setq A (push-nots-in A))
;;  (setq A (reduc A))                                ;1 Oct 92
    (setq A (CNF A))     ;At this point A is a set of ordinary clauses.
    (setq A (remove-&&-or-signs A)))

(defun collapse1 (CL &optional can)          ;10Sep92 Called by FIRE1-ETC. 
  ;;For Rules 1,1.1
  ;;Very much like collapse, but less general, and other changes.
  ;;This replaces the set variable A by {z% (P z%)} in a clause C of CL if 
  ;; (not (subset A {z% (P z%)})) is a literal of C and simplify the result.
  ;;We obtain a changed version of CL.
  ;;Changed clauses are returned and also are placed in *CL*.
  ;;23Sep92 If the parameter CAN is not nil then require that a clause C
  ;; be eligible-for-collapse before collapsing it.
  (let (R C* R*)
;;(prnt CL) (break)
     (cond ((null CL) nil)
	   ((and (setq C* (car CL)) can 
		 (not (eligible-for-collapse (cadr C*))))
	    (cons (car CL) (collapse1 (cdr CL) can)))
	   ((equal (cadr C*) (setq R (collapse-cc (cadr C*))))
	    (cons (car CL) (collapse1 (cdr CL) can)))
	   (R (setq R* (new-parentage 'res R C* 'L0 'R0 1))
	      (setq *CL* (append *CL* (list R*)))
	      (cond ((eq '[] (setq R (collapse-cc (cadr C*)))) 
		     (return-from collapse1 '[])))
	      (cons R* (collapse1 (cdr CL) can)))
	   (t (format t "~%ALARM in Collapse1~%") nil) )))

(defun collapse2 (RR CL)                ;17Nov92 Called by A-guns-sv
  ;;If C1 is in RR and C1 is eligible for collapse, and 
  ;; if C1 has not partner in CL with an A-gun, 
  ;; then collapse ALL C's in CL with the same set-of value as in C1.
  ;;Eg, IMV, (collapse2 (CS '(32 33 34)) (cs '(1 2 3 4 11 12 13 14 32 33 34)))
  ;;=nil but (collapse2 (CS '(42)) (cs '(1 2 3 4 11 12 13 14 32 42)))
  ;;= (1 2 3 4 11 12 13 14 43 44) where 32 and 42 are replaced by 43 and 44.
  (let (L A pz CL2 C* B)
    (cond ((some #'(lambda (C) 
      (and (eligible-for-collapse (cadr C))
	   (every #'(lambda (C3) (some #'(lambda (L) (some #'(lambda (L3)
		     (and (partners2 C C3 L L3) (setq C* C)))
							   (cadr C3)))
				       (cadr C)))
		  RR)))
		 CL)
	   (setq L (car (cadr C*)))    ;L= (not (subset A (set-of z Pz)))
	   (setq A (cadadr L))
	   (setq B (caddr (cadr L)))   ;B= (set-of z Pz)
	   (setq Pz (cadr B))
	   (setq CL2 (collapse-with-a-value CL A Pz)))
	  (t CL))))

(defun collapse-with-value (CL A A-value &optional u)  ;29oct92  u 6Jan93
  ;;replace A by A-value in ALL OF CL, simplify, reclause, etc, as needed
  ;;eg, if A= A, A-value= (set-of z% (& (P z%) (Q z%))), and
  ;;CL= (((D) (E)) 
  ;;     ((not (subset A (set-of z% (P z%)))) 
  ;;      (not (element tm1 A)) (element tm2 A)))
  ;;then (collapse-with-value CL A A-value) 
  ;;  = ((98 ((D) (E)))
  ;;     (99 ((not (element tm1 (set-of z% (& (P z%) (Q z%)))))
  ;;               (element tm2 (set-of z% (& (P z%) (Q z%)))))))
  ;;  =   after simplification and reclausification
  ;;    ((98 ((D) (E)))
  ;;     (101 ((NOT (P TM1)) (NOT (Q TM1)) (P TM2)) (RES (99 1) (0 1)))
  ;;     (102 ((NOT (P TM1)) (NOT (Q TM1)) (Q TM2)) (RES (99 1) (0 1))))
  (let (R RR RR* n)
    (setq n 100)
    (mapcan #'(lambda (C) 
      (cond ((eq '[] (setq R (collapse-cc (cadr C) A A-value u))) 
	     (return-from collapse-with-value '[]))
	    ((equal (cadr C) R) (list C))
	    (R (setq RR (reclausify R))
	       (setq RR* (mapcar #'(lambda (R2) 
		   (setq n (1+ n))
                   (new-parentage 'res R2 C 'L0 'R0 1 n))
				 RR))
	       (setq *CL* (append *CL* RR*))                     ;29Dec92
	       RR*)
	    (t (format t "~%ALARM: COLLAPSE-CC returned NIL~%") nil)))
	    CL)))

(defun collapse-cc (C &optional A2 A-value u)   ;10Sep92  29oct92
  ;;Called by COLLAPSE1 and COLLAPSE-WITH-VALUE 29oct92
  ;;C is an ordinary clause, a list of literals.
  (let (nL A sz C* C** Cs rest)
    (cond ((some #'(lambda (L)   ;L= (not (subset A (set-of z% (P z%))))
              (and (eq-car L 'not) (setq nL (cadr L))
		   (eq-car nL 'subset) 
		   (atom (setq A (cadr nL)))
		   (setq sz (caddr nL)) (eq-car sz 'set-of)
		   (cond (u (setq rest (remove L C :test 'equal))
			    ;;u=t means that it is sound to replace   6Jan93
			    ;; A in all clauses by A-value           
			    (setq C* (subst A-value A rest :test 'equal)))
			 ((and A-value (eq A2 A))         ;29oct92
			  (cond ((equal A-value sz)
				 (setq C* (remove L C :test 'equal)) t)
				(t (setq C*(subst A-value A C :test 'equal)))))
			 (t (setq C* (remove L C :test 'equal)) t) )))
		 C)
	   (cond ((null C*) (return-from collapse-cc '([]))))
	   (setq C** (subst sz A C* :test 'equal))
	   (setq C** (mapcar #'red* C**))
	   (setq C** (mapcan #'(lambda (L) (cond ((equal L '(not (true))) nil)
						 (t (list L)))) C**)) ;29oct92
	   (setq C** (factor-subsume (remove-ors C**)))
            ;;See file reduce-ve.lisp for the functions RED* and reduce-ve.
            ;;Reclausify in calling function
	   (collapse-cc C**))               ;should this have the other args?
	  (t C) )))

(defun collect-terms (sym L)                  ;24Nov92
  ;;Return the list of terms of the form (sym ...) occuring in L
  (cond ((atom L) nil)
	((eq-car L sym) (list L))
	(t (union-w (collect-terms sym (car L)) (collect-terms sym (cdr L))))))

(defun convert-for-strive (thm)               ;18Nov92 28Nov92
  ;;convert each atom (P x ... y) to (< (ff x ... y) 0)
  (let ()
    (cond ((atom thm) thm)
	  ((null (cdr thm)) thm)
	  ((member (car thm) '(all some & or -> not))
	   (cons (car thm) (mapcar #'convert-for-strive (cdr thm))))
	  ((member (car thm) '(< <=)) thm)
	  (t (list '< (cons (car thm) (cdr thm)) 0)) )))

(defun cross-product (SS)   ;15Jan93 called by SPLIT-S
  ;;eg, (cross-product '( ((A)) ((R) (B)) ((R) (D)) ((E))))
  ;;    =  ( ((A)(R)(E)) ((A)(B)(R)(E)) ((A)(B)(R)(E)) ((A)(B)(D)(E)) )
  (let (CC SxS SxS*)
    (cond ((null SS) nil)
	  ;;((or (prnt SS) (break) nil))
	  ((and (setq CC (car SS)) nil))
	  ((null (cdr SS)) (mapcar #'list CC))
	  (t (setq SxS (cross-product (cdr SS)))
	     (setq SxS*(mapcan #'(lambda (S) (mapcar #'(lambda (C) 
						(crunch (cons C S)))
						     CC)) 
			       SxS))) )))

(defun descendants (NN)             ;18Jan93 Call by remove-descendants-cr=cr
  ;;NN is a list of clause numbers in *CL*.
  ;; Return the numbers of their descendants
  (let (CD)
    (cond ((setq CD (crunch (mapcan #'(lambda (C) (mapcan #'(lambda (n) 
               (cond ((or (equal n (caadr  (caddr C))) 
			  (equal n (caaddr (caddr C))))
		      (list (car C)))))
							  NN))
				    *CL*)))
	   (union-w CD (descendants CD)))
	  (t nil))))

(defun crunch-cadr (CL)                   ;10Nov92 Called by SET-VAR
  ;;Remove clauses with duplicate Cadr's
  (cond ((null CL) nil)
	((null (cdr CL)) CL)
	((some #'(lambda (C) (equal (cadar CL) (cadr C))) (cdr CL))
	 (crunch-cadr (cdr CL)))
	(t (cons (car CL) (crunch-cadr (cdr CL)))) ))

(defun eligible-for-collapse (C)          ;23Sept92 6oct92 Called by COLLAPSE1
  ;;C is an ordinary (not numbered) clause
  ;;To be eligible, the first literal of C is a (not (subset A (set-of ...)))
  ;;  literal, and each literal L of the rest of C must either not contain A,
  ;;  or must be a not-element-of-A literals. (See NOT-ELEMENT-OF-A for 
  ;;  description).
  ;;29oct92.  Also each free variable in (set-of ...) must NOT be tied to
  ;;  the variable A.  
  (let (L1 B A C2 A-value)
    (or (null C) 
	(and (some #'(lambda (L) 
             (and (eq-car L 'not) 
		  (eq-car (setq B (cadr L)) 'subset) ;B=(subset A (set-of ...))
		  (atom (setq A (cadr B))) (eq A *A*)
		  (eq-car (caddr B) 'set-of)
		  (setq L1 L) (setq A-value (caddr B)))) ;A-value= (set-of ...)
		   C)
	     (or (setq C2 (remove L1 C :test 'equal)) t)
	     (every #'(lambda (L) (or (not (occur A L))
				      (not-element-of-A A L))) C2)
	     (not (tied-to A (remove 'z% (free-variables-in A-value)) 
			   (remove L1 C :test 'equal)))
	     (setq *A-value* A-value) ))))

(defun existentially-quantify (C  &optional (V (free-variables-in C)))
  ;;2 Oct 92  Called by NEGATE-&-REQUANTIFY
  (cond ((null V) C)
	(t (list 'some (car V) (existentially-quantify C (cdr V))))))

(defun exis-quan (Pz V)                        ;15Sept92 Called by RULE1o
  ;;existentially quantify Pz by all variables in V
  ;;Eg, (exis-quan '(P x y z) '(x z w)) =
  ;;    (some x (some z (P x y z)))
  (let (x)
    (cond ((null V) Pz)
	  ((not (occur (setq x (car V)) Pz)) (exis-quan Pz (cdr V)))
	  (t (list 'some x (exis-quan Pz (cdr V)))) )))

(defun find-pairs (CL)                        ;21oct92
  ;;CL is a list of numbered clauses. This produces a list of "pairs".
  ;;21oct92: *pairs* originally was a list of pairs of clauses, 
  ;;  ( [C11 C12] [C21 C22] ... ) where Ci1 is a partner-clause to Ci2.
  ;;We have generalized this to use triples instead of pairs,
  ;;  ( [C11 L1 (C12 ... C1k1)] [C21 L2 (C22 ... C2k2)] ...) 
  ;;  except that we just use the NUMBERS of the clauses and literals instead
  ;;  of the clauses and literals themselves, in the form
  ;;  ( [n1 i1 (m11 m12 ... m1k)] [n2 i2 (m21 m22 ... m2k)] ...)
  ;;For now, we define a collection (m11 m12 ... m1k) to the numbers of
  ;;  clauses that have the same literal as does clause n1. etc.
  ;;Also we require that n1 be "near" m11, etc.
  ;;eg, (find-pairs '((1 ((A) (B)) INPUT) (2 ((A) (C)) INPUT)
  ;;                  (3 ((D) (E)) INPUT) (4 ((F X) (G X)) INPUT)
  ;;                  (5 ((H X) (F X)) INPUT) (6 ((J X) (F y) (F X)) INPUT)))
  ;;      =  ((1 1 (2)) (2 1 (1)) (4 1 (5 6)) (5 2 (4 6)) (6 3 (4 5)))
  (let (fp n C CL-near Cs pr)
    (setq fp nil)
    (mapc #'(lambda (C*)                           ;C* is in CL
      (setq n (car C*)) (setq C (cadr C*))
      (setq CL-near (mapcan #'(lambda (C2)         ;nearby C's to C*
            (cond ((and (< (- n 4) (car C2)) (< (car C2) (+ n 4))) (list C2))))
			    (remove C* CL :test 'equal)))
      (mapc #'(lambda (L)                          ;L is in C
        ;;change this to exclude SET-OF literals   ;22oct92
	(setq Cs (mapcan #'(lambda (C2) 
                  (cond ((and (member L (cadr C2) :test 'equal)    ;22oct92
			      (<= (length (intersection (cadr C2) C)) 1))
			 (list (car C2)))))
			 CL-near))      
	(cond ((null Cs) nil)
	      (t (setq pr (list n (position-in L C) Cs))
		 (setq fp (append fp (list pr))))))
	    C))
	  CL)
    fp))

(defun find-targets (guns CL)       ;31oct92
  (let (gunCs CL2 C1 L1 gs Tgs Tgsl C2 L2 lb targets)
    (cond ((null CL) (return-from find-targets nil)))
    (cond ((> (caar (last CL)) 150)                 ;14Nov92 for debugging
	   (Format t "~%Halted in find-targets for debugging~%") 
	   (prnt *pairs*) (n) (break)))
    (setq gunCs (mapcar #'car guns))              ;the C1's of guns
    (setq CL2 CL)                                        ;11Nov92
    (setq targets (mapcar #'(lambda (g)               ;g= (C1 L1) is in guns
      (setq C1 (car g)) (setq L1 (cadr g)) (setq gs (car (top-skolem L1)))
      (setq Tgs (mapcan #'(lambda (C2) (mapcan #'(lambda (L2)     ;targets
                (cond ((and (not (equal C2 C1)) (occur gs L2)
			    (not (and (eq-car L2 'element) (atom (caddr L2))))
			    (not (occur 'set-of L2)))
		       (list (list C2 L2)))))
					       (cadr C2)))
			CL2)) 
      (setq Tgsl (mapcan #'(lambda (tg)                        ;labled-targets
                   (setq C2 (car tg)) (setq L2 (cadr tg))
		   (setq lb (label-targets C1 C2 L1 L2 gs))
		   (cond ((eq '[] lb) (return-from find-targets '[])))
		   ;;lb= ('legit C2 L2 R) or ('blocked C2 L2 L2S) or nil
		   (cond (lb (list lb))))
			 Tgs))
      (list C1 L1 Tgsl))
			  guns))
    (prnt-c CL 'find-targets)
    (prnt-t targets 'find-targets-results t)
    targets))

(defun finish-by-strive (CL)              ;15Jan93 Called by A-GUNS-SV
  ;;This version uses the clauses rather than the original theorem 13Jan93
  ;;Try to find [] by the STR+VE prover.  But first,
  ;; Remove ALL parents of any Rule1 or Rule11 literals, (not yet)
  ;;    and all parents of *set-theory-definitional-clauses*
  ;; Collapse ALL set-of literals
  ;; 29oct92: For collapsing, use an ELIGIBLE-FOR-COLLAPSE (set-of ...)
  ;;   to replace all A's in CL. 
  ;; Subsume all possible  (for now we will assume that this is already done)
  (let (anc all CL2 CL3 CL4 Thm RR H Thm* CL* a-value V SxS)
    (format t "~%CL= ~s~%" (mapcar #'car CL))          
    (setq CL2 (subsume-out CL CL))                                ;1oct92
    (setq anc (ancesters CL))
    (setq all (remove 0 (sort (append anc (mapcar #'car CL)) '<)))  
     ;;13Jan93 For now, use "all" clauses when calling STR+VE
    (prnt all)
    (setq CL3 (mapcar #'C all))
;;    (setq CL3 (subsume-out CL3 CL3))                              ;13Jan93
    (setq CL3 (remove-ancesters-c CL3 'collapse))                 ;14Jan93
    (format t "~%CL3= ~s~%" (mapcar #'car CL3))

    (setq CL2 (remove-ancesters CL2))     ;for finding a-value only 13Jan93
    (cond ((setq a-value (select-an-a-value-u CL2))               ;6Jan93
	   (setq CL4 (collapse-with-value CL3 *A* A-value t)))    ;13Jan93
	  ((setq a-value (select-an-a-value CL2))                 ;24Nov92
	   (setq CL4 (collapse-with-value CL3 *A* A-value)))      ;13Jan93
	  ((not (occur 'set-of CL2)) 
	   (setq CL4 CL2))
	  (t (return-from finish-by-strive nil)))                 ;6Jan93
    (prnt A-value)
    (format t "~%CL4= ~s~%" (mapcar #'car CL4))
    (setq CL4 (red* CL4))
    (setq SxS (split-s (mapcar #'cadr CL4)))                      ;15Jan93
    (cond ((every #'(lambda (S) 
(prnt S) 
      (setq Thm (negate-&-requantify S))
      (setq Thm (rewrite-subset-set-of thm))        ;5oct92
      (setq thm* (remove-constant-parentheses Thm))
      (setq thm* (universally-quantify thm*))       ;12oct92

      (setq Thm* (unique-ize-quan-vars Thm*))
      (setq Thm* (subst 'z+ 'z% thm*))
      (setq Thm* (subst '<= '<<= thm*))             ;20Nov92  For STR+VE
      (setq Thm* (convert-for-strive Thm*))         ;20Nov92
      ;;(print Thm*)

      (setq RR (user::prove thm* 40 400))                              ;24Nov92
      (cond ((equal ':TIME-LIMIT RR) 
	     (format t "~%:TIME-LIMIT~%") nil)
	    (RR
	     (format t "~%        Clauses Actually Involved In Deriving []~%")
	     (setq CL* (ancestry-multiple *CL* CL3)) ;2oct92 finish this
	     (format t "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx")
	     (number-print (crunch CL*)) 
	     '[])
	    (t nil))) SxS)
	   '[])
	  (t nil) )))

(defun fire (targets)              ;31oct92  20Nov92
  ;;targets is a list, each of the form (C1 L1 Tgsl), where
  ;;   Tgsl is a list, each of the form ('legit C2 L2 R) or('blocked C2 L2 L2S)
  ;;   where R is an ordinary resolvent and L2S is a subset of (cadr C2),
  ;;   each literal of L2S is a "blocking" literal.
  ;;Here we fire ONLY the LEGITimate ones.
  (let (C1 L1 Tgs1 gs C2 L2 R RR RR* R4 Rc Rc* RR2 RR4)
    (setq RR2 (mapcan #'(lambda (Tg)                         ;Tg is in Targets
       (setq C1 (car Tg)) (setq L1 (cadr Tg)) (setq gs (car (top-skolem L1)))
       (setq Tgsl (caddr Tg))
       (setq Tgsl (mapcan #'(lambda (tg)           ;just the Legit ones
              (cond ((eq-car tg 'legit) (list tg)))) Tgsl))
       (mapcan #'(lambda (u)                       ;u is in Tgsl
         (setq C2 (cadr u)) (setq L2 (caddr u)) (setq R (cadddr u))
	 (setq RR (reclausify R))
	 (setq RR* (mapcar #'(lambda (R4) (new-parentage 'r1 R4 C1 L1 C2 L2))
			   RR))
	 (setq *CL* (append *CL* RR*))
	 (cond ((eq '[] R) (return-from fire '[])))
	 (setq RR4 (mapcan #'(lambda (R4*)
		  (setq R4 (cadr R4*))
                  (cond ((and (eligible-for-collapse R4)
			      (not (equal R (setq Rc (collapse-cc R4)))))
			 (setq Rc* (new-parentage 'res Rc R4* 'L0 'R0 1))
			 (setq *CL* (append *CL* (list Rc*)))
			 (cond ((eq '[] Rc) (return-from fire '[])))
			 (list Rc*))
			(t (list R4*))))
		 RR*))
	 (cons C2 RR4))
	       Tgsl))
		      Targets))
    (setq RR2 (sort (crunch RR2) 'clause-number<))
    (update-pairs)                            ;14Nov92
    (setq fire-R*s (mapcar #'car RR2))
    (prnt fire-R*s)
    RR2
    ))

(defun fire11 (guns CL)               ;5sep92 Called by A-GUNS-SV11 12Nov92
  ;;GUNS is a set of pairs (C* L) where C* is a numbered clause in CL
  ;; and L is a type R11 literal in C*.
  ;;This applies Rule11 to each such pair, getting a new numbered clause R*,
  ;; deletes C* from CL, and adds R* to CL. (Also adds R* to *CL*)
  (let (pair C* L C R R* CL2 guns11)
    (cond ((null guns) CL)                    ;finished
	  ((and (setq pair (car guns)) (setq C* (car pair)) 
		(setq L (cadr pair)) (setq C (cadr C*)) nil))    ;name things
	  ((eq '[] (setq R (rule11 C L))) '[])                   ;12Nov92  []
	  (R
	   (setq R* (new-parentage 'r11 R C* L))      ;numbered resolvent
	   (setq CL2 (remove C* CL :test 'equal))     ;delete C* from CL
	   (setq CL2 (append CL2 (list R*)))          ;add R* to CL
	   (setq *CL* (append *CL* (list R*)))        ;add R* to *CL*
	   (setq guns11 (sv-guns (list R*) 'r11))     ;any new guns in R*?
	   (fire11 (append guns11 (cdr guns)) CL2))  ;add them to GUNS
	   
	  (t (format t "~%ALARM in FIRE11, Rule11 failed~%")
	     (fire11 (cdr guns) CL)) )))

(defun f-remove-parents (CL &optional (n-top (caar (last CL))))  ;17oct92 
  ;;10Nov92 19Nov92 Called by FINISH-BY-STRIVE
  ;;Remove from CL those clauses PRNT which are parents of a resolvent R in*CL*
  ;; if either Rule1 or Rule11 was used to produce R, or the other parent of R
  ;;  is in *set-theory-definitional-clauses* (ie, from definitional lemmas of
  ;;    subset type).
  ;;and return the other members of CL.
  (let (np anc n m L L2 C2 C3 C2S)
    (setq L (mapcar #'car *set-theory-definitional-clauses*))      ;17oct92
    (setq CL2 (mapcan #'(lambda (Prnt)              ;Prnt is a potential parent
;;(prnt prnt n-top) (break)
      (cond ((and (setq np (car Prnt))         ;np is the number of that parent
;;		  (not (equal '(input) (caddr Prnt)))
		  (some #'(lambda (R)     
		    (and (setq anc (caddr R))      ;anc= (action (n i) (m j))
			 (setq n (caadr anc)) 
			 (<= n n-top)              ;19Nov92
			 (setq m (cond ((atom (caddr anc)) nil) 
				       (t (caaddr anc))))
			 (or (equal np n) (equal np m))    ;R is Prnt's child
			 (or (and (equal n np) (member m L)) ;a set-def child
			     (and (equal m np) (member n L)) ;a set-def child
			     (eq-car anc 'r1) (eq-car anc 'r11))) )
                                                          ;a r1 or r11 child
			*CL*))
                                                   ;(setq L (cons Prnt L))
	     nil)
	    (t (list Prnt))))
	    CL))
     ;;Now drop the PARTNERS of those that were dropped.
    (setq L2 (mapcar #'car CL2))
    (setq CL3 (mapcan #'(lambda (C) 
      (setq n (car C))
      (cond ((some #'(lambda (pr)                  ;eg, pr= (4 1 (5 6))
		(and (setq m (car pr))                  ;m= 4
		     (< m n-top)                  ;19Nov92
		     (setq C2s (caddr pr))            ;C2S= (5 6)
		     (or (and (eq n m) (not (set-intersection L2 C2S)))
;;			 (and (member n C2S) (not (member m L2)))
)
))
		   *pairs*)
	     nil)
	    (t (list C)))) CL2))
    (format t "~%~s~%Remove-parents~%~s~%" (mapcar #'car CL)(mapcar #'car CL3))
    CL3))

(defun friends (C1 C)     ;15Jan93  Called by SPLIT-CLAUSE
  ;;eg, (friends '((P x)) '((Q x) (R y) (S x y) (U v))) = ((Q x) (R y) (S x y))
  (let (V C2 C2*)
    (cond ((not (setq V (free-variables-in C1))) C1)
	  ((not (setq C2 (infected V C))) C1)
	  (t (setq C2* (union-w C1 C2))
	     (friends C2* (set-difference C C2 :test 'equal))) )))

(defun generate-A-guns (CL &optional A-guns)    ;2oct92 Called by NEW-GUNS
  ;;Resolve literal of CL w/r *A* against (CL U CL)
  ;; until all sv-A-guns are generated, and return them (and put the clauses
  ;; generated into the global variable *R-clauses*) 
  ;;Any Resolvent generated is saved in *CL* and *R-clauses* (by RES-GENERATE).
  (let (CL* A Ag B)
       (setq CL* *guns-find-clauses*)                        ;10oct92
       (setq A *A*)                              ;for ease of writing 
       (mapcan #'(lambda (C) (mapcan #'(lambda (L) 
                    (cond ((and (eq-car L 'not)           ;(not (subset A tm))
				(or (and (eq-car (setq B (cadr L)) 'subset)
					 (occur A (cadr B))
					 (not (occur 'set-of B)))
				    (and (eq-car (setq B (cadr L)) 'set=)
					 (occur A B)))    ;(not (set= tm tm))
				(setq Ag (res-generate C L A CL* A-guns)))
			   (list Ag))
			  ((and (or (and (eq-car L 'subset) ;(subset tm A)
					 (occur A (caddr L))
					 (not (occur 'set-of L)))
				    (and (eq-car L 'set=)   ;(set= (f A)(g A))
					 (occur A L)))
				(setq Ag (res-generate C L A CL* A-guns)))
			   (list Ag))))
				     (cadr C))) 		 
	       CL) ))

(defun infected (V C)     ;15Jan93  Called by FRIENDS
  ;;which literals of C are "infected" by variables of V?
  (cond ((null C) nil)
	((occur-set V (car C))
	 (cons (car C) (infected V (cdr C))))
	(t (infected V (cdr C))) ))

(defun label-targets (C1 C2 L1 L2 gs)   ;31Oct92 17Nov92Called by FIND-TARGETS
  (let (R R*)
    (cond ((null (setq R (rule1 (cadr C1) (cadr C2) L1 L2)))
	   (format t "~%Label-Targets Problem~%") 
	   nil)
	  ((eq '[] R)
	   (setq R* (new-parentage 'R1 R C1 L1 C2 L2))
	   (setq *CL* (append *CL* (list R*)))
	   '[])
	  ((eq-car R 'occur-a) (list 'blocked C2 L2))    ;17Nov92
	  ((eq-car R 'tied-to) (list 'blocked C2 L2 (caddr R)))
	  (t (list 'legit C2 L2 R)))))

(defun n () (number-print *CL*))

(defun negate-&-requantify (CL)            ;1Oct92  Called by FINISH-BY-OTTER
  ;;CL is a set of clauses, equivalent to 
  ;; (& (or L11 ... L1n1) ... (or Lk1 ... Lknk)).  This changes it into
  ;; (SOME x's in CL (or (& ~L11 ... ~L1n1) ... (& ~Lk1 ... ~Lknk)))
   (orify (mapcar #'(lambda (C) 
      (existentially-quantify (andify (mapcar #'not-formula-r C))))
		     CL)))

(defun new-guns (CL)           ;31oct92 2oct92  Called by A-guns-sv
  ;;This collects the list of pairs (C L), for which L is a sv-literal 
  ;; of type R1. That is L can serve as the first part of Rule1
  ;;It also stores in *R-clauses* any new clause generated 
  ;; 
  (let (new-guns new-guns2 CL*)
    (setq new-guns (sv-guns CL 'r1))     ;2oct92
     ;;get the regular A-guns in CL
    (setq *R-clauses* nil)
    (setq CL* (mapcan #'(lambda (C) 
        (cond ((some #'(lambda (pr) (equal C (car pr))) new-guns) nil) 
	      (t (list C))))
			 CL))
    (cond ((eq '[] (setq new-guns2 (generate-A-guns CL*))) '[])
           ;;and generate any additional guns by resolution
	  (t (union-w new-guns new-guns2)))))

(defun new-parentage (rule R C1* L1 &optional C2* L2 rn)  ;called by many 
  (let (n m i j pr1 pr2 pr next-n R*)
    (setq n (car C1*))
    (setq i (cond ((eq 'R0 C2*) 1) 
		  (t (position-in L1 (cadr C1*)))))
    (setq pr1 (list n i))
    (setq m (cond ((eq 'R0 C2*) 0) (C2* (car C2*)) (t nil)))
    (setq j (cond ((eq 'R0 C2*) 1) (C2* (position-in L2 (cadr C2*))) (t nil)))
    (setq pr2 (cond (m (list m j)) (t 'single)))

    (setq pr (list rule pr1 pr2))
    (setq next-n (cond (rn rn) (t (1+ (caar (last *CL*))))))
    (setq R* (list next-n R pr))))

(defun not-element-of-A (A L &optional (P t))  ;called by ELIGIBLE-FOR-COLLAPSE
  ;;Does L either not contain A or have the form (not (element tm A)) where A 
  ;; does not occur in tm. And any descendant literal (via set-theory axioms)
  ;; has the same form.                                          6oct92
  ;;Similarly for (element tm A) if the parity P is nil.
  ;;eg, for L= (not (element x A)), (not (element x (union A))),
  ;;    (subset A tm), (subset (union A) tm), (not (subset tm (union A)))
  ;; we get (not-element-of-A L t) = T.
  (let ()
  (cond ((not (occur A L)) t)
	((eq-car L 'not) (not-element-of-a A (cadr L) (not P)))
	((and (not P) 
	      (or (and (eq-car L 'element) (not (occur A (cadr L)))
		       (eq 'pos (positive-negative (caddr L) A)))
		  (and (eq-car L 'subset)  (not (occur A (cadr L)))
		       (eq 'pos (positive-negative (caddr L) A)))
		  (and (eq-car L 'subset)  (not (occur A (caddr L)))
		       (eq 'neg (positive-negative (cadr L) A))))))
	((and P
	      (or (and (eq-car L 'subset)  (not (occur A (caddr L)))
		       (eq 'pos (positive-negative (cadr L) A)))
		  (and (eq-car L 'subset)  (not (occur A (cadr L)))
		       (eq 'neg (positive-negative (caddr L) A)))))) )))

(defun numbersp (L)                           ;called by SKO-EXPN
  (every #'numberp L))

(defun occur-instance (tm tm2)       ;called by RULE1
  ;;This should now be called "occur-unify"   10nov92
  (let (theta)
    (cond ((setq theta (unify tm2 tm)) theta)     ;NOTE NOTE 10nov92 unify
	  ((atom tm2) nil)
	  ((not (atom (car tm2))) nil)
	  ((some #'(lambda (tm3) (setq theta (occur-instance tm tm3))) 
			(cdr tm2))
	   theta))))

(defun occur-set (V A)              ;2Feb93  used in INFECTED
  (some #'(lambda (x) (occur x A)) V))

(defun partners2 (C1 C2 L1 L2)   ;17Nov92 Called by RESOLVE2-TARGETS
  ;;Are C1 and C2 partners on literal L1 and L2?
  ;;eg, (partners2 (C 33) (C 34) '(<<= (LS A) Y) '(<<= (LS A) Y)) = t
  (let (m n k)
    (and (equal L1 L2) (setq m (car C1)) (setq n (car C2))
	 (setq k (position-in L1 (cadr C1))) 
	 (some #'(lambda (pr) (and (eq (car pr) m) (eq (cadr pr) k)
				   (member n (caddr pr))))
	       *pairs*))))

(defun positive-negative (B A)             ;6oct92 Called by NOT-ELEMENT-OF-A
  (cond ((not (occur A B)) nil)
	((eq B A) 'pos)
	((atom B) nil)
	((member (car B) '(union)) (positive-negative (cadr B) A))
	((member (car B) '(intersection negation))
	 (pn-neg (positive-negative (cadr B) A))) ))

(defun pn-neg (Ans)                         ;6oct92 Called by POSITIVE-NEGATIVE
  (cond ((not Ans) nil)
	((eq 'pos ans) 'neg)
	((eq 'neg ans) 'pos)))

(defun prnt-c (CL name &optional sp)
  (let (L)
    (setq L (mapcar #'car CL)) 
    (cond (sp (format t   "~s ~s~%"name L))
	  (t  (format t "~%~s ~s~%"name L)))))

(defun prnt-t (targets name &optional sp) 
  (let (L) 
    (setq L (target-numbers targets))
    (cond (sp (format t   "~s ~s~%"name L))
	  (t  (format t "~%~s ~s~%"name L)))))



(defun put-set-of-first (R)                ;7oct92 Called by RULE1o
  (let (Lso)
    (cond ((some #'(lambda (L) (and (set-of-literal L) (setq Lso L))) R)
	   (setq R (remove Lso R :test 'equal))
	   (cons Lso R)))))

(defun quantify-select (u)              ;31Dec92 called by SELECT-AN-A-VALUE
  ;;u is a formula. Exist. quantify any free variable in u except *A* and z%.
  ;;And THEN remove constant parentheses.
  (let (V u2)
    (setq V (remove *A* (remove 'z% (remove 'z+ (free-variables-in u)))))
    (setq u* (existentially-quantify u V))            
    (setq u* (remove-constant-parentheses u*))))

(defun reclausify (C)    ;29oct92 Called by COLLAPSE-WITH-VALUE, COLLAPSE1
  ;;eg, (reclausify '(e (& a b c) f)) 
  ;; =  ((e a f) (e b f) (e c f))
  (let (L* L1 L2 C1 C2)
    (cond ((some #'(lambda (L) (and (eq-car L '&) (setq L* L))) C)
	   (setq L1 (cadr L*))                 ;L*= (& a b c d) L1= a
	   (setq L2 (andify (cddr L*)))        ;L2= (& b c d)
	   (setq C1 (subst L1 L* C :test 'equal))
	   (setq C2 (subst L2 L* C :test 'equal))
	   (union-w (reclausify C1) (reclausify C2)))
	  (t (list C)))))

(defun red-subset (CL)                        ;21Sept92 Called by SV
  ;;CL is a list of numbered clauses.  If A is *A*, then, in CL, change
  ;; (not (subset A D)) to (not (subset A (set-of z% (element z% D))))
  ;;Eg, (setq *A* 'A)  
  ;; (setq CL '((1 ((P x) (not (subset A (D))) (Q x))) (2  (((Q y))))))
  (mapcar #'red-subset-c CL))

(defun red-subset-c (C) 
  (let (C2 C* L L*)
    (cond ((equal (cadr C) (setq C2 (mapcar #'red-subset-L (cadr C)))) C)
	  (t (some #'(lambda (x y) (and (not (equal x y)) (setq L x)
					                  (setq L* y)))
		   (cadr C) C2)
	     (setq C2 (cons L* (remove L* C2 :test 'equal))) ;put L* first
	     (setq C* (new-parentage 'red C2 C L))
	     (setq *CL* (append *CL* (list C*)))
	     (setq *th-clauses* (subst C* C *th-clauses* :test 'equal))
	     C*))))

(defun red-subset-L (L) 
  (let (B D Ds)  
    (cond ((and (eq-car L 'not) (eq-car (setq B (cadr L)) 'subset)
		(eq (cadr B) *A*)	;L= (not (subset A D))
		(not (eq-car (setq D (caddr B)) 'set-of)))
	   (setq Ds (list 'set-of 'z% (list 'element 'z% D)))
;; (prnt B D Ds) (break)
	   (subst Ds D L :test 'equal))
	  (t L) )))
(defun remove-constant-parentheses (C)   ;2Oct92 Called by FINISH-BY-RESOLUTION
  ;;eg, ((f (a))) goes to ((f a))             ;C is a clause
  (mapcar #'rem-c2 C))

(defun rem-c2 (L) 
  (cond ((atom L) L)
	((not (atom (car L))) (format t "~%Alarm in rem-c2~%") L)
	((eq-car L 'opens) L)                     ;12oct92 Dont change (OPENS)
	((and (null (cdr L)) (atom (car L))) (car L))
	(t (cons (car L) (mapcar #'rem-c2 (cdr L))))))

(defun remove-all-parents (CL)                 ;not now used 28Nov92
  (let (LC LR LC*)
    (setq LC (mapcar #'car CL))
    (mapc #'(lambda (C) 
	 (cond ((equal '(input) (setq anc (caddr C))) nil)
                                                ;anc=(action (m i) (n j))
	       (t (setq LR (cons (caadr anc) LR))
		  (cond ((atom (caddr anc)) nil)
			(t (setq LR (cons (car (caddr anc)) LR)))))))
	  CL)
    (setq LR (sort (crunch LR) '<))
    (setq LC* (set-difference LC LR))
    (format t "~%~s~%Remove-all-parents~%~s~%" LC LC*)
    (mapcar #'C LC*) ))

(defun remove-ancesters (CL)        ;28Nov92
  ;;Called by A-GUNS-SV,FINISH-BY-STRIVE
  (let (LC ancesters LC*)
    (setq LC (mapcar #'car CL))
    (setq ancesters (ancesters CL))
    (setq LC* (sort (set-difference LC ancesters) '<))
    (format t "~%~s~%Remove-Ancesters~%~s~%" LC LC*)
    (mapcar #'C LC*)))

(defun remove-ancesters-c (CL &optional col)   ;13Jan93  col means collapse
  ;;Called by FINISH-BY-STRIVE
  (let (LC LC* C anc n1 n2 C1 C2 L1 L2 C1* C2*)
    (setq LC (mapcar #'car CL))
    (setq LC* LC)
    (mapc #'(lambda (n) 
	 (and col (setq C (C n)) (setq anc (caddr C)) ;anc= (Rule (n1 i)(m1 j))
	      (eq 'R1 (car anc)) 
	      (setq n1 (caadr anc)) (setq n2 (caaddr anc))
	      (setq C1 (C n1)) (setq C2 (C n2)) 
	      (setq L1 (nth (1- (cadadr anc))       (cadr C1)))
	      (setq L2 (nth (1- (cadr (caddr anc))) (cadr C2)))
	      (setq C1* (remove L1 (cadr C1) :test 'equal))
	      (setq C2* (remove L2 (cadr C2) :test 'equal))
	      (cond ((equal C1* C2*)
		     (setq LC* (remove n1 (remove n2 LC*))))
		    ((subsume-c C1* C2*)
		     (setq LC* (remove n2 LC*)))
		    (t nil))))
	  LC)
    (setq LC* (sort LC* '<))
    (format t "~%~s~%Remove-Ancesters-c~%~s~%" LC LC*)
    (mapcar #'C LC*)))

(defun remove-descendants-cr=cr (CL)   ;18Jan93
  ;;For each clause R of CL, if is R is a R1 child of C1 and C2, with Cr=Cr',
  ;; then remove from CL all descendants of C1 and C2.  
  (let (CL* anc m n i j C1 C2 L1 L2 CDn CD)
    (setq CL* CL) 
    (mapc #'(lambda (R) 
;;(prnt R) (break)
       (cond ((and (setq anc (caddr R))	              ;anc= (rule (m i) (n j))
		   (eq 'R1 (car anc))
		   (setq m (caadr anc)) (setq n (caaddr anc))
		   (setq i (cadadr anc))(setq j (cadr (caddr anc)))
		   (setq C1 (cadr (C m))) (setq C2 (cadr (C n)))
		   (setq L1 (nth (1- i) C1)) (setq L2 (nth (1- j) C2))
		   (setq C1 (remove L1 C1 :test 'equal))
		   (setq C2 (remove L2 C2 :test 'equal))
		   (equal C1 C2))                     ;ie, cr=cr'
	      (setq CDn (remove (car R) (descendants (list m n))))
	      (setq CD (mapcar #'C CDn))
	      (setq CL* (set-difference CL* CD :test 'equal)))
	     (t nil)))
	  CL)
    CL*))

(defun replace-A (A a-value theorem &optional (P t))   ;24Nov92    P is parity
  ;;This replaces the set-variable A by a-value in theorem, and
  ;; since a-value has a set-of structure, reduces it appropriately
  ;;Eg, (replace-A 'A '(set-of z% (& (P z%) (Q z%))) 
  ;;                  '(some A (-> (element t A) (subset A B))))
  ;; =  (-> (& (P t) (Q t)) (ALL x (-> (& (P x) (Q x)) (element x B))))
  (let (RA RA*)
    (cond ((atom theorem) theorem)
	  ((null a-value) theorem)
	  ((not (occur A theorem)) theorem)
	  ((and P (eq-car theorem 'some) (eq A (cadr theorem))) ;(some A (R A))
	   (setq RA (subst a-value A (caddr theorem) :test 'equal))
	   (setq RA* (red* RA)))
	  ((and (not P) (eq-car theorem 'all) (eq A (cadr theorem))) ;(all A .)
	   (setq RA (subst a-value A (caddr theorem) :test 'equal))
	   (setq RA* (red* RA)))
	  ((eq-car theorem 'not) 
	   (list 'not (replace-A A a-value (cadr theorem) (not P))))
	  ((eq-car theorem '->)
	   (imp-on (replace-A A a-value (cadr  theorem) (not P))
		   (replace-A A a-value (caddr theorem)      P)))
	  ((member (car theorem) '(& or all some))
	   (cons (car theorem) (mapcar #'(lambda (u) (replace-A A a-value u P))
				       (cdr theorem))))
	  (t theorem))))

(defun res-generate (C L A CL A-guns)  ;16Sep92 Called by GENERATE-A-GUNS 
  ;;9Sep92 Called by GENERATE-A-GUNS and by RES-GENERATE 15Sep92
  ;;L is a literal of C. (C L) is a potential A-gun.
  ;;Resolve C upon L against a Clause of CL until an A-gun (R* L3) is obtained.
  ;;Any Resolvent generated is saved in *CL* and *R-clauses*
  (let (R R* Ag C22 L22 B)
    (cond ((some #'(lambda (C2) (some #'(lambda (L2)    
;;(prnt C L C2 L2) (break)
         (cond ((and (not (occur 'set-of L))                      ;25Sep92
		     (setq R (res-sv (cadr C) (cadr C2) L L2))
		     (setq C22 C2) (setq L22 L2) 
                      ;;for use below (symmetric-resolvents)
		     (setq R* (new-parentage 'res R C L C2 L2))
		     (setq *CL* (append *CL* (list R*)))
		     (setq *R-clauses* (append *R-clauses* (list R*)))
		     (cond ((eq '[] R) (return-from res-generate '[])) (t t))
		     (or (some #'(lambda (L3) 
			   (and (sv-literal L3 'r1) 
				(not (member (setq Ag(list R* L3)) A-guns))
				))
			       R)
			 (some #'(lambda (L3) 
                           (or (and (eq-car L 'not) 
				    (or (and (eq-car (setq B (cadr L)) 'subset)
					     (occur A (cadr B))
					     (not (occur 'set-of B)))
					(and (eq-car (setq B (cadr L)) 'set=)
					     (occur A B)))
				    (setq Ag (res-generate R* L3 A CL A-guns))
				    )
			       (and (or (and (eq-car L 'subset)
					     (occur A (caddr L))
					     (not (occur 'set-of L)))
					(and (eq-car L 'set=)
					     (occur A L)))
				    (setq Ag (res-generate R* L3 A CL A-guns))
				    ) ))
			       R))) t)
	       (t        
		  (and R

		       (setq *CL* (remove (car (last *CL*)) *CL* :test 'equal))
		       (setq *R-clauses* (reverse (cdr (reverse *R-clauses*))))
		  nil))))
				      (cadr C2)))
		 CL)
	   (symmetric-resolvents C L C22 L22 R*) 
            ;;Add any symmetric resolvents to *CL* and *R-clauses*.
	   Ag)
	  (t nil))))

(defun res-sv (C1 C2 L1 L2)   ;called by Resolve1, Resolve2
 ;;The bulk of this is taken from resolution.lisp
 (let (theta R)
;;(prnt c1 c2 L1 L2) (break)

 (cond ((not (setq theta (unify-r (not-formula-r L2) L1)))
	(return-from res-sv nil)))
 (setq R (union-w (remove L1 C1 :test 'equal)
		  (remove L2 C2 :test 'equal))) 
 (setq R (crunch (apply-sub theta R)))
 (setq R (factor-subsume (remove-ors (mapcar #'red* R)))) 
 (cond ((equal '((false)) R) (return-from res-sv '[])))
 (cond ((and (null (cdr R))                     ;unit clause
	     (set-of-literal (car R)))
	(return-from res-sv '[])))
 (cond ((null R) (return-from res-sv '[])))
 (cond ((tautology R) ;(format t "~%R is a tautology ~s ~%" R) 
	(return-from res-sv nil)))
 (cond ((some #'(lambda (C*) (set-instance R (cadr C*))) *CLL*)   ;27Aug92
	(return-from res-sv nil)))
;;do subsumption  and  factors later
 R))

(defun resolve1-targets (C1 C2 L1 L2 CL)  ;17Nov92 Called by UNLOCK
  ;;Resolve C2 upon L2 against all clauses C3 of CL upon L3, getting
  ;; numbered resolvents R3 and return all gun-targets of the kind 
  ;;    (C1 L1 ((legit R2* L3* R3)) (blocked R2* L3* L3S*) ...).
  ;;  (We may have 0 or more of each of legit's and blocked).
  ;;eg, (resolve1-targets (C 33) (C 34) '(ELEMENT (ZS Y A) A)
  ;;       '(NOT (<<= (ZS Y A) Y)) (list (c 1) (c 2) (c 3) (c 4)))
  (let (RR guns targs)
    (cond ((not (setq RR (resolve1 C2 L2 CL))) nil)
	  (t (setq guns (list (list C1 L1)))
	     (setq targs (find-targets guns RR))
	     (subsume-targets targs)))))

(defun resolve2-targets (C1 C2 L1 L2 CL L2S)  ;17Nov92 Called by UNLOCK
  ;;For each L2* in L2S,
  ;; IF C1 and C2 are PARTNERS on L2*, ie, (partners2 C1 C2 L2* L2*), 
  ;;  Resolve C1 and C2 upon L2* against each clause C3 of CL upon L3,
  ;;  getting (numbered) resolvents R1*,R2*,
  ;;  and return all gun-targets of the kind
  ;;    (R1* L1 ((legit R2* L2 R3)) (blocked R2* L2 L2S*) ...).
  ;;  (We may have 0 or more of each of legit's and blocked).
  ;; IF C1 and C2 are NOT partners, then Resolve C2 upon L2* against 
  ;;  each clause C3 in CL upon L3, getting (numbered) resolvents R3
  ;;  and return all gun-targets of the kind
  ;;    (C1 L1 ((legit R2* L3* R3)) (blocked R2* L3* L3S*) ...).
  ;;  (We may have 0 or more of each of legit's and blocked). Eg,
  ;; (resolve2-targets (C 33) (C 34) '(ELEMENT (ZS Y A) A)
  ;;                                 '(NOT (<<= (ZS Y A) Y))
  ;;                                  (list (c 1) (c 2) (c 3) (c 4))
  ;;                                 '((<<= (LS A) Y)) )
  (let (targs RR1 RR2 R1 R2 guns CL* L1* L2*)
    (setq targs (mapcan #'(lambda (L2*)               ;L2* is in L2S
       (cond ((partners2 C1 C2 L2* L2*)
	      (setq RR2 (Resolve2 C1 C2 L2* L2* CL L1))
	      (mapcan #'(lambda (W)    ;W= (R1* R2* L1* L2*) in RR2
		  (setq R1 (car W)) (setq R2 (cadr W)) 
		  (setq L1* (caddr W))
		  (setq guns (list (list R1 L1*)))
		  (setq CL* (list R2))
		  (setq targs (find-targets guns CL*))
		  (subsume-targets targs))
		      RR2))
	     (t (setq RR1 (resolve1 C2 L2* CL))
		(setq guns (list (list C1 L1)))
		(setq targs (find-targets guns RR1))
		(subsume-targets targs)) ))
			L2S))))

(defun resolve1 (C2 L2 CL)       ;17Nov92 Called by RESOLVE1-TARGETS
  ;;Perform one round of resolvents on C2 upon L2 against all clauses
  ;; of CL, and return the list of generated (numbered) resolvents.
  ;;eg, (resolve1 (C 34) '(<<= (LS A) Y) (list (c 1) (c 2) (c 3) (c 4)))
  (let (RR R R*)
    (setq RR (mapcan #'(lambda (C3) (mapcan #'(lambda (L3)
       (cond ((not (setq R (res-sv (cadr C2) (cadr C3) L2 L3))) nil)
	     ((and (setq R* (new-parentage 'res R C2 L2 C3 L3))
		   (setq *CL* (append *CL* (list R*)))
		   (setq *R-clauses* (append *R-clauses* (list R*)))
		   (symmetric-resolvents C2 L2 C3 L3 R*) nil))
	     ((eq '[] R) (return-from resolve1 '[]))
	     (t (list R*))))                
					    (cadr C3)))
		     CL))
    (subsume-out RR RR) ))

(defun resolve2 (C1 C2 L2* L2* CL L1)       ;17Nov92 Called by RESOLVE2-TARGETS
  ;;C1 and C2 are partners. Ie, (partners2 C1 C2 L2* L2*).
  ;;Perform one round of resolvents on C1 and C2 upon L2*
  ;; against all clauses C3 in CL, getting (numbered) partner-resolvents 
  ;; R1*,R2*, and return the list of triples (R1* R2* L1*).
  ;; Where L1* is L1 theta, where theta unifies L2* and ~L3.
  ;;eg, (resolve2 (C 33) (C 34) '(<<= (LS A) Y) '(<<= (LS A) Y) 
  ;;                           (list (c 1) (c 2) (c 3) (c 4)))
  (let (RR2 R1 R2 R1* R2* prs m n i j L1*)
    (setq RR2 (mapcan #'(lambda (C3) (mapcan #'(lambda (L3)
       (cond ((not (setq R1 (res-sv (cadr C1) (cadr C3) L2* L3))) nil)
             ((not (setq R2 (res-sv (cadr C2) (cadr C3) L2* L3))) nil)
	     ((and (setq R1* (new-parentage 'res R1 C1 L2* C3 L3))
		   (setq *CL* (append *CL* (list R1*)))
		   (setq R2* (new-parentage 'res R2 C2 L2* C3 L3))
		   (setq *CL* (append *CL* (list R2*)))
		   (setq *R-clauses* (append *R-clauses* (list R1* R2*)))
		   (setq m (car R1*)) (setq n (car R2*))
		   (setq i (position-in L2* (cadr C1)))
		   (setq j (position-in L2* (cadr C2)))
		   (setq prs (list (List m i (list m))(List n j (list n))))
		   (setq *pairs* (append *pairs* prs)) 
		   (some #'(lambda (L11) 
                     (and (instance L11 L1) (setq L1* L11))) R1)
		   nil))
	     ((eq '[] R1) (return-from resolve1 '[]))
	     ((eq '[] R2) (return-from resolve1 '[]))
	     (t (list (list R1* R2* L1*))) ))
					     (cadr C3)))
		      CL)) ))

(defun rewrite-subset-set-of (A &optional (V (free-variables-in A)))
  ;;5oct92 Called by FINISH-BY-OTTER
  ;;This replaces a subformula (subset D (set-of z% (P z%))) of A by
  ;; (all x (-> (element x D) (P x)))
  (let (B E D x px pD AP)

    (cond ((atom A) A)
	  ((and (eq-car A 'subset) (eq-car (setq B (caddr A)) 'set-of))
	   (setq x (new-sym-h 'x V))
	   (setq px (subst x (cadr B) (caddr B)))
	   (setq px (rewrite-subset-set-of px))
	   (list 'all x (list '-> (list 'element x (cadr A)) px)))
	  ((and (eq-car A 'subset) (eq-car (setq B (caddr A)) 'union)
		(eq-car (setq E (cadr B)) 'set-of))
	    ;;A= (subset tm (union (set-of z% (P z%))))
            ;;to (all x (-> (e x tm) (some D (& (P D) (e x D)))))
	   (setq D (new-sym-h 'D V))
	   (setq x (new-sym-h 'x V))
	   (setq pD (subst D (cadr E) (caddr E)))      
	   (setq AP (list 'some D (and-on PD (list 'element x D))))
	   (setq AP (rewrite-subset-set-of AP))
	   (list 'all x (list '-> (list 'element x (cadr A)) AP)))
	  ((and (eq-car A 'element) (equal '(opens) (caddr A))) ;x e (opens)
	   (list 'open (cadr A)))                             ;open x  20oct92
	  ((not (atom (car A))) (format t "~%Alarm in rewrite-s...~%") nil)
;;((break))
	  (t (cons (car A) (mapcar #'(lambda (u) (rewrite-subset-set-of u V))
				   (cdr A)))) )))

(defun Rule1 (C1 C2 L1 L2)            ;called by SV7  10Nov92 17Nov92 changed
  ;;This is like the Rule1 function in hand.set-var but different.
  ;;A call to  (rule1 C1 C2 L1 L2) will result in the performing of Rule1 of
  ;; the SET-VAR procedure on the two clauses C1 and C2 upon literals L1 and L2
  ;; getting resolvent R.
  ;; Or nil if there is no Rule1-resolvent,
  ;; or (occur-A theta) or (tied-to theta V) if it is blocked (see below).
  ;;
  ;;For example if C1 = ((element (F (gs A s) u) A) (Cr  A  s  u))
  ;;               C2 = ((not  (P (gs A* s*) u*))   (Cr* A* s* u*))
  ;;               L1 =  (element (F (gs A s) u) A)
  ;;               L2 =  (not  (P (gs A* s*) u*))   then
  ;;  R = ((not (subset A (set-of z% (all x% (-> (= x% (F x% u)) (P x% u*))))))
  ;;       (Cr A s u) (Cr* A s u*))
  ;;
  ;;And if         C1 = ((element (gs A) A) (not (element (b) A)))
  ;;               C2 = ((not  (P (gs A*))) (not (element (b) A*)))
  ;;               L1 =  (element (gs A) A), L2 = (not  (P (gs A*))), then
  ;;  R = ((not (subset A (set-of z% (P z%)))) (not (element (b) A)))

  (let (A ts ga theta L2* C2* L2z L2z vL2z C2** Pz A0 LA0 c1A c2A R factors tt
	  tst)
;;(prnt C1 C2 L1 L2) (break)     
    (cond ((not (and (eq-car L1 'element) (atom (setq A (caddr L1))) 
		(or (eq A *A*) (not *A*)) ;is L1 an OK gun?
		;;L1= (e (F (gs A s) u) A), ga= (F(gs A s) u), ts=(gs A s)
		(setq ts (top-skolem (setq ga (cadr L1)))) (occur A ga))) nil)
	  ((not (and (or (not (occur 'set-of L2)) ;or L1 has form (subset D D)
			 (and (eq-car L1 'subset) (eq (cadr L1) (caddr L1))))
		     (setq theta (occur-instance ts L2)) ;theta= (A*/ A  s*/ s)
                      ;;Modified occur-instance is in file set-var4.lisp
		     (setq L2* (apply-sub theta L2)) ;L2*=(not(P (gs A s) u*))
		     (setq tst (apply-sub theta ts))
		     (cond ((occur A(setq L2z (subst 'z% tst L2 :test 'equal)))
			    (cond ((and (eq-car L2z 'not) 
					(eq-car (cadr L2z) 'element)
					(equal (caddr (cadr L2z)) A)) nil)
				  (t (return-from Rule1 
						  (list 'occur-a L2 theta))))) 
			   (t t)))) nil)

	  (t (setq C2* (apply-sub theta C2))
	     (setq L2z (subst 'z% ts L2* :test 'equal)) ;L2z=(not(P z% u*))
	     (setq L2z (subst 'z% (apply-sub theta ts) L2z :test 'equal))
                                              ;10nov92   L2z=(not(P z% u*))
	     (setq vL2z (remove 'z% (free-variables-in L2z)))
	     (setq vL2z (remove A vL2z))
	     (setq C2** (remove L2* C2* :test 'equal))
	     (setq C2** (mapcan #'(lambda (L)
                                      ;Remove L= (not (subset A (set-of ...)))
                   (cond ((and (eq-car L 'not) (eq-car (cadr L) 'subset)
			       (eq-car (caddr (cadr L)) 'set-of)) nil)
			 (t (list L)))) C2**))
	     (cond ((setq tt (tied-to A vL2z C2**)) 
;;		    (setq tt (append tt (list L2)))              ;;out 17Nov92 
		    (return-from rule1 (list 'tied-to L2 tt)))
		   (t t))
	     (setq Pz (negation L2z))	;(P z% u*)
	     (setq A0 (cond ((eq (car ts) (car ga)) (list 'set-of 'z% pz))
			    (t (list 'set-of 'z% (list 'all 'x%
			       (list '-> (list '= 'z% (subst 'x% ts ga))
			         		     (subst 'x% 'z% pz)))))))
	     ;;A0= (set-of z% (all x% (-> (= x% (F x% u)) (P x% u*))))
	     (setq LA0 (list 'not (list 'subset A A0))) 
	     ;;(not (subset A A0))
	     (setq c1A (remove L1  c1  :test 'equal)) ;(Cr  A s u)
	     (setq c2A (remove L2* c2* :test 'equal)) ;(Cr* A s u*)
	     (setq R (crunch (apply-sub theta (union-w c1a c2a)))) 
	     ;; R=  ((Cr A s u) (Cr* A s u*))
	     (cond ((not R)
		    (format t "~%[] found in Rule1 by collapsing LA0")
		    (format t "~%LA0= ~s~%" LA0)
		    (return-from rule1 '[])))
	     (setq R (add-LA LA0 R A t)) 
	     ;;combine LA0 with any other Set-Of-Literal in R
	     ;;((not(subset A (set-of z%(all x%(->(= x% (F x% u))(P x% u*))))))
	     ;; (Cr A s u) (Cr* A s u*))
	     (setq R (factor-subsume (remove-ors (mapcar #'red* R))))  ;3Sep92
	     (cond ((tautology-A R)	;18July90 not implemented
		    ;;this includes ordinary and set-var tautologies
		    (format t "~%R is a tautology ~s ~%" R) 
		    (return-from rule1 nil)))
	     (cond ((some #'(lambda (C*) (set-instance R (cadr C*))) *CLL*)
		    (return-from Rule1 nil)))                    ;3Sep92
	     (setq R (factor-subsume (remove-ors (mapcar #'red* R))))
	     (cond ((> (length (setq factors (factors R))) 1)
		    (format t "~%R has the factors ~s~%" factors)))
	     (setq R (put-set-of-first R))                        ;7oct92     
	     R) )))

(defun Rule11 (C1 L1)                  ;called by SV7  12Nov92
  ;;This is like the Rule11 function in hand.set-var but different
  ;;A call to  (rule1 C1 L1) will result in the performing of Rule11 of
  ;; the SET-VAR procedure on the clause C1 upon literal L1
  ;; getting resolvent R (or nil if there is no Rule11-resolvent).
  ;;For example if C1 = ((element (F (gs A s) u) A) (Cr  A  s  u))
  ;;               L1 =  (element (F (gs A s) u) A) then
  ;;  R = ((not (subset A (set-of z% (all x% (not (= x% (F x% u)))))))
  ;;       (Cr A s u))
  ;;And if         C1 = ((element (a*) A) (not (element (b) A)))
  ;;               L1 =  (element (a*) A), then
  ;;  R = ((not (subset A (set-of z%(not (= x% (a*)))))) (not (element (b) A)))
;;(prnt C1 L1) (break)

  (let (A ts ga  Pz A0 LA0 R factors)
    (cond ((and (eq-car L1 'element) (atom (setq A (caddr L1)))
		;;L1= (e (F u) A), ga= (F u)
		(setq pz 
		   (cond ((and (setq ts (top-skolem (setq ga (cadr L1))))
			       (occur A ga) NIL) ;27Aug92 with this NIL I have
			  (list 'all 'x%      ; temporarily prevented this case
				(list 'not (list '= 'z% (subst 'x% 'z% ga)))))
			 ;;(all x% (not (= z% (F x% u))))
			 ((and (setq ga (cadr L1)) 
			       (or (not (occur A ga))
				   (equal C1 (list L1))))  ;unit clause 12Nov92
                          ;;need to do the "(not (occur (car ts) CL))" case too
			  (list 'not (list '= 'z% ga)))
			 ;;(not (= z% (a)))        for second example
			 (t nil))))

	   (setq A0 (list 'set-of 'z% pz))
              ;;A0= (set-of z% (all x% (not (= z% (F x% u))))) or etc
	   (setq LA0 (list 'not (list 'subset A A0))) 
            ;;(not (subset A A0))
	   (setq R (remove L1  c1  :test 'equal)) ;(Cr A s u)
	   (cond ((not R) (return-from rule11 '[])))
	   (setq R (add-LA LA0 R A t)) 
            ;;combine LA0 with any other Set-Of-Literal in R
            ;;((not (subset A (set-of z% (all x% (not (= z% (F x% u)))))))
	    ;; (Cr A s u))
	   (setq R (factor-subsume (remove-ors (mapcar #'red* R)))) 
;;	   (setq R (factor-subsume R))
	   (cond ((tautology-A R)	;18July90 not implemented
		  ;;this includes ordinary and set-var tautologies
		  (format t "~%R is a tautology ~s ~%" R) 
		  (return-from rule11 nil)))
	   (cond ((and (not *A*)
		       (some #'(lambda (C*) (set-instance R (cadr C*))) *CLL*))
		  (return-from Rule11 nil)))                 ;3Sep92
	   (cond ((> (length (setq factors (factors R))) 1)
		  (format t "~%R has the factors ~s~%" factors)))

	   R) )))

(defun select-an-a-value (CL)            ;Called by FINISH-BY-STRIVE
  ;;From all the set-of literals of CL, of the form, 
  ;;    (not (subset A (set-of z% (P z%))))
  ;; select the one for which (P z%) is most "powerful".  Ie, for which
  ;; (P z%) -> (Q z%) for any other one with (Q z%) for (P z%).
  ;;Ie,(setq CL '((1 ((not (subset A (set-of z% (P z%)))) (L2)))
  ;;              (2 ((not (subset A (set-of z% (and (P z%) (Q z%)))) (L3))))))
  ;; (select-an-a-value CL) = (set-of z% (and (P z%) (Q z%)))
  (let (TS PS P-best thm1 thm2)
    (cond ((setq TS (mapcan #'(lambda (C) (mapcan #'(lambda (L) 
	     (collect-terms 'set-of L))
						  (cadr C)))
			    CL))
	   (setq PS (crunch (mapcar #'caddr TS)))
	   (setq PS (subst 'z+ 'z% PS))
	   (setq P-best nil)
	   (mapc #'(lambda (P)
	       (cond ((null P-best) (setq P-best P))
		     ((and (setq thm1 (imp-on P P-best))
			   (setq thm1 (Quantify-select thm1))
			   (equal '(T) (otter-prove thm1)))
		      (setq P-best P))
		     ((and (setq thm2 (imp-on P-best P))
			   (setq thm2 (Quantify-select thm2))
			   (equal '(T) (otter-prove thm2))))
		     (t (return-from select-an-a-value nil))))
		 PS)
	   (list 'set-of 'z% (subst 'z% 'z+ P-best)))
	  (t nil))))

(defun select-an-a-value-u (CL)            ;Called by FINISH-BY-STRIVE
  ;;6Jan93 Save the above version.  We will try this one for now.
  ;;From all the set-of literals of CL, of the form, 
  ;;    (not (subset A (set-of z% (P z%))))
  ;; select the one for which (P z%) is most "powerful".  Ie, for which
  ;; (6Jan93)  (P z%) -> all the others.
  ;; And return (apply-sub theta (P z%))   NOTE: eg, b/u  "Unify" version
  ;;Ie,(setq CL 
  ;; '((1 ((not (subset A (SET-OF Z% (<= Z% U)))) (L2)))
  ;;   (2 ((not (subset A (SET-OF Z% (& (<= Z% U) (<= (F Z%) 0))))) (L3)))
  ;;   (3 ((not (subset A (set-of z% (& (<= Z% U) (<= Z% (B1)))))) (L4)))))
  ;; (select-an-a-value CL) = (SET-OF Z% (& (<= Z% (B1)) (<= (F Z%) 0)))
  (let (TS PS p-rest thm1 theta P-best)
    (setq *global-QED-light* t)
    (cond ((and (setq TS (mapcan #'(lambda (C) (mapcan #'(lambda (L) 
		   (collect-terms 'set-of L))
						  (cadr C)))
			    CL))
		(setq PS (crunch (mapcar #'caddr TS)))
		(setq PS (subst 'z+ 'z% PS))
		(some #'(lambda (P) 
		   (and (setq p-rest (remove P PS :test 'equal))
			(setq thm1 (imp-on P (andify p-rest)))
			(setq thm1 (Quantify-select thm1))
			(setq theta (imply-prove thm1))  ;note, eg, b/u
			(setq P-best (apply-sub theta P))))
		      PS))
	   (list 'set-of 'z% (subst 'z% 'z+ P-best)))
	  (t nil))))

(defun set-instance (A B)  ;28Aug92  Called by RES-SV
  ;;Is some reordering of A an instance of B?
  ;;Eg, (set-instance '((P x) (Q (b))) '((Q y) (P z)))
  ;;3Sep92 Also allow commute of (& a b c) within (set-of z% (& a b c))
  (let (L* theta A* B*)
    (cond ((instance A B) t)
	  ((not (= (length A) (length B))) nil)
	  ((or (null (cdr A)) (null (cdr B))) nil)
	  ((some #'(lambda (L) 
                      (or (setq theta (instance (car A) (setq L* L)))
			  (and (set-of-equal (car A) (setq L* L)) 
			       (setq theta t))) )
		 B)
	   (setq A* (apply-sub theta (cdr A)))
	   (setq B* (apply-sub theta (remove L* B :test 'equal)))
	   (set-instance A* B*))
	  (t nil) )))

(defun set-of-equal (A B)     ;3Sep92  Called by set-instance
  ;;Determines whether A and B are equal except for the permulation
  ;; of conjuncts within a corresponding subterm of each.
  ;;Eg, (set-of-equal '(not (subset A (set-of z% (& (P z) (Q z) (R z)))))
  ;;                  '(not (subset A (set-of z% (& (Q z) (R z) (P z) ))))) = T
  ;;We may want to generalize this to set-of-instance later.
  (let (PzA PzB)
    (cond ((equal A B) t)
	  ((atom A) nil) ((atom B) nil)
	  ((not (equal (length A) (length B))) nil)
	  ((not (eq (car A) (car B))) nil)
	  ((eq-car A 'set-of)
	   (setq B (subst (cadr A) (cadr B) B))
	   (cond ((and (eq-car (setq PzA (caddr A)) '&) ;PzA= (& a b c)
		       (eq-car (setq PzB (caddr B)) '&) ;PzB= (& d e f)
		       (subset (cdr PzA) (cdr PzB))
		       (subset (cdr PzB) (cdr PzA)))
		  t)
		 (t nil)))
	  (t (every #'(lambda (u v) (set-of-equal u v)) A B)) )))

(defun set-of-literal (L)		;called by SELECT-AN-A-VALUE
  (let (B)
    (and (eq-car L 'not) (eq-car (setq B (cadr L)) 'subset)
	 (eq-car (caddr B) 'set-of))))

(defun set-var-variable (th-clauses CL)   ;10Sep92     Called by SET-VAR
  ;;Find the variable A which will be pursued as the set-variable in SV
  ;;It is located in th-clauses (clauses from the theorem itself) and
  ;; should be in a literal of the form (element tm A) (or negation),
  ;; or potientially in such a literal (ie, in a literal (subset ...),(= ...))
  ;;Eg, (set-var-variable '((5 ((NOT (SUBSET G17 (OPENS))) 
  ;;             (NOT (SET= D17 (UNION G17))) (OPEN D17)) (INPUT))) nil) = D17
  ;;For now, we will use a simple algorithm, but eventually we will need
  ;; to reach "down" by resolution to be sure that a (element tm A) is reached.
  ;; Also, we may need to find a new such variable after 1 or more old ones
  ;; have failed (in A-GUNS-SV), then we will need to use a list of the old A's
  (let (A L2)
    (cond ((some #'(lambda (C) (some #'(lambda (L) 
;;(prnt C L) (break)

             (or (and (eq-car L 'element) (atom (setq A (caddr L))))
		 (and (eq-car L 'not) (setq L2 (cadr L)) 
		      (or (and (eq-car L2 'subset) (atom (setq A (cadr L2))))
			  (and (eq-car L2 'SET=) 
			       (or (atom (setq A (cadr L2)))
				   (atom (setq A (caddr L2)))))))) )
				     (cadr C)))
		 th-clauses)
	   A) )))

(defun split-clause (C)            ;15Jan93
  ;;eg (split-clause '((P x) (Q y) (R x))) = (((P x) (R x)) ((Q y)))
  (let (C1 C2)
    (cond ((null C) nil)
	  ((and (setq C1 (friends (list (car C)) (cdr C)))
		(setq C2 (set-difference C C1 :test 'equal)) nil))
	  (C2 (cons C1 (split-clause C2)))
	  (t (list C1)) )))

(defun split-S (S)    ;15Jan93 Called by FINISH-BY-STRIVE
  ;;eg, (split-S '(((A x) (C x)) ((R y) (B z)) ((R y) (D z)) ((E x) (F x))))
  ;; =          ( ((A C) (R)         (E F)) 
  ;;              ((A C)   (B)   (D) (E F)))
  (let (SS SxS)
    (setq SS (mapcar #'split-clause S))
    (setq SxS (cross-product SS))
    (subsume-s SxS) ))

(defun subsume-s (SxS)  ;15Jan93 SxS is a family of sets of clauses
  ;;eg, (subsume-s )
  (let (SxS*)
    (setq SxS* (mapcan #'(lambda (S) 
     (cond ((some #'(lambda (S2) (and (not (equal S2 S)) (subset S2 S)))
		  SxS) nil)
	   (t (list S)))) SxS))
    (crunch SxS*) ))

(defun subsume-targets (targets)    ;11Nov92 17Nov92 Called by A-GUNS-SV etc
  ;;each member of targets has the form 
  ;; ( (a numbered clause) (a literal) (a list tgt's)) where each tgt has the 
  ;;  form (legit   (numbered clause) (literal) (a resolvent))
  ;;   or  (blocked (numbered clause) (literal) (a (blocked) resolvent))
  ;;For each target, this removes a tgt of it if its resolvent is subsumed 
  ;;  by the resolvent of another tgt (for that target).
  (let (C1 L1 tgts tgts2 RR RR2)
    (mapcar #'(lambda (target) 
		(setq C1 (car target)) (setq L1 (cadr target)) 
		(setq tgts (caddr target))
		(setq RR (mapcan #'(lambda (tgt) ;17Nov92
				     (cond ((eq-car tgt 'legit) (list (cadddr tgt)))))
				 tgts))
		(setq RR2 (subsume-out2 RR))
		(setq tgts2 (mapcan #'(lambda (tgt)
					(cond ((not (eq-car tgt 'legit)) (list tgt)) ;17Nov92
					      ((member (cadddr tgt) RR2 :test 'equal) (list tgt))))
				    tgts)) ;The subsummed set of tgt's
		(list C1 L1 tgts2))
	    targets) ))

(defun subsume-out (S-new S &optional P)       ;22Sep92  Called by A-GUNS-SV
  (let (S2)                                    ;7oct92   Used only for tracing
  (setq S2 (subsume-out* S-new S P))
  (format t "~%~s~%Subsume-out ~%~s~%" (mapcar #'car S)(mapcar #'car S2))
  S2))

(defun subsume-out* (S-new S &optional P)       ;22Sep92  Called by SUBSUME-OUT
  ;;Remove from S any clause subsumed by a clause of S-new
  ;;Slightly different than the SUBSUME-OUT in resolution.lisp
  (let (Cn c2)
    (cond ((not P) (setq S-new (crunch-cadr S-new))
	           (setq S     (crunch-cadr S    )) ))
    (cond ((null S-new) S)
	  ((and (setq Cn (car S-new)) nil))
	  ((some #'(lambda (C) (and (not (equal C Cn))
				    (subsume-c (cadr Cn) (cadr C))
				    (setq C2 C)))
		 S)
	   (subsume-out* S-new (remove C2 S :test 'equal) t))
	  (t (subsume-out* (cdr S-new) S t)) )))

(defun subsume-out2 (S)                 ;11Nov92 Called by subsume-targets
   (subsume-out2* S S))                 ;used for tracing

(defun subsume-out2* (S S*)         ;11Nov92 Called by subsume-out2
  ;;S is a list of un-numbered clauses.  Remove those subsumed by others.
  ;;This is essentially identical to subsume-out in Resolution.lisp
  (let (C2)
    (cond ((null S) S*)
	  ((some #'(lambda (C) 
             (and (not (equal C (car S)))
		  (or (subset (car S) C)    ;NOTE NOTE SUBSET 11Nov92
		      (subsume-c (car S) C))
		  (setq C2 C)))
		 S*)
	   (subsume-out2* S (remove C2 S* :test 'equal)))
	  (t (subsume-out2* (cdr S) S*)) )))

(defun sv-guns (CL rule)         ;29Aug92 12Nov92
  ;;This collects the list of airs (C L), for which L is a sv-literal 
  ;; of type RULE. 
  ;;That is L can serve as the first part (or the only part) of the
  ;; sv-var rule, RULE.
  (let (L*)
    (mapcan #'(lambda (C) 
       (cond ((eq 'r11 rule)      ;take only one R11 gun (iterate in A-GUNS-SV)
	      (cond ((some #'(lambda (L) (and (sv-literal L rule C CL)
					      (setq L* L)))
			   (cadr C))
		     (list (list C L*)))))
	     (t (mapcan #'(lambda (L)         ;take all R1 guns
                          (cond ((sv-literal L rule C CL)  ;12Nov92 C CL
				 (list (list C L)))))
			(cadr C)))))
	    CL)))

(defun sv-literal (L rule &optional C CL) ;29Aug92  Called by SV-GUNS 12Nov92 
  ;;This determines whether L is a sv-literal of type RULE              C CL
  ;;9Sep92 *A* is a global variable whose value is the set variable
  ;; that must be acted upon by Rule1.
  (let (A ga gs ts C2 CL2)
;;(break)
    (or (and (eq 'r1 rule) (eq-car L 'element) (atom (setq A (caddr L)))
	     ;;L= (e (F (gs A s) u) A), ga= (F (gs A s) u), ts=(gs A s)
	     (setq ts (top-skolem (setq ga (cadr L)))) (occur A ga)
	     (or (not *A*) (eq *A* A)))   ;if *A* is not nil then A must = *A*
	(and (not *A*) (eq 'r1P rule) (eq-car L 'not) 
	     (eq-car (cadr L) 'element) (atom (setq A (caddr (cadr L))))
	     (setq ts (top-skolem (setq ga (cadr (cadr L))))) (occur A ga))
	(and (eq 'r11 rule) (eq-car L 'element) 
	     (atom (setq A (caddr L))) 
	     (or (not (occur A (cadr L)))                      ;L= (e (F u) A)
		 (equal (cadr C) (list L))                           ;12Nov92
		 (and CL (setq C2  (subst 'funny L C  :test 'equal)) ;12Nov92
		         (setq CL2 (subst    C2  C CL :test 'equal))
			 (setq ts (top-skolem (setq ga (cadr L))))
			 (not (occur (car ts) CL2)))))
	(and (eq 'r11P rule) (eq-car L 'not) 
	     (eq-car (cadr L) 'element) (atom (setq A (caddr (cadr L))))
	     (not (occur A (cadr (cadr L)))))     ;L= (not (e (F u) A))

	)))

(defun symmetric-resolvents (C1 L1 C2 L2 R*) ;20oct92 Called by RES-GENERATE
  ;;Also called by GENERATE-TARGETS&FIRE      10nov92
  ;;A resolvent R* has been generated from C1 L1 C2 L2, and placed 
  ;; in *CL* and *R-clauses*.
  ;;If C1 has a symmetric partner, C1p in *pairs*, and L1p is the corresponding
  ;; literal to L1, then generate the "symmetric" resolvent Rp*
  ;; by calling (res-sv (cadr C1p) (cadr C2) L1p L2), and new-parentage,
  ;; and store Rp* in *CL* and *R-clauses*.
  ;;Similarly if C2 has a symmetric partner in *pairs*.
  ;;See Find-pairs for the form of *pairs*. For example, *pairs* might be 
  ;;   ((1 1 (2)) (2 1 (1)) (4 1 (5 6)) (5 2 (4 6)) (6 3 (4 5)))
  (let (Pards pard R R*p R*s)   
    (mapc #'(lambda (pr)                      ;eg, pr= (4 1 (5 6))
       (cond ((and (equal (car C1) (car pr))                        ;Found C1
		   (equal L1 (nth (1- (cadr pr)) (cadr C1))))       ; and  L1
	      (setq Pards (mapcan #'(lambda (n) 
                     (cond ((setq pard (assoc n *cl*)) (list pard))))
				  (caddr pr)))
               ;;21oct92 Should we replace *cl* here by the "current" set??
	      (setq Pards (mapcar #'(lambda (n) (assoc n *cl*)) (caddr pr)))
	      (setq R*s (mapcan #'(lambda (C1p)	         ;resolve the partners
		     (cond ((and (setq R (res-sv (cadr C1p) (cadr C2) L1 L2))
				 (not (some #'(lambda (C) (equal R (cadr C)))
					    *CL*)))
			    (setq R*p (new-parentage 'res R C1p L1 C2 L2))
			    (format t "~%Symmetric R*p= ~s~%" R*p)
			    (list R*p))))
				Pards))
	      (setq *CL* (append *CL* R*s))
	      (setq *R-clauses* (append *R-clauses* R*s)))
	      
	     ((and (equal (car C2) (car pr))                        ;Found C2
		   (equal L2 (nth (1- (cadr pr)) (cadr C2))))       ; and  L2
	      (setq Pards (mapcan #'(lambda (n) 
                     (cond ((setq pard (assoc n *cl*)) (list pard))))
				  (caddr pr)))
	      (setq R*s (mapcan #'(lambda (C2p)	         ;resolve the partners
		     (cond ((setq R (res-sv (cadr C1) (cadr C2p) L1 L2))
			    (setq R*p (new-parentage 'res R C1 L1 C2p L2))
			    (format t "~%Symmetric R*p= ~s~%" R*p)
			    (list R*p))))
				Pards))
	      (setq *CL* (append *CL* R*s))
	      (setq *R-clauses* (append *R-clauses* R*s))) ))
	   ;;are there new pairs that should be placed in *pairs*???
	   *pairs*)
    (setq *pairs* (append *pairs* (find-pairs (cons R* R*s))))  ;22oct92
    R*s))

(defun tautology-A (R &optional A)          ;19Sep92 called by Rule1, Rule1o
  ;;Ordinary tautology as well as set-of tautology (19Sep92 not done yet)
  ;;A is the set-variable being processed
  (let ()
    (cond ((tautology R) t))))
(defun tautology (R)                        ;12Nov92
  ;;Eg, (tautology '(A (b x) C (not (b x)) D)) = T.
  ;;And (tautology '((<<= (LS A) (LS A)))) T.  Similarly for <=
  (cond ((null R) nil)
	((member (negation (car R)) (cdr R) :test 'equal) t)
	((and (not (atom (car R))) (member (caar R) '(<<= <=))
	      (equal (cadar R) (caddar R))) t)
	(t (tautology (cdr R))) ))

(defun target-numbers (targets) 
  (let (n L tgs tgsn)
     (mapcar #'(lambda (tg)
       (setq n (caar tg)) (setq L (cadr tg)) (setq tgs (caddr tg))
       (setq tgsn (mapcar #'(lambda (t2)
                (list (caadr t2) (car t2)))
			  tgs))
       (list n L tgsn))
	     targets) ))

(defun tied-to (A V C &optional C* VP)  ;31oct92 Called by RULE1
  ;;eg, (tied-to 'A '(x y) '((P x z) (Q z y) (R y A))) 
  ;;                     =  ((P x z) (Q z y) (R y A))
  ;;and (tied-to 'A '(x) '((P w v) (P y z) (Q z A) (R x y) (Q xx yy)))
  ;;                     =  ((P y z) (Q z A) (R x y))
  (let (VC VL)               
    (setq VC (mapcan #'(lambda (L) 
                  (setq VL (free-variables-in L))
		  (cond ((set-intersection V (remove A VL) )
			 (setq C* (cons L C*)) VL)
			(t nil)))
		     C))
    (setq VC (crunch (union VC V)))
    (cond ((subset VC VP) 
	   (cond ((member A V) (crunch (reverse C*))) (t nil)))
	  (t (tied-to A VC C C* V)))))

(defun unblock (targets CL)		;17Nov92 Called by A-guns-sv
  ;;One round of resolution, restructuring the existing targets
  (let (targets2 C1 L1 tgsl C2 L2 L2S targs1 targs2)
    (setq targets2 (mapcan #'(lambda (tg)
	 (setq C1 (car tg)) (setq L1 (cadr tg)) (setq tgsl (caddr tg))
	 (mapcan #'(lambda (u) ;u in tgsl
	      (cond ((eq-car u 'legit) nil) ;only unblock the blocked
		    (t (setq C2 (cadr u)) (setq L2 (caddr u))
		       (setq L2S (cadddr u)) ;u= (blocked C2 L2 L2S)
		       (setq targs1 (resolve1-targets C1 C2 L1 L2 CL))
		       (setq targs2 (resolve2-targets C1 C2 L1 L2 CL L2S))
		       (union-w targs1 targs2)) ))
				       tgsl))
			   targets))
    (subsume-targets targets2)))

(defun update-pairs ()                           ;14Nov92
  (let (n CL pairs)
    (setq n (caar (last (sort *pairs* 'clause-number<))))
    (setq n (- n 6))
    (setq CL (mapcan #'(lambda (C) (cond ((> (car C) n) (list C))))
		     *CL*))
    (setq pairs (find-pairs CL))
    (setq *pairs* (append *pairs* pairs))))

#|(defun unify-A-sets (AA)  ;called by FINISH-BY-STRIVE
  ;;AA is a set of sets of the form (set-of z% (P z%))
  ;;This function attemps to find a (mgu) substitution theta such that
  ;; for some A0 in AA, (A0 theta) is a subset of (A theta) for all A in AA.
  ;; Return (A0 theta) is there is one, else nil.
  ;;Eg, (unify-A-sets '((SET-OF Z% (& (<= Z% U) (<= (F Z%) 0)))
  ;;                    (SET-OF Z% (<= Z% U))
  ;;                    (SET-OF Z% (& (<= Z% U) (<= Z% (B1))))))
  ;;    returns         (SET-OF Z% (& (<= Z% (B1)) (<= (F Z%) 0)))
  (let (PP P)
    (setq PP (crunch (mapcar #'caddr AA)))
    (setq PP (subst 'z+ 'z% PP))
    (cond ((setq P (unify-subsubsume PP))

$$  kill this ?? ))))|#

(defun universally-quantify (C  &optional (V (free-variables-in C)))
  ;;1 Oct 92  Called by NEGATE-&-REQUANTIFY
  (cond ((null V) C)
	(t (list 'all (car V) (universally-quantify C (cdr V))))))

(defun update-pairs ()			;14Nov92
  (let (n CL pairs)
    (setq n (caar (last (sort *pairs* 'clause-number<))))
    (setq n (- n 6))
    (setq CL (mapcan #'(lambda (C) (cond ((> (car C) n) (list C))))
		     *CL*))
    (setq pairs (find-pairs CL))
    (setq *pairs* (crunch (append *pairs* pairs)))))

;;------------------------------------------------------------
;;Examples from "A Maximal Method for Set Variables in Automatic
;; Theorem Proving", MI9, 1977, pp. 56-78.
;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;;17oct92 We have renumbered some of the Lemmas L1 L2 etc below, and 
;; added others, naming them Ld1 Ld2 etc.
;;These are the defintional type lemmas that refer to set theory concepts,
;; and normally are not added as hypotheses for a theorem being proved,
;; but used appropiately anyway.

(setq Ld1 '(all A (all B (-> (all xd1 (-> (element xd1 A) (element xd1 B))) 
			     (subset A B)))))
(setq Ld2 '(all A (all B (-> (subset A B)
			     (all x (-> (element x A) (element x B))) ))))
(setq Ld3 '(all A (-> (open A) (element A (OPENS)))))
(setq Ld4 '(all A (-> (element A (OPENS)) (open A))))
(setq Ld5 '(all A (all B (-> (& (subset A B) (subset B A)) (set= A B)))))
(setq Ld6 '(all A (all B (-> (set= A B) (subset A B)))))
(setq Ld7 '(all H (all x (-> (some Ad7 (& (element Ad7 H) (element x Ad7))) 
			     (element x (union H))))))
(setq Ld8 '(all H (all x (-> (element x (union H))
			     (some Ad8 (& (element Ad8 H) (element x Ad8)))))))
(setq Ld9  '(all H (all B (-> (all Dd9 (-> (element Dd9 H) (subset Dd9 B))) 
			      (subset (union H) B)))))
(setq Ld10 '(all H (all B (-> (subset (union H) B)
			      (all D (-> (element D H) (subset D B))) ))))
(setq Ld11
      '(all H (all B (-> (all xd11 (-> (element xd11 B) 
				       (some A (& (element A H) (element xd11 A)))))
			 (subset B (union H))))))
(setq Ld12
      '(all H (all B (-> (subset B (union H))
			 (all x (-> (element x B) 
				    (some Ad12 (& (element Ad12 H) 
						  (element x Ad12)))))
			 ))))
(setq Ld13 '(ALL a (ALL b (all x (-> (element x (nt a b)) 
				     (& (<<= a x) (<<= x b)))))))
(setq Ld14 '(ALL a (ALL b (all x (-> (& (<<= a x) (<<= x b))
				     (element x (nt a b)))))))

(setq *set-theory-definitional-lemmas* 
  '(Ld1 Ld2 Ld3 Ld4 Ld5 Ld6 Ld7 Ld8 Ld9 Ld10 Ld11 Ld12 Ld13 Ld14))
(setq *guns-find-lemmas* '(Ld5 Ld7 Ld9 Ld11 Ld1))

;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq Example1 '(some A (all x (-> (element x A) (<= x 0)))))	 

(setq Example2 '(-> (element D F)
		    (some G (all A (-> (element A G) 
				       (some B (& (element B F) 
						  (subset A B))))))))

(setq Example3 '(-> (P b)		;(all D (subset D D))
		    (some A (& (all x (-> (element x A) (P x))) 
			       (some y (element y A))))))

(setq Example4 '(-> (& (not (= a b)) (not (= c b)) (not (= b a)) (not (= b c)))
		    (some D (& (not (element a D)) (element b D) 
			       (not (element c D))))))

(setq Example5 '(-> (P b) (some A (or (& (all x1 (-> (element x1 A) (P x1))) 
					 (some y1 (element y1 A)))
				      (& (all x2 (-> (element x2 A) (Q x2))) 
					 (some y2 (element y2 A))) ))))

(setq Example5A '(some A (all x (all y (-> (& (element x A) (element y A))
					   (P x y))))))
(setq Example5B '(some A (all L (some y (all z (& (-> (element z A) (P z y))
						  (Q L y)))))))
(setq Example6
      '(some A (-> (& (P c) (not (= c b)))
		   (& (all x (-> (element x A) (P x)))
		      (some y (& (element y A) (not (element b A))))))))

(setq Example7 '(-> (& (P 1) (<<= 0 1) (<<= 1 2)) 
		    (some x (& (<<= 0 x) (<<= x 2) (P x)))))

;;Example8  (Regarding Open Sets)

(setq L1 '(all D (-> (some G (& (subset G OPENS) (set= D (union G)))) 
		     (open D))))

;;%%$$%% 18Jan93 temporary to test.  (Later: I'll keep this version.)
(setq L1 '(all D (-> (some G (& (all E (-> (element E G) (open E)))
				(set= D (union G)))) 
		     (open D))))
(setq L2 '(all A (all B (-> (all x (-> (element x A) (element x B))) 
			    (subset A B)))))
(setq L3 '(all A (-> (open A) (element A (OPENS)))))
(setq L4 '(all H (all x (-> (some A (& (element A H) (element x A))) 
			    (element x (union H))))))
(setq L5 '(all A (all B (-> (& (subset A B) (subset B A)) (set= A B)))))
(setq L6 '(all H (all B (-> (all D (-> (element D H) (subset D B))) 
			    (subset (union H) B)))))
(setq L7 
      '(all H (all B (-> (all x (-> (element x B) 
				    (some A (& (element A H) (element x A)))))
			 (subset B (union H))))))
(setq L8 
      '(all D (all A (-> (& (element A (OPENS)) (subset A D))
			 (element A (set-of z% (& (element z% OPENS) 
						  (subset z% D))))))))
(setq H8 L1)				;17oct92  Used
(setq Theorem8 '(all B (-> (all x (-> (element x B) ;17oct92  Used
				      (some D (& (open D)(element x D)
						 (subset D B)))))
			   (open B))))
(setq Example8 (list '-> H8 theorem8))	;not used

(setq Example5p '(-> (P (f b)) 
		     (some s (some A (& (all x (-> (element x A) (P x)))
					(element (f s) A))))))
(setq Example6p 
 '(-> (& (P a1 b) (all E (Q b E)))
      (some A (& (all g (-> (element g A) (some x (& (P g x) (P x A)))))
		 (element a1 A)))))
(setq Example7p '(some A (not (element b A))))
(setq Example8p 
 '(-> (& (< -2 0) (all u (all v (-> (< u 0) (not (= u (abs v)))))))
      (some A (& (all y (not (element (abs y) A))) (element -2 A)))))
(setq Example9p
 '(some A (& (all x (all y (-> (& (element x A) (element y A))
			       (& (<= 0 (/ x 2)) (<= (/ x 2) y) 
				  (<= y (* 2 x))))))
	     (element 0 A))))
(setq H9p1 '(not (< 0 0))) (setq H9p2 '(not (< (/ 0 2) 0)))
(setq H9p3 '(not (< (* 2 0) 0)))

;;------------------------------------------------------------
;;Example 11. IMV    6oct92 we are using << for < for now.

(setq LC1 '(all x (-> (not (<<= (f x) 0)) ;(<< 0 (f x)) 
		      (some t (& (not (<<= x t)) ;(<< t x) 
				 (all s (-> (& (not (<<= s t)) (<<= s x))
					    (not (<<= (f s) 0)))))))))
(setq LC2 '(all x (-> (not (<<= 0 (f x))) ;(<< (f x) 0) 
		      (some t (& (not (<<= t x)) ;(<< x t)
				 (all s (-> (& (not (<<= t s)) (<<= x s))
					    (not (<<= 0 (f s))))))))))
;;24Nov92 Corrected error (not (<<= 0 (f x))) instead of (not (<<= (f x) 0))

(setq LUB '(all A (-> (& (some u (all t (-> (element t A) (<<= t u))))
			 (some r (element r A)))
		      (some L (& (all x (-> (element x A) (<<= x L)))
				 (all y (-> (all z (-> (& (element z A) 
							  (<<= z l)) ;29Dec92
						       (<<= z y)))
					    (<<= l y))))))))
(setq LUB* '(all A (-> (& (some u (all t (-> (element t A) (<<= t u))))
			  (some r (element r A)))
		       (some L (& (all x (-> (element x A) (<<= x L)))
				  (all y (-> (all z (-> (element z A) 
							(<<= z y)))
					     (<<= l y))))))))  
(setq H11 (list '& LC1 LC2))

(setq Th11 '(-> (& (<<= a1 b1) (<<= (f a1) 0) (<<= 0 (f b1)))
		(some x (& (<<= (f x) 0) (<<= 0 (f x))))))
(setq Th11* '(-> (& (<<= a1 b1) (<<= (f a1) 0) (<<= 0 (f b1)))
		 (some x (& (<<= (f x) 0) (<<= 0 (f x)) 
			    (<<= a1 x) (<<= x b1)))))

(setq theorem11 (list '-> LUB Th11))
(setq theorem11* (list '-> LUB* Th11*))
(setq Example11 (list '-> H11 theorem11))
(setq Example11* (list '-> H11 theorem11*))   ;This is like JAR paper IMV
;;-----------------------------------------------------------------------------
;;Heine-Borel Theorem
(setq Theorem-HB '(ALL G (-> (subset (nt a1 b1) (union G))
			     (some H (& (subset H G) (finite H) 
					(subset (nt a1 b1) (union H)))))))
(setq Example12 (list '-> LUB Theorem-HB))

;;-----------------------------------------------------------------------------
(defun s1 () (time (set-var Example1))) (defun ss1 () (untrace) (lsv) (s1))
(defun s2 () (time (set-var Example2))) (defun ss2 () (untrace) (lsv) (s2))
(defun s3 () (time (set-var Example3))) (defun ss3 () (untrace) (lsv) (s3))
(defun s4 () (time (set-var Example4))) (defun ss4 () (untrace) (lsv) (s4))
(defun s5 () (time (set-var Example5))) (defun ss5 () (untrace) (lsv) (s5))
(defun s5A ()(time (set-var Example5A))) (defun ss5A () (untrace) (lsv) (s5A))
;;s5A needs the F(g1(A),g2(A),...,gn(A)) generalization of Rule1
(defun s5B ()(time (set-var Example5B))) (defun ss5B () (untrace) (lsv) (s5B))
(defun s6 () (time (set-var Example6))) (defun ss6 () (untrace) (lsv) (s6))
(defun s7 () (time (set-var Example7))) (defun ss7 () (untrace) (lsv) (s7))
(defun s8 () (time (set-var theorem8 H8 nil 'G)))
(defun s8 () (time (set-var Example8)))  ;2Feb93 either version of Ex8 works
(defun ss8 () (untrace) (lsv) (s8))
(defun s5p () (time (set-var Example5p))) (defun ss5p () (untrace (lsv)(s5p)))
(defun s6p () (time (set-var Example6p))) (defun ss6p () (untrace (lsv)(s6p)))
(defun s7p () (time (set-var Example7p))) (defun ss7p () (untrace (lsv)(s7p)))
(defun s8p () (time (set-var Example8p))) (defun ss8p () (untrace (lsv)(s8p)))

(defun s11 ()(time (set-var Example11))) (defun ss11 () (untrace) (lsv) (s11))
(defun s11* ()(time (set-var Example11*)))                     ;31Dec92
(defun ss11* () (untrace) (lsv) (s11*))                       ;31Dec92
(defun s11h () (time (set-var theorem11 H11 nil 'A)))
(defun ss11h () (untrace) (lsv) (s11h))
(defun s12 ()(time (set-var Example12))) (defun ss12 () (untrace) (lsv) (s12))
;;-----------------------------------------------------------------------------

(defun sss (n &optional (L '((ss1)(s2)(s3)(s4)(s5)(s6)(s7)(s8))) (m 1))
  (let (ans u)
    (cond ((null L) nil)
	  ((zerop n) nil)
	  (t (cons (list m (eval (car L))) (sss (1- n) (cdr L) (1+ m)))) )))

(defun tr () (trace set-var sv a-guns-sv new-guns find-targets fire
                    unblock resolve1-targets resolve2-targets collapse2
		    finish-by-strive f-remove-parents remove-ancesters
                    hyper-res))
(defun trr () (trace red-subset a-guns-sv11 resolve1 resolve2))

;;15Jan93 Successful Runs with set-var6.lisp
;;(sss 8)   18Nov92  (Using Str+ve for this batch)

;;(s5p) []  (s5p) []  (s6p) []  (s7p) []  (s8p) []     20Jan93

;;(sss 8) ((1 []) (2 []) (3 []) (4 []) (5 []) (6 []) (7 []) (8 [])) 2Feb93
