
;;;**********************************************************************
;;;
;;;                     SKOLEM.LISP    In KCL
;;;
;;;               W. W. Bledsoe      12 Feb 1987         16 May 1988
;;;                                                
;;;
;;;                 Called by  IMPLY.LISP
;;;
;;;
;;;***********************************************************************

(defvar *function-type-list* '(f g h))

(DEFUN SKOLEMIZE (A &OPTIONAL V (P T))
  ;;Used-in PROVE-Executor CLAIM DEFINE REDUC USE
  ;;V is a list of existential variables, usually nil
  ;;P is the parity
        (setq A (unique-bound-variables A))         ;in Prove4.lisp
	(SETQ A (CONSTANT-IZE-FREE-VARIABLES-IN A))
        (SKOLEM A V P))

(DEFUN SKOLEM (A V P)  ;USED-IN (CLAIM DEFINE REDUC SKOLEMIZE USE)
  ;; A IS A FORMULA -- V IS A LIST OF EXISTENTIAL VARIABLES -- P IS THE
  ;;  PARITY 
  (COND ((ATOM A) A)
	((OR (AND P (EQ 'SOME (CAR A))) (AND (NOT P) (EQ 'ALL (CAR A))))
	 (SKOLEM (CADDR A) (CONS (CADR A) V) P))
	((OR (AND P (EQ 'ALL (CAR A))) (AND (NOT P) (EQ 'SOME (CAR A))))
	 (SKOLEM (SUBST= (MAKE-SKOLEM-EXPRESSION (CADR A) V)   
                        (CADR A) (CADDR A))
		 V P))
		 
	((EQ 'NOT (CAR A)) (LIST 'NOT (SKOLEM (CADR A) V (NOT P))))
	((EQ '-> (CAR A))
	 (LIST '-> (SKOLEM (CADR A) V (NOT P)) (SKOLEM (CADDR A) V P)))
	((MEMBER (CAR A) '(& OR))
	 (n-ary-&or (CONS (CAR A) 
                          (MAPCAR #'(lambda(Q) (SKOLEM Q V P)) (CDR A)))))
	(T A)))

(DEFUN SKOLEM* (A V P)                              ;27June92
  ;;Like Skolem, except that the emerging variables retain their quantifiers
  (COND ((ATOM A) A)
	((OR (AND P (EQ 'SOME (CAR A))) (AND (NOT P) (EQ 'ALL (CAR A))))
	 (list (car A) (cadr A) (SKOLEM* (CADDR A) (CONS (CADR A) V) P)))
	((OR (AND P (EQ 'ALL (CAR A))) (AND (NOT P) (EQ 'SOME (CAR A))))
	 (SKOLEM* (SUBST= (MAKE-SKOLEM-EXPRESSION (CADR A) V)   
                        (CADR A) (CADDR A))
		 V P))
		 
	((EQ 'NOT (CAR A)) (LIST 'NOT (SKOLEM* (CADR A) V (NOT P))))
	((EQ '-> (CAR A))
	 (LIST '-> (SKOLEM* (CADR A) V (NOT P)) (SKOLEM* (CADDR A) V P)))
	((MEMBER (CAR A) '(& OR))
	 (n-ary-&or (CONS (CAR A) 
                          (MAPCAR #'(lambda(Q) (SKOLEM* Q V P)) (CDR A)))))
	(T A)))

(defun skolem-ALL (A V &optional P)          ;used in pbr.lisp 17Jan92
  ;; A is a formula -- V is a list of exist vars -- P is the parity 
  ;;Like SKOLEM except it only handles ALL, and treats atoms as variables
  (cond ((atom A) A)
	((or (and P (eq 'SOME (car A))) (and (not P) (eq 'ALL (car A))))
	 (list (car A) (cadr A) (skolem-ALL (caddr A) (cons (cadr A) V) p)))
	((OR (AND P (EQ 'ALL (CAR A))) (AND (NOT P) (EQ 'SOME (CAR A))))
	 (skolem-ALL (subst= (make-skolem-expression (cadr A) V)   
                        (cadr A) (caddr A))
		 V P))
	((eq 'not (car A)) (list 'not (skolem-ALL (cadr A) V (not P))))
	((eq '-> (car A))
	 (list '-> (skolem-all (cadr A) V (not P)) (skolem-all (caddr A) V P)))
	((member (car A) '(& or))
	 (n-ary-&or (cons (car A) 
                          (mapcar #'(lambda(Q) (skolem-all Q V P)) (cdr A)))))
	(t A)))

(defun skolem-functions-for (A)     ;20June92  27June92
  ;;eg, (skolem-functions-for '(all x (some y (P x y)))) 
  ;;                       =   (all x (P x (yS x)))
  ;;    ( "" (all x (all y (some z (P x y z))))) 
  ;;                       =   (all x (all y (P x y (zS x y))))
  (sk-fns-for A (free-variables-in A)))

(defun sk-fns-for (A V &optional SF)  ;called by skolem-functions-for 27Jun92
  (cond ((atom A) A)
	((and (not SF) (eq-car A 'all))
	 (sk-fns-for A V (skolem A V nil)))
	((eq-car A 'all)
	 (list 'all (cadr A) 
	       (sk-fns-for (Caddr A) V SF)))
	((member (car A) '(& or))
	 (cons (car A) (mapcar #'(lambda (u) (sk-fns-for u V SF))
			       (cdr A))))
	(SF SF)
	(t A)))

(DEFUN CONSTANT-IZE-FREE-VARIABLES-IN (A)     ;USED-IN (SKOLEMIZE)
  (let (fcn-lt)
   (cond ((occur-in '(lim cont cont-n cont-de diff diff-l + * -) A)
          (setq fcn-lt t)))
   (MAPC #'(lambda(X) (cond ((and fcn-lt (member x *function-type-list*)) nil)
     ;;Leave function symbols, such as f,g,h, as variables, if A contains
     ;; one of the symbols like + * lim etc.
                            (t (SETQ A (SUBST= (LIST X) X A)))))
	 (FREE-VARIABLES-IN A))
   A))

(DEFUN CRUNCH (L)         ;USED-IN (ALL-VARIABLES-IN AND-C ATOMS-IN
			  ;CALLED-IN CRUNCH FREE-VARIABLES-IN VARIABLES-IN)
  (COND ((NULL L) L)
	((MEMBER (CAR L) (CDR L) :test #'equal) (CRUNCH (CDR L)))
	(T (CONS (CAR L) (CRUNCH (CDR L))))))

(DEFUN FREE-VARIABLES-IN (A)        
  ;;USED-IN (CONSTANT-IZE-FREE-VARIABLES-IN DEFINE FULLY-QUANTIFY)
  ;;The same as variables-in, except that here we omit quantified vars
  (CRUNCH (F-VARS-IN A)))
 
(DEFUN F-VARS-IN (A)                ;USED-IN (FREE-VARIABLES-IN)
  ;; CALLED BY FREE-VARIABLES-IN 
  (COND ((NUMBERP A) NIL)
        ((null A) nil)
        ((member A '([] & or -> not)) nil)
	((ATOM A) (LIST A))
	((MEMBER (CAR A) 
                '(ALL SOME SET-OF UNION-OF-SETS INTERSECTION-OF-SETS lambda
		      product sum))
	 (F-VARS-IN (SUBST= (LIST (CADR A)) (CADR A) (CADDR A))))
	((ATOM (CAR A))
	 (COND ((NULL (CDR A)) NIL) (T (MAPCAN 'F-VARS-IN (CDR A)))))
	(T (MAPCAN 'F-VARS-IN A))))

(DEFUN MAKE-SKOLEM-EXPRESSION (X V &optional)  ;USED-IN (SKOLEM)

  (COND (V (CONS (new-symbol (MAKNAM  (LIST X 'S))) V))
	(T (LIST X))))

(defun n-ary-&or (A)   ;Used by Skolem.          Replaced by below 15May92
  ;;A has the form (& A B ...) or (or A B ...).
  ;;eg,  (& A (& B C) D) goes to (& A B C D)
  (let (s a* a**)
      (setq s (car a))
      (cond ((setq a* (assoc-r s (cdr a)))   ;See replace.lisp for assoc-r
	     ;;(& a b (& c d) e) goes to (& a b c d e), etc.  a* = (& c d)
	     (setq a** (remove-one a* a)) ;a** = (& a b e)
	     (setq a** (cons s (append (cdr a*) (cdr a**))))
	     (n-ary-&or a**))
	    (t A))))

(defun n-ary-&or (A &optional s)   ;Used by Skolem                    ;15May92
  ;;A has the form (& A B ...) or (or A B ...).
  ;;eg,  (& A (& B C) D) or (& (& A B) C) goes to (& A B C D)
  (cond ((atom A) A)
	(s (cond ((eq-car A s)
		  (cond ((null (cdr A)) nil) ;A=(&)
			((null (cddr A)) (cadr A)) ;A=(& a)
			((eq s '&) (and-on (n-ary-&or (cadr A) s)
					   (n-ary-&or (remove (cadr A) A 
						      :test 'equal) s)))
			(t (or-on (n-ary-&or (cadr A) s)
					   (n-ary-&or (remove (cadr A) A 
						      :test 'equal) s)))))
		 (t A)))
	((member (setq s (car A)) '(& or))
	 (n-ary-&or A s))
	(t A)))

(defun assoc-r (a L)     ;Called by REPLACE-TERM, REDUCE2
  ;;Return the first member of L of the form (a ...) if there is one, else NIL
  ;;This is different from ASSOC because L may have atomic members
  (cond ((atom L) nil) 
        ((atom (car L)) (assoc-r a (cdr L)))
        ((equal a (caar L)) (car L))
        (t (assoc-r a (cdr L))) ))

(defun new-symbol (x)  ;called by UNIQUE-BOUND-VARIABLES,MAKE-SKOLEM-EXPRESSION
  ;;replace x by a sym not occuring in %theorem-variables%
  (prog (n x*)
    (setq x* x) (setq n 1)
    TOP
    (cond ((member x* %theorem-variables%) ;we did have OCCUR instead of MEMBER
           (setq x* (maknam (list x n)))   ; here, but it is VERY slow.
           (setq n (1+ n))                
           (go TOP))
          (t (return x*)) )))

(defun subst= (a b c)  ;replace a by b in c
   ;;So that (subst= '(y) '(x) '(P x))                correctly gives   (P x).
   ;;Whereas (subst  '(y) '(x) '(P x) :test 'equal) incorrectly gives   (P y).
   (cond ((equal b c) a)
         ((atom c) c)
         (t (mapcar #'(lambda (x) (subst= a b x)) c))))

(defun occur-in (L C)  ;does a member of the list L occur in C
  (some #'(lambda (s) (occur s C)) L))

(defun remove-one (x L)   ;remove the first x from L 
  (cond ((equal x L) nil)
        ((atom L) L)
        ((equal x (car L)) (cdr L))
        (t (cons (car L) (remove-one x (cdr L)))) ))

