

;;;                STRUVE-GUIDE.LISP                        16Sept93
;;;
;;;

(defun lsg () (load "struve-guide.lisp"))

;;;This requires the loading of pc.lisp and its list of files.
;;;The idea is to "guide" the use of STRUVE in proving theorems
;;; that require at least one non-set-theory-lemma.  

(defvar *hist=* nil)
(setq *Lemmas-sko* (mapcar #'(lambda (u)(skolemize u nil nil)) lemmas-ex4L*))

(defun struve-guide (H C &optional (nbc 0))
  ;;This acts as follows:
  ;;(1) First try STRUVE 
  ;;    Remove non-set-theory, and unusable hypotheses first.
  ;;(2) Then try to Back-chain (and complete the proof then) (sko form)
  ;;(3) Then try to set-theory rewrite C, and try again
  ;;(4) If all that fails, try STRUVE with a longer compute time.
  (let ( (N 5000) C-de H-de A-de W H* H** A* A** H-sko C-sko)
    (cond ((eq t C) t)
	  ((eq-car C '->) (struve-guide (and-on (cadr C) H) (caddr C) nbc))
          ((progn (setq C-de (demodu H C))
		  (setq H-de (demodu H H))
		  (setq C-de (apply-defns C-de))
		  (setq H-de (apply-defns H-de))
		  (setq H-de (set-theory-part H-de))
		  (setq W (discard C-de H-de)) 
		  (setq H* (car W)) (setq H** (cadr W))
		  (setq A* (imp-on H* C-de)) (setq A** (imp-on H** C-de))
		  (prnt nbc) nil))
	  ((eq t (time (struve-prove A** 10 10))) t)  ;10 for today 25Sep93
	  ((> nbc 2) nil)
	  ((and (> nbc 0) (not (equal A** A*))
		(eq t (time (struve-prove A* N N)))) t)
	  ((and (setq H-sko (skolemize H nil nil)) 
		(setq C-sko (skolemize C nil t)) 
		nil))
	  ((eq t (back-chain-set-theory *Lemmas-sko* H-sko C-sko H C nbc)) t)
	  ((and (= nbc 0) (not (equal A** A*))
		(eq t (time (struve-prove A* N N)))) t)
	  (t nil) )))
	  	  
(defun back-chain-set-theory (lems H-sko C-sko H C nbc)
  (let ()
    (cond ((null lems) nil) 
	  ((eq t (bcst (car lems) H-sko C-sko H C nbc)) t)
	  (t (back-chain-set-theory (cdr lems) H-sko C-sko H C nbc)))))

(defun bcst (L H-sko C-sko H C nbc &optional (PREM t))
  ;;called by BACK-CHAIN-SET-THEORY
  (let (theta V V2 PREM2 PREM3 PREM4 nn C* H*)  
    (cond ((and (setq theta (unify* L C-sko H-sko PREM))
		(progn (setq V (free-variables-in (imp-on H C)))
		       (setq PREM2 (apply-sub theta PREM))
		       (setq V2 (remove t (free-variables-in PREM2)))
		       (setq PREM3 (unsko-part PREM2 V))
		       (setq PREM4 (existentially-quantify V2 PREM3))
		       (prnt 'bcs PREM4) t)
		(eq t (struve-guide H PREM4 (1+ nbc)))) t)
	  ((and (eq-car L '=) (not (member L *hist=* :test 'equal))
		(or (and (setq theta (sub=g (cadr  L) C-sko)) (setq nn 1))
		    (and (setq theta (sub=g (caddr L) C-sko)) (setq nn 2)))
		(setq C* (apply-sub theta C-sko))
		(setq H* (apply-sub theta L))
		(or 
		 (and (eq nn 1) 
		      (setq C* (subst (caddr H*)(cadr H*) C* :test 'equal)))
		 (and (eq nn 2) 
		      (setq C* (subst (cadr H*)(caddr H*) C* :test 'equal))))
		(setq *hist=* (cons L *hist=*))
		(progn (setq V (free-variables-in (imp-on H C)))
		       (setq PREM2 (and-on C* PREM))
		       (setq PREM2 (apply-sub theta PREM2))
		       (setq V2 (remove t (free-variables-in PREM2)))
		       (setq PREM3 (unsko-part PREM2 V))
		       (setq PREM4 (existentially-quantify V2 PREM3))
		       (prnt 'bcst '= PREM4) t)
		(eq t (struve-guide H PREM4 (1+ nbc)))) t)
	  ((eq-car L '->) (setq PREM (and-on (cadr L) PREM)) 
	   (bcst (caddr L) H-sko C-sko H C nbc PREM))

	  ((eq-car L '&)  
	   (cond ((some #'(lambda (Li) 
                               (eq t (bcst Li H-sko C-sko H C nbc PREM)))
			(cdr L)) t)
		 (t nil)))

;;	  ((eq-car L 'or) (prnt L) (break) (Handle later))

	  ((and (eq-car L '->) (eq-car (caddr L) '->))
           ;;L= (-> (L1 (-> L2 C1))) ~> (-> (& L1 L2) C1)
	   (setq L* (list '-> (and-on (cadr L) (cadr (caddr L)))
			      (caddr (caddr L))))
	   (bcst L* H-sko C-sko H C nbc PREM))
	  (t nil))))

(defun atoms-of (A x)  ;called by MATCH~
  ;;atoms of A containing the symbol x (predicate, function, constant symbol)
  (let (AA AA1 AA2)
    (cond ((atom A) nil)
	  ((eq-car A x) (list A))
	  ((and (eq-car A 'not) (eq-car (cadr A) x)) (list A))
	  ((member (car A) '(& or)) 
	   (mapcan #'(lambda (u) (atoms-of u x)) (cdr A)))
	  ((and (eq-car A 'not) (setq AA (atoms-of (cadr A) x)))
	   (mapcar #'(lambda (u) (not-fcn u)) AA))
	  ((and (eq-car A '->))
	   (setq AA1 (atoms-of (cadr  A) x))
	   (setq AA2 (atoms-of (caddr A) x))
	   (crunch (append (mapcar #'(lambda (u) (not-fcn u)) AA1) AA2)))
	  ((member (car A) '(all some)) (atoms-of (caddr A) x))
	  ((occur x A) (list A))
	  (t nil))))

(defun discard (C H)  ;Called by STRUVE-GUIDE  
  (let (syms-C syms-H syms-HC syms-HC*  H* H** AAC AAH AA W)
    (setq H* (cond ((eq-car H '&) (cdr H)) (t (list H))))
    (setq syms-C (syms C)) (setq syms-C (append (car syms-C) (cadr syms-C)))
    (setq syms-H (crunch (mapcan #'(lambda (Hi) 
                           (setq W (syms Hi)) (append (car W) (cadr W)))
				 H*)))
    (setq syms-HC (remove '= (crunch (append syms-H syms-C))))
    (setq AAC (crunch (mapcan #'(lambda (x) (atoms-of (not-fcn C) x)) syms-C)))

     ;;RULE 1: For each x in Sym-HC, if only one Hi contains x, remove Hi fr H 

     ;;RULE 1: For each x in Sym-HC, if only one Hi contains x, remove Hi fr H 


    (setq H* (discard-Rule1 H* syms-HC AAC))

     ;;RULE2: For each x in Syms-HC, if x occurs in Hi and if x does
     ;;       not occur in an Hj in a "negation" way, remove Hi from H.

    (setq syms-H* (crunch (mapcan #'(lambda (Hi) 
                           (setq W (syms Hi)) (append (car W) (cadr W)))
				  H*)))
    (setq syms-HC* (remove '= (crunch (append syms-C syms-H*))))
    (setq H** (discard-Rule2 H* syms-HC* AAC))
    
    (prnt H* H**)
    (list (andify H*) (andify H**)) ))

(defun discard-Rule1 (H Syms AAC)                      ;Called by DISCARD
  (let ((H* H) (syms* syms) Hx)
    (mapc #'(lambda (x) 
             (setq Hx (mapcan #'(lambda (Hi) (cond ((occur x Hi) (list Hi)))) 
			      H))
	     (cond ((and Hx (null (cdr Hx)))
		    (prnt x H* Hx)
		    (setq H* (remove (car Hx) H* :test 'equal))
		    (setq syms* (remove x syms*))
		    (prnt H* syms*) (break) t
		    )))
	  syms)
    (cond ((equal H* H) H*)
	  (t (discard-Rule1 H* syms* AAC)) )))

(defun discard-Rule2 (H syms AAC)                  ;Called by DISCARD
  (let ((H** H) (syms* syms) AAH AA)
    (setq AAH (crunch (mapcan #'(lambda (x) (atoms-of H x)) syms)))
    (setq AA (crunch (append AAC AAH)))
    (mapc #'(lambda (x) 
             (setq Hx (mapcan #'(lambda (Hi) (cond ((occur x Hi) (list Hi)))) 
			      H))
	     (cond ((and Hx (null (cdr Hx)))
		    (setq H** (remove (car Hx) H** :test 'equal))
		    (setq syms* (remove x syms*)))                ;not likely
		   ((not (match~2 (atoms-of Hx x) AA))
		    (setq H** (set-difference H** Hx)) 
		    (setq syms* (remove x syms*)))))
	  syms)
    (cond ((equal H** H) H**)
	  (t (discard-Rule2 H** syms* AAC)) )))

(defun existentially-quantify (vL A)              ;called by RULE1
  ;;existentially quantify A on each variable in vL
  (cond ((null vL) A)
	(t (setq A (list 'some (car vL) A))
	   (existentially-quantify (cdr vL) A))))

(defun half-of (L &optional (Ln (length L)))
  (cond ((null L) nil)
	((<= (length L) (* .5 Ln)) nil)
	(t (cons (car L) (half-of (cdr L) Ln)))))

(defun match~ (HH AAC H)                     ;called by DISCARD
  ;;HH= (Wx Wy ... Wz), where each Wx has the form
  ;;Wx= [x [(Hx1 (Ax11... Ax1m))...(Hxr (Axr1... Axrn)) ...]], where
  ;;  Each Axsk of (Hxs (Axs1 ... Axsm)), is an atom of Hxs which contains x,
  ;;  ie, Axsk= (A x) or Axk= (not (A x))
  ;;Discard such an (Hxs (Axs1 ... Axsm)) if any one of these Axsk's 
  ;;  does NOT have its negation (not (A x)) or (A x) in another 
  ;;  (Hxr (Axr1 ... Axrm)) of Wx, and remove Hxs from H.
  (let (HHm HAH HAH* AA3 AA1 AA2)
    (setq HHm (mapcan #'(lambda (W)    
         (setq HAH* (mapcan #'(lambda (HA) ;HA= (Hxs (Axs1 ... Axsm))
                      (setq AA1 (cadr HA)) ;AA1 =    (Axs1 ... Axsm)
		      (setq HAH (remove HA (cadr W) :test 'equal))
		      (setq AA3 (mapcan-cadr HAH))
		      (setq AA2 (crunch (append AA3 AAC)))
		      (cond ((match~2 AA1 AA2) (list HA))
			    (t (setq H (remove (car HA) H :test 'equal)) nil)))
			    (cadr W)))
	 (cond ((cdr HAH*) (list (list (car w) HAH*)))
	       (t (setq H (remove (caaadr W) H :test 'equal)) nil)))
		      HH))
    (list HHm H)))

(defun mapcan-cadr (HAH)
  (cond ((null HAH) nil)
	(t (append (cadr (car HAH)) (mapcan-cadr (cdr HAH))))))


(defun match~2 (AA1 AA2) ;Called by MATCH~
  ;;If the negation of every atom of AA1 appears in AA2, return t.
  (every #'(lambda (A) 
;;                      (member    (not-fcn A) AA2 :test 'equal)
                        (member-st (not-fcn A) AA2)
        ) AA1))

(defun member-st (A L) ;called by MATCH~2
  ;;eg, (member-st '(not (e [] B)) '((f x) (subset B C))) = T
  (or (member A L :test 'equal) 
      (some #'(lambda (u) (equal-st A u)) L)))

(defun equal-st (A B) ;called by MEMBER-ST
  (or (equal A B) (equal-st2 A B) (equal-st2 B A)))

(defun equal-st2 (A B) ;called by EQUAL-ST
   (or (and (eq-car A 'e) (eq-car B 'subset)        ;(e x E) (subset D E)
	    (equal (caddr A) (caddr B)))             
       (and (eq-car A 'e) (eq-car-not B 'subset)    ;(e x D) (not (subset D E))
	    (equal (caddr A) (cadr (cadr B))))
       (and (eq-car-not A 'e) (eq-car B 'subset)    ;(not (e x D)) (subset D E)
	    (equal (caddr (cadr A)) (cadr B)))
       (and (eq-car-not A 'e) (eq-car-not B 'subset)     ;(not (e x E)) 
	    (equal (caddr (cadr A)) (caddr (cadr B)))) ;(not (e D E))
       ))

(defun eq-car-not (A B) (and (eq-car A 'not) (eq-car (cadr A) B)))

(defun set-theory-part (H) ;called by STRUVE-GUIDE.
  (cond ((eq-car H '&) 
	 (andify (mapcan #'(lambda (u) (cond ((set-theory u) (list u))))
			 (cdr H))))
	((set-theory H) H)
	(t nil)))

(defun sub=g (H C)  ;Called by BCST       20Sept93
  ;;If C* is a subformula of C, and theta= (unify H C*), return theta.
  ;;eg, (sub=g '(Res-L S6 (+ N11 N21)) '(e [] (res-L (S) (+ (n1) (n2)))))
  ;;    =  ((S6 S) (N11 n1) (N21 n2))
  (let (theta)
    (cond ((setq theta (unify H C)) theta)
	  ((atom C) nil)
	  ((and (not (atom (car C))) (sub=g H (car C))))
	  ((sub=g H (cdr C)))
	  (t nil))))

(defun unify* (L C-sko H-sko PREM) ;called by BCST
  (let ()
    (cond ((setq theta (unify L C-sko)) theta) ;if unify works return THETA
	  ((and (setq H2 (and-on C-sko PREM))  ;else sub=, and try again
		(setq E (=s-in (cdr H2)))
		(setq W (sub== E L C-sko)))
	   (unify (car W) (cadr W)))
	  (t nil))))

(defun unsko-part (A L) 
  (let (A*)
    (setq A* A)
    (mapc #'(lambda (x) (setq A* (subst x (list x) A* :test 'equal))) L)
    A*))

;;--------------------------------------------------------------------

(setq H472a
 '(& (-> (E (SNG L) (RES-L S N1)) (SUBSET S2 (RES-L S N1)))
     (E (SNG L) (RES-L S N1)) (E [] (RES-L S2 N2)) (E [] (RES-L S1 N1))
     (= C1 (~ C (SNG L))) (= C2 (SNG L))
     (= S1 (U (~ S (SNG C)) (SNG C1))) (= S2 (U (~ S (SNG C)) (SNG C2)))
     (E L C) (E C S) (NU C) (CLAUSES-L S) (FNT S) (UNSAT-L S)
     (< 0 (EXL S)) (NOT (E [] S))))

(setq step472 '(-> (E (SNG L) (RES-L S N1))
		   (SUBSET (RES-L S2 N2) (RES-L (RES-L S N1) N2))))

(defun sg2 () (time (struve-guide H472a step472)))
