
;;;*********************************************************************
;;;
;;;                      sv-aux.Lisp
;;;                                                         18Feb93
;;;  Auxiliary Fucntions for the file sv-marty.lisp
;;;

(setq *factors* nil)                            ;Should this be change?

(DEFUN ALL-VARIABLES-IN (A)   ;used in NEW-BINDINGS, (DEFINE,REDUC)
  ;; THIS IS THE SET OF ALL VARIABLES - GROUND OR NOT - IN A 
  (CRUNCH (ALL-VARS-IN (COPY-TREE A))))

(DEFUN ALL-VARS-IN (A)    ;used in ALL-VARIABLES-IN
  (cond ((atom a) 
         (cond ((numberp a) nil)
	       ((member a *constant-list*) nil) 
	       (t (list a))))
        ((atom (car a))
         (cond ((null (cdr a)) 
                (cond ((member (car a) *constant-list*) nil) (t a)))
	       (t (mapcan #'all-vars-in (cdr a)))))
        (t (append (all-vars-in (car A)) (all-vars-in (cdr A))))))

(defun andify (l)			;used in VE
  ;;convert a list into a conjunction
  (cond ((null l) nil)
	((null (cdr l)) (car l))
        (t (cons '& l))))

(DEFUN AND-ON (A B)
  ;; e g, (and-on (& c d) (& e f))  =  (& c d e f)
  (COND ((EQ T A) B)
	((EQ T B) A)
	((NULL A) B)
	((NULL B) A)
        ((equal B '(empty)) A)
	(T (CONS '&
		 (UNION-W (COND ((AND (NOT (ATOM A)) (EQ '& (CAR A))) (CDR A))
			      (T (LIST A)))
			(COND ((AND (NOT (ATOM B)) (EQ '& (CAR B))) (CDR B))
			      (T (LIST B))))))))

 (DEFUN AND-ON-A (A B)    ;used by RULE1
   ;; e g, (and-on-a (& c d) (& e f))  =  ((& c d e f) t), and
   ;;    (and-on-a '(& (p x) (r z)) '(& (p (b)) (p y)))
   ;;  = ((& (R Z) (P (B)) (P Y)) ((X . Y)))

   (let (A* B* U)
      (COND ((EQ T A) B) ((EQ T B) A) ((NULL A) B) ((NULL B) A)
	    ((equal B '(empty)) A)
	    (t (setq A* (COND ((eq-car A '&) (cdr A)) (T (LIST A))))
	       (setq B* (COND ((eq-car B '&) (cdr B)) (T (LIST B))))
	       (setq U (union-w A* B*))
	       (setq U (variant-crunch U))
	       (list (cons '& (car U)) (cadr U))))))

(DEFUN APPLY-substitution (THETA D)	;widely used
       ;; the substitution THETA is applied to the form D
       (COND ((EQ T THETA) D)
	     ((TWO-XS THETA NIL)
	      (PROG (X2)
		    (SETQ X2 (TWO-XS THETA))
		    (RETURN (reduc (AND-ON (APPLY-SUBSTITUTION (CAR X2) D)
				    (APPLY-SUBSTITUTION (CADR X2) D))))))
	     (T (SUBLIS= THETA D))))

(defun binary->unary< (A)                 ;7sep92
  ;;eg, (binary->unary< '(<= 1 2)) = (<= -1)
  (cond ((atom A) A)
	((and (not (occur '< A)) (not (occur '<= A))) A)
	((and (member (car A) '(< <=)) (cddr A)) ;A= (< x y)
	 (list (car A) (minus-on (cadr A) (caddr A))))
	(t (cons (binary->unary< (car A)) (binary->unary< (cdr A))))))


(defun bound-variables-in (A)     ;Called by PROVE-executor
  (cond ((atom A) nil)
        ((member (car A) '(all some set-of union-of-sets intersection-of-sets))
         (cons (cadr A) (bound-variables-in (caddr A))))
        (t (union-w (bound-variables-in (car A)) 
		    (bound-variables-in (cdr A))))))

(DEFUN COMPOSE (THETA SIGMA)  ;widely used  
  ;; this returns the composition of the two substitutions THETA & SIGMA
  (COND ((or (null theta) (null sigma)) nil)
        ((EQ T THETA) SIGMA)
        ((EQ T SIGMA) THETA)
        (T (SETQ THETA
                 (COND ((TWO-XS SIGMA NIL)
                        (PROG (X2)
                              (SETQ X2 (TWO-XS SIGMA))
                              (RETURN (UNION-W (COMPOSE THETA (CAR X2))
                                             (COMPOSE THETA (CADR X2))))))
                       (T (MAPCAR #'(lambda(U)
  				    (CONS (CAR U) (SUBLIS= SIGMA (CDR U))))
				  THETA))))
	   (SETQ THETA (UNION-W THETA SIGMA))
	   (MAPCAN #'(lambda(U)
		     (COND ((EQUAL (CAR U) (CDR U)) NIL) (T (LIST U))))
		   THETA))))

(defun convert-implication-signs (A) 
  ;;Ie, (-> P Q) goes to (or (not P) Q)
  (cond ((atom A) A)
        ((eq '-> (car A)) 
         (list 'or (list 'not (convert-implication-signs (cadr A)))
                   (convert-implication-signs (caddr A))))
        ((eq 'not (car A)) (list 'not (convert-implication-signs (cadr A))))

        ((member (car A) '(& or))
         (cons (car A) (mapcar #'convert-implication-signs (cdr A))))

	((member (car A) '(all some))
	 (list (car A) (cadr A) (convert-implication-signs (caddr A))))

        (t A)))

(DEFUN CRUNCH (L)   ;widely used  
  ;;removes duplication from a list
  (COND ((NULL L) L)
	((MEMBER (CAR L) (CDR L) :test #'equal) (CRUNCH (CDR L)))
	(T (CONS (CAR L) (CRUNCH (CDR L))))))

(defun explode (name)                       ;shaun's  Jan 1990
  (cond ((numberp name) name)
        (t   (mapcar #'(lambda (x) 
               (cond ((string= "0" x) 0)
                     ((string= "1" x) 1)
                     ((string= "2" x) 2)
                     ((string= "3" x) 3)
                     ((string= "4" x) 4)
                     ((string= "5" x) 5)
                     ((string= "6" x) 6)
                     ((string= "7" x) 7)
                     ((string= "8" x) 8)
                     ((string= "9" x) 9)
                     (t (intern x))))
        (mapcar #'string (coerce (symbol-name name) 'list))))))

(defun eq-car (A B)   ;is A non-nil and (Car A) is B.
  (and (not (atom A)) (eq (car A) B)))

(defun eq-cadr (A x)  ;called by sub=<  Is (cadr A) equal to x?
  (and (not (atom A)) (not (atom (cdr A))) (eq x (cadr A))))

(defun factor-subsume (C &optional (C* C)) ;Called by STR+VE-CHAIN
  ;;If C* is a variant-factor of C which subsumes C, the return C*
  ;;Return the smallest such C*.  If no change, return C itself.
  (let (theta L1 C2 C3)
    (cond ((null C) C*)
	  ((null (cdr C)) C*)
	  ((and (setq L1 (car C)) (setq C2 (remove1 L1 C*)) nil))
	  ((some #'(lambda (L) 
                     (and (setq theta (variant L1 L))
			  (setq C3 (crunch (sublis theta c2)))
			  (subset C3 C2))) 
		 (cdr C))            ;Success, try for more
	   (factor-subsume C3 C3))
	  (t (factor-subsume (cdr C) C*)))))

(DEFUN imp-ON (A B)    ;widely used    VE-the-variable
  (cond ((null A) B) ((eq t A) B) ((eq t B) t) ((equal '(empty) A) B) 
	((null B) (reduc (list 'not A))) 
        (t (list '-> A B))))

(DEFUN MAKNAM (L)                ;Fuzes the list L into an Atom
  ;;eg, (maknam '(a b 35)) = ab35
  (LET ((*PRINT-CASE* :UPCASE))
    (INTERN (WITH-OUTPUT-TO-STRING (V)
    	            (MAPC #'(LAMBDA (X) (PRIN1 X V))  L)
		        V))))

(DEFUN MINUS-ON (A B) 
  (cond ((or (null b) (equal 0 b)) 
         (cond ((or (null a) (equal 0 a)) 0) (t a)))
        ((or (null a) (equal 0 a)) (reduc (list '- b)))
	(t (REDUC (LIST '+ A (LIST '- B))))))
    


(defun negation (A)         ;called by STR+VE4
  ;;(< A B) goes to (<= B A), etc.  
  (cond ((atom A) (list 'not A))
	((eq '<  (car A)) (list '<= (neg-on-ve (cadr A))))
	((eq '<= (car A)) (list '<  (neg-on-ve (cadr A))))
	(t (reduce-ve (list 'not A)))))

(defun number-print (L &optional lt)     ;called by hand-resolution, hand-r
  (cond ((null L) nil)
        ((and (not lt) (cddar L) (not (eq 'input (caddar L))))
         (setq lt t)
         (format t "~%----------------------------~%")
	 (format t "~%~1,1t ~s ~2,6t ~s~%" (caar L) (cadar L))
         (format t "~48,22t ~s~%" (cond ((cddar L)(caddar L)) (t 'input)))
	 (number-print (cdr L) lt))
	(t  (format t "~%~1,1t ~s ~2,6t ~s~%" (caar L) (cadar L))
            (cond ((cddar L) 
                   (cond ((eq 'input (caddar L)) nil)
			 (t (format t "~48,22t ~s~%" (caddar L))))))
	    (number-print (cdr L) lt))))

 (defun occur (a b)   ;does x occur in A?
    (cond ((equal a b) t)
          ((atom b) nil)
          (t (some #'(lambda (x) (occur a x)) b))))

(defun orify (l)			;used in VE
  ;;convert a list into a conjunction
  (cond ((null (cdr l)) (car l))
        (t (cons 'or l))))

(DEFUN OR-ON (A B)   ;USED-IN (IMPLY-RULES OR-ON-LIST)
  (COND ((NOT A) B)
	((NOT B) A)
	(T (CONS 'OR
		 (UNION-W (COND ((AND (NOT (ATOM A)) (EQ 'OR (CAR A))) (CDR A))
			      (T (LIST A)))
			(COND ((AND (NOT (ATOM B)) (EQ 'OR (CAR B))) (CDR B))
			      (T (LIST B))))))))

(defun position-in (x L &optional (n 1))                 ;called by SV5
  ;;eg (position-in '(f x) '(P (g a) (f x) h q)) = 3
  (cond ((atom L) (format t "~%Position-in failed on ~s~%" x) nil)
	((equal x (car L)) n)
	(t (position-in x (cdr L) (1+ n)))))

(defvar *x* nil)(defvar *y* nil)(defvar *z* nil)(defvar *u* nil)
(defvar *v* nil)
(defmacro prnt (x &optional y z u v)
  `(let ((*x* ', x)(*y* ', y)(*z* ', z)(*u* ', u)(*v* ', v))
     (and *print-light* ,x (or (format t "~%~s~s ~s~%" *x* '= ,x) t)
	                ,y (or (format t "~%~s~s ~s~%" *y* '= ,y) t)
                        ,z (or (format t "~%~s~s ~s~%" *z* '= ,z) t)
                        ,u (or (format t "~%~s~s ~s~%" *u* '= ,u) t)
                        ,v (or (format t "~%~s~s ~s~%" *v* '= ,v) t)
			)))
(defun push-nots-in (A) 
  ;;Assume that A has no -> signs.
  ;;Ie, (not (& A (or B C))) goes to  (or (not A) (& (not B) (not C)))
  (cond ((atom A) A)
        ((eq 'not (car A))
         (cond ((atom (cadr A)) A)
               ((eq 'not (caadr A))       ;(not (not A)) goes to A
		(push-nots-in (cadadr A)))  

               ((eq '& (caadr A))
                (cons 'or (mapcar #'(lambda (x) 
                            (push-nots-in (list 'not x))) (cdadr A))))

               ((eq 'or (caadr A))
                (cons '& (mapcar #'(lambda (x) 
                            (push-nots-in (list 'not x))) (cdadr A))))
	       
	       ((eq 'all  (caadr A))
		(list 'some (cadr (cadr A))
		      (push-nots-in (list 'not (caddr (cadr A))))))

	       ((eq 'some (caadr A))
		(list 'all  (cadr (cadr A))
		      (push-nots-in (list 'not (caddr (cadr A))))))

	       ((eq '<  (caadr A))
		(list '<= (caddr (cadr A)) (cadr (cadr A))))

	       ((eq '<= (caadr A))
		(list '<  (caddr (cadr A)) (cadr (cadr A))))

               (t A)))
        ((member (car A) '(& or)) 
         (cons (car A) (mapcar #'push-nots-in (cdr A))))
	((member (car A) '(all some))
	 (list (car A) (cadr A) (push-nots-in (caddr A))))

        (t A)) )

(DEFUN REMOVE1 (X L)  ;This removes the first occurance of x from L
       (COND ((NULL L) NIL) ((EQUAL X (CAR L)) (CDR L))
             (T (CONS (CAR L) (REMOVE1 X (CDR L))))))

(defun remove-ors (C)             ;called by RDD
  ;;eg, If C= (D E (or F G) H) the (remove-ors C) = (D E F G H)
  (mapcan #'(lambda (L) (cond ((eq-car L 'or) (cdr L)) (t (list L)))) C))
 

(DEFUN SET-DIFFERENCE-W (L L*)   ;Woody's version     6-30-89
                                  ;USED-IN (USED-INS NEW-VARIABLES NEW-VARS)
  (MAPCAN #'(lambda(X) (COND ((MEMBER X L* :test #'equal) NIL) 
                             (T (LIST X)))) L))

(defun set-intersection (L L*)   ;27mar90
     (mapcan #'(lambda(X) (if (member X L* :test #'equal) (list x) nil)) L))

(defun skolem-expression (A)        ;called by TOP-SKOLEM
  (and (not (atom A)) (member 'S (last (cdr (explode (car A)))))))  ;27Aug92

(defun subset (L L2)     ;Is L included in L2?
  (every #'(lambda (x) (member x L2 :test 'equal)) L))

(defun top-skolem (A)             ;called by RULE1
 ;;A has the form F(g(A),...) where g is a skolem function and F is not
 ;;Return g(A)
 (let (tx)
   (cond ((atom A) nil)
	 ((not (atom (car A))) nil)
	 ((skolem-expression A) A)
	 ((some #'(lambda (x) (cond ((setq tx (top-skolem x))))) (cdr A)) 
	  tx))))

(DEFUN UNION-W (A B)			;a version of UNION

       (COND ((NULL B) A)
	     ((MEMBER (CAR B) A :test #'equal) (UNION-W A (CDR B)))
	     (T (UNION-W (APPEND A (LIST (CAR B))) (CDR B)))))
(defun unique-bound-variables (A)
  ;;called by SKOLEMIZE, DEFIN, DEFINEC
  ;;For each subformula Q of A of the form (ALL x P) or (SOME x P), replace x 
  ;; in Q by a variable not occuring in %theorem-variables% , or the rest of A.

  (let (x A* B B* W x*)
    (cond ((atom A) A)
	  ((member (car A)
		   '(all some set-of union-of-sets intersection-of-sets))
	   (setq x (cadr A)) (setq B (caddr A))
	   (setq B* (unique-bound-variables B))
	   (cond ((setq w (member x %theorem-variables%))
		  (cond ((and (equal B* B) (equal (cadr W) A)) A)
			(t (setq x* (new-symbol x))
                           (setq B* (subst x* x B*))
			   (setq A* (list (car A) x* B*))
			   (setq %theorem-variables% 
				 (append (list x* A*) %theorem-variables%))
			   A*) ))
		 ((equal B* B)
		  (setq %theorem-variables% 
			(append (list x A) %theorem-variables%))
		  A)
		 (t (setq x* (new-symbol x))
                    (setq B* (subst x* x B*))
                    (setq A* (list (car A) x* B*))
		    (setq %theorem-variables% 
			  (append (list x* A*) %theorem-variables%))
		    A*)))

          (t (cons (unique-bound-variables (car A))
                   (unique-bound-variables (cdr A)))) )))

(defun unique-ize-quan-vars (F &optional theta)          ;called by PR
  (let (W)
    (setq W (unique-iqv F))
    (cond (theta W) (t (car W)))))

(defun unique-iqv (F)          ;called by UNIQUE-IZE-QUAN-VARS
  ;;eg, (unique-iqv '(ALL x (& (P x) (some y (Q x y)))))
  (let (L x y F2 F* W theta L*)
    (cond ((atom F) (list F nil))
	  ((member (car F) '(all some)) 
	   (setq x (cadr F)) (setq L (explode x))
	   (setq y (maknam (list (car L) (setq *n* (1+ *n*)))))
	   (setq F2 (subst y x (caddr F) :test 'equal))
	   (setq W (unique-iqv F2))
	   (setq theta (cons (cons x y) (cadr W)))
	   (list (list (car F) y (car W)) theta))
	  ((member (car F) '(& or -> not))
	   (setq L (mapcar #'unique-iqv (cdr F)))
	   (setq L* (mapcar #'car L))
	   (setq theta (mapcan #'cadr L))
	   (list (cons (car F) L*) theta))
	  (t (list F nil)) )))

(defun variant-crunch (L &optional (L* L) (theta t))     ;called by AND-ON-A
  ;;eg, ((p x) (r z) (p (b)) (p y))  goes to ((r z) (p (b)) (p y))
 (let (sigma)
  (cond ((null (cdr L)) (list (crunch (apply-sub theta L*)) theta))
	((some #'(lambda (u) 
                   (and (not (equal (car L) u)) 
			(setq sigma (variant (car L) u))))
	       (cdr L))
	 (variant-crunch (cdr L) L* (compose sigma theta)))
	(t (variant-crunch (cdr L) L* theta)))))
